@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -37,10 +37,6 @@
  *   16Sep95 Hans Lermen <lermen@elserv.ffm.fgan.de>
  */
 
-#define LDT_ENTRIES 8192
-#define LDT_ENTRY_SIZE 8
-#define MAX_SELECTORS 8192
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -60,35 +56,18 @@
 #include "config.h"
 #include "emu.h"
 #include "cpu.h"
+#include "timers.h"
 #include "dpmi.h"
 #include "utilities.h"
+#include "dosemu_config.h"
+#include "hma.h"
+#include "dis8086.h"
 
 #define MHP_PRIVATE
 #include "mhpdbg.h"
 
 #define makeaddr(x,y) ((((unsigned long)x) << 4) + (unsigned long)y)
 
-#if 0 /* modify_ldt already defined in dpmi.c */
-_syscall3(int, modify_ldt, int, func, void *, ptr, unsigned long, bytecount)
-#endif
-extern int modify_ldt(int func, void *ptr, unsigned long bytecount);
-
-/* externs */
-extern struct mhpdbgc mhpdbgc;
-extern int a20;
-extern void mhp_send(void), mhp_close(void), mhp_putc(char c1);
-
-#if 0
-/* NOTE: the below is already defined with #include "emu.h"
- *       Must NOT redefine it, else vm86plus won't work !!!
- */
-extern struct vm86_struct vm86s;
-#endif
-
-extern int  dis_8086(unsigned int, const unsigned char *,
-                     unsigned char *, int, unsigned int *, unsigned int *,
-                     unsigned int, int);
-
 /* prototypes */
 static void* mhp_getadr(unsigned char *, unsigned int *, unsigned int *, unsigned int *);
 static void mhp_regs  (int, char *[]);
@@ -123,7 +102,6 @@ static void mhp_bclog   (int, char *[]);
 static void print_log_breakpoints(void);
 
 static unsigned int lookup(unsigned char *, unsigned int *, unsigned int *);
-static inline int get_ldt(void *);
 
 /* static data */
 static unsigned int linmode = 0;
@@ -138,7 +116,7 @@ static char lastldt[32];
 
 static struct symbol_entry symbol_table[MAXSYM];
 static unsigned int last_symbol = 0;
-static unsigned int addrmax;
+static unsigned long addrmax;
 
 static struct symbl2_entry symbl2_table[MAXSYM];
 static unsigned int last_symbol2 = 0;
@@ -260,11 +238,6 @@ static int mhp_addaxlist_value(int v)
   return 0;
 }
 
-static inline int get_ldt(void *buffer)
-{
-  return modify_ldt(0, buffer, LDT_ENTRIES * LDT_ENTRY_SIZE);
-}
-
 static unsigned char * getname(unsigned int addr)
 {
    int i;
@@ -351,7 +324,7 @@ static void mhp_rusermap(int argc, char *argv[])
   mhp_printf("reading map file %s\n", argv[2]);
   last_symbol2 = 0;
   for(;;) {
-     if(!fgets(bytebuf, 100, ifp)) {
+     if(!fgets(bytebuf, sizeof (bytebuf), ifp)) {
         mhp_printf("error: could not find following in %s:\n%s\n",
                    argv[2], srchfor);
         return;
@@ -362,7 +335,7 @@ static void mhp_rusermap(int argc, char *argv[])
         break;
   }
   for(;;) {
-     if(!fgets(bytebuf, 100, ifp))
+     if(!fgets(bytebuf, sizeof (bytebuf), ifp))
         break;
      if(bytebuf[5] != ':')
         continue;
@@ -388,8 +361,7 @@ static void mhp_rusermap(int argc, char *argv[])
 
 static void mhp_rmapfile(int argc, char *argv[])
 {
-  PRIV_SAVE_AREA
-  FILE * ifp;
+  FILE * ifp = NULL;
   unsigned char bytebuf[IBUFS];
   unsigned long a1;
   char *map_fname = DOSEMU_MAP_PATH;
@@ -397,9 +369,16 @@ static void mhp_rmapfile(int argc, char *argv[])
   if (argc >= 2) {
     map_fname = argv[1];
   }
-  enter_priv_off();
-  ifp = fopen(map_fname, "r");
-  leave_priv_setting();
+  if (map_fname == NULL && dosemu_proc_self_exe != NULL) {
+    /* try to get symbols on the fly */
+    map_fname = malloc(strlen(dosemu_proc_self_exe) + 60);
+    strcpy(map_fname, "nm ");
+    strcat(map_fname, dosemu_proc_self_exe);
+    strcat(map_fname, " | grep -v '\\(compiled\\)\\|\\(\\.o$\\)\\|\\( a \\)' | sort");
+    ifp = popen(map_fname, "r");
+  } else if (map_fname != NULL) {
+    ifp = fopen(map_fname, "r");
+  }
   if (!ifp) {
      mhp_printf("unable to open map file %s\n", map_fname);
      return;
@@ -407,7 +386,7 @@ static void mhp_rmapfile(int argc, char *argv[])
   mhp_printf("Reading map file %s\n", map_fname);
   last_symbol = 0;
   while (last_symbol < MAXSYM) {
-     if(!fgets(bytebuf, 100, ifp))
+     if(!fgets(bytebuf, sizeof (bytebuf), ifp))
         break;
      if (!strlen(bytebuf) || !isxdigit(*bytebuf))
         continue;
@@ -417,7 +396,7 @@ static void mhp_rmapfile(int argc, char *argv[])
       sscanf(bytebuf, "%lx %c %40s", &a1, &symbol_table[last_symbol].type,
               (char *)&symbol_table[last_symbol].name);
 #ifdef __ELF__
-     if (a1 < 0x08000000) continue;
+     if (a1 < 0x00000001) continue;
 #else
      if (a1 < 0x20000000)
         continue;
@@ -427,10 +406,15 @@ static void mhp_rmapfile(int argc, char *argv[])
      symbol_table[last_symbol].addr = a1;
      last_symbol++;
   }
-  fclose(ifp);
+  if (map_fname != argv[1] && map_fname != DOSEMU_MAP_PATH) {
+    pclose(ifp);
+    free(map_fname);
+  }
+  else
+    fclose(ifp);
   addrmax = symbol_table[last_symbol-1].addr;
   mhp_printf("%d symbol(s) processed\n", last_symbol);
-  mhp_printf("highest address %08x(%s)\n", addrmax, getname(addrmax));
+  mhp_printf("highest address %08lx(%s)\n", addrmax, getname(addrmax));
 }
 
 enum {
@@ -535,6 +519,7 @@ static void mhp_setreg(unsigned char * regn, unsigned long val)
 
 static void mhp_go(int argc, char * argv[])
 {
+   unfreeze_dosemu();
    if (!mhpdbgc.stopped) {
       mhp_printf("already in running state\n");
    } else {
@@ -678,7 +663,6 @@ static void mhp_dis(int argc, char * argv[])
 
 static void mhp_dump_to_file(int argc, char * argv[])
 {
-   PRIV_SAVE_AREA
    unsigned int nbytes;
    unsigned long seekval;
    unsigned char * buf = 0;
@@ -701,21 +685,14 @@ static void mhp_dump_to_file(int argc, char * argv[])
       return;
    }
 
-   if (config.secure && ((seekval+nbytes) >0x10ffff)) {
-      mhp_printf("secure is ON, can\'t access above 1MEG\n");
-      return;
-   }
-
-   enter_priv_off();
    fd = open(argv[3], O_WRONLY | O_CREAT | O_TRUNC, 00775);
-   leave_priv_setting();
 
    if (fd < 0) {
-      mhp_printf("cannot open/create file %s\n%s\n", argv[3], sys_errlist[errno]);
+      mhp_printf("cannot open/create file %s\n%s\n", argv[3], strerror(errno));
       return;
    }
    if (write(fd, buf, nbytes) != nbytes) {
-      mhp_printf("write error: %s\n", sys_errlist[errno]);
+      mhp_printf("write error: %s\n", strerror(errno));
    }
    close(fd);
 }
@@ -736,7 +713,6 @@ static void mhp_mode(int argc, char * argv[])
 
 static void mhp_disasm(int argc, char * argv[])
 {
-   extern int a20; /* a20 is defined in dosemu/xms.c */
    int rc;
    unsigned int nbytes;
    unsigned long org;
@@ -806,8 +782,8 @@ static void mhp_disasm(int argc, char * argv[])
              mhp_printf ("%s:\n", getname2(seg,off));
        }
        refseg = seg;
-       rc = dis_8086(org+bytesdone, buf+bytesdone, frmtbuf, def_size,
-                  &refseg, &ref, (IN_DPMI ? dpmi_mhp_getselbase(refseg) : 0), 16);
+       rc = dis_8086(buf+bytesdone, frmtbuf, def_size, &ref,
+                  (IN_DPMI ? dpmi_mhp_getselbase(refseg) : refseg * 16));
        for (i=0;i<rc;i++) {
            sprintf(&bytebuf[i*2], "%02X", *(buf+bytesdone+i) );
            bytebuf[(i*2)+2] = 0x00;
@@ -900,7 +876,7 @@ static void mhp_enter(int argc, char * argv[])
       return;
 
    if (!strcmp(argv[1],"-")) {
-     if ((int)zapaddr == -1) {
+     if (zapaddr == (void *)-1) {
         mhp_printf("Address invalid, no previous 'e' command with address\n");
         return;
      }
@@ -935,15 +911,15 @@ static void mhp_enter(int argc, char * argv[])
 
 static void* mhp_getadr(unsigned char * a1, unsigned int * s1, unsigned int *o1, unsigned int *lim)
 {
+   static char buffer[0x10000];
    unsigned char * srchp;
    unsigned int seg1;
    unsigned int off1;
    unsigned long ul1;
    int selector = 0;
    int use_ldt = IN_DPMI;
-   unsigned long * lp;
-   unsigned long base_addr, limit;
-   char buffer[0x10000];
+   unsigned int * lp;
+   unsigned int base_addr, limit;
 
    *lim = 0xFFFFFFFF;
 
@@ -1019,7 +995,7 @@ static void* mhp_getadr(unsigned char * a1, unsigned int * s1, unsigned int *o1,
      return (void *)0;
    }
 
-   lp = (unsigned long *) buffer;
+   lp = (unsigned int *) buffer;
    lp += (seg1 & 0xFFF8) >> 2;
 
    base_addr = (*lp >> 16) & 0x0000FFFF;
@@ -1041,7 +1017,7 @@ static void* mhp_getadr(unsigned char * a1, unsigned int * s1, unsigned int *o1,
    }
 
    *lim = limit - off1;
-   return (void *) base_addr + off1;
+   return (void *)(uintptr_t)(base_addr + off1);
 }
 
 int mhp_setbp(unsigned long seekval)
@@ -1426,13 +1402,12 @@ int mhp_bpchk(unsigned char * a1)
 int mhp_getcsip_value()
 {
   int  seg, off, limit;
-  if (IN_DPMI) return (int)mhp_getadr("cs:eip", &seg, &off, &limit);
+  if (IN_DPMI) return (int)(uintptr_t)mhp_getadr("cs:eip", &seg, &off, &limit);
   else return (LWORD(cs) << 4) + LWORD(eip);
 }
 
 void mhp_modify_eip(int delta)
 {
-  extern int dpmi_mhp_modify_eip(int delta);
   if (IN_DPMI) dpmi_mhp_modify_eip(delta);
   else LWORD(eip) +=delta;
 }
@@ -1445,10 +1420,9 @@ void mhp_cmd(const char * cmd)
 
 static void mhp_print_ldt(int argc, char * argv[])
 {
-  char buffer[0x10000];
-  extern unsigned long *ldt_buffer;
-  unsigned long *lp, *lp_=ldt_buffer;
-  unsigned long base_addr, limit;
+  static char buffer[0x10000];
+  unsigned int *lp, *lp_=(unsigned int *)ldt_buffer;
+  unsigned int base_addr, limit;
   int type, type2, i;
   unsigned int seg;
   int lines=0, cache_mismatch;
@@ -1485,7 +1459,7 @@ static void mhp_print_ldt(int argc, char * argv[])
     else lines = 16;
   }
 
-  lp = (unsigned long *) buffer;
+  lp = (unsigned int *) buffer;
   lp += (seg & 0xFFF8) >> 2;
   lp_ += (seg & 0xFFF8) >> 2;
   
@@ -1504,7 +1478,7 @@ static void mhp_print_ldt(int argc, char * argv[])
         if (type2 & 0x08)
            limit *= 4096;
         if (type & 0x10)
-           mhp_printf ("%04x: %08lx %08lx %s%d %d %c%c%c%c%c%s\n",
+           mhp_printf ("%04x: %08x %08x %s%d %d %c%c%c%c%c%s\n",
               i | 0x07,
               base_addr,
               limit,
@@ -1518,7 +1492,7 @@ static void mhp_print_ldt(int argc, char * argv[])
               (lp_[1] &Abit)?'a':' ',
               cache_mismatch ?" (cache mismatch)":"");
         else
-           mhp_printf ("%04x: %08lx %08lx System(%02x)%s\n",
+           mhp_printf ("%04x: %08x %08x System(%02x)%s\n",
               i, base_addr, limit, type,
               cache_mismatch ?" (cache mismatch)":"");
     }
