@@ -1,7 +1,7 @@
 /*
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -11,7 +11,7 @@
  * routines for DOSEMU.
  *
  * If you have set up the joystick in Linux, everything should work like it
- * does in real DOS ie. you can have two joysticks (untested) or one joystick,
+ * does in real DOS i.e. you can have two joysticks (untested) or one joystick,
  * in which this code will do the appropriate (and correct) axis and button
  * mapping or of course, no joysticks, which is also supported :).
  *
@@ -23,7 +23,8 @@
  * Copyright (c) 2002 Clarence Dang <dang@kde.org>
  * (not to be confused with DANG, which is the Dosemu Alterer Novices Guide :)
  *
- * Last Modified: 20 Nov 2002 (Version 2)
+ *
+ * Last Modified: $Date: 2007-05-04 01:59:48 -0400 (Fri, 04 May 2007) $ (Version 2)
  */
 
 
@@ -346,15 +347,16 @@ void joy_init (void)
 	{
 		int axis;
 
+		joy_fd [joynum] = -1;
+
+		joy_numaxes [joynum] = 0;
+		joy_numbuttons [joynum] = 0;
+		
 		/* set to initial values */
 		for (axis = 0; axis < 4; axis++)
 			joy_axis [joynum][axis] = JOY_AXIS_INVALID;
 
-		joy_port_x [joynum] = joy_port_y [joynum] = 0;
-		joy_fd [joynum] = -1;
-
-		joy_numbuttons [joynum] = 0;
-		joy_numaxes [joynum] = 0;
+		joy_port_x [joynum] = joy_port_y [joynum] = -1;
 	}
 
 
@@ -364,13 +366,12 @@ void joy_init (void)
 	for (joynum = 0; joynum < 2; joynum++)
 	{
 		/* does the user even want this joystick? */
-		if (strlen (config.joy_device [joynum]) == 0)
+		if (!config.joy_device[joynum] || strlen (config.joy_device [joynum]) == 0)
 		{
 		#ifdef JOY_INIT_DEBUG
 			joy_init_printf ("joystick 0x%X: will not be initialised\n", joynum);
 		#endif
-			continue;	/* I wonder if a real PC would treat the 2nd joystick as
-								the 2nd joystick if the 1st failed to init... */
+			continue;
 		}
 
 		/* open joystick device */
@@ -539,6 +540,25 @@ void joy_init (void)
 	#endif	/* USE_PTHREADS */
 	}
 
+	/* no joysticks at all! */
+	if (joy_fd [JOY_0] < 0 && joy_fd [JOY_1] < 0)
+	{
+	#ifdef JOY_INIT_DEBUG
+		joy_init_printf ("WARNING! No joysticks enabled!\n");
+	#endif
+		joy_status = 0;
+		
+		/* 
+		 * We don't need to register the port - leave it for direct port
+		 * access (Sourceforge Support Request #1178900).  If direct port
+		 * access is not used, the DOSEMU port emulation will return the
+		 * correct value of 0xff (port_not_avail_inb()).
+		 */
+		return;
+	}
+	else
+		joy_status = 1;
+
 	/* handle joystick port/game card routines */
 	/* DANG_FIXTHIS does this code work for ports other than 0x201?
 	 */
@@ -554,31 +574,12 @@ void joy_init (void)
 	io_device.irq          = EMU_NO_IRQ;
 	io_device.fd           = -1;
 
-	/*
-	 * DANG_BEGIN_REMARK
-	 *
-	 * Register the joystick ports so that correct port values are returned
-	 * for programs that try to detect the joystick (or the lack of one).
-	 *
-	 * DANG_END_REMARK
-	 */
 	if (port_register_handler (io_device, 0) != 0)
 	{
 	#ifdef JOY_INIT_DEBUG
 		joy_init_printf ("ERROR! Couldn't register joystick port handler!\n");
 	#endif
 	}
-
-	/* no joysticks at all! */
-	if (joy_fd [JOY_0] < 0 && joy_fd [JOY_1] < 0)
-	{
-	#ifdef JOY_INIT_DEBUG
-		joy_init_printf ("WARNING! No joysticks enabled!\n");
-	#endif
-		joy_status = 0;
-	}
-	else
-		joy_status = 1;
 }
 
 void joy_uninit (void)
@@ -703,7 +704,7 @@ static inline int joy_latency_over (void)
  * Update the button status for each joystick.
  *
  * We must perform "button mapping" if only the first joystick is enabled
- * ie. we are required to map the "excessive" buttons (>2) from the first
+ * i.e. we are required to map the "excessive" buttons (>2) from the first
  * joystick onto the second:
  *
  *	a) 3rd button of 1st joy --> 1st button of 2nd joy
@@ -756,7 +757,7 @@ static inline void joy_emu_button_set (const int joynum, const int button, const
  * Update the axis status for each joystick.
  *
  * We must perform "axis mapping" if only the first joystick is enabled
- * ie. we are required to map the "excessive" axes (>2) from the first
+ * i.e. we are required to map the "excessive" axes (>2) from the first
  * joystick onto the second:
  *
  *	a) 3rd axis of 1st joy --> 2st axis of 2nd joy
@@ -766,9 +767,6 @@ static inline void joy_emu_button_set (const int joynum, const int button, const
  *
  * DANG_END_FUNCTION
  */
-
-/* DANG_FIXTHIS I've lost my joystick Y-cable (lets you connect two joysticks to one gameport) so the code to handle two joysticks is totally untested!
- */
 static inline void joy_emu_axis_set (const int joynum, const int axis, const int value)
 {
 	int anum = axis;
@@ -1195,7 +1193,7 @@ static int joy_linux_read_axis_new_threaded (const int joynum, const int axis,
 /*
  * DANG_BEGIN_FUNCTION joy_bios_read
  *
- * This is the int15 function 0x84 handler (ie. BIOS joystick emulation),
+ * This is the int15 function 0x84 handler (i.e. BIOS joystick emulation),
  * called from src/base/async/int.c.
  *
  * The real BIOS actually reads its values straight from the joystick port
