@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*  DANG_BEGIN_MODULE
@@ -134,7 +134,6 @@
 #include "bitops.h"
 #include "pic.h"
 #include "memory.h"
-/* #include <sys/vm86.h> */
 #include <sys/time.h>
 #include "cpu.h"
 #include "emu.h"
@@ -144,15 +143,10 @@
 #include "serial.h"
 #include "int.h"
 #include "ipx.h"
-#ifdef X86_EMULATOR
-#include "cpu-emu.h"
-#endif
 
 #undef us
 #define us unsigned
-void timer_tick(void);
-void pic_activate();
-extern void timer_int_engine(void);
+static void pic_activate(void);
 
 static unsigned long pic1_isr;         /* second isr for pic1 irqs */
 static unsigned long pic_irq2_ivec = 0;
@@ -172,6 +166,21 @@ static Bit16u cb_ip = 0;
  * occurs.
  */
 
+/* PIC "registers", plus a few more */
+
+static unsigned long pic_irr;          /* interrupt request register */
+static unsigned long pic_isr;          /* interrupt in-service register */
+static unsigned int pic_iflag;        /* interrupt enable flag: en-/dis- =0/0xfffe */
+static unsigned int pic_icount;       /* iret counter (to avoid filling stack) */
+static unsigned long pic_irqall = 0xfffe;       /* bits for all IRQs set. */
+
+static unsigned long pic0_imr = 0xf800;  /* interrupt mask register, pic0 */
+static unsigned long pic1_imr = 0x0670;         /* interrupt mask register, pic1 */
+static unsigned long pic_imr = 0xfff8;          /* interrupt mask register */
+static unsigned int pic_stack[32];     /* list of active irqd */
+static unsigned int pic_sp = 0;	       /* pointer to pic_stack */ 
+static unsigned int pic_vm86_count = 0;   /* count of times 'round the vm86 loop*/
+static unsigned int pic_dpmi_count = 0;   /* count of times 'round the dpmi loop*/
 static unsigned long pic1_mask = 0x07f8; /* bits set for pic1 levels */
 static unsigned long   pic_smm = 0;      /* 32=>special mask mode, 0 otherwise */
 
@@ -179,17 +188,8 @@ static unsigned long   pic_pirr;         /* pending requests: ->irr when icount=
 static unsigned long   pic_wirr;             /* watchdog timer for pic_pirr */
 static unsigned long   pic_wcount = 0;       /* watchdog for pic_icount  */
 unsigned long	pic_icount_od = 1;           /* overdrive for pic_icount_od */
-#if 0
-unsigned long   pic0_imr = 0xf800;  /* IRQs 3-7 on pic0 start out disabled */
-unsigned long   pic1_imr = 0x07f8;  /* IRQs 8-15 on pic1 start out disabled */
-unsigned long   pic_imr = 0xfff8;   /* interrupt mask register, enable irqs 0,1 */
-unsigned long   pice_imr = -1;      /* interrupt mask register, dos emulator */ 
-unsigned long   pic_stack[32];      /* stack of interrupt levels */
-unsigned long   pic_sp = 0;         /* pointer to stack */
-unsigned long   pic_vm86_count=0;   /* counter for trips around vm86 loop */
-unsigned long   pic_dpmi_count=0;   /* counter for trips around dpmi loop */
-         hitimer_t pic_sys_time=NEVER; /* system time for scheduling interrupts */
-#endif
+unsigned long	pic_irqs_active = 0;
+
 static   hitimer_t pic_ltime[33] =     /* timeof last pic request honored */
                 {NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
                  NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER, NEVER,
@@ -215,6 +215,8 @@ static struct lvldef pic_iinfo[32] = {
 {PNULL,PNULL,0x00}, {PNULL,PNULL,0x00}, {PNULL,PNULL,0x00}, {PNULL,PNULL,0x00}
 };
 
+static void do_irq(int ilevel);
+
 /*
  * run_irq()       checks for and runs any interrupts requested in pic_irr
  * do_irq()        runs the dos interrupt for the current irq
@@ -293,6 +295,7 @@ static unsigned char pic1_cmd;
 static void p_pic_print(char *s1, int v1, char *s2)
 {
 static int oldi=0, oldc=0, header_count=0;
+int pic_ilevel=find_bit(pic_isr);
 char ci,cc;
 
   if (pic_icount > oldc) cc='+';
@@ -309,10 +312,10 @@ char ci,cc;
   if(header_count>15) header_count=0;
   
   if(s2)
-  log_printf(1, "PIC: %c%2ld %c%2ld %08lx %08lx %08lx %s%02d%s\n",
+  log_printf(1, "PIC: %c%2d %c%2d %08lx %08lx %08lx %s%02d%s\n",
      cc, pic_icount, ci, pic_ilevel, pic_isr, pic_imr, pic_irr, s1, v1, s2);
   else
-  log_printf(1, "PIC: %c%2ld %c%2ld %08lx %08lx %08lx %s\n",
+  log_printf(1, "PIC: %c%2d %c%2d %08lx %08lx %08lx %s\n",
      cc, pic_icount, ci, pic_ilevel, pic_isr, pic_imr, pic_irr, s1);
 
 }
@@ -332,7 +335,7 @@ char ci,cc;
 
  * DANG_END_REMARK pic_print
  */
-inline void pic_push(int val)
+static inline void pic_push(int val)
 {
     if(pic_sp<32){
        pic_stack[pic_sp++]=val;
@@ -341,7 +344,7 @@ inline void pic_push(int val)
     }
 }
 
-inline int pic_pop()
+static inline int pic_pop(void)
 {
     if(pic_sp) {
        return pic_stack[--pic_sp];
@@ -352,7 +355,7 @@ inline int pic_pop()
 }
 
 
-void set_pic0_base(unsigned char int_num)
+static void set_pic0_base(unsigned char int_num)
 {
   unsigned char int_n;
   int_n        = int_num & 0xf8;         /* it's not worth doing a loop */
@@ -368,7 +371,7 @@ void set_pic0_base(unsigned char int_num)
 }
 
 
-void set_pic1_base(unsigned char int_num)
+static void set_pic1_base(unsigned char int_num)
 {
   unsigned char int_n;
   int_n        = int_num & 0xf8;         /* it's not worth doing a loop */
@@ -408,14 +411,18 @@ static char                icw_max_state;          /* number of icws expected
 int ilevel;			  /* level to reset on outb 0x20  */
 
 port -= 0x20;
-ilevel=pic_ilevel;
-if(pic_sp) {
-   if(ilevel!=pic_stack[pic_sp-1]) {
-      pic_print(2,"ilevel=",ilevel," differed from pic_stack");
-   }
-   ilevel=pic_stack[pic_sp-1];
+ilevel = 32;
+if (pic_isr)
+  ilevel=find_bit(pic_isr);
+if (ilevel != 32 && !test_bit(ilevel, &pic_irqall)) {
+  /* this is a fake IRQ, don't allow to reset its ISR bit */
+  pic_print(1, "Protecting ISR bit for lvl ", ilevel, " from spurious EOI");
+  ilevel = 32;
 }
 
+if (in_dpmi)
+  dpmi_return_request();	/* we have to leave the signal context */
+
 if(!port){                          /* icw1, ocw2, ocw3 */
   if(value&0x10){                   /* icw1 */
     icw_max_state = (value & 1) + 1;
@@ -430,15 +437,9 @@ if(!port){                          /* icw1, ocw2, ocw3 */
     pic0_cmd=3;
     }
   else if((value&0xb8) == 0x20) {    /* ocw2 */
-     if(!pic_sp) {
-	pic_print(2,"EOI with nothing on pic_stack! ilevel=",ilevel,"");
-     } else if(pic_stack[pic_sp-1]!=ilevel) {
-	pic_print(2,"Stack value differed from ilevel! stack=",pic_stack[pic_sp-1],"");
-     }
     /* irqs on pic1 require an outb20 to each pic. we settle for any 2 */
      if(!clear_bit(ilevel,&pic1_isr)) {
        clear_bit(ilevel,&pic_isr);  /* the famous outb20 */
-       pic_ilevel=pic_pop();  /* only pop stack when pic_isr gets cleared */
        pic_print(1,"EOI resetting bit ",ilevel, " on pic0");
        }
      else
@@ -468,14 +469,18 @@ static char /* icw_state, */     /* !=0 => port 1 does icw 2,3,(4) */
 int ilevel;			  /* level to reset on outb 0x20  */
 
 port -= 0xa0;
-ilevel=pic_ilevel;
-if(pic_sp) {
-   if(ilevel!=pic_stack[pic_sp-1]) {
-      pic_print(2,"ilevel=",ilevel," differed from pic_stack");
-   }
-   ilevel=pic_stack[pic_sp-1];
+ilevel = 32;
+if (pic_isr)
+  ilevel=find_bit(pic_isr);
+if (ilevel != 32 && !test_bit(ilevel, &pic_irqall)) {
+  /* this is a fake IRQ, don't allow to reset its ISR bit */
+  pic_print(1, "Protecting ISR bit for lvl ", ilevel, " from spurious EOI");
+  ilevel = 32;
 }
 
+if (in_dpmi)
+  dpmi_return_request();	/* we have to leave the signal context */
+
 if(!port){                            /* icw1, ocw2, ocw3 */
   if(value&0x10){                     /* icw1 */
     icw_max_state = (value & 1) + 1;
@@ -489,15 +494,9 @@ if(!port){                            /* icw1, ocw2, ocw3 */
     pic1_cmd=3;
     }
   else if((value&0xb8) == 0x20) {    /* ocw2 */
-     if(!pic_sp) {
-	pic_print(2,"EOI with nothing on pic_stack! ilevel=",ilevel,"");
-     } else if(pic_stack[pic_sp-1]!=ilevel) {
-	pic_print(2,"Stack value differed from ilevel! stack=",pic_stack[pic_sp-1],"");
-     }
     /* irqs on pic1 require an outb20 to each pic. we settle for any 2 */
      if(!clear_bit(ilevel,&pic1_isr)) {
        clear_bit(ilevel,&pic_isr);  /* the famous outb20 */
-       pic_ilevel=pic_pop();  /* only pop stack when pic_isr gets cleared */
        pic_print(1,"EOI resetting bit ",ilevel, " on pic0");
        }
      else
@@ -545,33 +544,6 @@ Bit8u read_pic1(ioport_t port)
                          return((unsigned char)get_pic1_irr());
 }
 
-
-/* DANG_BEGIN_FUNCTION pic_mask,pic_unmask
- *
- * The pic maintains an additional interrupt mask which is not visible
- * to the DOS process.  This is normally cleared (enabling an interrupt)
- * when an interrupt is initialized, but dosemu code may choose to
- * use this mask internally.  One possible use is to implement the interrupt
- * gate controlled by the OUT2 bit of the 16550A UART's Modem Control 
- * Register.  This mask is cleared by pic_unmaski() and set by pic_maski()
- *
- * DANG_END_FUNCTION
- */
-void pic_unmaski(int level)
-{
- pic_print(2,"Unmasking lvl= ",level,"");
- if(pic_iinfo[level].func != (void*)0) clear_bit(level,&pice_imr);
- pic_set_mask;
-}
-
-
-void pic_maski(int level)
-{
-  pic_print(2,"Masking lvl= ",level,"");
-  set_bit(level,&pice_imr);
-}
-
-
 /* DANG_BEGIN_FUNCTION pic_seti
  *
  * pic_seti is used to initialize an interrupt for dosemu.  It requires
@@ -591,17 +563,26 @@ void pic_maski(int level)
  *
  * DANG_END_FUNCTION
  */
-void pic_seti(unsigned int level, void (*func), unsigned int ivec,
-  void (*callback))
+void pic_seti(unsigned int level, int (*func)(int), unsigned int ivec,
+  void (*callback)(void))
 {
   if(level>=32) return;
+  if(pic_iinfo[level].func) {
+    if(pic_iinfo[level].func != func) {
+      error("Attempt to register more than one handler for IRQ level %i (%p %p)\n",
+        level, pic_iinfo[level].func, func);
+      config.exitearly = 1;
+    } else {
+      error("Handler for IRQ level %i was registered more than once! (%p)\n",
+        level, func);
+    }
+  }
   pic_iinfo[level].func = func;
   if(callback) {
     pic_iinfo[level].callback = callback;
     set_bit(level, &pic_irqall);
   }
   if(level>15) pic_iinfo[level].ivec = ivec;
-  if(func == (void*)0) pic_maski(level);
 }
 
 
@@ -648,16 +629,11 @@ void pic_seti(unsigned int level, void (*func), unsigned int ivec,
 void run_irqs(void)
 /* find the highest priority unmasked requested irq and run it */
 {
-       int old_ilevel;
        int int_request;
-       int priority;
 
        /* don't allow HW interrupts in force trace mode */
        if (vm86s.vm86plus.vm86dbg_TFpendig) return;
 
-       /* Old hack, to be removed */
-       if (in_dpmi && in_dpmi_timer_int) return;
-
        /* check for and find any requested irqs.  Having found one, we atomic-ly
         * clear it and verify it was there when we cleared it.  If it wasn't, we
         * look for the next request.  There are two in_service bits for pic1 irqs.
@@ -665,36 +641,33 @@ void run_irqs(void)
         * irq code actually runs, will reset the bits.  We also reset them here,
         * since dos code won't necessarily run.
         */
-
-       old_ilevel=pic_ilevel;                          /* save old pic_ilevl   */
-       priority = pic_smm + old_ilevel;                /* check spec. mask mode       */
-
        while((int_request = pic_irr & ~(pic_isr | pic_imr)) != 0) { /* while something to do*/
-               int local_pic_ilevel;
+               int local_pic_ilevel, old_ilevel, ret;
 
-	       if (!isset_IF() && !in_dpmi)
-	    	       goto exit;                      /* exit if ints are disabled */
+	       if (!isset_IF()) {
+		       set_VIP();
+	    	       return;                      /* exit if ints are disabled */
+	       }
 
-               do {
-                       local_pic_ilevel = find_bit(int_request);    /* find out what it is  */
-                       /* In case int_request has no bits set */
-                       if (local_pic_ilevel == -1)
-                               goto exit;
-                       if (local_pic_ilevel > priority)  /* priority check */
-                               goto exit;
-               } while (clear_bit(local_pic_ilevel, &pic_irr) == 0);   /* dbl check & clear req */
-               pic_ilevel = local_pic_ilevel;
-               set_bit(local_pic_ilevel, &pic_isr);     /* set in-service bit */
-               set_bit(local_pic_ilevel, &pic1_isr);    /* pic1 too */
-               pic1_isr &= pic_isr & pic1_mask;         /* isolate pic1 irqs */
-               pic_iinfo[local_pic_ilevel].func();      /* run the function */
-               local_pic_ilevel = pic_ilevel;
-               clear_bit(local_pic_ilevel, &pic_isr);   /* clear in_service bit */
-               clear_bit(local_pic_ilevel, &pic1_isr);  /* pic1 too */
+               local_pic_ilevel = find_bit(int_request);    /* find out what it is  */
+	       old_ilevel = find_bit(pic_isr);
+               /* In case int_request has no bits set */
+               if (local_pic_ilevel == -1)
+                       return;
+               if (local_pic_ilevel >= old_ilevel + pic_smm)  /* priority check */
+                       return;
+
+               if (pic_irqs_active && local_pic_ilevel >= find_bit(pic_irqs_active))
+                       return;
+
+               clear_bit(local_pic_ilevel, &pic_irr);
+	       /* pic_isr bit is set in do_irq() */
+               ret = (pic_iinfo[local_pic_ilevel].func ?
+	    	      pic_iinfo[local_pic_ilevel].func(local_pic_ilevel) : 1);      /* run the function */
+	       if (ret) {
+		       do_irq(local_pic_ilevel);
+	       }
        }
- exit:
-       /* whether we did or didn't :-( get one, we must still reset pic_ilevel */
-       pic_ilevel=old_ilevel;
 }
 
    
@@ -725,48 +698,32 @@ void run_irqs(void)
  *
  * DANG_END_FUNCTION
  */
-int do_irq()
+static void do_irq(int ilevel)
 {
- int intr;
- unsigned long ilevel;
- unsigned char * ssp;
- unsigned long sp;
-#ifdef X86_EMULATOR
- unsigned char *tmp_ssp;
- int tmp;
-#endif
+    int intr;
+
+    set_bit(ilevel, &pic_isr);     /* set in-service bit */
+    set_bit(ilevel, &pic1_isr);    /* pic1 too */
+    pic1_isr &= pic_isr & pic1_mask;         /* isolate pic1 irqs */
 
-    if(pic_ilevel==32) return 0;
-    ilevel=pic_ilevel;
     intr=pic_iinfo[ilevel].ivec;
 
     if(ilevel==PIC_IRQ9)      /* unvectored irq9 just calls int 0x0a.. */
       if(!IS_REDIRECTED(intr)) {intr=0x0a;pic1_isr&= 0xffef;} /* & one EOI */
-    {
-     if (test_bit(ilevel, &pic_irqall)) pic_push(ilevel);
+
+     if (test_bit(ilevel, &pic_irqall)) {
+       pic_push(ilevel);
+       set_bit(ilevel, &pic_irqs_active);
+       pic_icount++;
+     }
 
      if (!in_dpmi || in_dpmi_dos_int) {
-      ssp = (unsigned char *)(LWORD(ss)<<4);
-      sp = (unsigned long) LWORD(esp);
-
  /* save the real return address on the stack. Make iret return to
   * PIC_SEG:PIC_OFF so we can catch it.
-  * change esp first to protect the stack we're about to use
   */
-      LWORD(esp) = (LWORD(esp) - 4) & 0xffff;  
-#ifdef X86_EMULATOR
-      tmp_ssp = ssp+sp;
-      tmp = E_MUNPROT_STACK(tmp_ssp);
-#endif
-      pushw(ssp, sp, REG(cs));
-      pushw(ssp, sp, LWORD(eip));
-#ifdef X86_EMULATOR
-      if (tmp) E_MPROT_STACK(tmp_ssp);
-#endif
+      fake_call_to(PIC_SEG, PIC_OFF);
       if(debug_level('r')>7) r_printf("PIC: setting iret trap at %04x:%04lx\n",
         REG(cs), REG(eip));
-      REG(cs) = PIC_SEG;
-      LWORD(eip) = PIC_OFF;
      }
 
      if (pic_iinfo[ilevel].callback)
@@ -778,35 +735,6 @@ int do_irq()
          run_int(intr);
        }
      }
-
-      pic_icount++;
-      pic_wcount++;
-
- /* enter PIC loop - we can continue only when pic_isr has been cleared */
-      while(!fatalerr && test_bit(ilevel,&pic_isr))
-      {
-	if (debug_level('r')>2)
-		r_printf("------ PIC: intr loop ---%06lx--%06lx----\n",
-			pic_vm86_count,pic_dpmi_count);
-	if (in_dpmi ) {
-          ++pic_dpmi_count;
-	  pic_print(2, "Initiating DPMI irq lvl ", ilevel, " in do_irq");
-	  run_dpmi();
-	  }
-	else {
-	  ++pic_vm86_count;
-	  pic_print(2, "Initiating VM86 irq lvl ", ilevel, " in do_irq");
-          run_vm86();
-          }
-        pic_isr &= pic_irqall;    /*  levels 0 and 16-31 are Auto-EOI  */
-        serial_run();           /*  delete when moved to timer stuff */
-        pic_run();
-      }
-      pic_sti();
-      return(0);
-
-    }   /* else */ 
-    return(1);
 }
 
 /* DANG_BEGIN_FUNCTION pic_resched
@@ -822,11 +750,25 @@ int do_irq()
  * DANG_END_FUNCTION
  */
 void
-pic_resched() 
+pic_resched(void) 
 {
-unsigned long pic_newirr;
-    if(pic_icount)
+unsigned long pic_newirr, pic_last_ilevel;
+    if(pic_icount) {
        pic_icount--;
+       pic_last_ilevel = pic_pop();
+       if (pic_icount != pic_sp)
+         error("pic_icount=%i != pic_sp=%i\n", pic_icount, pic_sp);
+       if (!test_bit(pic_last_ilevel, &pic_irqs_active)) {
+         error("PIC: bit %li is not set, irqs_active=%lx\n",
+	   pic_last_ilevel, pic_irqs_active);
+	 pic_irqs_active = 0;
+       }
+       clear_bit(pic_last_ilevel, &pic_irqs_active);
+    } else if (pic_irqs_active) {
+      error("pic_icount=%i and pic_irqs_active=%lx are out of sync!\n",
+        pic_icount, pic_irqs_active);
+      pic_irqs_active = 0;
+    }
     if(pic_icount<=pic_icount_od) {
 	pic_newirr=pic_pirr&~pic_irr&~pic_isr;
         pic_irr|=pic_newirr;
@@ -834,6 +776,8 @@ unsigned long pic_newirr;
         pic_wirr&=~pic_newirr;        /* clear watchdog timer */
 	pic_activate();
     }
+    if (!pic_icount)
+      pic_wcount = 0;
 }
 
 /* DANG_BEGIN_FUNCTION pic_request
@@ -856,19 +800,7 @@ int pic_request(int inum)
 static char buf[81];
   int ret=PIC_REQ_NOP;
 
-  if (pic_iinfo[inum].func == (void *)0)
-    return ret; 
-
-  REG(eflags) |= VIP;
-  if (in_dpmi)
-    dpmi_eflags |= VIP;
-
-#if 1		/* use this result mouse slowdown in winos2 */
-  if (((pic_irr|pic_isr)&(1<<inum)) || (pic_icount>pic_icount_od))
-#else          /* this makes mouse work under winos2, but sometimes */
-	       /* results in internal stack overflow  */
-  if(pic_isr&(1<<inum) || pic_irr&(1<<inum))
-#endif
+  if ((pic_irr|pic_isr)&(1<<inum))
     {
     if (pic_pirr&(1<<inum)){
      ret=PIC_REQ_LOST;
@@ -909,7 +841,6 @@ void pic_untrigger(int inum)
     pic_pirr &= ~(1<<inum);
     pic_irr &= ~(1<<inum);
     pic_wirr &= ~(1<<inum);
-    pic_ltime[inum] = 0;
 }
 
 /* DANG_BEGIN_FUNCTION pic_iret
@@ -933,17 +864,12 @@ void pic_untrigger(int inum)
  * DANG_END_FUNCTION
  */
 void
-pic_iret()
+pic_iret_dpmi(void)
 {
-unsigned char * ssp;
-unsigned long sp;
-
-  if(in_dpmi && !in_dpmi_dos_int) {
 /* Even if cs:ip now points to PIC_SEG:PIC_OFF hlt, it means nothing while
  * in protected mode, so we can't modify it here.
  */
     pic_resched();
-    if(!pic_irr) dpmi_eflags &= ~VIP;
     pic_print(2,"IRET in dpmi, loops=",pic_dpmi_count," ");
     pic_dpmi_count=0;
     if ((cb_cs || cb_ip) && !pic_icount) {
@@ -952,20 +878,23 @@ unsigned long sp;
      /* we will have an extra pic_iret() after this, but no problems
       * since pic_icount==0
       */
-      fake_call(REG(cs), LWORD(eip));
-      LWORD(eip) = cb_ip;
-      REG(cs) = cb_cs;
+      fake_call_to(cb_cs, cb_ip);
       cb_cs = cb_ip = 0;
     }
-  }
-  else {
+}
+
+void
+pic_iret(void)
+{
 /* if we've really come from an irq, cs:ip will point to PIC_SEG:PIC_OFF */
 /* if we haven't come from an irq and cs:ip points as above, we're going to
  * die anyway, since our next instruction is a hlt.
- */ 
-    if(REG(cs) == PIC_SEG && LWORD(eip) == PIC_OFF) {
+ */
+      if (REG(cs) != PIC_SEG || LWORD(eip) != PIC_OFF)
+        error("pic_iret called from %#04x:%#04x\n", REG(cs), LWORD(eip));
+      if (!pic_icount)
+        r_printf("pic_iret called when pic_icount==0 !!\n");
       pic_resched();
-      if(!pic_irr) REG(eflags)&=~(VIP);
       pic_print(2,"IRET in vm86, loops=", in_dpmi?
       pic_dpmi_count : pic_vm86_count," ");
       pic_vm86_count=0;
@@ -977,14 +906,14 @@ unsigned long sp;
 	cb_cs = cb_ip = 0;
       }
       else {
-	ssp = (unsigned char *)(LWORD(ss)<<4);
+	unsigned char * ssp;
+	unsigned long sp;
+	ssp = SEG2LINEAR(LWORD(ss));
 	sp = (unsigned long) LWORD(esp);
 	LWORD(eip) = popw(ssp, sp);
 	REG(cs) = popw(ssp, sp);
 	LWORD(esp) = (LWORD(esp) + 4) & 0xffff;
       }
-    }
-  }
 }
 
  
@@ -1021,6 +950,15 @@ unsigned long pic_newirr;
   pic_print(2,"pic_sys_time set to ",pic_sys_time," ");
   pic_dos_time = pic_itime[32];
   if(pic_icount<=pic_icount_od) pic_activate();
+  if(config.pic_watchdog > 0) {
+    if(pic_icount && !pic_isr) {
+      if(++pic_wcount >= config.pic_watchdog) {
+        r_printf("PIC: force reschedule\n");
+        pic_resched();
+        pic_wcount = 0;
+      }
+    }
+  }
 }      
 
 
@@ -1041,10 +979,21 @@ unsigned long pic_newirr;
  * DANG_END_FUNCTION
  */
  
-int pic_pending(int ilevel)
+int pic_pending(void)
 {
-    return (pic_irr|pic_pirr|~pic_imr)&(1<<ilevel);
+    return (pic_irr & ~(pic_isr | pic_imr | pic_irqs_active));
 }
+
+int pic_irq_active(int num)
+{
+    return test_bit(num, &pic_isr);
+}
+
+int pic_irq_masked(int num)
+{
+    return test_bit(num, &pic_imr);
+}
+
 /* DANG_BEGIN_FUNCTION pic_activate
  * 
  * pic_activate requests any interrupts whose scheduled time has arrived.
@@ -1052,7 +1001,7 @@ int pic_pending(int ilevel)
  * pic_dos_time is advanced to the earliest time scheduled.
  * DANG_END_FUNCTION
  */
-void pic_activate()
+static void pic_activate(void)
 {
 hitimer_t earliest;
 int timer, count;
@@ -1072,6 +1021,7 @@ int timer, count;
    }
    if(count) pic_print(2,"Activated ",count, " interrupts.");
    pic_print(2,"Activate ++ dos time to ",earliest, " ");
+   pic_print(2,"pic_sys_time is ",pic_sys_time," ");
    /*if(!pic_icount)*/ pic_dos_time = pic_itime[32] = earliest;
 }
 
@@ -1113,9 +1063,6 @@ void pic_sched(int ilevel, int interval)
   *	[1u-15sec] range check?
   */
   if(interval > 0 && interval < 0x3fffffff) {
-     if (interval < 100) {
-       if (in_dpmi) run_dpmi();  /* we're surely out of int8 here */
-     }
      if(pic_ltime[ilevel]==NEVER) {
 	pic_itime[ilevel] = pic_itime[32] + interval;
      } else {
@@ -1132,16 +1079,35 @@ void pic_sched(int ilevel, int interval)
 
 void pic_set_callback(Bit16u cs, Bit16u ip)
 {
-  r_printf("PIC: setting callback to %x:%x (pic_icount=%ld)\n",
+  r_printf("PIC: setting callback to %x:%x (pic_icount=%u)\n",
     cs, ip, pic_icount);
   cb_ip = ip;
   cb_cs = cs;
 }
 
+void pic_sti(void)
+{
+  pic_iflag = 0;
+  pic_set_mask;
+}
+
+void pic_cli(void)
+{
+  pic_iflag = pic_irqall;
+  pic_set_mask;
+}
+
+int CAN_SLEEP(void)
+{
+  return (!(pic_icount || pic_isr || (REG(eflags) & VIP) || signal_pending ||
+    (pic_sys_time > pic_dos_time)));
+}
+
 void pic_init(void)
 {
   /* do any one-time initialization of the PIC */
   emu_iodev_t  io_device;
+  emu_hlt_t    hlt_hdlr;
 
   /* 8259 PIC (Programmable Interrupt Controller) */
   io_device.read_portb   = read_pic0;
@@ -1163,8 +1129,15 @@ void pic_init(void)
   io_device.read_portb   = read_pic1;
   io_device.write_portb  = write_pic1;
   port_register_handler(io_device, 0);
+
+  hlt_hdlr.name       = "PIC";
+  hlt_hdlr.start_addr = PIC_ADD - BIOS_HLT_BLK;
+  hlt_hdlr.end_addr = hlt_hdlr.start_addr;
+  hlt_hdlr.func       = (emu_hlt_func)pic_iret;
+  hlt_register_handler(hlt_hdlr);
 }
 
 void pic_reset(void)
 {
+  pic_set_mask;
 }
