@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* dos emulator, Matthias Lautner
@@ -10,11 +10,15 @@
  * floppy disks, dos partitions or their images (files) (maximum 8 heads)
  */
 
+#ifdef __linux__
+#define _LARGEFILE64_SOURCE 1
+#endif
+
 #include <sys/types.h>
 #include <unistd.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <malloc.h>
+#include <stdlib.h>
 #include <ctype.h>
 #include <string.h>
 #include <sys/ioctl.h>
@@ -27,26 +31,28 @@
 #include <sys/time.h>
 
 #include "emu.h"
+#include "int.h"
+#include "port.h"
+#include "bios.h"
 #include "disks.h"
 #include "timers.h"
 #include "utilities.h"
 #include "priv.h"
-#include "dosemu_select.h"
 #include "int.h"
 #include "fatfs.h"
 #include "utilities.h"
+#include "dos2linux.h"
+#include "redirect.h"
 #ifdef X86_EMULATOR
 #include "cpu-emu.h"
 #endif
 
-#define llseek libless_llseek
-
 static int disks_initiated = 0;
 
 #define FDISKS config.fdisks
 #define HDISKS config.hdisks
 
-inline void disk_close(void);
+static void set_part_ent(struct disk *dp, char *tmp_mbr);
 
 #define USE_FSYNC 1
 
@@ -133,7 +139,7 @@ int
 read_sectors(struct disk *dp, char *buffer, long head, long sector,
 	     long track, long count)
 {
-  loff_t  pos;
+  off64_t  pos;
   long already = 0;
   long tmpread;
 
@@ -141,10 +147,10 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
   pos = DISK_OFFSET(dp, head, sector, track) + dp->header;
   d_printf("DISK: %s: Trying to read %ld sectors at T/S/H %ld/%ld/%ld",
 	   dp->dev_name,count,track,sector,head);
-#ifdef __linux__
-  d_printf("%+Ld at pos %Ld\n", dp->header, pos);
+#if defined(__linux__) && defined(__i386__)
+  d_printf("%+lld at pos %lld\n", dp->header, pos);
 #else
-  d_printf(" at pos %ld\n", pos);
+  d_printf("%+ld at pos %ld\n", dp->header, pos);
 #endif
 
   /* reads beginning before that actual disk/file */
@@ -206,17 +212,13 @@ read_sectors(struct disk *dp, char *buffer, long head, long sector,
     tmpread *= SECTOR_SIZE;
   }
   else {
-#ifdef __linux__
-    if(pos != llseek(dp->fdesc, pos, SEEK_SET)) {
-#else
-    if(pos != lseek(dp->fdesc, pos, SEEK_SET)) {
-#endif
+    if(pos != lseek64(dp->fdesc, pos, SEEK_SET)) {
       error("Sector not found in read_sector, error = %s!\n", strerror(errno));
       return -DERR_NOTFOUND;
     }
 #ifdef X86_EMULATOR
     if (config.cpuemu>1)
-	tmpread = e_dos_read(dp->fdesc, buffer, count * SECTOR_SIZE - already);
+	tmpread = dos_read(dp->fdesc, buffer, count * SECTOR_SIZE - already);
     else
 #endif
     tmpread = RPT_SYSCALL(read(dp->fdesc, buffer, count * SECTOR_SIZE - already));
@@ -236,7 +238,7 @@ int
 write_sectors(struct disk *dp, char *buffer, long head, long sector,
 	      long track, long count)
 {
-  loff_t  pos;
+  off64_t  pos;
   long tmpwrite, already = 0;
 
   if (dp->rdonly) {
@@ -253,8 +255,8 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
   pos = DISK_OFFSET(dp, head, sector, track) + dp->header;
   d_printf("DISK: %s: Trying to write %ld sectors T/S/H %ld/%ld/%ld",
 	   dp->dev_name,count,track,sector,head);
-#ifdef __linux__
-  d_printf(" at pos %Ld\n", pos);
+#if defined(__linux__) && defined(__i386__)
+  d_printf(" at pos %lld\n", pos);
 #else
   d_printf(" at pos %ld\n", pos);
 #endif
@@ -295,11 +297,7 @@ write_sectors(struct disk *dp, char *buffer, long head, long sector,
     tmpwrite *= SECTOR_SIZE;
   }
   else {
-#ifdef __linux__
-    if(pos != llseek(dp->fdesc, pos, SEEK_SET)) {
-#else
-    if(pos != lseek(dp->fdesc, pos, SEEK_SET)) {
-#endif
+    if(pos != lseek64(dp->fdesc, pos, SEEK_SET)) {
       error("Sector not found in write_sector!\n");
       return -DERR_NOTFOUND;
     }
@@ -327,14 +325,14 @@ image_auto(struct disk *dp)
 
   if (dp->fdesc == -1) {
     warn("WARNING: image filedesc not open\n");
-    dp->fdesc = open(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
+    dp->fdesc = open64(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
     /* The next line should only be done in case the open succeeds, 
        but that should be the normal case, and allows somewhat better
        code for the if (how sick can you get, since the open is going to
        take a lot more time anyways :-) )
     */
     dp->rdonly = dp->wantrdonly;
-    dp->fdesc = open(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0);
+    dp->fdesc = open64(dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0);
     if (dp->fdesc == -1) {
       /* We should check whether errno is EROFS, but if not the next open will
          fail again and the following lseek will throw us out of dos. So we win
@@ -342,29 +340,28 @@ image_auto(struct disk *dp)
          this does work (should be impossible), we can at least try to 
          continue. (again how sick can you get :-) )
        */
-      dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
+      dp->fdesc = open64(dp->dev_name, O_RDONLY, 0);
       dp->rdonly = 1;
     }
   }
 
-  lseek(dp->fdesc, 0, SEEK_SET);/* Use lseek here, as nobody want
-				   > 2^^31 bytes for the image file */
+  lseek64(dp->fdesc, 0, SEEK_SET);
   if (RPT_SYSCALL(read(dp->fdesc, header, HEADER_SIZE)) != HEADER_SIZE) {
     error("could not read full header in image_init\n");
     leavedos(19);
   }
 
   if (strncmp(header, IMAGE_MAGIC, IMAGE_MAGIC_SIZE)
-		&& (*((long *)header) != DEXE_MAGIC) ) {
+		&& (*((uint32_t *)header) != DEXE_MAGIC) ) {
     error("IMAGE %s header lacks magic string - cannot autosense!\n",
 	  dp->dev_name);
     leavedos(20);
   }
 
-  dp->heads = *(long *) &header[7];
-  dp->sectors = *(long *) &header[11];
-  dp->tracks = *(long *) &header[15];
-  dp->header = *(long *) &header[19];
+  dp->heads = *(int *) &header[7];
+  dp->sectors = *(int *) &header[11];
+  dp->tracks = *(int *) &header[15];
+  dp->header = *(int *) &header[19];
 
   d_printf("IMAGE auto_info disk %s; h=%d, s=%d, t=%d, off=%ld\n",
 	   dp->dev_name, dp->heads, dp->sectors, dp->tracks,
@@ -395,15 +392,41 @@ hdisk_auto(struct disk *dp)
 
 void dir_auto(struct disk *dp)
 {
+  if (dp->floppy) {
+    switch (dp->default_cmos) {
+      case THREE_INCH_288MFLOP:
+        dp->heads = 2;
+        dp->tracks = 80;
+        dp->sectors = 5760/80/2;
+	break;
+      case THREE_INCH_FLOPPY:
+        dp->heads = 2;
+        dp->tracks = 80;
+        dp->sectors = 2880/80/2;
+	break;
+      case THREE_INCH_720KFLOP:
+        dp->heads = 2;
+        dp->tracks = 80;
+        dp->sectors = 1440/80/2;
+	break;
+      case FIVE_INCH_FLOPPY:
+        dp->heads = 2;
+        dp->tracks = 80;
+        dp->sectors = 2400/80/2;
+	break;
+    }
+    dp->start = 0;
+    dp->rdonly = 1;	// should be for HDD too, but...
+  } else {
   /*
    * We emulate an entire disk with 1 partition starting at t/h/s 0/1/1.
    * You are free to change the geometry (e.g. to change the partition size).
    */
-  dp->sectors = 63;
-  dp->heads = 255;
-  dp->tracks = 50;
-
-  dp->start = dp->sectors;
+    dp->sectors = 63;
+    dp->heads = 255;
+    dp->tracks = 50;
+    dp->start = dp->sectors;
+  }
 
   d_printf(
     "DIR auto_info disk %s; h=%d, s=%d, t=%d, start=%ld\n",
@@ -430,35 +453,39 @@ void dir_setup(struct disk *dp)
   pi->pre_secs = dp->sectors;
   pi->num_secs = (dp->tracks * dp->heads - 1) * dp->sectors;
 
-  dp->header = -(SECTOR_SIZE * (loff_t) (pi->pre_secs));
-
-  pi->mbr_size = SECTOR_SIZE;
-  pi->mbr = malloc(pi->mbr_size);
-  mbr = pi->mbr;
-
-  memset(mbr, 0, SECTOR_SIZE);
-  /*
-   * mov ax,0fffeh
-   * int 0e6h
-   */
-  mbr[0x00] = 0xb8;
-  mbr[0x01] = 0xfe;
-  mbr[0x02] = 0xff;
-  mbr[0x03] = 0xcd;
-  mbr[0x04] = 0xe6;
-  mbr[PART_INFO_START + 0x00] = PART_BOOT;
-  mbr[PART_INFO_START + 0x01] = pi->beg_head;
-  mbr[PART_INFO_START + 0x02] = pi->beg_sec + ((pi->beg_cyl & 0x300) >> 2);
-  mbr[PART_INFO_START + 0x03] = pi->beg_cyl;
-  mbr[PART_INFO_START + 0x04] = pi->num_secs < 1 << 16 ? 0x04 : 0x06;
-  mbr[PART_INFO_START + 0x05] = pi->end_head;
-  mbr[PART_INFO_START + 0x06] = pi->end_sec + ((pi->end_cyl & 0x300) >> 2);
-  mbr[PART_INFO_START + 0x07] = pi->end_cyl;
-  *(unsigned *) (mbr + PART_INFO_START + 0x08) = pi->pre_secs;
-  *(unsigned *) (mbr + PART_INFO_START + 0x0c) = pi->num_secs;
-  mbr[SECTOR_SIZE - 2] = 0x55;
-  mbr[SECTOR_SIZE - 1] = 0xaa;
+  if (dp->floppy) {
+    dp->header = 0;
+    pi->mbr_size = 0;
+    pi->mbr = NULL;
+  } else {
+    dp->header = -(SECTOR_SIZE * (off64_t) (pi->pre_secs));
+    pi->mbr_size = SECTOR_SIZE;
+    pi->mbr = malloc(pi->mbr_size);
+    mbr = pi->mbr;
 
+    memset(mbr, 0, SECTOR_SIZE);
+    /*
+     * mov ax,0fffeh
+     * int 0e6h
+     */
+    mbr[0x00] = 0xb8;
+    mbr[0x01] = 0xfe;
+    mbr[0x02] = 0xff;
+    mbr[0x03] = 0xcd;
+    mbr[0x04] = 0xe6;
+    mbr[PART_INFO_START + 0x00] = PART_BOOT;
+    mbr[PART_INFO_START + 0x01] = pi->beg_head;
+    mbr[PART_INFO_START + 0x02] = pi->beg_sec + ((pi->beg_cyl & 0x300) >> 2);
+    mbr[PART_INFO_START + 0x03] = pi->beg_cyl;
+    mbr[PART_INFO_START + 0x04] = pi->num_secs < 1 << 16 ? 0x04 : 0x06;
+    mbr[PART_INFO_START + 0x05] = pi->end_head;
+    mbr[PART_INFO_START + 0x06] = pi->end_sec + ((pi->end_cyl & 0x300) >> 2);
+    mbr[PART_INFO_START + 0x07] = pi->end_cyl;
+    *(unsigned *) (mbr + PART_INFO_START + 0x08) = pi->pre_secs;
+    *(unsigned *) (mbr + PART_INFO_START + 0x0c) = pi->num_secs;
+    mbr[SECTOR_SIZE - 2] = 0x55;
+    mbr[SECTOR_SIZE - 1] = 0xaa;
+  }
   d_printf("partition table entry for device %s is:\n", dp->dev_name);
   d_printf(
     "beg head %d, sec %d, cyl %d = end head %d, sec %d, cyl %d\n",
@@ -472,9 +499,6 @@ void dir_setup(struct disk *dp)
   );
 
   dp->fatfs = NULL;
-  fatfs_init(dp);
-
-  if(dp->fatfs && !redir_state) set_int21_revectored(redir_state = 1);
 }
 
 /* XXX - relies upon a file of SECTOR_SIZE in PARTITION_PATH that which
@@ -489,11 +513,9 @@ void dir_setup(struct disk *dp)
 void
 partition_setup(struct disk *dp)
 {
-  PRIV_SAVE_AREA
   int part_fd, i;
   unsigned char tmp_mbr[SECTOR_SIZE];
   char *hd_name;
-  void set_part_ent(struct disk *, char *);
 
 #define PART_BYTE(p,b)  *((unsigned char *)tmp_mbr + PART_INFO_START + \
 			  (PART_INFO_LEN * (p-1)) + b)
@@ -510,19 +532,19 @@ partition_setup(struct disk *dp)
   hd_name[8] = '\0';			/* i.e.  /dev/hda6 -> /dev/hda */
 #endif
 
-  enter_priv_on();
   part_fd = SILENT_DOS_SYSCALL(open(hd_name, O_RDONLY));
-  leave_priv_setting();
   if (part_fd == -1) {
-    if (dp->removeable) return;
-    error("opening device %s to read MBR for PARTITION %s\n",
-	  hd_name, dp->dev_name);
-    leavedos(22);
+    if (dp->floppy) return;
+    PNUM = 1;
+    set_part_ent(dp, tmp_mbr);
+    tmp_mbr[0x1fe] = 0x55;
+    tmp_mbr[0x1ff] = 0xaa;
+  } else {
+    RPT_SYSCALL(read(part_fd, tmp_mbr, SECTOR_SIZE));
+    close(part_fd);
+    d_printf("Using MBR from %s for PARTITION %s (part#=%d).\n",
+             hd_name, dp->dev_name, PNUM);
   }
-  RPT_SYSCALL(read(part_fd, tmp_mbr, SECTOR_SIZE));
-  close(part_fd);
-  d_printf("Using MBR from %s for PARTITION %s (part#=%d).\n",
-           hd_name, dp->dev_name, PNUM);
   free(hd_name);
 
   /* check for logical partition, if so simulate as primary part#1 */
@@ -553,7 +575,7 @@ partition_setup(struct disk *dp)
 			     dp->part_info.beg_cyl) +
 		 (SECTOR_SIZE * (dp->part_info.pre_secs - 1)));
 #else
-  dp->header = -(SECTOR_SIZE * (loff_t) (dp->part_info.pre_secs));
+  dp->header = -(SECTOR_SIZE * (off64_t) (dp->part_info.pre_secs));
 #endif
 
   dp->part_info.mbr_size = SECTOR_SIZE;
@@ -592,8 +614,7 @@ partition_setup(struct disk *dp)
  *       put in the dp->part_info.number'th entry in the part table.
  */
 
-void
-set_part_ent(struct disk *dp, char *tmp_mbr)
+static void set_part_ent(struct disk *dp, char *tmp_mbr)
 {
   long	length;		/* partition length in sectors		*/
   long	end;		/* last sector number offset		*/
@@ -619,18 +640,28 @@ set_part_ent(struct disk *dp, char *tmp_mbr)
 	HEAD(end), SECT(end), CYL(end));
 
   /* get address of where to put new part table entry */
-  p = tmp_mbr + PART_INFO_START + (PART_INFO_LEN * (dp->part_info.number-1)); 
+  p = tmp_mbr + PART_INFO_START + (PART_INFO_LEN * (dp->part_info.number-1));
 
+  /* initialize with LBA values */
+  p[1] = p[5] = 254;
+  p[2] = p[6] = 255;
+  p[3] = p[7] = 255;
+  p[4] = 0xe;
+    
   p[0] = PART_BOOT;						/* bootable  */
-  p[1] = HEAD(dp->start);					/* beg head  */
-  p[2] = SECT(dp->start) | ((CYL(dp->start) >> 2) & 0xC0);	/* beg sect  */
-  p[3] = CYL(dp->start) & 0xFF;					/* beg cyl   */
-  p[4] = (length < 32*1024*1024/SECTOR_SIZE)? 0x04 : 0x06;	/* dos sysid */
-  p[5] = HEAD(end);						/* end head  */
-  p[6] = SECT(end) | ((CYL(end) >> 2) & 0xC0);			/* end sect  */
-  p[7] = CYL(end) & 0xFF;					/* end cyl   */
-  *((long *)(p+8)) = dp->start;					/* pre sects */
-  *((long *)(p+12)) = length;					/* len sects */
+  if (CYL(dp->start) <= 1023) {
+    p[1] = HEAD(dp->start);					/* beg head  */
+    p[2] = SECT(dp->start) | ((CYL(dp->start) >> 2) & 0xC0);	/* beg sect  */
+    p[3] = CYL(dp->start) & 0xFF;				/* beg cyl   */
+  }
+  if (CYL(end) <= 1023) {
+    p[4] = (length < 32*1024*1024/SECTOR_SIZE)? 0x04 : 0x06;	/* dos sysid */
+    p[5] = HEAD(end);						/* end head  */
+    p[6] = SECT(end) | ((CYL(end) >> 2) & 0xC0);		/* end sect  */
+    p[7] = CYL(end) & 0xFF;					/* end cyl   */
+  }
+  *((uint32_t *)(p+8)) = dp->start;				/* pre sects */
+  *((uint32_t *)(p+12)) = length;				/* len sects */
 }
 
 void
@@ -657,20 +688,30 @@ disk_close(void)
   }
 }
 
+static void disk_sync(void)
+{
+  struct disk *dp;
+
+  if (!disks_initiated) return;  /* just to be safe */
+  for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
+    if (dp->removeable && dp->fdesc >= 0) {
+      d_printf("DISK: Syncing disk %s\n",dp->dev_name);
+      (void) fsync(dp->fdesc);
+    }
+  }
+}
+
 
 #ifdef __linux__
 void
 disk_open(struct disk *dp)
 {
-  PRIV_SAVE_AREA
   struct floppy_struct fl;
 
   if (dp == NULL || dp->fdesc >= 0)
     return;
     
-  enter_priv_on();
-  dp->fdesc = SILENT_DOS_SYSCALL(open(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
-  leave_priv_setting();
+  dp->fdesc = SILENT_DOS_SYSCALL(open64(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->wantrdonly ? O_RDONLY : O_RDWR, 0));
 
   /* FIXME:
    * Why the hell was the below handling restricted to non-removeable disks?
@@ -679,25 +720,16 @@ disk_open(struct disk *dp)
    */
   if ( /*!dp->removeable &&*/ (dp->fdesc < 0)) {
     if (errno == EROFS || errno == ENODEV) {
-      enter_priv_on();
-      dp->fdesc = DOS_SYSCALL(open(dp->dev_name, O_RDONLY, 0));
-      leave_priv_setting();
+      dp->fdesc = DOS_SYSCALL(open64(dp->dev_name, O_RDONLY, 0));
       if (dp->fdesc < 0) {
-        d_printf("ERROR: (disk) can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
+        error("ERROR: (disk) can't open %s for read nor write: %s\n", dp->dev_name, strerror(errno));
         /* In case we DO get more clever, we want to share that code */
-        goto fail;
       } else {
         dp->rdonly = 1;
         d_printf("(disk) can't open %s for read/write. Readonly used.\n", dp->dev_name);
       }
     } else {
       d_printf("ERROR: (disk) can't open %s: %s\n", dp->dev_name, strerror(errno));
-    fail:
-#if 0
-      /* We really should be more clever here */
-      fatalerr = 5;
-#endif
-      return;
     }
   }
   else dp->rdonly = dp->wantrdonly;
@@ -718,7 +750,7 @@ disk_open(struct disk *dp)
 	      fl.sect = *((unsigned char *)&ATAPI_buf0[0x18]);
 	      fl.head = *((unsigned char *)&ATAPI_buf0[0x1a]);
 	      tns = *((unsigned short *)&ATAPI_buf0[0x13]);
-	      if (tns==0) tns = *((unsigned long *)&ATAPI_buf0[0x20]);
+	      if (tns==0) tns = *((uint32_t *)&ATAPI_buf0[0x20]);
 	      fl.track = tns/(fl.sect*fl.head);
 	}
 	else {	/* no disk available */
@@ -743,7 +775,7 @@ disk_open(struct disk *dp)
       dp->tracks = 0;
       return;
     }
-    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, sys_errlist[errno]);
+    error("can't get floppy parameter of %s (%s)\n", dp->dev_name, strerror(errno));
     fatalerr = 5;
     return;
   }
@@ -788,6 +820,24 @@ disk_close_all(void)
   disks_initiated = 0;
 }
 
+static Bit8u floppy_DOR = 0xc;
+
+static Bit8u floppy_io_read(ioport_t port)
+{
+  if (port == 0x3f2)
+    return floppy_DOR;
+  return 0xff;
+}
+
+static void floppy_io_write(ioport_t port, Bit8u value)
+{
+  if (port == 0x3f2) {
+    floppy_DOR = value;
+    if ((value & 0x30) == 0)
+      disk_close();
+  }
+}
+
 /*
  * DANG_BEGIN_FUNCTION disk_init
  *
@@ -799,9 +849,9 @@ disk_close_all(void)
 void
 disk_init(void)
 {
-  PRIV_SAVE_AREA
-  struct disk *dp=NULL;
+  struct disk *dp;
   struct stat stbuf;
+  int i;
 
 #ifdef SILLY_GET_GEOMETRY
   int s;
@@ -811,39 +861,46 @@ disk_init(void)
 
   disks_initiated = 1;  /* disk_init has been called */
   if (config.bootdisk) {
-    enter_priv_on();
-    bootdisk.fdesc = open(bootdisk.dev_name,
-			  bootdisk.rdonly ? O_RDONLY : O_RDWR, 0);
-    leave_priv_setting();
+    bootdisk.fdesc = open64(bootdisk.type == DIR_TYPE ? "/dev/null" : bootdisk.dev_name,
+			    bootdisk.rdonly ? O_RDONLY : O_RDWR, 0);
     if (bootdisk.fdesc < 0) {
-      if (errno == EROFS) {
-        enter_priv_on();
-        bootdisk.fdesc = open(bootdisk.dev_name, O_RDONLY, 0);
-        leave_priv_setting();
+      if ((errno == EROFS || errno == EACCES) && bootdisk.type != DIR_TYPE) {
+        bootdisk.fdesc = open64(bootdisk.dev_name, O_RDONLY, 0);
         if (bootdisk.fdesc < 0) {
-          error("can't open bootdisk %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
-          leavedos(23);
+          error("can't open bootdisk %s for read nor write: %s\n", bootdisk.dev_name, strerror(errno));
+          config.exitearly = 1;
         } else {
           bootdisk.rdonly = 1;
           d_printf("(disk) can't open bootdisk %s for read/write. Readonly did work though\n", bootdisk.dev_name);
         }
       } else {
-        error("can't open bootdisk %s: %s\n", dp->dev_name, strerror(errno));
-        leavedos(23);
+        error("can't open bootdisk %s: %s\n", bootdisk.dev_name, strerror(errno));
+        config.exitearly = 1;
       }
     }
     else bootdisk.rdonly = bootdisk.wantrdonly;
     bootdisk.removeable = 0;
+    bootdisk.floppy = 1;
+    bootdisk.drive_num = 0;
+    bootdisk.serial = 0xB00B00B0;
+    if (bootdisk.type == DIR_TYPE) {
+      disk_fptrs[bootdisk.type].autosense(&bootdisk);
+      disk_fptrs[bootdisk.type].setup(&bootdisk);
+    }
   }
 
   /*
    * Open floppy disks
    */
   ATAPI_buf0[0] = 0;
-  for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
+  for (i = 0; i < FDISKS; i++) {
+    dp = &disktab[i];
+    dp->floppy = 1;
+    dp->drive_num = i;
+    dp->serial = 0xF10031A0 + dp->drive_num;	// sernum must be unique!
     if (stat(dp->dev_name, &stbuf) < 0) {
       error("can't stat %s\n", dp->dev_name);
-      leavedos(24);
+      config.exitearly = 1;
     }
     if (S_ISBLK(stbuf.st_mode))
       d_printf("ISBLK ");
@@ -860,17 +917,14 @@ disk_init(void)
       continue;
     }
 #endif
-    enter_priv_on();
-    dp->fdesc = open(dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
-    leave_priv_setting();
+    dp->fdesc = open64(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name,
+	    dp->rdonly ? O_RDONLY : O_RDWR, 0);
     if (dp->fdesc < 0) {
-      if (errno == EROFS || errno == EACCES) {
-        enter_priv_on();
-        dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
-        leave_priv_setting();
+      if ((errno == EROFS || errno == EACCES) && bootdisk.type != DIR_TYPE) {
+        dp->fdesc = open64(dp->dev_name, O_RDONLY, 0);
         if (dp->fdesc < 0) {
-          error("can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
-          leavedos(25);
+          error("can't open %s for read nor write: %s\n", dp->dev_name, strerror(errno));
+          config.exitearly = 1;
         } else {
           dp->rdonly = 1;
           d_printf("(disk) can't open %s for read/write. Readonly did work though\n", dp->dev_name);
@@ -878,10 +932,14 @@ disk_init(void)
       } else {
         error("can't open %s: %s\n", dp->dev_name, strerror(errno));
         /* 'leavedos' here hangs the machine */
-        leavedos(25);
+        config.exitearly = 1;
       }
     }
     else dp->rdonly = dp->wantrdonly;
+    if (dp->type == DIR_TYPE) {
+      disk_fptrs[dp->type].autosense(dp);
+      disk_fptrs[dp->type].setup(dp);
+    }
   }
 
   if (!FDISKS && use_bootdisk) {
@@ -890,38 +948,53 @@ disk_init(void)
     FDISKS++;	/* now we have one */
   }
 
+  if (FDISKS) {
+    emu_iodev_t  io_device;
+
+    io_device.read_portb   = floppy_io_read;
+    io_device.write_portb  = floppy_io_write;
+    io_device.read_portw   = NULL;
+    io_device.write_portw  = NULL;
+    io_device.read_portd   = NULL;
+    io_device.write_portd  = NULL;
+    io_device.handler_name = "Floppy Drive";
+    io_device.start_addr   = 0x03F0;
+    io_device.end_addr     = 0x03F7;
+    io_device.irq          = 6;
+    io_device.fd           = -1;
+    port_register_handler(io_device, 0);
+  }
+
   /*
    * Open hard disks
    */
-  for (dp = hdisktab; dp < &hdisktab[HDISKS]; dp++) {
+  for (i = 0; i < HDISKS; i++) {
+    dp = &hdisktab[i];
+    dp->floppy = 0;
+    dp->drive_num = i | 0x80;
+    dp->serial = 0x4ADD1B0A + dp->drive_num;	// sernum must be unique!
     if(dp->type == IMAGE)  {
 	if (dp->dexeflags & DISK_DEXE_RDWR) {
 	  d_printf("IMAGE: dexe, RDWR access allowed for %s\n",dp->dev_name);
-	  enter_priv_on();
 	}
 	else {
 	  d_printf("IMAGE: Using user permissions\n");
-	  enter_priv_off();
 	}
     }
-    else enter_priv_on();
-    dp->fdesc = open(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
-    leave_priv_setting();
+    dp->fdesc = open64(dp->type == DIR_TYPE ? "/dev/null" : dp->dev_name, dp->rdonly ? O_RDONLY : O_RDWR, 0);
     if (dp->fdesc < 0) {
       if (errno == EROFS || errno == EACCES) {
-        if (dp->type == IMAGE) enter_priv_off(); else enter_priv_on();
-        dp->fdesc = open(dp->dev_name, O_RDONLY, 0);
-        leave_priv_setting();
+        dp->fdesc = open64(dp->dev_name, O_RDONLY, 0);
         if (dp->fdesc < 0) {
-          error("can't open %s for read nor write: %s (you should never see this message)\n", dp->dev_name, strerror(errno));
-          leavedos(26);
+          error("can't open %s for read nor write: %s\n", dp->dev_name, strerror(errno));
+          config.exitearly = 1;
         } else {
           dp->rdonly = 1;
           d_printf("(disk) can't open %s for read/write. Readonly did work though\n", dp->dev_name);
         }
       } else {
         error("can't open %s: #%d - %s\n", dp->dev_name, errno, strerror(errno));
-        leavedos(26);
+        config.exitearly = 1;
       }
     }
     else dp->rdonly = dp->wantrdonly;
@@ -947,7 +1020,7 @@ disk_init(void)
 #ifdef SILLY_GET_GEOMETRY
     if (RPT_SYSCALL(read(dp->fdesc, buf, 512)) != 512) {
       error("can't read disk info of %s\n", dp->dev_name);
-      leavedos(27);
+      config.exitearly = 1;
     }
 
     dp->sectors = *(us *) & buf[0x18];	/* sectors per track */
@@ -959,7 +1032,7 @@ disk_init(void)
      */
     s = *(us *) & buf[19];
     if (!s) {
-      s = *(unsigned long *) &buf[32];
+      s = *(uint32_t *) &buf[32];
       d_printf("DISK: zero # secs, so DOS 4 # secs = %d\n", s);
     }
     s += *(us *) & buf[28];	/* + hidden sectors */
@@ -974,7 +1047,7 @@ disk_init(void)
     memcpy(label, &buf[0x2b], 11);
     label[11] = 0;
     g_printf("VOLUME serial #: 0x%08x, LABEL: %s\n",
-	     *(unsigned long *) &buf[39], label);
+	     *(unsigned int *) &buf[39], label);
 
     if (s % (dp->sectors * dp->heads) != 0) {
       error("incorrect track number of %s\n", dp->dev_name);
@@ -984,8 +1057,34 @@ disk_init(void)
   }
 }
 
-int
-checkdp(struct disk *disk)
+void disk_reset(void)
+{
+  struct disk *dp;
+  int i;
+
+  subst_file_ext(NULL);
+  for (i = 0; i < 26; i++)
+    ResetRedirection(i);
+  set_int21_revectored(redir_state = 1);
+  if (config.bootdisk && bootdisk.type == DIR_TYPE) {
+    if (bootdisk.fatfs) fatfs_done(&bootdisk);
+    fatfs_init(&bootdisk);
+  }
+  for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
+    if(dp->type == DIR_TYPE) {
+      if (dp->fatfs) fatfs_done(dp);
+      fatfs_init(dp);
+    }
+  }
+  for (dp = hdisktab; dp < &hdisktab[HDISKS]; dp++) {
+    if(dp->type == DIR_TYPE) {
+      if (dp->fatfs) fatfs_done(dp);
+      fatfs_init(dp);
+    }
+  }
+}
+
+static int checkdp(struct disk *disk)
 {
   if (disk == NULL) {
     d_printf("DISK: null dp\n");
@@ -999,12 +1098,11 @@ checkdp(struct disk *disk)
     return 0;
 }
 
-void
-int13(u_char i)
+int int13(void)
 {
   unsigned int disk, head, sect, track, number;
   int res;
-  loff_t  pos;
+  off64_t  pos;
   char *buffer;
   struct disk *dp;
   int checkdp_val;
@@ -1015,12 +1113,11 @@ int13(u_char i)
   else if (disk < FDISKS) {
       dp = &disktab[disk];
     switch (HI(ax)) {
-      #define DISKETTE_MOTOR_TIMEOUT (*((unsigned char *)0x440))
-      /* NOTE: we don't need this counter, but older games seem to rely
-       * on it. We count it down in INT08 (bios.S) --SW, --Hans
+      /* NOTE: we use this counter for closing. Also older games seem to rely
+       * on it. We count it down in INT08 (bios.S) --SW, --Hans, --Bart
        */
-      case 0: case 2: case 3: case 5: case 10: case 11:
-        DISKETTE_MOTOR_TIMEOUT = 3*18;  /* set timout to 3 seconds */
+      case 0: case 2: case 3: case 5: case 10: case 11: case 0x42: case 0x43:
+        WRITE_BYTE(BIOS_MOTOR_TIMEOUT, 37);  /* set timout to 2 seconds */
         break;
     }
   }
@@ -1029,7 +1126,7 @@ int13(u_char i)
   else
     dp = NULL;
 
-  d_printf("INT13: disk=%#x ah=%#x dp=%p\n", LO(dx), HI(ax), dp);
+  d_printf("INT13: ax=%04x cx=%04x dx=%04x\n", LWORD(eax), LWORD(ecx), LWORD(edx));
 
   /* this is a bad hack to ensure that the cached blocks aren't.
    * Linux only checks disk change on open()
@@ -1063,10 +1160,8 @@ int13(u_char i)
       track |= (HI(dx) & 0xc0) << 4;
     buffer = SEG_ADR((char *), es, bx);
     number = LO(ax);
-#ifdef X86_EMULATOR
-    e_printf("DISK %d read [h:%d,s:%d,t:%d](%d)->%04x:%04x\n",
+    d_printf("DISK %d read [h:%d,s:%d,t:%d](%d)->%04x:%04x\n",
 	     disk, head, sect, track, number, REG(es), LWORD(ebx));
-#endif
 
     if (checkdp_val || head >= dp->heads ||
 	sect >= dp->sectors || track >= dp->tracks) {
@@ -1136,7 +1231,7 @@ int13(u_char i)
     if (dp->rdonly) {
       W_printf("write protect!\n");
       show_regs(__FILE__, __LINE__);
-      if (dp->removeable)
+      if (dp->floppy)
 	HI(ax) = DERR_WP;
       else
 	HI(ax) = DERR_WRITEFLT;
@@ -1187,19 +1282,11 @@ int13(u_char i)
 		  dp->heads, dp->sectors, dp->tracks);
       break;
     }
-#ifdef __linux__
-    pos = (long long ) ((track * dp->heads + head) * dp->sectors + sect) << 9;
-#else
-    pos = (long) ((track * dp->heads + head) * dp->sectors + sect) << 9;
-#endif
+    pos = (off64_t) ((track * dp->heads + head) * dp->sectors + sect) << 9;
     /* XXX - header hack */
     pos += dp->header;
 
-#ifdef __linux__
-    if (pos != llseek(dp->fdesc, pos, 0)) {
-#else
-    if (pos != lseek(dp->fdesc, pos, 0)) {
-#endif
+    if (pos != lseek64(dp->fdesc, pos, 0)) {
       HI(ax) = DERR_NOTFOUND;
       REG(eflags) |= CF;
       error("test: sector not found 6\n");
@@ -1233,6 +1320,9 @@ int13(u_char i)
       case 18:
 	LO(bx) = THREE_INCH_FLOPPY;
 	break;
+      case 36:
+	LO(bx) = THREE_INCH_288MFLOP;
+	break;
       case 0:
 	LO(bx) = dp->default_cmos;
 	dp->tracks = 80;
@@ -1326,7 +1416,7 @@ int13(u_char i)
   case 0x15:			/* Get type */
     d_printf("disk gettype %#x\n", disk);
     if (dp != NULL && disk >= 0x80) {
-      if (dp->removeable) {
+      if (dp->floppy) {
 	HI(ax) = 1;		/* floppy disk, change detect (1=no, 2=yes) */
 	d_printf("disk gettype: floppy\n");
 	LWORD(edx) = 0;
@@ -1415,6 +1505,7 @@ int13(u_char i)
     buffer = (char *)(((unsigned long)diskaddr->buf_seg << 4)+diskaddr->buf_ofs);
     number = diskaddr->blocks;
     diskaddr->blocks = 0;
+    d_printf("DISK read [h:%d,s:%d,t:%d](%d)\n", head, sect, track, number);
 
     if (checkdp(dp) || track >= dp->tracks) {
       error("Sector not found, AH=0x42!\n");
@@ -1488,7 +1579,7 @@ int13(u_char i)
     if (dp->rdonly) {
       d_printf("DISK is write protected!\n");
       show_regs(__FILE__, __LINE__);
-      if (dp->removeable)
+      if (dp->floppy)
 	HI(ax) = DERR_WP;
       else
 	HI(ax) = DERR_WRITEFLT;
@@ -1554,7 +1645,7 @@ int13(u_char i)
     }
 
     params->flags = IMEXT_INFOFLAG_CHSVALID;
-    if (dp->removeable)
+    if (dp->floppy)
       params->flags |= IMEXT_INFOFLAG_REMOVABLE;
     params->tracks = dp->tracks;
     params->heads = dp->heads;
@@ -1589,24 +1680,57 @@ int13(u_char i)
 	  LWORD(eax));
     show_regs(__FILE__, __LINE__);
     CARRY;
-    return;
+    break;
   }
+  return 1;
 }
 
-#define FLUSH_DELAY 2
-
-/* flush disks every FLUSH_DELAY seconds
- * XXX - make this configurable later
- */
+/* flush disks every config.fastfloppy ticks */
 void
 floppy_tick(void)
 {
   static int ticks = 0;
 
   if (++ticks >= config.fastfloppy) {
-    disk_close();
+    disk_sync();
     if (debug_level('d') > 2)
       d_printf("FLOPPY: flushing after %d ticks\n", ticks);
     ticks = 0;
   }
 }
+
+fatfs_t *get_fat_fs_by_serial(unsigned long serial)
+{
+  struct disk *dp;
+  if (bootdisk.type == DIR_TYPE && bootdisk.fatfs && bootdisk.serial == serial)
+    return bootdisk.fatfs;
+  for (dp = disktab; dp < &disktab[FDISKS]; dp++) {
+    if(dp->type == DIR_TYPE && dp->fatfs && dp->serial == serial)
+      return dp->fatfs;
+  }
+  for (dp = hdisktab; dp < &hdisktab[HDISKS]; dp++) {
+    if(dp->type == DIR_TYPE && dp->fatfs && dp->serial == serial)
+      return dp->fatfs;
+  }
+  return NULL;
+}
+
+fatfs_t *get_fat_fs_by_drive(unsigned char drv_num)
+{
+  struct disk *dp = NULL;
+  int num = drv_num & 0x7f;
+  if (!drv_num && config.bootdisk)
+    dp = &bootdisk;
+  else if (drv_num & 0x80) {
+    if (num >= HDISKS)
+      return NULL;
+    dp = &hdisktab[num];
+  } else {
+    if (num >= FDISKS)
+      return NULL;
+    dp = &disktab[num];
+  }
+  if (dp->type == DIR_TYPE)
+    return dp->fatfs;
+  return NULL;
+}
