@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 #include <stdio.h>
@@ -38,11 +38,11 @@ char *emu_disasm(int sga, unsigned int ip)
      refseg = REG(cs);
    }
    else {
-     cp = (unsigned char *)ip;
+     cp = (unsigned char *)(uintptr_t)ip;
      refseg = 0;	/* ??? */
    }
 
-   rc = dis_8086((unsigned long)cp, cp, frmtbuf, 0, &refseg, &ref, 0, 1);
+   rc = dis_8086(cp, frmtbuf, 0, &ref, refseg * 16);
 
    p = buf;
    for (i=0; i<rc && i<8; i++) {
@@ -126,7 +126,7 @@ show_regs(char *file, int line)
   /* display the 10 bytes before and after CS:EIP.  the -> points
    * to the byte at address CS:EIP
    */
-  if (((unsigned) sp < 0xa0000) && ((unsigned)sp >10)) {
+  if ((sp < (unsigned char *)0xa0000) && (sp > (unsigned char *)10)) {
 	  g_printf("STACK: ");
 	  sp -= 10;
 	  for (i = 0; i < 10; i++)
@@ -136,7 +136,7 @@ show_regs(char *file, int line)
 		  g_printf("%02x ", *sp++);
 	  g_printf("\n");
   }
-  if (((unsigned) cp < 0xa0000) && ((unsigned)cp>10)) {
+  if ((cp < (unsigned char *)0xa0000) && (cp>(unsigned char *)10)) {
 	  g_printf("OPS  : ");
 	  cp -= 10;
 	  for (i = 0; i < 10; i++)
@@ -164,29 +164,31 @@ show_ints(int min, int max)
   }
 }
 
+#if MAX_SELECTORS != 8192
+#error MAX_SELECTORS needs to be 8192
+#endif
 
-#define GetSegmentBaseAddress(s)	(((s) >= (MAX_SELECTORS << 3))? 0 :\
-					Segments[(s) >> 3].base_addr)
-#define IsSegment32(s)			(((s) >= (MAX_SELECTORS << 3))? 0 :\
-					Segments[(s) >> 3].is_32)
+#define GetSegmentBaseAddress(s)	((unsigned long)Segments[(s) >> 3].base_addr)
+#define IsSegment32(s)			(Segments[(s) >> 3].is_32)
 
-void DPMI_show_state(struct sigcontext_struct *scp)
+char *DPMI_show_state(struct sigcontext_struct *scp)
 {
+    static char buf[4096];
     unsigned char *csp2, *ssp2;
-    D_printf("eip: 0x%08lx  esp: 0x%08lx  eflags: 0x%08lx\n"
-	     "\ttrapno: 0x%02lx  errorcode: 0x%08lx  cr2: 0x%08lx\n"
+    sprintf(buf, "eip: 0x%08x  esp: 0x%08x  eflags: 0x%08lx\n"
+	     "\ttrapno: 0x%02x  errorcode: 0x%08lx  cr2: 0x%08lx\n"
 	     "\tcs: 0x%04x  ds: 0x%04x  es: 0x%04x  ss: 0x%04x  fs: 0x%04x  gs: 0x%04x\n",
 	     _eip, _esp, _eflags, _trapno, _err, _cr2, _cs, _ds, _es, _ss, _fs, _gs);
-    D_printf("EAX: %08lx  EBX: %08lx  ECX: %08lx  EDX: %08lx\n",
+    sprintf(buf, "%sEAX: %08x  EBX: %08x  ECX: %08x  EDX: %08x\n", buf,
 	     _eax, _ebx, _ecx, _edx);
-    D_printf("ESI: %08lx  EDI: %08lx  EBP: %08lx\n",
+    sprintf(buf, "%sESI: %08x  EDI: %08x  EBP: %08x\n", buf,
 	     _esi, _edi, _ebp);
     /* display the 10 bytes before and after CS:EIP.  the -> points
      * to the byte at address CS:EIP
      */
     if (!((_cs) & 0x0004)) {
       /* GTD */
-      csp2 = (unsigned char *) _eip - 10;
+      csp2 = (unsigned char *) _rip - 10;
     }
     else {
       /* LDT */
@@ -197,21 +199,27 @@ void DPMI_show_state(struct sigcontext_struct *scp)
      * area, we fall into another fault which likely terminates dosemu.
      */
 #ifdef X86_EMULATOR
-    if (!config.cpuemu || (_trapno!=0x0b && _trapno!=0x0c)) {
-#else
-    if (1) {
+    if (!config.cpuemu || (_trapno!=0x0b && _trapno!=0x0c))
 #endif
+    {
       int i;
-      D_printf("OPS  : ");
-      for (i = 0; i < 10; i++)
-        D_printf("%02x ", *csp2++);
-      D_printf("-> ");
-      for (i = 0; i < 10; i++)
-        D_printf("%02x ", *csp2++);
-      D_printf("\n");
+      sprintf(buf, "%sOPS  : ", buf);
+      if (!(_cs & 0x0004) ||
+	  (uintptr_t)csp2 < 0x110000 ||
+	  ((uintptr_t)csp2 > config.dpmi_base &&
+	   (uintptr_t)csp2 < config.dpmi_base + config.dpmi * 1024)) {
+	for (i = 0; i < 10; i++)
+	  sprintf(buf, "%s%02x ", buf, *csp2++);
+	sprintf(buf, "%s-> ", buf);
+	for (i = 0; i < 10; i++)
+	  sprintf(buf, "%s%02x ", buf, *csp2++);
+	sprintf(buf, "%s\n", buf);
+      } else {
+	sprintf(buf, "%sCS:EIP points to invalid memory\n", buf);
+      }
       if (!((_ss) & 0x0004)) {
         /* GDT */
-        ssp2 = (unsigned char *) _esp - 10;
+        ssp2 = (unsigned char *) _rsp - 10;
       }
       else {
         /* LDT */
@@ -220,14 +228,23 @@ void DPMI_show_state(struct sigcontext_struct *scp)
         else
 	  ssp2 = (unsigned char *) (GetSegmentBaseAddress(_ss) + _LWORD(esp) ) - 10;
       }
-      D_printf("STACK: ");
-      for (i = 0; i < 10; i++)
-        D_printf("%02x ", *ssp2++);
-      D_printf("-> ");
-      for (i = 0; i < 10; i++)
-        D_printf("%02x ", *ssp2++);
-      D_printf("\n");
+      sprintf(buf, "%sSTACK: ", buf);
+      if (!(_ss & 0x0004) ||
+	  (uintptr_t)ssp2 < 0x110000 ||
+	  ((uintptr_t)ssp2 > config.dpmi_base &&
+	   (uintptr_t)ssp2 < config.dpmi_base + config.dpmi * 1024)) {
+	for (i = 0; i < 10; i++)
+	  sprintf(buf, "%s%02x ", buf, *ssp2++);
+	sprintf(buf, "%s-> ", buf);
+	for (i = 0; i < 10; i++)
+	  sprintf(buf, "%s%02x ", buf, *ssp2++);
+	sprintf(buf, "%s\n", buf);
+      } else {
+	sprintf(buf, "%sSS:ESP points to invalid memory\n", buf);
+      }
     }
+
+    return buf;
 }
 
 /* @@@ MOVE_END @@@ 32768 */
