@@ -1,8 +1,8 @@
-/* 
+/*
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -17,6 +17,8 @@
  *
  * email: Steffen Winterfeldt <wfeldt@suse.de>
  *
+ * FAT12 support by Stas Sergeev <stsp@users.sourceforge.net>
+ *
  */
 
 
@@ -25,7 +27,7 @@
  * Debug level.
  * 0 - normal / 1 - useful / 2 - too much
  */
-#define DEBUG_FATFS	1
+#define DEBUG_FATFS	2
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -45,6 +47,7 @@
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
+#include "emu.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -55,11 +58,14 @@
 #include <fcntl.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <errno.h>
 
-#include "emu.h"
 #include "disks.h"
 #include "fatfs.h"
 #include "doshelpers.h"
+#include "cpu-emu.h"
+#include "dos2linux.h"
+#include "utilities.h"
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
@@ -82,14 +88,12 @@ static int read_file(fatfs_t *, unsigned, unsigned, unsigned);
 static int read_dir(fatfs_t *, unsigned, unsigned, unsigned);
 static unsigned next_cluster(fatfs_t *, unsigned);
 static void build_boot_blk(fatfs_t *);
-static void make_i1342_blk(unsigned char *, unsigned, unsigned, unsigned, unsigned);
-
-static char *bootfile = 0;
+static void make_i1342_blk(struct ibm_ms_diskaddr_pkt *b, unsigned start, unsigned blks, unsigned seg, unsigned ofs);
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 void fatfs_init(struct disk *dp)
 {
-  fatfs_t *f;
+  fatfs_t *f0, *f;
   unsigned u;
 
   if(dp->fatfs) fatfs_done(dp);
@@ -115,19 +119,38 @@ void fatfs_init(struct disk *dp)
   f->ffn_obj = 1;			/* this object doesn't exist */
 
   f->dir = dp->dev_name;
-  f->fat_id = 0xf8;
-  f->serial = 0x12345678;
+  if (dp->floppy) {
+    switch (dp->default_cmos) {
+      case THREE_INCH_288MFLOP:
+      case THREE_INCH_FLOPPY:
+        f->fat_id = 0xf0;
+	break;
+      case FIVE_INCH_FLOPPY:
+      case THREE_INCH_720KFLOP:
+        f->fat_id = 0xf9;
+	break;
+    }
+    f->fat_type = FAT_TYPE_FAT12;
+  } else {
+    f->fat_id = 0xf8;
+    f->fat_type = FAT_TYPE_FAT16;
+  }
+  f->serial = dp->serial;
   f->secs_track = dp->sectors;
+  f->bytes_per_sect = SECTOR_SIZE;
   f->heads = dp->heads;
   f->reserved_secs = 1;
   f->hidden_secs = dp->start;
-  f->total_secs = dp->part_info.num_secs;
+  f->total_secs = dp->floppy ? dp->tracks * dp->heads * dp->sectors :
+      dp->part_info.num_secs;
   f->fats = 2;
   for(u = 1; u <= 64; u <<= 1) {
     if(u * 0xfff0u > f->total_secs) break;
   }
   f->cluster_secs = u;
-  f->fat_secs = ((f->total_secs / u + 2) * 2 + 0x1ff) >> 9;
+  f->fat_secs = f->fat_type == FAT_TYPE_FAT12 ? 
+	((f->total_secs / u + 2) * 3 + 0x3ff) >> 10 :
+	((f->total_secs / u + 2) * 2 + 0x1ff) >> 9;
   f->root_secs = 4;
 
   f->root_entries = f->root_secs << 4;
@@ -135,6 +158,8 @@ void fatfs_init(struct disk *dp)
   f->last_cluster = (f->total_secs - f->reserved_secs - f->fats * f->fat_secs
                     - f->root_secs) / f->cluster_secs + 1;
 
+  f->drive_num = dp->drive_num;
+
   f->obj = NULL;
   f->objs = f->alloc_objs = 0;
 
@@ -158,6 +183,7 @@ void fatfs_init(struct disk *dp)
 
   f->obj[0].name = f->dir;
   f->obj[0].is.dir = 1;
+  f0 = hdisktab[0].fatfs;
   scan_dir(f, 0);	/* set # of root entries accordingly ??? */
   if(f->boot_sec == NULL) {
     build_boot_blk(f);
@@ -199,7 +225,8 @@ int fatfs_read(fatfs_t *f, unsigned char *buf, unsigned pos, int len)
 
   while(l) {
     if((i = read_sec(f, pos))) return i;
-    memcpy(buf, f->sec, 0x200);
+    MEMCPY_2DOS(buf, f->sec, 0x200);
+    e_invalidate(buf, 0x200);
     buf += 0x200; pos++; l--;
   }
 
@@ -252,20 +279,53 @@ int read_sec(fatfs_t *f, unsigned pos)
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 int read_fat(fatfs_t *f, unsigned pos)
 {
-  unsigned u, u0 = pos << 8, u1;
+  unsigned epfs, u, u0, u1 = 0, i = 0, nbit = 0, lnb = 0, boffs, bioffs, wb;
 
   fatfs_deb("dir %s, reading fat sector %d\n", f->dir, pos);
 
-  if(f->got_all_objs && u0 >= f->first_free_cluster) {
-    memset(f->sec, 0, 0x200);
-    return 0;
-  }
+  memset(f->sec, 0, 0x200);
 
-  for(u = 0; u < 0x100; u++) {
+  if (f->fat_type == FAT_TYPE_FAT12) {
+    epfs = f->bytes_per_sect * 2 / 3;	// 341
+    boffs = ((f->bytes_per_sect * 2) % 3) * 4;	// 4
+  } else {
+    epfs = f->bytes_per_sect / 2;
+    boffs = 0;
+  }
+  u0 = pos * epfs + ((pos * boffs) / 12);
+  bioffs = (pos * boffs) % 12;
+  if(f->got_all_objs && u0 >= f->first_free_cluster)
+    return 0;
+
+  for(u = 0;; u++) {
     u1 = next_cluster(f, u + u0);
     fatfs_deb2("cluster %u follows %u\n", u1, u + u0);
-    f->sec[2 * u] = u1;
-    f->sec[2 * u + 1] = u1 >> 8;
+    if (f->fat_type == FAT_TYPE_FAT12) {
+      u1 &= 0xfff;
+      lnb = 12;
+    } else {
+      lnb = 16;
+    }
+    if (bioffs) {
+      fatfs_deb2("... offset %u bits\n", bioffs);
+      u1 >>= bioffs;
+      lnb -= bioffs;
+      bioffs = 0;
+    }
+    f->sec[i] |= (u1 << nbit) & 0xff;
+    wb = min(8 - nbit, lnb);
+    u1 >>= wb;
+    lnb -= wb;
+    nbit += wb;
+    i += nbit >> 3;
+    nbit &= 7;
+    if (i >= SECTOR_SIZE) break;
+    if (!lnb) continue;
+    f->sec[i] |= u1;
+    nbit += lnb;
+    i += nbit >> 3;
+    nbit &= 7;
+    if (i >= SECTOR_SIZE) break;
   }
 
   return 0;
@@ -296,19 +356,13 @@ int read_boot(fatfs_t *f)
 
   if(f->boot_sec) {
     memcpy(b, f->boot_sec, 0x200);
-  }
-  else {
-    memset(b, 0, 0x200);
-    b[0x00] = 0xeb;	/* jmp $ */
-    b[0x01] = 0xfe;
-    b[0x02] = 0x90;
-    b[0x1fe] = 0x55;
-    b[0x1ff] = 0xaa;
+  } else {
+    build_boot_blk(f);
   }
 
   memcpy(b + 0x03, "DOSEMU10", 8);
-  b[0x0b] = 0x00;
-  b[0x0c] = 0x02;
+  b[0x0b] = f->bytes_per_sect;
+  b[0x0c] = f->bytes_per_sect >> 8;
   b[0x0d] = f->cluster_secs;
   b[0x0e] = f->reserved_secs;
   b[0x0f] = f->reserved_secs >> 8;
@@ -342,7 +396,7 @@ int read_boot(fatfs_t *f)
     b[0x22] = f->total_secs >> 16;
     b[0x23] = f->total_secs >> 24;
   }
-  b[0x24] = 0x80;
+  b[0x24] = f->drive_num;
   b[0x25] = 0x00;
   b[0x26] = 0x29;
   b[0x27] = f->serial;
@@ -350,7 +404,10 @@ int read_boot(fatfs_t *f)
   b[0x29] = f->serial >> 16;
   b[0x2a] = f->serial >> 24;
   memcpy(b + 0x2b, f->label, 11);
-  memcpy(b + 0x36, "FAT16   ", 8);
+  if (f->fat_type == FAT_TYPE_FAT12)
+    memcpy(b + 0x36, "FAT12   ", 8);
+  else
+    memcpy(b + 0x36, "FAT16   ", 8);
 
   return 0;
 }
@@ -386,7 +443,12 @@ void make_label(fatfs_t *f)
 
   if(*s == ' ') s++, i--;
 
-  if(i <= 11 && i > 0) memcpy(f->label, s, i);
+  if(i <= 11 && i > 0) {
+    memcpy(f->label, s, i);
+    while ((s = strchr(f->label, '/')))
+      *s = ' ';
+    strupr(f->label);
+  }
 }
 
 
@@ -425,9 +487,9 @@ void scan_dir(fatfs_t *f, unsigned oi)
   struct dirent* dent;
   struct stat sb;
   char *s, *name, *buf, *buf_ptr;
-  unsigned u, scans = 1;
-  int i, j, fd, size;
-  char *sf[3] = { NULL, NULL, NULL };
+  unsigned u;
+  int i, fd, size, sfs = 0;
+  char *sf[] = { NULL, NULL, NULL, NULL, NULL };
 
   // just checking...
   if(!o->is.dir || o->size || !o->name || o->is.scanned) {
@@ -439,13 +501,6 @@ void scan_dir(fatfs_t *f, unsigned oi)
 
   o->is.scanned = 1;
 
-  name = full_name(f, oi, "");
-
-  if(!name) {
-    fatfs_msg("file name too complex: object %u\n", oi);
-    return;
-  }
-
   if(oi) {
     for(i = 0; i < 2; i++) {
       if((u = new_obj(f))) {	/* ".", ".." */
@@ -465,39 +520,30 @@ void scan_dir(fatfs_t *f, unsigned oi)
   }
   else {
     /* look for "boot.blk" and read it */
-    dir = opendir(name);
-    if(dir != NULL) {
-      while((dent = readdir(dir))) {
-        if(!strcasecmp(dent->d_name, "boot.blk")) {
-          if((s = full_name(f, oi, dent->d_name))) {
-            if(!stat(s, &sb)) {
-              if(S_ISREG(sb.st_mode) && sb.st_size == 0x200) {
+    s = full_name(f, oi, "boot.blk");
+    if (s && access(s, R_OK) == 0 && !stat(s, &sb) &&
+	S_ISREG(sb.st_mode) && sb.st_size == 0x200) {
                 if((fd = open(s, O_RDONLY)) != -1) {
                   if((f->boot_sec = malloc(0x200)) && read(fd, f->boot_sec, 0x200) != 0x200) {
                     free(f->boot_sec);
                     f->boot_sec = NULL;
                   }
                   close(fd);
+	          fatfs_msg("fatfs: boot block taken from boot.blk\n");
                 }
-              }
-            }
-          }
-        }
-      }
-      closedir(dir);
     }
 
     /* look for "IO.SYS" & "MSDOS.SYS" */
-    dir = opendir(name = full_name(f, oi, ""));
-    if(dir != NULL) {
-      while((dent = readdir(dir))) {
-        if(!strcasecmp(dent->d_name, "io.sys")) f->sys_type |= 1;
-        if(!strcasecmp(dent->d_name, "msdos.sys")) f->sys_type |= 2;
-        if(!strcasecmp(dent->d_name, "ibmbio.com")) {
+    s = full_name(f, oi, "io.sys");
+    if (s && access(s, R_OK) == 0 && stat(s, &sb) == 0 &&
+	S_ISREG(sb.st_mode) && sb.st_size > 0)
+	    f->sys_type |= 1;
+    if (access(full_name(f, oi, "msdos.sys"), R_OK) == 0) f->sys_type |= 2;
+    s = full_name(f, oi, "ibmbio.com");
+    if (s && access(s, R_OK) == 0 && stat(s, &sb) == 0 &&
+	S_ISREG(sb.st_mode) && sb.st_size > 0) {
 	  f->sys_type |= 4;
-          if((s = full_name(f, oi, dent->d_name))) {
-            if(!stat(s, &sb)) {
-              if(S_ISREG(sb.st_mode)) {
+          if(S_ISREG(sb.st_mode)) {
                 if((fd = open(s, O_RDONLY)) != -1) {
                   buf = malloc(sb.st_size + 1);
 		  size = read(fd, buf, sb.st_size);
@@ -513,76 +559,86 @@ void scan_dir(fatfs_t *f, unsigned oi)
                   free(buf);
                   close(fd);
                 }
-              }
-            }
           }
-	}
-        if(!strcasecmp(dent->d_name, "ibmdos.com")) f->sys_type |= 8;
-        if(!strcasecmp(dent->d_name, "ipl.sys")) f->sys_type |= 0x10;
-        if(!strcasecmp(dent->d_name, "kernel.sys")) f->sys_type |= 0x20;
+    }
+    if (access(full_name(f, oi, "ibmdos.com"), R_OK) == 0) f->sys_type |= 8;
+    if (access(full_name(f, oi, "ipl.sys"), R_OK) == 0) f->sys_type |= 0x10;
+    if (access(full_name(f, oi, "kernel.sys"), R_OK) == 0) f->sys_type |= 0x20;
+    else {
+      char *libdir = getenv("DOSEMU_LIB_DIR");
+      if (libdir) {
+	char *kernelsyspath = assemble_path(libdir, "drive_z/kernel.sys", 0);
+	if (access(kernelsyspath, R_OK) == 0)
+	  f->sys_type |= 0x20;
+	free(kernelsyspath);
       }
-      closedir(dir);
     }
 
     if((f->sys_type & 3) == 3) {
       f->sys_type = 3;			/* MS-DOS */
       sf[0] = "io.sys";
       sf[1] = "msdos.sys";
-      scans = 3;
+      sfs = 2;
     }
-    if(f->sys_type & 0x40) {
+    if((f->sys_type & 0x4c) == 0x4c) {
       f->sys_type = 0x40;		/* PC-DOS */
       sf[0] = "ibmbio.com";
       sf[1] = "ibmdos.com";
-      scans = 3;
+      sfs = 2;
     }
     if((f->sys_type & 0x0c) == 0x0c) {
       f->sys_type = 0x0c;	/* DR-DOS */
       sf[0] = "ibmbio.com";
       sf[1] = "ibmdos.com";
-      scans = 3;
+      sfs = 2;
     }
     if((f->sys_type & 0x30) == 0x10) {
       f->sys_type = 0x10;	/* FreeDOS, orig. Patv kernel */
       sf[0] = "ipl.sys";
-      scans = 2;
+      sfs = 1;
     }
 
     if((f->sys_type & 0x30) == 0x20) {
       f->sys_type = 0x20;	/* FreeDOS, FD maintained kernel */
       sf[0] = "kernel.sys";
-      scans = 2;
-      bootfile = malloc(strlen(name) + strlen(sf[0]) + 1);
-      sprintf(bootfile, "%s%s", name, sf[0]);
+      sfs = 1;
     }
 
+#define TRY_ADD(x) \
+    sf[sfs] = x; \
+    if (access(full_name(f, oi, sf[sfs]), R_OK) == 0) \
+      sfs++
+
+    TRY_ADD("command.com");
+    TRY_ADD("config.sys");
+    TRY_ADD("autoexec.bat");
+
+    for (i = 0; i < sfs; i++)
+      add_object(f, oi, sf[i]);
     fatfs_msg("system type is 0x%x\n", f->sys_type);
   }
 
-  fatfs_deb2("going to scan directory %u times\n", scans);
-
-  for(i = 0; i < scans; i++) {
-    dir = opendir(name = full_name(f, oi, ""));
-
-    if(dir == NULL) {
-      fatfs_msg("cannot read directory \"%s\"\n", name);
+  name = full_name(f, oi, "");
+  if(!name) {
+    fatfs_msg("file name too complex: object %u\n", oi);
+    return;
+  }
+  dir = opendir(name);
+  if(dir == NULL) {
+    fatfs_msg("cannot read directory \"%s\"\n", name);
+  } else {
+    while((dent = readdir(dir))) {
+      for(i = 0; i < sfs; i++)
+        if(!strcasecmp(dent->d_name, sf[i]))
+	  break;
+      if(i == sfs)
+        add_object(f, oi, dent->d_name);
     }
-    else {
-      while((dent = readdir(dir))) {
-        if(i == scans - 1) {
-          for(j = 0; j < i; j++) if(!strcasecmp(dent->d_name, sf[j])) break;
-          if(i == j) add_object(f, oi, dent->d_name);
-        }
-        else {
-          if(!strcasecmp(dent->d_name, sf[i])) add_object(f, oi, dent->d_name);
-        }
-        o = f->obj + oi;
-      }
 
-      closedir(dir);
-    }
+    closedir(dir);
   }
 
+  o = f->obj + oi;
   if(oi == 0) {
     if(*f->label != ' ') {
       if((u = new_obj(f))) {	/* volume label */
@@ -597,6 +653,7 @@ void scan_dir(fatfs_t *f, unsigned oi)
         if(!stat(f->dir, &sb)) {
           f->obj[u].time = dos_time(&sb.st_mtime);
         }
+	fatfs_deb2("added label \"%s\"\n", f->label);
       }
     }
   }
@@ -654,6 +711,7 @@ void add_object(fatfs_t *f, unsigned parent, char *name)
   struct stat sb;
   obj_t tmp_o = {{0}, 0};
   unsigned u;
+#if 0
   static int first = 1;
   static int exxx = 0;
   static char esys[16] = "";
@@ -679,6 +737,7 @@ void add_object(fatfs_t *f, unsigned parent, char *name)
       exxx -= 1;
     }
   }
+#endif
   if(!(strcmp(name, ".") && strcmp(name, ".."))) return;
 
   if(!(s = full_name(f, parent, name))) {
@@ -686,13 +745,27 @@ void add_object(fatfs_t *f, unsigned parent, char *name)
     return;
   }
 
+  fatfs_deb("trying to add \"%s\":\n", s);
   if(stat(s, &sb)) {
-    fatfs_deb("file not found: \"%s\"\n", s);
-    return;
+    int found = 0;
+    if (strcmp(name, "kernel.sys") == 0) {
+      char *libdir = getenv("DOSEMU_LIB_DIR");
+      fatfs_deb("does not exist\n");
+      if (libdir) {
+	s = assemble_path(libdir, "drive_z/kernel.sys", 0);
+	fatfs_deb("trying to add \"%s\":\n", s);
+	found = (stat(s, &sb) == 0);
+	free(s);
+      }
+    }
+    if (!found) {
+      fatfs_deb("file not found\n");
+      return;
+    }
   }
 
   if(!(S_ISDIR(sb.st_mode) || S_ISREG(sb.st_mode))) {
-    fatfs_deb("entry ignored: \"%s\"\n", s);
+    fatfs_deb("entry ignored\n");
     return;
   }
 
@@ -726,7 +799,7 @@ void add_object(fatfs_t *f, unsigned parent, char *name)
 
   memcpy(f->obj + u, &tmp_o, sizeof *f->obj);
 
-  fatfs_deb("added %s: \"%s\"\n", tmp_o.is.dir ? "directory" : "file", s);
+  fatfs_deb("added as a %s\n", tmp_o.is.dir ? "directory" : "file");
 }
 
 
@@ -751,8 +824,9 @@ unsigned make_dos_entry(fatfs_t *f, obj_t *o, unsigned char **e)
   memset(dos_ent, 0, sizeof dos_ent);
 
   s = o->name;
+#if 0
   if (o->is.faked_sys) s = "config.sys";
-
+#endif
   if(o->is.this_dir) {
     s = ".";
     o = f->obj + o->parent;
@@ -864,7 +938,8 @@ void assign_clusters(fatfs_t *f, unsigned max_clu, unsigned max_obj)
       f->got_all_objs = 1;
       fatfs_msg("assign_clusters: file system full\n");
     }
-    fatfs_deb("assign_clusters: obj %u, start %u, len %u\n", u, f->obj[u].start, f->obj[u].len);
+    fatfs_deb("assign_clusters: obj %u, start %u, len %u (%s)\n",
+	u, f->obj[u].start, f->obj[u].len, f->obj[u].name);
   }
 
   if(u == f->objs) {
@@ -1039,23 +1114,32 @@ unsigned next_cluster(fatfs_t *f, unsigned clu)
 void fdkernel_boot_mimic(void)
 {
   int f, size;
-  unsigned int loadaddress = 0x600;
+  char *bootfile;
+  void *loadaddress = (void *)0x600;
+  fatfs_t *fs = get_fat_fs_by_drive(HI(ax));
 
-  if (!bootfile) {
+  if (!fs || !(bootfile = full_name(fs, 0, fs->obj[1].name))) {
     error("BOOT-helper requested, but no systemfile available\n");
     leavedos(99);
   }
   if ((f = open(bootfile, O_RDONLY)) == -1) {
-    error("cannot open DOS system file %s\n", bootfile);
-    leavedos(99);
+    char *libdir = getenv("DOSEMU_LIB_DIR");
+    if (libdir)
+      bootfile = assemble_path(libdir, "drive_z/kernel.sys", 0);
+    if (!libdir || (f = open(bootfile, O_RDONLY)) == -1) {
+      error("cannot open DOS system file %s\n", bootfile);
+      if (libdir) free(bootfile);
+      leavedos(99);
+    }
+    if (libdir) free(bootfile);
   }
   size = lseek(f, 0, SEEK_END);
   lseek(f, 0, SEEK_SET);
-  read(f, (void *)loadaddress, size);
+  dos_read(f, loadaddress, size);
   close(f);
-  LWORD(cs) = LWORD(ds) = LWORD(es) = loadaddress >> 4;
+  LWORD(cs) = LWORD(ds) = LWORD(es) = FP_SEG32(loadaddress);
   LWORD(eip) = 0;
-  LWORD(ebx) = 0x80;	/* boot drive */
+  LWORD(ebx) = fs->drive_num;	/* boot drive */
   LWORD(ss) = 0x1FE0;
   LWORD(esp) = 0x7c00;	/* temp stack */
 }
@@ -1065,17 +1149,17 @@ void fdkernel_boot_mimic(void)
  */
 void build_boot_blk(fatfs_t *f)
 {
-  /* boot block from "fatfs_boot.asm" */
-  static unsigned char boot_prog[] = {
-    #include "fatfs_boot.h"
-  };
-
   /*
    * Make sure this messages are not too long; they should not extend
    * beyond 0x7ded (incl. final '\0').
    */
-  char *msg = "\r\nSorry, could not load an operating system.\r\nHave a nice day!\r\n";
-  char *msg1 = "\r\nSorry, there is no operating system.\r\nHave a nice day!\r\n";
+  char *msg_f = "\r\nSorry, could not load an operating system from\r\n%s\r\n\r\n"
+"Please try to install FreeDOS from dosemu-freedos-*-bin.tgz\r\n\r\n"
+"Press any key to return to Linux...\r\n";
+  char *msg1_f = "\r\nSorry, there is no operating system here:\r\n%s\r\n\r\n"
+"Please try to install FreeDOS from dosemu-freedos-*-bin.tgz\r\n\r\n"
+"Press any key to return to Linux...\r\n";
+  char *msg, *msg1;
 
   int i;
   unsigned r_o, d_o, t_o;
@@ -1084,17 +1168,19 @@ void build_boot_blk(fatfs_t *f)
 
   if(!(f->boot_sec = malloc(0x200))) return;
 
+  asprintf(&msg, msg_f, f->dir);
+  asprintf(&msg1, msg1_f, f->dir);
   b = f->boot_sec;
   memset(b, 0, 0x200);
   b[0x00] = 0xeb;	/* jmp 0x7c40 */
   b[0x01] = 0x3e;
   b[0x02] = 0x90;
-  memcpy(b + 0x40, boot_prog, sizeof boot_prog);
-  strcpy(b + 0x40 + sizeof boot_prog, msg);
+  memcpy(b + 0x40, boot_prog, boot_prog_end - boot_prog);
+  strcpy(b + 0x40 + (boot_prog_end - boot_prog), msg);
   b[0x1fe] = 0x55;
   b[0x1ff] = 0xaa;
 
-  t_o = (0x40 + sizeof boot_prog + strlen(msg) + 1 + 3) & ~3;
+  t_o = (0x40 + boot_prog_end - boot_prog + strlen(msg) + 1 + 3) & ~3;
   d0 = b + t_o;
   d1 = d0 + 0x14;
 
@@ -1102,7 +1188,7 @@ void build_boot_blk(fatfs_t *f)
   b[0x3e] = t_o;
   b[0x3f] = t_o >> 8;
 
-  t_o = 0x7c00 + 0x40 + sizeof boot_prog - 5;
+  t_o = 0x7c00 + 0x40 + (boot_prog_end - boot_prog) - 5;
 
   /* 1st root directory sector */
   r_o = f->fats * f->fat_secs + f->reserved_secs + f->hidden_secs;
@@ -1121,18 +1207,20 @@ void build_boot_blk(fatfs_t *f)
   0x0e	/* di */
   0x10	/* bp */
   0x12	/* list entries */
-  0x13	/* unused */
+  0x13	/* drive */
   0x14	/* start of disk_tab list */
 #endif
 
+  d0[0x13] = f->drive_num;
+
   switch(f->sys_type) {
     case 0x03:
     case 0x40:
       i = read_data(f, 0);
       if(i || f->sec[0] != 'M' || f->sec[1] != 'Z') {
         /* for IO.SYS, MS-DOS version < 7 */
-        make_i1342_blk(d1 + 0x00, r_o, 1, 0, 0x500);
-        make_i1342_blk(d1 + 0x10, d_o, 4, 0, 0x700);
+        make_i1342_blk((struct ibm_ms_diskaddr_pkt *)(d1 + 0x00), r_o, 1, 0, 0x500);
+        make_i1342_blk((struct ibm_ms_diskaddr_pkt *)(d1 + 0x10), d_o, 4, 0, 0x700);
         d0[0x12] = 2;		/* 2 entries */
 
         d0[0x02] = 0x70;	/* start seg */
@@ -1141,13 +1229,13 @@ void build_boot_blk(fatfs_t *f)
         d0[0x06] = d_o;		/* bx */
         d0[0x07] = d_o >> 8;
         d0[0x09] = 0xf8;	/* ch */
-        d0[0x0a] = 0x80;	/* dl */
+        d0[0x0a] = f->drive_num;	/* dl */
 
         fatfs_msg("made boot block suitable for MS-DOS, version < 7\n");
       }
       else {
         /* for IO.SYS, MS-DOS version >= 7 */
-        make_i1342_blk(d1, d_o, 4, 0, 0x700);
+        make_i1342_blk((struct ibm_ms_diskaddr_pkt *)d1, d_o, 4, 0, 0x700);
         d0[0x12] = 1;		/* 1 entry */
 
         d0[0x01] = 0x02;	/* start ofs */
@@ -1180,21 +1268,21 @@ void build_boot_blk(fatfs_t *f)
 
     case 0x0c:			/* PC-DOS */
       /* for IBMBIO.COM, PC-, DR-, Open-, NOVELL-DOS  */
-      make_i1342_blk(d1, d_o, (f->obj[1].size + 0x1ff) >> 9, 0, 0x700);
+      make_i1342_blk((struct ibm_ms_diskaddr_pkt *)d1, d_o, (f->obj[1].size + 0x1ff) >> 9, 0, 0x700);
       d0[0x12] = 1;		/* 1 entry */
 
       d0[0x02] = 0x70;	/* start seg */
-      d0[0x0a] = 0x80;	/* dl */
+      d0[0x0a] = f->drive_num;	/* dl */
 
       fatfs_msg("made boot block suitable for PC-, DR-, Open-, NOVELL-DOS\n");
       break;
 
     case 0x10:			/* FreeDOS, orig. Patv kernel */
-      make_i1342_blk(d1, d_o, (f->obj[1].size + 0x1ff) >> 9, 0x2000, 0x0);
+      make_i1342_blk((struct ibm_ms_diskaddr_pkt *)d1, d_o, (f->obj[1].size + 0x1ff) >> 9, 0x2000, 0x0);
       d0[0x12] = 1;		/* 1 entry */
 
       d0[0x03] = 0x20;		/* start seg */
-      d0[0x0a] = 0x80;		/* dl */
+      d0[0x0a] = f->drive_num;		/* dl */
 
       fatfs_msg("made boot block suitable for DosC\n");
       break;
@@ -1204,7 +1292,7 @@ void build_boot_blk(fatfs_t *f)
 			   and above fdkernel_boot_mimic() function */
       b[0x40] = 0xb8;	/* mov ax,0fdh */
       b[0x41] = DOS_HELPER_BOOTSECT;
-      b[0x42] = 0x00;
+      b[0x42] = f->drive_num;
       b[0x43] = 0xcd;	/* int 0e6h */
       b[0x44] = 0xe6;
 
@@ -1212,8 +1300,8 @@ void build_boot_blk(fatfs_t *f)
       break;
 
     default:			/* no system */
-      strcpy(b + 0x40 + sizeof boot_prog, msg1);
-      t_o = (0x40 + sizeof boot_prog + strlen(msg1) + 1 + 3) & ~3;
+      strcpy(b + 0x40 + (boot_prog_end - boot_prog), msg1);
+      t_o = (0x40 + (boot_prog_end - boot_prog) + strlen(msg1) + 1 + 3) & ~3;
       d0 = b + t_o;
 
       t_o += 0x7c00;
@@ -1231,6 +1319,8 @@ void build_boot_blk(fatfs_t *f)
       fatfs_msg("boot block has no boot program\n");
       break;
   }
+  free(msg);
+  free(msg1);
 }
 
 
@@ -1238,17 +1328,12 @@ void build_boot_blk(fatfs_t *f)
  * Set up disk transfer blocks for int 0x13, ah = 0x42 (0x10 bytes each).
  * Don't forget to zero the buffer before calling make_i1342_blk!
  */
-void make_i1342_blk(unsigned char *b, unsigned start, unsigned len, unsigned seg, unsigned ofs)
+void make_i1342_blk(struct ibm_ms_diskaddr_pkt *b, unsigned start, unsigned blks, unsigned seg, unsigned ofs)
 {
-  b[0x00] = 0x10;
-  b[0x02] = len;
-  b[0x03] = len >> 8;
-  b[0x04] = ofs;
-  b[0x05] = ofs >> 8;
-  b[0x06] = seg;
-  b[0x07] = seg >> 8;
-  b[0x08] = start;
-  b[0x09] = start >> 8;
-  b[0x0a] = start >> 16;
-  b[0x0b] = start >> 24;
+  b->len = 0x10;
+  b->blocks = blks;
+  b->buf_ofs = ofs;
+  b->buf_seg = seg;
+  b->block_lo = start;
+  b->block_hi = 0;
 }
