@@ -1,16 +1,19 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 #include <stdlib.h>
 #include <unistd.h>
-#if GLIBC_VERSION_CODE >= 2000
-  #include <sys/io.h>
-#endif
+#include <fcntl.h>
+#include <grp.h>
+#include <sys/io.h>
 #include "emu.h"
 #include "priv.h"
+#include "dosemu_config.h"
+#include "mapping.h"
+#include "utilities.h"
 #ifdef X86_EMULATOR
 #include "cpu-emu.h"
 #endif
@@ -143,6 +146,8 @@ int priv_iopl(int pl)
 #ifdef X86_EMULATOR
   if (config.cpuemu) e_priv_iopl(pl);
 #endif
+  if (ret == 0)
+    current_iopl = pl;
   return ret;
 }
 
@@ -176,9 +181,6 @@ gid_t get_orig_gid(void)
   return gid;
 }
 
-/* BIG NOTE:
- * The below 'priv_drop' should only be called from forked child processes !!
- */
 int priv_drop(void)
 {
   if (setreuid(uid,uid) || setregid(gid,gid))
@@ -186,6 +188,10 @@ int priv_drop(void)
       error("Cannot drop root uid or gid!\n");
       return 0;
     }
+  cur_euid = euid = uid;
+  cur_egid = egid = gid;
+  skip_priv_setting = 1;
+  if (uid) can_do_root_stuff = 0;
   return 1;
 }
 #endif /* __linux__ */
@@ -215,19 +221,46 @@ void priv_init(void)
   gid  = cur_gid  = getgid();
   egid = cur_egid = getegid();
 
-  if ((euid != 0) && (uid == euid))
+  /* must store the /proc/self/exe symlink contents before dropping
+     privs! */
+  dosemu_proc_self_exe = readlink_malloc("/proc/self/exe");
+  /* For Fedora we must also save a file descriptor to /proc/self/maps */
+  dosemu_proc_self_maps_fd = open("/proc/self/maps", O_RDONLY);
+  if (under_root_login)
+  {
+    /* check for sudo and set to original user */
+    char *s = getenv("SUDO_GID");
+    if (s) {
+      gid = cur_gid = atoi(s);
+      if (gid) {
+        setregid(gid, egid);
+      }
+    }
+    s = getenv("SUDO_UID");
+    if (s) {
+      uid = cur_uid = atoi(s);
+      if (uid) {
+        skip_priv_setting = under_root_login = 0;
+	using_sudo = 1;
+	s = getenv("SUDO_USER");
+	if (s) {
+	  initgroups(s, gid);
+	  setenv("USER", s, 1);
+	}
+        setreuid(uid, euid);
+      }
+    }
+  }
+  
+  if (!can_do_root_stuff)
     {
       skip_priv_setting = 1;
-      can_do_root_stuff = 0;
-      fprintf(stderr,"\nRunning unpriviledged in low feature mode\n");
     }
 
   num_groups = getgroups(0,0);
   groups = malloc(num_groups * sizeof(gid_t));
   getgroups(num_groups,groups);
 
-#ifndef RUN_AS_ROOT
   if (!skip_priv_setting) _priv_off();
-#endif
 }
 
