@@ -1,17 +1,18 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* 
  * Various sundry utilites for dos emulator.
  *
  */
+#include "emu.h"
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdarg.h>
-#include <malloc.h>
 #include <sys/time.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -19,17 +20,18 @@
 #include <fcntl.h>
 #include <string.h>
 #include <ctype.h>
-#include "Linux/vt.h"
-#include <sys/ioctl.h>
 #include <errno.h>
+#include <assert.h>
+#include <dlfcn.h>
 
-#include "emu.h"
 #include "machcompat.h"
 #include "bios.h"
 #include "timers.h"
 #include "pic.h"
 #include "dpmi.h"
+#include "debug.h"
 #include "utilities.h"
+#include "dosemu_config.h"
 #ifdef USE_MHPDBG
 #include "mhpdbg.h"
 #endif
@@ -140,7 +142,7 @@ char *chrprintable(char c)
   return strprintable(buf);
 }
 
-static int vlog_printf(int flg, const char *fmt, va_list args)
+int vlog_printf(int flg, const char *fmt, va_list args)
 {
   int i;
   static int is_cr = 1;
@@ -266,16 +268,18 @@ int log_printf(int flg, const char *fmt, ...)
 void verror(const char *fmt, va_list args) 
 {
 	char fmtbuf[1025];
+	va_list orig_args;
+	__va_copy(orig_args, args);
 
 	if (fmt[0] == '@') {
 		vlog_printf(10, fmt+1, args);
-		vfprintf(stderr, fmt+1, args);
+		vfprintf(stderr, fmt+1, orig_args);
 	}
 	else {
 		fmtbuf[sizeof(fmtbuf)-1] = 0;
 		snprintf(fmtbuf, sizeof(fmtbuf)-1, "ERROR: %s", fmt);
 		vlog_printf(10, fmtbuf, args);
-		vfprintf(stderr, fmtbuf, args);
+		vfprintf(stderr, fmtbuf, orig_args);
 	}
 	
 }
@@ -290,8 +294,8 @@ void error(const char *fmt, ...)
 
 
 /* write string to dos? */
-void
-p_dos_str(char *fmt,...) {
+int
+p_dos_str(const char *fmt,...) {
   va_list args;
   static char buf[1025];
   char *s;
@@ -306,6 +310,7 @@ p_dos_str(char *fmt,...) {
   g_printf("CONSOLE MSG: '%s'\n",buf);
   while (*s) 
 	char_out(*s++, READ_BYTE(BIOS_CURRENT_SCREEN_PAGE));
+  return i;
 }
         
 /* some stuff to handle reading of /proc */
@@ -448,7 +453,7 @@ char *assemble_path(char *dir, char *file, int append_pid)
 	s = malloc(strlen(dir)+1+strlen(file)+strlen(pid)+1);
 	if (!s) {
 		fprintf(stderr, "out of memory, giving up\n");
-		longjmp(NotJEnv, 0x4d);
+		siglongjmp(NotJEnv, 0x4d);
 	}
 	sprintf(s, "%s/%s%s", dir, file, pid);
 	return s;
@@ -461,8 +466,7 @@ char *mkdir_under(char *basedir, char *dir, int append_pid)
 	if (dir) s = assemble_path(basedir, dir, append_pid);
 	if (!exists_dir(s)) {
 		if (mkdir(s, S_IRWXU)) {
-			fprintf(stderr, "can't create local %s directory, giving up\n", s);
-			longjmp(NotJEnv, 0x42);
+			fprintf(stderr, "can't create local %s directory\n", s);
 		}
 	}
 	return s;
@@ -527,6 +531,12 @@ void subst_file_ext(char *ptr)
 		     while (*r) { *r=toupper(*r); r++; } }
     static int subst_sys=2;
 
+    if (ptr == NULL) {
+      /* reset */
+      subst_sys = 2;
+      return;
+    }
+
     /* skip leading drive name and \ */
     if (ptr[1]==':' && ptr[2]=='\\') ptr+=3;
     else if (ptr[0]=='\\') ptr++;
@@ -609,7 +619,7 @@ void sigalarm_onoff(int on)
   static struct itimerval itv_oldp;
 #endif
   static struct itimerval itv;
-  static int is_off = 0;
+  static volatile int is_off = 0;
   if (on) {
     if (is_off--) {
     	setitimer(ITIMER_REAL, &itv_old, NULL);
@@ -628,105 +638,207 @@ void sigalarm_onoff(int on)
   }
 }
 
-int is_console(int fd)
+void sigalarm_block(int block)
 {
-  /* normally the below ioctl will return the next available not opened VT.
-   * The fd maybe _any_ valid fd of an opened VT.
-   * Hence, if the below ioctl fails, we assume the fd is not the one
-   * of an VT. ... simple;-)
-   */
-  int nextvt;
-  if( ioctl( fd, VT_OPENQRY, &nextvt ) ) return 0;
+  static volatile int is_blocked = 0;
+  sigset_t blockset;
+  if (block) {
+    if (!is_blocked++) {
+      sigemptyset(&blockset);
+      sigaddset(&blockset, SIGALRM);
+      sigprocmask(SIG_BLOCK, &blockset, NULL);
+    }
+  }
+  else if (is_blocked--) {
+    sigemptyset(&blockset);
+    sigaddset(&blockset, SIGALRM);
+    sigprocmask(SIG_UNBLOCK, &blockset, NULL);
+  }
+}
+
+/* dynamic readlink, adapted from "info libc" */
+char *readlink_malloc (const char *filename)
+{
+  int size = 50;
+  int nchars;
+  char *buffer;
+
+  do {
+    size *= 2;
+    nchars = 0;
+    buffer = malloc(size);
+    if (buffer != NULL) {
+      nchars = readlink(filename, buffer, size);
+      if (nchars < 0) {
+        free(buffer);
+        buffer = NULL;
+      }
+    }
+  } while (nchars >= size);
+  if (buffer != NULL)
+    buffer[nchars] = '\0';
+  return buffer;
+}
+
+char * strupr(char *s)
+{
+	char *p = s;
+	while (*p) {
+		*p = toupper(*p);
+		p++;
+	}
+	return s;
+}
+
+char * strlower(char *s)
+{
+	char *p = s;
+	while (*p) {
+		*p = tolower(*p);
+		p++;
+	}
+	return s;
+}
+
+void dosemu_error(char *fmt, ...)
+{
+    va_list args;
+    va_start(args, fmt);
+    verror(fmt, args);
+    va_end(args);
+    gdb_debug();
+}
+
+#ifdef USE_DL_PLUGINS
+void *load_plugin(const char *plugin_name)
+{
+    char *fullname = malloc(strlen(dosemu_proc_self_exe) +
+			    strlen(plugin_name) + 20);
+    void *handle;
+    char *slash;
+
+    strcpy(fullname, dosemu_proc_self_exe);
+    slash = strrchr(fullname, '/');
+    if (slash == NULL) return NULL;
+    sprintf(slash + 1, "libplugin_%s.so", plugin_name);
+    handle = dlopen(fullname, RTLD_LAZY);
+    free(fullname);
+    if (handle != NULL)
+	return handle;
+    asprintf(&fullname, "%s/dosemu/libplugin_%s.so",
+	     LIB_DEFAULT, plugin_name);
+    handle = dlopen(fullname, RTLD_LAZY);
+    free(fullname);
+    if (handle != NULL)
+	return handle;
+    error("%s support not compiled in or not found:\n", plugin_name);
+    error("%s\n", dlerror());
+    return handle;
+}
+#else
+void *load_plugin(const char *plugin_name)
+{
+    return NULL;
+}
+#endif
+
+/* Ring buffer implementation */
+
+void rng_init(struct rng_s *rng, size_t objnum, size_t objsize)
+{
+  rng->buffer = calloc(objnum, objsize);
+  rng->objnum = objnum;
+  rng->objsize = objsize;
+  rng->tail = rng->objcnt = 0;
+}
+
+int rng_destroy(struct rng_s *rng)
+{
+  int ret = rng->objcnt;
+  free(rng->buffer);
+  rng->objcnt = 0;
+  return ret;
+}
+
+int rng_get(struct rng_s *rng, void *buf)
+{
+  if (!rng->objcnt)
+    return 0;
+  if (buf)
+    memcpy(buf, rng->buffer + rng->tail, rng->objsize);
+  rng->tail += rng->objsize;
+  rng->tail %= rng->objsize * rng->objnum;
+  rng->objcnt--;
   return 1;
 }
 
-
-/*
- * getting tired of needless 'llseek() versus lseek64()' war
- * ... calling the kernel directly  --Hans 2000/03/26
- */
-long long libless_llseek(int fd, long long offset, int origin)
+int rng_peek(struct rng_s *rng, int idx, void *buf)
 {
-  long long result;
-  int offlo = offset;
-  int offhi = offset >> 32;
-  int __res;
-  __asm__ __volatile__(
-    "int $0x80\n"
-    :"=a" (__res)
-    :"a" ((int)140), "b" (fd), "c" (offhi), "d" (offlo),
-                    "S" ((int)&result), "D" (origin)
-  );
-  errno = 0;
-  if (__res < 0) {
-    errno = -__res;
-    result = -1;
+  int obj_pos;
+  if (rng->objcnt <= idx)
+    return 0;
+  obj_pos = (rng->tail + idx * rng->objsize) % (rng->objnum * rng->objsize);
+  assert(buf);
+  memcpy(buf, rng->buffer + obj_pos, rng->objsize);
+  return 1;
+}
+
+int rng_put(struct rng_s *rng, void *obj)
+{
+  int head_pos, ret = 1;;
+  head_pos = (rng->tail + rng->objcnt * rng->objsize) % (rng->objnum * rng->objsize);
+  assert(head_pos <= (rng->objnum - 1) * rng->objsize);
+  memcpy(rng->buffer + head_pos, obj, rng->objsize);
+  rng->objcnt++;
+  if (rng->objcnt > rng->objnum) {
+    rng_get(rng, NULL);
+    ret = 0;
   }
-  return result;
+  assert(rng->objcnt <= rng->objnum);
+  return ret;
 }
 
-
-#if (GLIBC_VERSION_CODE >= 2000) && defined(PORTABLE_BINARY)
-/*
- * Under glibc we can't make portable static binaries due to libnss stuff.
- * As we only need to look at /etc/passwd (and which non-libnss system won't
- * have it?) we work around this problem by doing getpwuid() getpwnam()
- * ourselves, if we have no /etc/nsswitch.conf.
- *                                                -- Hans, 981201
- */
-
-static struct passwd *our_getpw(const char *name, uid_t uid)
+int rng_put_const(struct rng_s *rng, int value)
 {
-  FILE *f;
-  static struct passwd pw;
-  static char buf[1024];
-  char *p;
-
-  f = fopen("/etc/passwd", "r");
-  if (!f) return 0;
-  buf[sizeof(buf)-1] = 0;
-  while (fgets(buf, sizeof(buf)-1, f)) {
-    pw.pw_name = strtok(buf, ":");
-    if (!pw.pw_name) continue;
-    if (name && strcmp(name, pw.pw_name)) continue;
-    pw.pw_passwd = strtok(0, ":");
-    if (!pw.pw_passwd) continue;
-    p = strtok(0, ":");
-    if (!p) continue;
-    pw.pw_uid = (uid_t)strtoul(p,0,0);
-    if (!name && pw.pw_uid != uid) continue;
-    p = strtok(0, ":");
-    if (!p) continue;
-    pw.pw_gid = (gid_t)strtoul(p,0,0);
-    /* we don't need more, hence we skip the rest */
-    fclose(f);
-    return &pw;
+  if (rng->objsize > sizeof(value)) {
+    dosemu_error("rng_put_const for obj size=%i\n", rng->objsize);
+    return 0;
   }
-  fclose(f);
-  return 0;
+  return rng_put(rng, &value);
 }
 
-static int have_libnss()
+int rng_poke(struct rng_s *rng, int idx, void *buf)
 {
-  static int have_it = -1;
-  if (have_it >= 0 ) return have_it;
-  have_it = exists_file("/etc/nsswitch.conf") ? 1: 0;
-  return have_it;
+  int obj_pos;
+  if (rng->objcnt <= idx)
+    return 0;
+  obj_pos = (rng->tail + idx * rng->objsize) % (rng->objnum * rng->objsize);
+  memcpy(rng->buffer + obj_pos, buf, rng->objsize);
+  return 1;
 }
 
-#undef getpwuid
-#undef getpwnam
-
-struct passwd *our_getpwuid(uid_t uid)
+int rng_add(struct rng_s *rng, int num, void *buf)
 {
-  if (have_libnss()) return getpwuid(uid);
-  return our_getpw(0, uid);
+  int i, ret = 0;
+  for (i = 0; i < num; i++)
+    ret += rng_put(rng, (unsigned char *)buf + i * rng->objsize);
+  return ret;
 }
 
-struct passwd *our_getpwnam(const char *name)
+int rng_remove(struct rng_s *rng, int num, void *buf)
 {
-  if (have_libnss()) return getpwnam(name);
-  return our_getpw(name, 0);
+  int i, ret = 0;
+  for (i = 0; i < num; i++)
+    ret += rng_get(rng, buf ? (unsigned char *)buf + i * rng->objsize : NULL);
+  return ret;
 }
 
-#endif /* GLIBC_VERSION_CODE >= 2000 */
+int rng_count(struct rng_s *rng)
+{
+  return rng->objcnt;
+}
+
+void rng_clear(struct rng_s *rng)
+{
+  rng->objcnt = 0;
+}
