@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* DANG_BEGIN_MODULE
@@ -33,11 +33,13 @@
 #include <sys/ioctl.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <string.h>
+#include <errno.h>
+#include <Linux/serial.h>
 
 #include "config.h"
 #include "emu.h"
 #include "mouse.h"
-#include "serial.h"
 #include "ser_defs.h"
 
 
@@ -75,15 +77,17 @@ static inline void tx_buffer_dump(int num)
 /* This function updates the flow control status depending on buffer condition */
 static inline void flow_control_update(int num)
 {
-  static int control;
-  if (com[num].rx_buf_bytes == 0) {			/* buffer empty? */
+  int control;
+  if (RX_BUF_BYTES(num) == 0) {			/* buffer empty? */
     control = TIOCM_RTS;
     ioctl(com[num].fd, TIOCMBIS, &control);		/* Raise RTS */
   }
   else {						/* Buffer has chars */
     control = TIOCM_RTS;
     ioctl(com[num].fd, TIOCMBIC, &control);		/* Lower RTS */
+#if 0
     com[num].rx_timer = RX_READ_FREQ;	/* Reset rcv read() timer */
+#endif
   }
 }
 
@@ -93,17 +97,30 @@ static inline void flow_control_update(int num)
  * a circular buffer because this way, a single read() can easily
  * put data straight into our internal receive buffer!
  */
-static inline void rx_buffer_slide(int num)
+void rx_buffer_slide(int num)
 {
-  static int i;
-  
+  if (com[num].rx_buf_start == 0)
+    return;
   /* Move existing chars in receive buffer to the start of buffer */
-  for(i = 0; i < com[num].rx_buf_bytes; i++)
-    com[num].rx_buf[i] = com[num].rx_buf[com[num].rx_buf_start++];
+  memmove(com[num].rx_buf, com[num].rx_buf + com[num].rx_buf_start,
+    RX_BUF_BYTES(num));
 
   /* Update start and end pointers in buffer */        
+  com[num].rx_buf_end -= com[num].rx_buf_start;
   com[num].rx_buf_start = 0;
-  com[num].rx_buf_end = i;
+}
+
+void tx_buffer_slide(int num)
+{
+  if (com[num].tx_buf_start == 0)
+    return;
+  /* Move existing chars in receive buffer to the start of buffer */
+  memmove(com[num].tx_buf, com[num].tx_buf + com[num].tx_buf_start,
+    TX_BUF_BYTES(num));
+
+  /* Update start and end pointers in buffer */        
+  com[num].tx_buf_end -= com[num].tx_buf_start;
+  com[num].tx_buf_start = 0;
 }
 
 
@@ -117,7 +134,7 @@ static inline void rx_buffer_slide(int num)
  */
 void uart_fill(int num)
 {
-  static int size;
+  int size = 0;
   
   /* Return if in loopback mode */
   if (com[num].MCR & UART_MCR_LOOP) return;
@@ -131,9 +148,10 @@ void uart_fill(int num)
    * contains enough data for a full FIFO (at least 16 bytes).
    * The receive buffer is a sliding buffer.
    */
-  if (com[num].rx_buf_bytes < com[num].rx_fifo_size) {
+  if (RX_BUF_BYTES(num) < com[num].rx_fifo_size) {
+#if 0
     if (com[num].rx_timer == 0) {
-    
+#endif    
       /* Slide the buffer contents to the bottom */
       rx_buffer_slide(num);
     
@@ -143,18 +161,26 @@ void uart_fill(int num)
       size = RPT_SYSCALL(read(com[num].fd, 
                               &com[num].rx_buf[com[num].rx_buf_end],
                               RX_BUFFER_SIZE - com[num].rx_buf_end));
-    
+      if (size < 0)
+        return;
+      if(s3_printf) s_printf("SER%d: Got %i bytes, %i in buffer\n",num,
+        size, RX_BUF_BYTES(num));
+#if 0    
       if (size == 0) { 				/* No characters read? */
         com[num].rx_timer = RX_READ_FREQ;	/* Reset rcv read() timer */
       }
-      else if (size > 0) {		/* Note that size is -1 if error */
+      else
+#endif
+      if (size > 0) {		/* Note that size is -1 if error */
         com[num].rx_timeout = TIMEOUT_RX;	/* Reset timeout counter */
-        com[num].rx_buf_bytes += size;		/* No. of chars in buffer */
+        com[num].rx_buf_end += size;
       }
+#if 0
     }
+#endif
   }
-  
-  if (com[num].rx_buf_bytes) {		/* Is data waiting in the buffer? */
+
+  if (RX_BUF_BYTES(num)) {		/* Is data waiting in the buffer? */
     if (com[num].fifo_enable) {		/* Is it in 16550 FIFO mode? */
 
       com[num].LSR |= UART_LSR_DR;		/* Set recv data ready bit */
@@ -164,7 +190,7 @@ void uart_fill(int num)
        * Reset the receive FIFO counter up to a value of 16 (configurable)
        * if we are not within an interrupt.
        */
-      com[num].rx_fifo_bytes = com[num].rx_buf_bytes;
+      com[num].rx_fifo_bytes = RX_BUF_BYTES(num);
       if (com[num].rx_fifo_bytes > com[num].rx_fifo_size)
 	      com[num].rx_fifo_bytes = com[num].rx_fifo_size;
       
@@ -193,6 +219,8 @@ void uart_fill(int num)
 void uart_clear_fifo(int num, int fifo)
 {
   /* DANG_FIXTHIS Should clearing UART cause THRE int if it's enabled? */
+
+  if(s1_printf) s_printf("SER%d: Clear FIFO.\n",num);
   
   /* Clear the receive FIFO */
   if (fifo & UART_FCR_CLEAR_RCVR) {
@@ -201,7 +229,6 @@ void uart_clear_fifo(int num, int fifo)
     com[num].LSRqueued &= ~(UART_LSR_ERR | UART_LSR_DR);
     com[num].rx_buf_start = 0;		/* Beginning of rec FIFO queue */
     com[num].rx_buf_end = 0;		/* End of rec FIFO queue */
-    com[num].rx_buf_bytes = 0;		/* Number of bytes in recv queue */
     com[num].rx_fifo_bytes = 0;         /* Number of bytes in emulated FIFO */
     com[num].rx_timeout = 0;		/* Receive intr already occured */
     com[num].int_condition &= ~(LS_INTR | RX_INTR);  /* Clear LS/RX conds */
@@ -215,12 +242,12 @@ void uart_clear_fifo(int num, int fifo)
     com[num].LSRqueued &= ~(UART_LSR_TEMT | UART_LSR_THRE);
     com[num].tx_buf_start = 0;		/* Start of xmit FIFO queue */
     com[num].tx_buf_end = 0;		/* End of xmit FIFO queue */
-    com[num].tx_buf_bytes = 0;		/* Number of bytes in xmit FIFO */
     com[num].tx_trigger = 0;            /* Transmit intr already occured */
     com[num].tx_overflow = 0;		/* Not in overflow state */
     com[num].int_condition &= ~TX_INTR;	/* Clear TX int condition */
     tx_buffer_dump(num);		/* Clear transmit buffer */
   }
+  com[num].IIR &= ~UART_IIR_FIFO;
 
   serial_int_engine(num, 0);		/* Update interrupt status */
 }
@@ -235,15 +262,20 @@ void ser_termios(int num)
 {
   speed_t baud;
   long int rounddiv;
+  struct serial_struct ser_info;
   
   /* The following is the same as (com[num].dlm * 256) + com[num].dll */
   #define DIVISOR ((com[num].dlm << 8) | com[num].dll)
 
   /* Return if not a tty */
   if (tcgetattr(com[num].fd, &com[num].newset) == -1) {
-    if(s1_printf) s_printf("SER%d: Line Control: NOT A TTY.\n",num);
+    if(s1_printf) s_printf("SER%d: Line Control: NOT A TTY (%s).\n",num,strerror(errno));
     return;
   }
+  ioctl(com[num].fd, TIOCGSERIAL, &ser_info);
+  ser_info.flags |= ASYNC_LOW_LATENCY;
+  ioctl(com[num].fd, TIOCSSERIAL, &ser_info);
+
   s_printf("SER%d: LCR = 0x%x, ",num,com[num].LCR);
 
   /* Set the word size */
@@ -394,7 +426,7 @@ void ser_termios(int num)
   /* The following does the actual system calls to set the line parameters */
   cfsetispeed(&com[num].newset, baud);
   cfsetospeed(&com[num].newset, baud);
-  tcsetattr(com[num].fd, TCSANOW, &com[num].newset);
+  tcsetattr(com[num].fd, TCSADRAIN, &com[num].newset);
 
 #if 0
   /* Many mouse drivers require this, they detect for Framing Errors
@@ -423,7 +455,7 @@ void ser_termios(int num)
  */
 static int get_rx(int num)
 {
-  static int val;
+  int val;
   com[num].rx_timeout = TIMEOUT_RX;		/* Reset timeout counter */
 
   /* If the Received-Data-Ready bit is clear then return a 0
@@ -438,13 +470,15 @@ static int get_rx(int num)
      * buffer is bigger) for compatibility purposes.  Note, that the 
      * following code is optimized for speed rather than compactness.
      */
+    com[num].rx_fifo_bytes = RX_BUF_BYTES(num);
+    if (com[num].rx_fifo_bytes > com[num].rx_fifo_size)
+	      com[num].rx_fifo_bytes = com[num].rx_fifo_size;
      
     /* Is the receive FIFO empty? */
     if (com[num].rx_fifo_bytes == 0) return 0;
 
     val = com[num].rx_buf[com[num].rx_buf_start];	/* Get byte */
     com[num].rx_buf_start++;
-    com[num].rx_buf_bytes--;		/* Bytes waiting in rcv queue */
     com[num].rx_fifo_bytes--;		/* Emulated 16-byte limitation */
  
     /* Did the FIFO become "empty" right now? */
@@ -464,12 +498,11 @@ static int get_rx(int num)
     else if ((com[num].IIR & UART_IIR_ID) == UART_IIR_RDI) {
       
       /* Did the FIFO drop below the trigger level? */
-      if (com[num].rx_buf_bytes < com[num].rx_fifo_trigger) {
+      if (RX_BUF_BYTES(num) < com[num].rx_fifo_trigger) {
         com[num].int_condition &= ~RX_INTR;	/* Clear receive condition */
 
         /* DANG_FIXTHIS Is this safe to put this here? */
-	/* I don't think so  - SS */
-//        serial_int_engine(num, 0);		/* Update interrupt status */
+        serial_int_engine(num, 0);		/* Update interrupt status */
       }
       else {		
         /* No, the FIFO didn't drop below trigger, but clear timeout bit */
@@ -480,27 +513,25 @@ static int get_rx(int num)
   }
   
   /* The following code executes only if in non-FIFO mode */
-  
-  /* Clear data waiting status and interrupt condition flag */
-  com[num].LSR &= ~UART_LSR_DR;
-  com[num].LSRqueued &= ~UART_LSR_DR;
+
+  /* Get byte from internal receive queue */
+  val = com[num].rx_buf[com[num].rx_buf_start];
+  /* Update receive queue pointer and number of chars waiting */
+  if (RX_BUF_BYTES(num)) {
+    com[num].rx_buf_start++;
+  }
   com[num].int_condition &= ~RX_INTR;
-  
+  if (!RX_BUF_BYTES(num)) {
+    /* Clear data waiting status and interrupt condition flag */
+    com[num].LSR &= ~UART_LSR_DR;
+    com[num].LSRqueued &= ~UART_LSR_DR;
+  }
   /* If receive interrupt is set, then update interrupt status */
   if ((com[num].IIR & UART_IIR_ID) == UART_IIR_RDI) {
     /* DANG_FIXTHIS Is this safe to put this here? */
     serial_int_engine(num, 0);			/* Update interrupt stuats */
   }
-  
-  /* Get byte from internal receive queue */
-  val = com[num].rx_buf[com[num].rx_buf_start];
-  
-  /* Update receive queue pointer and number of chars waiting */
-  if (com[num].rx_buf_bytes) {
-    com[num].rx_buf_start++;
-    com[num].rx_buf_bytes--;
-  }
-  
+
   return val;		/* Return received byte */
 }
 
@@ -541,6 +572,7 @@ get_msr(int num)
 {
   int val;
 
+  modstat_engine(num);				/* Get the fresh MSR status */
   val = com[num].MSR;				/* Save old MSR value */
   com[num].MSR &= UART_MSR_STAT; 		/* Clear delta bits */
   com[num].int_condition &= ~MS_INTR;		/* MSI condition satisfied */
@@ -564,7 +596,7 @@ get_msr(int num)
 static int
 get_lsr(int num)
 {
-  static int val;
+  int val;
 
   val = com[num].LSR;			/* Save old LSR value */
   com[num].int_condition &= ~LS_INTR;	/* RLSI int condition satisfied */
@@ -590,13 +622,11 @@ get_lsr(int num)
  */
 static void put_tx(int num, int val)
 {
-  static int rtrn;
-  
+  int rtrn;
+#if 0  
   /* Update the transmit timer */
   com[num].tx_timer += com[num].tx_char_time;
-
-  com[num].TX = val;			/* Mainly used in overflow cases */
-  com[num].tx_trigger = 1;		/* Time to trigger next tx int */
+#endif
   com[num].int_condition &= ~TX_INTR;	/* TX interrupt condition satisifed */
   com[num].LSR &= ~UART_LSR_TEMT;	/* TEMT not empty */
 
@@ -604,6 +634,9 @@ static void put_tx(int num, int val)
   if ((com[num].IIR & UART_IIR_ID) == UART_IIR_THRI)
     serial_int_engine(num, 0);		/* Update interrupt status */
 
+  com[num].TX = val;			/* Mainly used in overflow cases */
+  com[num].tx_trigger = 1;		/* Time to trigger next tx int */
+
   /* Loop-back writes.  Parity is currently not calculated.  No
    * UART diagnostics programs including COMTEST.EXE, that I tried,
    * complained about parity problems during loopback tests anyway!
@@ -619,7 +652,7 @@ static void put_tx(int num, int val)
     if (com[num].fifo_enable) {		/* Is it in FIFO mode? */
     
       /* Is the FIFO full? */
-      if (com[num].rx_buf_bytes >= com[num].rx_fifo_size) {
+      if (RX_BUF_BYTES(num) >= com[num].rx_fifo_size) {
         if(s3_printf) s_printf("SER%d: Func put_tx loopback overrun requesting LS_INTR\n",num);
         com[num].LSR |= UART_LSR_OE;		/* Indicate overrun error */
         com[num].LSRqueued |= UART_LSR_OE;	/* Update queued LSR bits */
@@ -630,14 +663,13 @@ static void put_tx(int num, int val)
         /* Put char into recv FIFO */
         com[num].rx_buf[com[num].rx_buf_end] = val;
         com[num].rx_buf_end++;
-        com[num].rx_buf_bytes++;
-        com[num].rx_fifo_bytes = com[num].rx_buf_bytes;
+        com[num].rx_fifo_bytes = RX_BUF_BYTES(num);
 
 	/* If the buffer touches the top, slide chars to bottom of buffer */
         if ((com[num].rx_buf_end - 1) >= RX_BUFFER_SIZE) rx_buffer_slide(num);
         
         /* Is it the past the receive FIFO trigger level? */
-        if (com[num].rx_buf_bytes >= com[num].rx_fifo_trigger) {
+        if (RX_BUF_BYTES(num) >= com[num].rx_fifo_trigger) {
           com[num].rx_timeout = 0;
           com[num].LSRqueued |= UART_LSR_DR;	/* Flag Data Ready bit */
           if(s3_printf) s_printf("SER%d: Func put_tx loopback requesting RX_INTR\n",num);
@@ -665,39 +697,45 @@ static void put_tx(int num, int val)
     return;
   }
   /* Else, not in loopback mode */
-  
+
   if (com[num].fifo_enable) {			/* Is FIFO enabled? */
-    if (com[num].tx_buf_bytes >= TX_BUFFER_SIZE) {	/* Is FIFO already full? */
+    if (TX_BUF_BYTES(num) >= TX_BUFFER_SIZE) {	/* Is FIFO already full? */
       /* try to write the character directly to the tty, hoping it
        * will be queued there. Hmmm... */
       rtrn = RPT_SYSCALL(write(com[num].fd,&com[num].tx_buf[com[num].tx_buf_start],1));
       if (rtrn != 1) {				/* Did transmit fail? */
         com[num].tx_overflow = 1;		/* Set overflow flag */
       }
-      else {					
+      else {
+        /* char written */
+        com[num].tx_buf_start++;
         /* Squeeze char into FIFO */
+        tx_buffer_slide(num);
         com[num].tx_buf[com[num].tx_buf_end] = val;
-        
         /* Update FIFO queue pointers */
-        com[num].tx_buf_end = (com[num].tx_buf_end + 1) % TX_BUFFER_SIZE;
-        com[num].tx_buf_start = (com[num].tx_buf_start + 1) % TX_BUFFER_SIZE;
-        
+        com[num].tx_buf_end++;
       }
     } 
     else {					/* FIFO not full */
       com[num].tx_buf[com[num].tx_buf_end] = val;  /* Put char into FIFO */
-      com[num].tx_buf_end = (com[num].tx_buf_end +1) % TX_BUFFER_SIZE;
-      com[num].tx_buf_bytes++;
+      com[num].tx_buf_end++;
     } 
   } 
   else { 				/* Not in FIFO mode */
     rtrn = RPT_SYSCALL(write(com[num].fd, &val, 1));   /* Attempt char xmit */
     if (rtrn != 1) 				/* Did transmit fail? */
       com[num].tx_overflow = 1; 		/* Set overflow flag */
+#if 0
+    /* This slows the transfer. There might be better way of avoiding
+     * queueing. */
+    else tcdrain(com[num].fd);
+#endif
   }
   if (!com[num].fifo_enable ||
-       com[num].tx_buf_bytes >= (TX_BUFFER_SIZE-1)) 	/* Is FIFO full? */
+       TX_BUF_BYTES(num) >= (TX_BUFFER_SIZE-1)) 	/* Is FIFO full? */
       com[num].LSR &= ~UART_LSR_THRE;		/* THR full */
+
+  transmit_engine(num);
 }
 
 
@@ -762,7 +800,10 @@ put_fcr(int num, int val)
 static void
 put_lcr(int num, int val)
 {
-  com[num].LCR = val;			/* Set new LCR value */
+  int changed = com[num].LCR ^ val;    /* bitmask of changed bits */
+  
+  com[num].LCR = val;                  /* Set new LCR value */
+  
   if (val & UART_LCR_DLAB) {		/* Is Baudrate Divisor Latch set? */
     s_printf("SER%d: LCR = 0x%x, DLAB high.\n", num, val);
     com[num].DLAB = 1;			/* Baudrate Divisor Latch flag */
@@ -778,6 +819,18 @@ put_lcr(int num, int val)
     com[num].DLAB = 0;			/* Baudrate Divisor Latch flag */
     ser_termios(num);			/* Sets new line settings */
   }
+
+  if (changed & UART_LCR_SBC) {
+    /* there is change of break state */
+    if (val & UART_LCR_SBC) {
+      s_printf("SER%d: Setting BREAK state.\n", num);
+      tcdrain(com[num].fd);
+      ioctl(com[num].fd, TIOCSBRK);
+    } else {
+      s_printf("SER%d: Clearing BREAK state.\n", num);
+      ioctl(com[num].fd, TIOCCBRK);
+    }
+  }
 }
 
 
@@ -787,16 +840,17 @@ put_lcr(int num, int val)
 static void
 put_mcr(int num, int val)
 {
-  static int newmsr, delta;
-  static int changed;
-  static int control;
+  int newmsr, delta;
+  int changed;
+  int control;
   changed = com[num].MCR ^ val;			/* Bitmask of changed bits */
   com[num].MCR = val & UART_MCR_VALID;		/* Set valid bits for MCR */
 
   if (val & UART_MCR_LOOP) {		/* Is Loopback Mode set? */
     /* If loopback just enabled, clear FIFO and turn off DTR & RTS on line */
     if (changed & UART_MCR_LOOP) {		/* Was loopback just set? */
-      uart_clear_fifo(num,UART_FCR_CLEAR_CMD);	/* Clear FIFO's */
+      if (com[num].fifo_enable)
+        uart_clear_fifo(num,UART_FCR_CLEAR_CMD);	/* Clear FIFO's */
       control = TIOCM_DTR | TIOCM_RTS;		/* DTR and RTS to be cleared */
       ioctl(com[num].fd, TIOCMBIC, &control);	/* Clear DTR and RTS */
     }
@@ -824,10 +878,17 @@ put_mcr(int num, int val)
   }
   else {				/* It's not in Loopback Mode */
     /* Was loopback mode just turned off now?  Then reset FIFO */
-    if (changed & UART_MCR_LOOP) uart_clear_fifo(num,UART_FCR_CLEAR_CMD);
+    if (changed & UART_MCR_LOOP) {
+      if (com[num].fifo_enable)
+        uart_clear_fifo(num,UART_FCR_CLEAR_CMD);
+    }
 
     /* Set interrupt enable flag according to OUT2 bit in MCR */
     com[num].int_enab = (val & UART_MCR_OUT2) ? 1 : 0;
+    if (com[num].int_enab) {
+      if(s3_printf) s_printf("SER%d: Update interrupt status after MCR update\n",num);
+      serial_int_engine(num, 0);	/* Update interrupt status */
+    }
 
     /* Force RTS & DTR reinitialization if the loopback state has changed */
     if (UART_MCR_LOOP) changed |= UART_MCR_RTS | UART_MCR_DTR;
@@ -932,7 +993,16 @@ put_msr(int num, int val)
 int
 do_serial_in(int num, ioport_t address)
 {
-  static int val;
+  int val;
+
+  /* delayed open happens here */
+  if (com[num].fd == -1) {
+    if (ser_open(num) >= 0)
+      ser_set_params(num);
+  }
+  if (com[num].fd < 0)
+    return 0;
+
   switch (address - com[num].base_port) {
   case UART_RX:		/* Read from Received Byte Register */	
 /*case UART_DLL:*/      /* or Read from Baudrate Divisor Latch LSB */
@@ -1001,6 +1071,9 @@ do_serial_in(int num, ioport_t address)
     val = 0;
     break;
   }
+
+  serial_run();		/* See if some work is to be done */
+
   return val;
 }
 
@@ -1015,6 +1088,15 @@ do_serial_in(int num, ioport_t address)
 int
 do_serial_out(int num, ioport_t address, int val)
 {
+
+  /* delayed open happens here */
+  if (com[num].fd == -1) {
+    if (ser_open(num) >= 0)
+      ser_set_params(num);
+  }
+  if (com[num].fd < 0)
+    return 0;
+
   switch (address - com[num].base_port) {
   case UART_TX:		/* Write to Transmit Holding Register */
 /*case UART_DLL:*/	/* or write to Baudrate Divisor Latch LSB */
@@ -1045,7 +1127,7 @@ do_serial_out(int num, ioport_t address, int val)
         com[num].tx_trigger = 1;
       }
       com[num].IER = (val & 0xF);	/* Write to IER */
-      if(s1_printf) s_printf("SER%d: IER = 0x%x\n", num, val);
+      if(s1_printf) s_printf("SER%d: Write IER = 0x%x\n", num, val);
       serial_int_engine(num, 0);		/* Update interrupt status */
     }
     break;
@@ -1085,5 +1167,8 @@ do_serial_out(int num, ioport_t address, int val)
              address,num);
     break;
   }
+
+  serial_run();		/* See if some work is to be done */
+
   return 0;
 }
