@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* 
@@ -31,25 +31,24 @@ Modified by O.V.Zhirov, July 1998
 */
 
 
-
 #ifdef DOSEMU
+#include "config.h"
+#include "utilities.h"
 #include "mangle.h"
+#include "mfs.h"
+#include "dos2linux.h"
 #include "emu.h"
+#include "translate.h"
+#include <wctype.h>
 #else
 #include "includes.h"
 #include "loadparm.h"
 #endif
 
-extern int DEBUGLEVEL;
-extern int case_default;
-extern BOOL case_mangle;
-
-extern int codepage;
-
 /****************************************************************************
 provide a checksum on a string
 ****************************************************************************/
-int str_checksum(char *s)
+static int str_checksum(char *s)
 {
   int res = 0;
   int c;
@@ -70,8 +69,6 @@ the name is either a full Unix name or an 8 character candidate
 unsigned long is_dos_device(const char *fname)
 {
   char *p;
-  extern unsigned char *lol;
-  extern int lol_nuldev_off;
   unsigned char *dev;
   unsigned long devfar;
   int i;
@@ -112,6 +109,9 @@ unsigned long is_dos_device(const char *fname)
     for (i = 0; i < 8; i++)
     {
       char c1 = fname[i];
+      /* hopefully no device names with cyrillic characters exist ... */
+      if ((unsigned char)c1 >= 128) break;
+      if ((unsigned char)dev[0xa + i] >= 128) break;
       if (c1 == '.' || c1 == '\0')
       {
         /* check if remainder of device name consists of spaces or nulls */
@@ -123,7 +123,7 @@ unsigned long is_dos_device(const char *fname)
         }
         break;
       }
-      if (toupperDOS(c1) != toupperDOS(dev[0xa + i]))
+      if (toupper(c1) != toupper(dev[0xa + i]))
         break;
     }
     if (i == 8)
@@ -131,40 +131,15 @@ unsigned long is_dos_device(const char *fname)
     if (dev[0] == 0xff || dev[1] == 0xff)
       return 0;
     memcpy(&devfar, dev, 4);
-    dev = (char *)((((dev[3] << 8) | dev[2]) << 4) + ((dev[1] << 8) | dev[0]));
+    dev = MK_FP32((dev[3] << 8) | dev[2], (dev[1] << 8) | dev[0]);
   } while (1);
 }
 
 
-BOOL valid_dos_char[256];
-
-#define VALID_DOS_PCHAR(p) (valid_dos_char[*(unsigned char *)(p)])
-
-/****************************************************************************
-initialise the valid dos char array
-****************************************************************************/
-void valid_initialise(void)
-{
-  static BOOL initialised = False;    
-  int i;
-
-  if (initialised)
-    return;
-
-  for (i=0;i<256;i++)
-    valid_dos_char[i] = is_valid_DOS_char(i);
-
-#ifdef DOSEMU
-  valid_dos_char['?'] = True;
-#endif
-  
-  initialised = True;
-}
-
 /****************************************************************************
 return True if a name is in 8.3 dos format
 ****************************************************************************/
-BOOL is_8_3(char *fname)
+static BOOL is_8_3(char *fname)
 {
   int len = strlen(fname);
   char *dot_pos = strchr(fname,'.');
@@ -187,7 +162,6 @@ BOOL is_8_3(char *fname)
 
   {
     char *p = fname;
-    valid_initialise();
 
 #ifdef KANJI
     dot_pos = 0;
@@ -262,7 +236,7 @@ int mangled_stack_len = 0;
 /****************************************************************************
 create the mangled stack
 ****************************************************************************/
-void create_mangled_stack(int size)
+static void create_mangled_stack(int size)
 {
   if (mangled_stack)
     {
@@ -278,7 +252,7 @@ void create_mangled_stack(int size)
 /****************************************************************************
 push a mangled name onto the stack
 ****************************************************************************/
-void push_mangled_name(char *s)
+static void push_mangled_name(char *s)
 {
   int i;
   char *p;
@@ -298,12 +272,12 @@ void push_mangled_name(char *s)
       }
 
   safe_memcpy(mangled_stack[1],mangled_stack[0],
-	      sizeof(fstring)*MIN(mangled_stack_len,mangled_stack_size-1));
+	      sizeof(fstring)*min(mangled_stack_len,mangled_stack_size-1));
   strcpy(mangled_stack[0],s);
   p = strrchr(mangled_stack[0],'.');
   if (p && (!strhasupperDOS(p+1)) && (strlen(p+1) < 4))
     *p = 0;
-  mangled_stack_len = MIN(mangled_stack_size,mangled_stack_len+1);
+  mangled_stack_len = min(mangled_stack_size,mangled_stack_len+1);
 }
 
 /****************************************************************************
@@ -323,7 +297,6 @@ BOOL check_mangled_stack(char *s, char *MangledMap)
     {
       check_extension = True;
       StrnCpy(extension,p,4);
-      strlowerDOS(extension); /* XXXXXXX */
     }
 
   for (i=0;i<mangled_stack_len;i++)
@@ -406,14 +379,14 @@ void mangle_name_83(char *s, char *MangledMap)
   int baselen = 0;
   int extlen = 0;
 
-  valid_initialise();
-
+#ifndef DOSEMU
   if (MangledMap && *MangledMap) {
     if (do_fwd_mangled_map(s, MangledMap))
       return;
   }
+#endif
 
-  p = strrchr(s,'.');  
+  p = strrchr(s,'.');
   if (p && (strlen(p+1)<4) )
     {
       BOOL all_normal = (strisnormal(p+1)); /* XXXXXXXXX */
@@ -424,9 +397,6 @@ void mangle_name_83(char *s, char *MangledMap)
 	  *p = '.';
 	}
     }
-      
-
-  strupperDOS(s);
 
   DEBUG(5,("Mangling name %s to ",s));
 
@@ -470,38 +440,64 @@ void mangle_name_83(char *s, char *MangledMap)
 }
 
 
+/****************************************************************************
+convert a filename from a UNIX to a DOS character set.
+****************************************************************************/
+
+BOOL name_ufs_to_dos(char *dest, const char *src)
+{
+  mbstate_t unix_state;
+
+  int retval = 1;
+
+  wchar_t symbol;
+  size_t slen = strlen(src), result;
+  wchar_t wdst[slen + 1], *wdest = wdst;
+
+  memset(&unix_state, 0, sizeof unix_state);
+
+  *dest = '\0';
+  result = mbsrtowcs(wdest, &src, slen + 1, &unix_state);
+  if (result == (size_t) -1)
+    return 0;
+
+  while (*wdest) {
+    symbol = *wdest++;
+    if (symbol > 0xffff)
+      symbol = '_';
+    *dest = unicode_to_dos_table[symbol];
+    if (!VALID_DOS_PCHAR(dest) && strchr(" +,;=[]",*dest)==0)
+      retval = 0;
+    dest++;
+  }
+  *dest = '\0';
+  return retval;
+}
+
 /****************************************************************************
 convert a filename to 8.3 format. return True if successful.
 ****************************************************************************/
-BOOL name_convert(char *OutName,char *InName,BOOL mangle, char *MangledMap)
+BOOL name_convert(char *Name, BOOL mangle)
 {
-  /* initially just copy it */
-#ifdef KANJI
-  strcpy(OutName, kj_dos_format (InName, False));
-#else
-  strcpy(OutName,InName);
-#endif
-
   /* check if it's already in 8.3 format */
-  if (is_8_3(OutName))
+  if (is_8_3(Name))
     return(True);
 
   if (!mangle)
     return(False);
 
-  DEBUG(5,("Converted name %s ",OutName));
+  DEBUG(5,("Converted name %s",Name));
 
   /* mangle it into 8.3 */
-  push_mangled_name(OutName);  
-  mangle_name_83(OutName, MangledMap);
-
-  DEBUG(5,("to %s\n",OutName));
+  push_mangled_name(Name);
+  mangle_name_83(Name, NULL);
 
+  DEBUG(5,("to %s\n",Name));
   
   return(True);
 }
 
-
+#ifndef DOSEMU
 static char *mangled_match(char *s, /* This is null terminated */
                            char *pattern, /* This isn't. */
                            int len) /* This is the length of pattern. */
@@ -649,7 +645,6 @@ BOOL do_fwd_mangled_map(char *s, char *MangledMap)
 /******************************************************/
 
 
-#ifndef DOSEMU
 static BOOL do_rev_mangled_map(char *s, char *MangledMap)
 {
   /* Does the reverse of do_fwd_mangled_map().  Look in there for
