@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* header file to make mangle.c fit with dosemu 
@@ -21,7 +21,6 @@ Modified by O.V.Zhirov, July 1998
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
-#include <malloc.h>
 #include <string.h>
 #include <stddef.h>
 #include <ctype.h>
@@ -31,13 +30,6 @@ Modified by O.V.Zhirov, July 1998
 #define DEBUG(level,message)
 #define PTR_DIFF(p1,p2) ((ptrdiff_t)(((char *)(p1)) - (char *)(p2)))
 
-#ifndef MIN
-#define MIN(a,b) ((a)<(b)?(a):(b))
-#endif
-#ifndef MAX
-#define MAX(a,b) ((a)>(b)?(a):(b))
-#endif
-
 #define strnorm(s) strlowerDOS(s)
 #define strisnormal(s) (!strhasupperDOS(s))
 
@@ -61,16 +53,22 @@ typedef char pstring[1024];
 
 
 /* prototypes */
-unsigned long is_dos_device(const char *fname);
-void mangle_name_83(char *s, char *MangledMap);
+extern unsigned long is_dos_device(const char *fname);
+extern void mangle_name_83(char *s, char *MangledMap);
+extern BOOL name_ufs_to_dos(char *dest, const char *src);
 extern BOOL do_fwd_mangled_map(char *s, char *MangledMap);
-extern BOOL name_convert(char *OutName,char *InName,BOOL mangle, char *MangledMap);
+extern BOOL name_convert(char *Name,BOOL mangle);
 extern BOOL is_mangled(char *s);
 extern BOOL check_mangled_stack(char *s, char *MangledMap);
 
 /* prototypes, found in util.c */
+void init_all_DOS_tables(void);
+extern unsigned char unicode_to_dos_table[0x10000];
+extern unsigned short dos_to_unicode_table[0x100];
+extern unsigned char upperDOS_table[0x100];
+
 BOOL isupperDOS(int c);
-int  toupperDOS(int c);
+#define toupperDOS(c) (upperDOS_table[(unsigned char)(c)])
 BOOL islowerDOS(int c);
 int  tolowerDOS(int c);
 void strupperDOS(char *s);
@@ -88,9 +86,14 @@ int strcasecmpDOS(char *s1, char *s2);
 
 char *StrnCpy(char *dest,const char *src,int n);
 void array_promote(char *array,int elsize,int element);
+BOOL strequalDOS(const char *s1, const char *s2);
 BOOL strequal(char *s1,char *s2);
 
 
+extern BOOL valid_dos_char[256];
+
+#define VALID_DOS_PCHAR(p) (valid_dos_char[*(unsigned char *)(p)])
+
 
 #ifndef MANGLE
 #define MANGLE 1
