@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -11,11 +11,11 @@ emulator.
 Andrew.Tridgell@anu.edu.au 30th March 1993
 */
 
-#ifdef DOSEMU
+#include <sys/stat.h>
+#include <dirent.h>
+#include <utime.h>
+
 /* definitions to make mach emu code compatible with dosemu */
-#ifdef __linux__
-#include <sys/vfs.h>
-#endif
 #include "emu.h"
 
 #if 0
@@ -28,7 +28,7 @@ typedef unsigned char boolean_t;
 #endif
 
 #ifndef MAX_DRIVE
-#define MAX_DRIVE 26
+#define MAX_DRIVE 33
 #endif
 
 #define USE_DF_AND_AFS_STUFF
@@ -67,7 +67,6 @@ typedef struct vm86_regs state_t;
 #define SETHIGH(x,y) 	(*(x) = (*(x) & ~0xff00) | ((MASK8(y))<<8))
 #define SETLOW(x,y) 	(*(x) = (*(x) & ~0xff) | (MASK8(y)))
 #define SETWORD(x,y)	(*(x) = (*(x) & ~0xffff) | (MASK16(y)))
-#endif
 
 /*
  * Copyright (c) 1991 Carnegie Mellon University
@@ -99,7 +98,11 @@ typedef struct vm86_regs state_t;
  *	V86 DOS disk emulation header file
  *
  * HISTORY:
- * $Log: mfs.h,v $
+ *
+ * Log: mfs.h,v
+ * Revision 1.1  2003/06/23 00:02:08  bartoldeman
+ * Initial revision
+ *
  * Revision 1.2  1995/05/23  06:04:49  root
  * fix for redirector open existing file function
  *
@@ -249,12 +252,19 @@ typedef struct vm86_regs state_t;
 struct dir_ent {
   char name[8];			/* dos name and ext */
   char ext[3];
+  char d_name[256];             /* unix name as in readdir */
   u_short mode;			/* unix st_mode value */
   u_short hidden;
   u_short long_path;            /* directory has long path */
-  long size;			/* size of file */
+  int size;			/* size of file */
   time_t time;			/* st_mtime */
-  struct dir_ent *next;
+  int attr;
+};
+
+struct dir_list {
+  int nr_entries;
+  int size;
+  struct dir_ent *de;
 };
 
 struct dos_name {
@@ -262,67 +272,19 @@ struct dos_name {
   char ext[3];
 };
 
-typedef struct far_record {
-  u_short offset;
-  u_short segment;
-} far_t;
+struct mfs_dirent
+{
+  char *d_name;
+  char *d_long_name;
+};
 
-#define DOSVER_31_33	1
-#define DOSVER_41	2
-#define DOSVER_50	3
-#define DOSVER_60	4
-
-typedef u_char *sdb_t;
-
-#define sdb_drive_letter(sdb)	(*(u_char  *)&sdb[sdb_drive_letter_off])
-#define sdb_template_name(sdb)	((u_char   *)&sdb[sdb_template_name_off])
-#define sdb_template_ext(sdb)	((u_char   *)&sdb[sdb_template_ext_off])
-#define	sdb_attribute(sdb)	(*(u_char  *)&sdb[sdb_attribute_off])
-#define sdb_dir_entry(sdb)	(*(u_short *)&sdb[sdb_dir_entry_off])
-#define sdb_p_cluster(sdb)	(*(u_short *)&sdb[sdb_p_cluster_off])
-#define	sdb_file_name(sdb)	((char     *)&sdb[sdb_file_name_off])
-#define	sdb_file_ext(sdb)	((char     *)&sdb[sdb_file_ext_off])
-#define	sdb_file_attr(sdb)	(*(u_char  *)&sdb[sdb_file_attr_off])
-#define	sdb_file_time(sdb)	(*(u_short *)&sdb[sdb_file_time_off])
-#define	sdb_file_date(sdb)	(*(u_short *)&sdb[sdb_file_date_off])
-#define sdb_file_st_cluster(sdb)(*(u_short *)&sdb[sdb_file_st_cluster_off])
-#define sdb_file_size(sdb)	(*(u_long  *)&sdb[sdb_file_size_off])
-
-typedef u_char *sft_t;
-
-#define sft_handle_cnt(sft) 	(*(u_short *)&sft[sft_handle_cnt_off])
-#define sft_open_mode(sft)  	(*(u_short *)&sft[sft_open_mode_off])
-#define sft_attribute_byte(sft) (*(u_char  *)&sft[sft_attribute_byte_off])
-#define sft_device_info(sft)  	(*(u_short *)&sft[sft_device_info_off])
-#define	sft_dev_drive_ptr(sft)	(*(u_long  *)&sft[sft_dev_drive_ptr_off])
-#define	sft_start_cluster(sft)	(*(u_short *)&sft[sft_start_cluster_off])
-#define	sft_time(sft)		(*(u_short *)&sft[sft_time_off])
-#define	sft_date(sft)		(*(u_short *)&sft[sft_date_off])
-#define	sft_size(sft)		(*(u_long  *)&sft[sft_size_off])
-#define	sft_position(sft)	(*(u_long  *)&sft[sft_position_off])
-#define sft_rel_cluster(sft)	(*(u_short *)&sft[sft_rel_cluster_off])
-#define sft_abs_cluster(sft)	(*(u_short *)&sft[sft_abs_cluster_off])
-#define	sft_directory_sector(sft) (*(u_short *)&sft[sft_directory_sector_off])
-#define	sft_directory_entry(sft)  (*(u_char  *)&sft[sft_directory_entry_off])
-#define	sft_name(sft)		( (char    *)&sft[sft_name_off])
-#define	sft_ext(sft)		( (char    *)&sft[sft_ext_off])
-
-#define	sft_fd(sft)		(*(u_char *)&sft[sft_fd_off])
-
-typedef u_char *cds_t;
-
-#define	cds_current_path(cds)	((char	   *)&cds[cds_current_path_off])
-#define	cds_flags(cds)		(*(u_short *)&cds[cds_flags_off])
-#define cds_DBP_pointer(cds)	(*(far_t *)&cds[cds_DBP_pointer_off])
-#define cds_cur_cluster(cds)	(*(u_short *)&cds[cds_cur_cluster_off])
-#define	cds_rootlen(cds)	(*(u_short *)&cds[cds_rootlen_off])
-#define drive_cds(dd) ((cds_t)(((int)cds_base)+(cds_record_size*(dd))))
-
-#define CDS_FLAG_REMOTE		0x8000
-#define CDS_FLAG_READY		0x4000
-#define CDS_FLAG_NOTNET		0x0080
-#define CDS_FLAG_SUBST		0x1000
-#define CDS_DEFAULT_ROOT_LEN	2
+struct mfs_dir
+{
+  DIR *dir;
+  struct mfs_dirent de;
+  int fd;
+  unsigned int nr;
+};
 
 #define FAR(x) (Addr_8086(x.segment, x.offset))
 #define FARPTR(x) (Addr_8086((x)->segment, (x)->offset))
@@ -331,45 +293,13 @@ typedef u_short *psp_t;
 
 #define PSPPTR(x) (Addr_8086(x, 0))
 
-typedef u_char *sda_t;
-
-#define	sda_current_dta(sda)	((char *)(FARPTR((far_t *)&sda[sda_current_dta_off])))
-#define sda_cur_psp(sda)		(*(u_short *)&sda[sda_cur_psp_off])
-#define sda_cur_drive(sda)		(*(u_char *)&sda[sda_cur_drive_off])
-#define sda_filename1(sda)		((char  *)&sda[sda_filename1_off])
-#define	sda_filename2(sda)		((char  *)&sda[sda_filename2_off])
-#define sda_sdb(sda)			((sdb_t    )&sda[sda_sdb_off])
-#define	sda_cds(sda)		((cds_t)(FARPTR((far_t *)&sda[sda_cds_off])))
-#define sda_search_attribute(sda)	(*(u_char *)&sda[sda_search_attribute_off])
-#define sda_open_mode(sda)		(*(u_char *)&sda[sda_open_mode_off])
-#define sda_rename_source(sda)		((sdb_t    )&sda[sda_rename_source_off])
-#define sda_user_stack(sda)		((char *)(FARPTR((far_t *)&sda[sda_user_stack_off])))
-
-/*
- *  Data for extended open/create operations, DOS 4 or greater:
- */
-#define sda_ext_act(sda)		(*(u_short *)&sda[sda_ext_act_off])
-#define sda_ext_attr(sda)		(*(u_short *)&sda[sda_ext_attr_off])
-#define sda_ext_mode(sda)		(*(u_short *)&sda[sda_ext_mode_off])
-
-#define psp_parent_psp(psp)		(*(u_short *)&psp[0x16])
-#define psp_handles(psp)		((char *)(FARPTR((far_t *)&psp[0x34])))
-
-#define lol_cdsfarptr(lol)		(*(far_t *)&lol[lol_cdsfarptr_off])
-#define lol_last_drive(lol)		(*(u_char *)&lol[lol_last_drive_off])
-#define lol_nuldev(lol)		        (&lol[lol_nuldev_off])
-
-typedef u_char *lol_t;
-
-#ifdef OLD_OBSOLETE
-typedef struct lol_record {
-  u_char filler1[22];
-  far_t cdsfarptr;
-  u_char filler2[6];
-  u_char last_drive;
-} *lol_t;
-
-#endif
+struct drive_info 
+{
+  char *root;
+  int root_len;
+  boolean_t read_only;
+};
+extern struct drive_info drives[MAX_DRIVE];
 
 /* dos attribute byte flags */
 #define REGULAR_FILE 	0x00
@@ -405,6 +335,40 @@ typedef struct lol_record {
 #define CANCEL_REDIRECTION 4
 #define EXTENDED_GET_REDIRECTION 5
 
-#ifdef X86_EMULATOR
-extern int e_dos_read(int fd, char *data, int cnt);
-#endif
+
+/* #define MAX_PATH_LENGTH 57 */
+/* 2001/01/05 Manfred Scherer
+ * With the value 57 I can create pathlength until 54.
+ * In native DOS I can create pathlength until 63.
+ * With the value 66 it should be possible to create
+ * paths with length 63.
+ * I've tested it on my own system, and I found the value 66
+ * is right for me.
+ */
+
+#define MAX_PATH_LENGTH 66
+
+
+extern int build_ufs_path_(char *ufs, const char *path, int drive,
+                           int lowercase);
+extern boolean_t find_file(char *fpath, struct stat *st, int drive);
+extern boolean_t is_hidden(char *fname);
+extern int get_dos_attr(const char *fname,int mode,boolean_t hidden);
+extern int set_fat_attr(int fd,int attr);
+extern int set_dos_attr(char *fname,int mode,int attr);
+extern int dos_utime(char *fpath, struct utimbuf *ut);
+extern int get_unix_attr(int mode, int attr);
+extern void time_to_dos(time_t clock, u_short *date, u_short *time);
+extern time_t time_to_unix(u_short dos_date, u_short dos_time);
+extern void extract_filename(const char *filestring0, char *name, char *ext);
+extern struct mfs_dir *dos_opendir(const char *name);
+extern struct mfs_dirent *dos_readdir(struct mfs_dir *);
+extern int dos_closedir(struct mfs_dir *dir);
+extern void get_volume_label(char *fname, char *fext, char *lfn, int drive);
+extern int dos_rename(const char *filename1, const char *filename2, int drive, int lfn);
+extern int dos_mkdir(const char *filename, int drive, int lfn);
+extern int dos_rmdir(const char *filename, int drive, int lfn);
+
+extern void register_cdrom(int drive, int device);
+extern void unregister_cdrom(int drive);
+extern int get_volume_label_cdrom(int drive, char *name);
