@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* xms.c for the DOS emulator
@@ -33,18 +33,21 @@
 #include "config.h"
 #include "memory.h"
 #include "xms.h"
-#include "dosio.h"
+#include "hlt.h"
+#include "int.h"
+#include "hma.h"
 #include "machcompat.h"
+#include "dos2linux.h"
+#include "cpu-emu.h"
+#include "smalloc.h"
 
 #undef  DEBUG_XMS
 
 static int umb_find_unused(void);
-/* 128*1024 is the amount of memory currently reserved in dos.c above
- * the 1 MEG mark.  ugly.  fix this.
- */
+static void xms_control(void);
 
 /*
-static char RCSxms[] = "$Header: /usr/src/dosemu0.60/dosemu/RCS/xms.c,v 2.11 1995/05/06 16:25:30 root Exp root $";
+static char RCSxms[] = "$Id: xms.c 1769 2007-05-04 05:59:48Z bartoldeman $";
 */
 
 #define	 XMS_GET_VERSION		0x00
@@ -71,18 +74,14 @@ static char RCSxms[] = "$Header: /usr/src/dosemu0.60/dosemu/RCS/xms.c,v 2.11 199
 #define XMS_OUT_OF_SPACE		0xa0
 #define XMS_INVALID_HANDLE		0xa2
 
-
-int a20 = 0;
-int freeHMA = 1;		/* is HMA free? */
+static int a20_local, a20_global, freeHMA;	/* is HMA free? */
 
 static struct Handle handles[NUM_HANDLES + 1];
 static int handle_count = 0;
 
-static int xms_grab_int15again = 0;
-
 struct EMM get_emm(unsigned int, unsigned int);
 void show_emm(struct EMM);
-static void xms_query_freemem(int), xms_allocate_EMB(int), xms_free_EMB(void),
+static unsigned char xms_query_freemem(int), xms_allocate_EMB(int), xms_free_EMB(void),
  xms_move_EMB(void), xms_lock_EMB(int), xms_EMB_info(int), xms_realloc_EMB(int);
 
 static int FindFreeHandle(int);
@@ -107,7 +106,8 @@ static int FindFreeHandle(int);
 #else
 /*  EMS page frame now is configurable via /etc/dosemu.conf */
 #define SPACE_64K   0x10000
-#define IN_EMM_SPACE(addr) (config.ems_size && (int)(addr) >= EMM_BASE_ADDRESS \
+#define IN_EMM_SPACE(addr) ((config.ems_size || config.pm_dos_api) \
+			    && (int)(addr) >= EMM_BASE_ADDRESS \
                             && (int)(addr) < (EMM_BASE_ADDRESS+SPACE_64K))
 
 #define IN_HARDWARE_PAGES(addr) ( config.must_spare_hardware_ram\
@@ -136,116 +136,29 @@ static struct umb_record {
 #define Debug2(args)		x_Stub args
 /* #define dbg_fd stderr */
 
-static boolean_t
-umb_memory_empty(addr, size)
-     vm_address_t addr;
-     int size;
-{
-  int i;
-  int *memory = (int *) addr;
-
-  for (i = 0; i < (size / 4); i++) {
-    if ((memory[i] != 0xffffffff) && (memory[i] != 0)) {
-      return FALSE;
-    }
-  }
-#ifdef DEBUG_XMS
-  Debug0((dbg_fd, "Found free UMB region: %p\n", (void *) addr));
-#endif
-  return (TRUE);
-}
-
 static int
 umb_setup(void)
 {
   int i;
+  size_t addr_start;
+  int size, umb;
 
   for (i = 0; i < UMBS; i++) {
     umbs[i].in_use = FALSE;
   }
 
-  if (config.max_umb) {
-    int addr_start, size, umb;
+  memcheck_addtype('U', "Upper Memory Block (UMB, XMS 3.0)");
 
-    memcheck_addtype('U', "Upper Memory Block (UMB, XMS 3.0)");
+  addr_start = 0x00000;     /* start address */
+  while ((size = memcheck_findhole(&addr_start, 1024, 0x100000)) != 0) {
+    Debug0((dbg_fd, "findhole - from 0x%5.5zX, %dKb\n", addr_start, size/1024));
+    memcheck_reserve('U', addr_start, size);
 
-    addr_start = 0x00000;     /* start address */
-    while ((size = memcheck_findhole(&addr_start, 1024, 0x100000)) != 0) {
-      Debug0((dbg_fd, "findhole - from 0x%5.5X, %dKb\n", addr_start, size/1024));
-      memcheck_reserve('U', addr_start, size);
-
-      umb = umb_find_unused();
-      umbs[umb].in_use = TRUE;
-      umbs[umb].free = TRUE;
-      umbs[umb].addr = (vm_address_t) addr_start;
-      umbs[umb].size = size;
-    }
-  }
-  else {
-    int umb,tumb;
-    vm_address_t addr;
-
-    if (config.ems_size) {
-      /* set up EMS page frame */
-      umb = umb_find_unused();
-      
-      umbs[umb].in_use = TRUE;
-      umbs[umb].free = FALSE;
-      umbs[umb].addr = (caddr_t) EMM_BASE_ADDRESS;
-      umbs[umb].size = 0x10000;
-    }
-    
-    /* set up DOSEMU trampolines */
     umb = umb_find_unused();
-    tumb=umb;
     umbs[umb].in_use = TRUE;
-    umbs[umb].free = FALSE;
-    umbs[umb].addr = (caddr_t) (BIOSSEG * 16);
-    umbs[umb].size = 0x10000;
-    
-    /* set up 32K VGA BIOS mem if necessary */
-    if (config.mapped_bios) {
-      umb = umb_find_unused();
-      
-      umbs[umb].in_use = TRUE;
-      umbs[umb].free = FALSE;
-      umbs[umb].addr = (caddr_t) VBIOS_START;
-      umbs[umb].size = 0x8000;
-      
-      if (config.vbios_seg == 0xe000) {
-	Debug0((dbg_fd, "Fixing UMB for E000\n"));
-	umb = umb_find_unused();
-	umbs[umb].in_use = TRUE;
-	umbs[umb].free = FALSE;
-	umbs[umb].addr = (caddr_t) 0xc0000;
-	umbs[umb].size = 0x8000;
-      }
-    }
-
-    umb=tumb;
-    for (addr = UMB_BASE; addr < (vm_address_t) (UMB_BASE + UMB_SIZE);
-	 addr += UMB_PAGE) {
-      if (IN_EMM_SPACE(addr) || IN_EMU_SPACE(addr) || IN_HARDWARE_PAGES(addr)) {
-#ifdef DEBUG_XMS
-	Debug0((dbg_fd, "skipped UMB region: %p EMM-%d EMU=%d hardw=%d\n",(void *) addr, 
-		IN_EMM_SPACE(addr),IN_EMU_SPACE(addr),IN_HARDWARE_PAGES(addr)));
-#endif
-	continue;
-      }
-      if (umb_memory_empty(addr, UMB_PAGE)) {
-	if ((umbs[umb].addr + umbs[umb].size) == addr) {
-	  umbs[umb].size += UMB_PAGE;
-	}
-	else {
-	  umb = umb_find_unused();
-	  umbs[umb].in_use = TRUE;
-	  umbs[umb].free = TRUE;
-	  umbs[umb].addr = addr;
-	  umbs[umb].size = UMB_PAGE;
-	  Debug0((dbg_fd, "New UMB region: %p\n", (void *) addr));
-	}
-      }
-    }
+    umbs[umb].free = TRUE;
+    umbs[umb].addr = (vm_address_t)(uintptr_t) addr_start;
+    umbs[umb].size = size;
   }
 
   for (i = 0; i < UMBS; i++) {
@@ -261,7 +174,7 @@ umb_setup(void)
 			     FALSE)),
 		"vm_allocate of umb block.");
 #else
-      Debug0((dbg_fd, "umb_setup: addr %p size 0x%04x\n",
+      Debug0((dbg_fd, "umb_setup: addr %p size 0x%04zx\n",
 	      (void *) addr, size));
 #endif
     }
@@ -286,7 +199,7 @@ static int
 umb_find(vm_address_t segbase)
 {
   int i;
-  vm_address_t addr = (vm_address_t) ((int) segbase * 16);
+  vm_address_t addr = SEG2LINEAR(segbase);
 
   for (i = 0; i < UMBS; i++) {
     if (umbs[i].in_use &&
@@ -359,11 +272,23 @@ static void umb_cleanup(int umb)
     }
   } while (updated);
 }
+
+static void umb_free_all(void)
+{
+  int i;
+  for (i = 0; i < UMBS; i++) {
+    if (umbs[i].in_use && !umbs[i].free) {
+      umbs[i].free = TRUE;
+      umb_cleanup(i);
+    }
+  }
+}
+
 /* why int? */
 static int
 umb_free(int segbase)
 {
-  int umb = umb_find((vm_address_t)segbase);
+  int umb = umb_find((vm_address_t)(uintptr_t)segbase);
 
   if (umb != UMB_NULL) {
     umbs[umb].free = TRUE;
@@ -391,47 +316,123 @@ umb_query(void)
 
 /* end of stuff from Mach */
 
+static smpool mp = SM_EMPTY_POOL;
+
 void
-xms_init(void)
+xms_reset(void)
+{
+  umb_free_all();
+  config.xms_size = 0;
+}
+
+
+void xms_helper(void)
 {
   int i;
 
-  if (!config.xms_size) {
-    CARRY;
-    REG(eax) = 0x8600;
+  if (!config.ext_mem)
     return;
-  }
 
+  LWORD(eax) = 0;	/* report success */
+
+  if (config.xms_size)
+    return;
+
+  config.xms_size = EXTMEM_SIZE >> 10;
   x_printf("XMS: initializing XMS... %d handles\n", NUM_HANDLES);
 
   freeHMA = 1;
-
-  umb_setup();
-
-  a20 = 0;
-  xms_grab_int15 = 0;
+  a20_global = a20_local = 0;
 
   handle_count = 0;
-  for (i = 0; i < NUM_HANDLES + 1; i++)
+  for (i = 0; i < NUM_HANDLES + 1; i++) {
+    if (handles[i].valid && handles[i].addr)
+      smfree(&mp, handles[i].addr);
     handles[i].valid = 0;
+  }
+
+  smdestroy(&mp);
+  sminit(&mp, ext_mem_base, config.xms_size * 1024);
 }
 
 void
-xms_control(void)
+xms_init(void)
 {
-  /* the int15 get memory size call (& block copy, unimplemented) are
-   * supposed to be unrevectored until the first non-version XMS call
-   * to "maintain compatibility with existing device drivers."
-   * hence this blemish.  if Microsoft could act, say, within 3 years of
-   * a problem's first appearance, maybe we wouldn't have these kinds
-   * of compatibility hacks.
-   */
-  if (HI(ax) != 0) {
-    if (!xms_grab_int15again)
-      xms_grab_int15 = 1;
-    xms_grab_int15again = 1;
-  }
+  emu_hlt_t hlt_hdlr;
 
+  umb_setup();
+
+  hlt_hdlr.name = "XMS";
+  hlt_hdlr.start_addr = XMSControl_ADD - BIOS_HLT_BLK;
+  hlt_hdlr.end_addr = hlt_hdlr.start_addr + 5;
+  hlt_hdlr.func = (emu_hlt_func)xms_control;
+  hlt_register_handler(hlt_hdlr);
+}
+
+static void XMS_RET(int err)
+{
+  /* careful here: BL should NOT be set to 0 if there is no error,
+     in particular for "get handle information" that is wrong,
+     but also for locking a handle.
+     An exception is xms_query_freemem, which sets BL to zero
+     itself, and does not use AX to report an error */
+  LWORD(eax) = err ? 0 : 1;
+  if (err)
+    LO(bx) = err;
+}
+
+static void xms_control(void)
+{
+  int is_umb_fn = 0;
+
+ /* First do the UMB functions */
+ switch (HI(ax)) {
+  case XMS_ALLOCATE_UMB:
+    {
+      int size = LWORD(edx) << 4;
+      vm_address_t addr = umb_allocate(size);
+      is_umb_fn = 1;
+      Debug0((dbg_fd, "Allocate UMB memory: %#x\n", size));
+      if (addr == (vm_address_t) 0) {
+        int avail=umb_query();
+
+	Debug0((dbg_fd, "Allocate UMB Failure\n"));
+	XMS_RET(avail ? 0xb0 : 0xb1);
+	LWORD(edx) = avail >> 4;
+      }
+      else {
+	Debug0((dbg_fd, "Allocate UMB Success\n"));
+	LWORD(eax) = 1;
+	LWORD(ebx) = (unsigned long)addr >> 4;
+	LWORD(edx) = size >> 4;
+      }
+      Debug0((dbg_fd, "umb_allocated: %#x0:%#x0\n",
+	      (unsigned) LWORD(ebx),
+	      (unsigned) LWORD(edx)));
+      /* retval = UNCHANGED; */
+      break;
+    }
+
+  case XMS_DEALLOCATE_UMB:
+    {
+      is_umb_fn = 1;
+      umb_free(LWORD(edx));
+      XMS_RET(0);			/* no error */
+      Debug0((dbg_fd, "umb_freed: 0x%04x\n",
+	      (unsigned) LWORD(edx)));
+      /* retval = UNCHANGED; */
+      break;
+    }
+
+  case 0x12:
+    is_umb_fn = 1;
+    x_printf("XMS realloc UMB segment 0x%04x size 0x%04x\n",
+	     LWORD(edx), LWORD(ebx));
+    XMS_RET(0x80);		        /* function unimplemented */
+    break;
+ }
+
+ if (config.xms_size && !is_umb_fn) {
   switch (HI(ax)) {
   case 0:			/* Get XMS Version Number */
     LWORD(eax) = XMS_VERSION;
@@ -446,12 +447,11 @@ xms_control(void)
     if (freeHMA) {
       x_printf("XMS: allocating HMA size 0x%04x\n", LWORD(edx));
       freeHMA = 0;
-      LWORD(eax) = 1;
+      XMS_RET(0);			/* no error */
     }
     else {
       x_printf("XMS: HMA already allocated\n");
-      LWORD(eax) = 0;
-      LO(bx) = 0x91;
+      XMS_RET(0x91);
     }
     break;
 
@@ -459,47 +459,46 @@ xms_control(void)
     x_printf("XMS release HMA\n");
     if (freeHMA) {
       x_printf("XMS: HMA already free\n");
-      LWORD(eax) = 0;
-      LO(bx) = 0x93;
+      XMS_RET(0x93);
     }
     else {
       x_printf("XMS: freeing HMA\n");
-      LWORD(eax) = 1;
       freeHMA = 1;
+      XMS_RET(0);			/* no error */
     }
     break;
 
   case 3:			/* Global Enable A20 */
     x_printf("XMS global enable A20\n");
-    if (!a20)
+    if (!a20_global)
       set_a20(1);		/* map in HMA */
-    a20 = 1;
-    LWORD(eax) = 1;
+    a20_global = 1;
+    XMS_RET(0);			/* no error */
     break;
 
   case 4:			/* Global Disable A20 */
     x_printf("XMS global disable A20\n");
-    if (a20)
+    if (a20_global)
       set_a20(0);
-    a20 = 0;
-    LWORD(eax) = 1;
+    a20_global = 0;
+    XMS_RET(0);			/* no error */
     break;
 
   case 5:			/* Local Enable A20 */
     x_printf("XMS LOCAL enable A20\n");
-    a20++;
-    if (a20 == 1)
+    if (!a20_local)
       set_a20(1);
-    LWORD(eax) = 1;
+    a20_local++;
+    XMS_RET(0);			/* no error */
     break;
 
   case 6:			/* Local Disable A20 */
     x_printf("XMS LOCAL disable A20\n");
-    if (a20)
-      a20--;
-    if (a20 == 0)
+    if (a20_local)
+      a20_local--;
+    if (!a20_local)
       set_a20(0);
-    LWORD(eax) = 1;
+    XMS_RET(0);			/* no error */
     break;
 
     /* DOS, if loaded as "dos=high,umb" will continually poll
@@ -515,77 +514,31 @@ xms_control(void)
     break;
 
   case 9:			/* Allocate Extended Memory Block */
-    xms_allocate_EMB(OLDXMS);
+    XMS_RET(xms_allocate_EMB(OLDXMS));
     break;
 
   case 0xa:			/* Free Extended Memory Block */
-    xms_free_EMB();
+    XMS_RET(xms_free_EMB());
     break;
 
   case 0xb:			/* Move Extended Memory Block */
-    xms_move_EMB();
+    XMS_RET(xms_move_EMB());
     break;
 
   case 0xc:			/* Lock Extended Memory Block */
-    xms_lock_EMB(1);
+    XMS_RET(xms_lock_EMB(1));
     break;
 
   case 0xd:			/* Unlock Extended Memory Block */
-    xms_lock_EMB(0);
+    XMS_RET(xms_lock_EMB(0));
     break;
 
   case 0xe:			/* Get EMB Handle Information */
-    xms_EMB_info(OLDXMS);
-    return;
-
-  case 0xf:			/* Reallocate Extended Memory Block */
-    xms_realloc_EMB(OLDXMS);
+    XMS_RET(xms_EMB_info(OLDXMS));
     break;
 
-#define state (&_regs)
-  case XMS_ALLOCATE_UMB:
-    {
-      int size = WORD(state->edx) << 4;
-      vm_address_t addr = umb_allocate(size);
-
-      Debug0((dbg_fd, "Allocate UMB memory: %#x\n", size));
-      if (addr == (vm_address_t) 0) {
-        int avail=umb_query();
-
-	Debug0((dbg_fd, "Allocate UMB Failure\n"));
-	SETWORD(&(state->eax), 0);
-	if(avail) SETLOW(&(state->ebx), 0xb0);
-	else SETLOW(&(state->ebx), 0xb1);
-	SETWORD(&(state->edx), avail >> 4);
-      }
-      else {
-	Debug0((dbg_fd, "Allocate UMB Success\n"));
-	SETWORD(&(state->eax), 1);
-	SETWORD(&(state->ebx), (int) addr >> 4);
-	SETWORD(&(state->edx), size >> 4);
-      }
-      Debug0((dbg_fd, "umb_allocated: %#x0:%#x0\n",
-	      (unsigned) WORD(state->ebx),
-	      (unsigned) WORD(state->edx)));
-      /* retval = UNCHANGED; */
-      break;
-    }
-
-  case XMS_DEALLOCATE_UMB:
-    {
-      umb_free(WORD(state->edx));
-      SETWORD(&(state->eax), 1);
-      Debug0((dbg_fd, "umb_freed: 0x%04x\n",
-	      (unsigned) WORD(state->edx)));
-      /* retval = UNCHANGED; */
-      break;
-    }
-
-  case 0x12:
-    x_printf("XMS realloc UMB segment 0x%04x size 0x%04x\n",
-	     LWORD(edx), LWORD(ebx));
-    LWORD(eax) = 0;		/* failure */
-    LO(bx) = 0x80;		/* function unimplemented */
+  case 0xf:			/* Reallocate Extended Memory Block */
+    XMS_RET(xms_realloc_EMB(OLDXMS));
     break;
 
     /* the functions below are the newer, 32-bit XMS 3.0 interface */
@@ -595,24 +548,33 @@ xms_control(void)
     break;
 
   case 0x89:			/* Allocate Extended Memory */
-    xms_allocate_EMB(NEWXMS);
+    XMS_RET(xms_allocate_EMB(NEWXMS));
     break;
 
   case 0x8e:			/* Get EMB Handle Information */
-    xms_EMB_info(NEWXMS);
+    XMS_RET(xms_EMB_info(NEWXMS));
     break;
 
   case 0x8f:			/* Reallocate EMB */
-    xms_realloc_EMB(NEWXMS);
+    XMS_RET(xms_realloc_EMB(NEWXMS));
     break;
 
   default:
     x_printf("XMS: Unimplemented XMS function AX=0x%04x", LWORD(eax));
     show_regs(__FILE__, __LINE__);		/* if you delete this, put the \n on the line above */
-    LWORD(eax) = 0;		/* failure */
-    LO(bx) = 0x80;		/* function not implemented */
-
+    XMS_RET(0x80);		/* function not implemented */
   }
+ }
+ /* If this is the UMB request which came via the external himem.sys,
+  * dont pass it back. */
+ if (!config.xms_size) {
+   if (is_umb_fn)
+     LWORD(esp) += 4;
+   else
+     x_printf("XMS: skipping external request, ax=0x%04x, dx=0x%04x\n",
+	      LWORD(eax), LWORD(edx));
+ }
+ fake_retf(0);
 }
 
 static int
@@ -642,10 +604,10 @@ ValidHandle(unsigned short h)
     return 0;
 }
 
-static void
+static unsigned char
 xms_query_freemem(int api)
 {
-  unsigned long totalBytes = 0, subtotal;
+  unsigned long totalBytes = 0, subtotal, largest;
   int h;
 
   /* the new XMS api should actually work with the function as it
@@ -654,8 +616,8 @@ xms_query_freemem(int api)
   if (api == NEWXMS)
     x_printf("XMS: new XMS API query_freemem()!\n");
 
-  /* total XMS mem and largest block under DOSEMU will always be the
-   * same, thanks to the wonderful operating environment
+  /* total XMS mem and largest block can be different because of
+   *  fragmentation
    */
 
   totalBytes = 0;
@@ -665,8 +627,14 @@ xms_query_freemem(int api)
   }
 
   subtotal = config.xms_size - (totalBytes / 1024);
+  if (debug_level('x')) {
+    if (smget_free_space(&mp) != subtotal * 1024)
+      x_printf("XMS smalloc mismatch!!!\n");
+  }
   /* total free is max allowable XMS - the number of K already allocated */
 
+  largest = smget_largest_free_area(&mp) / 1024;
+
   if (api == OLDXMS) {
     /* old XMS API uses only AX, while new API uses EAX. make
        * sure the old API sees 64 MB if the count >= 64 MB (instead of seeing
@@ -675,20 +643,27 @@ xms_query_freemem(int api)
        * know of 64MB of the > 64MB available memory.
        */
 
+    if (largest > 65535)
+      largest = 65535;
     if (subtotal > 65535)
       subtotal = 65535;
-    LWORD(eax) = LWORD(edx) = subtotal;
+    LWORD(eax) = largest;
+    LWORD(edx) = subtotal;
     x_printf("XMS query free memory(old): %dK %dK\n", LWORD(eax),
 	     LWORD(edx));
   }
   else {
-    REG(eax) = REG(edx) = subtotal;
+    REG(eax) = largest;
+    REG(edx) = subtotal;
     x_printf("XMS query free memory(new): %ldK %ldK\n",
 	     REG(eax), REG(edx));
   }
+  /* the following line is NOT superfluous!! (see above) */
+  LO(bx) = 0;
+  return 0;
 }
 
-static void
+static unsigned char
 xms_allocate_EMB(int api)
 {
   unsigned long totalBytes, subtotal;
@@ -712,15 +687,12 @@ xms_allocate_EMB(int api)
   if(kbsize > subtotal)
   {
     x_printf("XMS: out of memory (only %ldK available)\n",subtotal);
-    LWORD(eax) = 0;
-    LO(bx) = 0xa0; /* Out of memory */
-    return;
+    return 0xa0; /* Out of memory */
   }
   
   if (!(h = FindFreeHandle(FIRST_HANDLE))) {
     x_printf("XMS: out of handles\n");
-    LWORD(eax) = 0;
-    LO(bx) = 0xa1;
+    return 0xa1;
   }
   else {
     handles[h].num = h;
@@ -734,7 +706,7 @@ xms_allocate_EMB(int api)
        * to mean that reserving a handle of size 0 gives it no address
        */
     if (handles[h].size)
-      handles[h].addr = malloc((long)handles[h].size);
+      handles[h].addr = smalloc(&mp, handles[h].size);
     else {
       x_printf("XMS WARNING: allocating 0 size EMB\n");
       handles[h].addr = 0;
@@ -749,25 +721,23 @@ xms_allocate_EMB(int api)
       LWORD(edx) = h;		/* handle # */
     else
       REG(edx) = h;
-
-    LWORD(eax) = 1;		/* success */
+    return 0;
   }
 }
 
-static void
+static unsigned char
 xms_free_EMB(void)
 {
   unsigned short int h = LWORD(edx);
 
   if (!ValidHandle(h)) {
     x_printf("XMS: free EMB bad handle %d\n", h);
-    LWORD(eax) = 0;
-    LO(bx) = 0xa2;
+    return 0xa2;
   }
   else {
 
     if (handles[h].addr)
-      free(handles[h].addr);
+      smfree(&mp, handles[h].addr);
     else
       x_printf("XMS WARNING: freeing handle w/no address, size 0x%08lx\n",
 	       handles[h].size);
@@ -775,11 +745,11 @@ xms_free_EMB(void)
     handle_count--;
 
     x_printf("XMS: free'd EMB %d\n", h);
-    LWORD(eax) = 1;
+    return 0;
   }
 }
 
-static void
+static unsigned char
 xms_move_EMB(void)
 {
   char *src, *dest;
@@ -787,29 +757,21 @@ xms_move_EMB(void)
 
   x_printf("XMS move extended memory block\n");
   show_emm(e);
-  LWORD(eax) = 1;		/* start with success */
-#if 0
-  LWORD(ebx) = 0xff00;		/* start with success */
-  LO(bx) = 0x00; /* do not touch BH, 
-		    else Win98 boot failure. */
-  /* also don't touch BL - the XMS spec only specifies  *
-   * BL = 0x00 for function 7 (query A20)		*/
-#endif
 
   if (e.SourceHandle == 0) {
     src = (char *) (((e.SourceOffset >> 16) << 4) + \
 		    (e.SourceOffset & 0xffff));
   }
   else {
-    if (handles[e.SourceHandle].valid == 0)
+    if (handles[e.SourceHandle].valid == 0) {
       x_printf("XMS: invalid source handle\n");
-    src = handles[e.SourceHandle].addr + e.SourceOffset;
-    if (((unsigned long int)handles[e.SourceHandle].addr + handles[e.SourceHandle].size) <
-        ((unsigned long int)src + e.Length) ) {
-      e.Length = (unsigned long int)handles[e.SourceHandle].addr + handles[e.SourceHandle].size - (unsigned long int)src;
-      LWORD(eax) = 0;
-      LO(bx) = 0xa7;		/* invalid Length */
+      return 0xa3;
     }
+    src = handles[e.SourceHandle].addr + e.SourceOffset;
+    if (src > handles[e.SourceHandle].addr + handles[e.SourceHandle].size)
+      return 0xa4;              /* invalid source offset */
+    if (src + e.Length > handles[e.SourceHandle].addr + handles[e.SourceHandle].size)
+      return 0xa7;              /* invalid Length */
   }
 
   if (e.DestHandle == 0) {
@@ -817,31 +779,33 @@ xms_move_EMB(void)
 		     (e.DestOffset & 0xffff));
   }
   else {
-    if (handles[e.DestHandle].valid == 0)
+    if (handles[e.DestHandle].valid == 0) {
       x_printf("XMS: invalid dest handle\n");
+      return 0xa5;
+    }
     dest = handles[e.DestHandle].addr + e.DestOffset;
-    if (((unsigned long int)handles[e.DestHandle].addr + handles[e.DestHandle].size) <
-        ((unsigned long int)dest + e.Length) ) {
-      e.Length = (unsigned long int)handles[e.DestHandle].addr + handles[e.DestHandle].size - (unsigned long int)dest;
-      LWORD(eax) = 0;
-      LO(bx) = 0xa7;		/* invalid Length */
+    if (dest > handles[e.DestHandle].addr + handles[e.DestHandle].size)
+      return 0xa6;              /* invalid dest offset */
+    if (dest + e.Length > handles[e.DestHandle].addr + handles[e.DestHandle].size) {
+      return 0xa7;		/* invalid Length */
     }
   }
 
   x_printf("XMS: block move from %p to %p len 0x%lx\n",
 	   (void *) src, (void *) dest, e.Length);
-
-  memmove(dest, src, e.Length);
-
+  memmove_dos2dos(dest, src, e.Length);
+  if ((size_t)dest < 0x110000)
+    e_invalidate(dest, e.Length);
   x_printf("XMS: block move done\n");
 
-  return;
+  return 0;
 }
 
-static void
+static unsigned char
 xms_lock_EMB(int flag)
 {
   int h = LWORD(edx);
+  unsigned addr;
 
   if (flag)
     x_printf("XMS lock EMB %d\n", h);
@@ -858,52 +822,46 @@ xms_lock_EMB(int flag)
         handles[h].lockcount--;
       else {
         x_printf("XMS: Unlock handle %d already at 0\n", h);
-        LWORD(eax) = 0;
-        LO(bx) = 0xaa;		/* Block is not locked */
-	return;
+        return 0xaa;		/* Block is not locked */
       }
 
-    LWORD(edx) = (int) handles[h].addr >> 16;
-    LWORD(ebx) = (int) handles[h].addr & 0xffff;
-
-    LWORD(eax) = 1;
+    addr = handles[h].addr - (char *)ext_mem_base + LOWMEM_SIZE + HMASIZE;
+    LWORD(edx) = addr >> 16;
+    LWORD(ebx) = addr & 0xffff;
+    return 0;
   }
   else {
     x_printf("XMS: invalid handle %d, can't (un)lock\n", h);
-    LWORD(eax) = 0;
-    LO(bx) = 0xa2;		/* invalid handle */
+    return 0xa2;		/* invalid handle */
   }
 }
 
-static void
+static unsigned char
 xms_EMB_info(int api)
 {
   int h = LWORD(edx);
 
   if (ValidHandle(h)) {
     if (api == OLDXMS) {
-      HI(bx) = handles[h].lockcount;
       LO(bx) = NUM_HANDLES - handle_count;
       LWORD(edx) = handles[h].size / 1024;
-      LWORD(eax) = 1;
       x_printf("XMS Get EMB info(old) %d\n", h);
     }
     else {			/* api == NEWXMS */
-      HI(bx) = handles[h].lockcount;
       LWORD(ecx) = NUM_HANDLES - handle_count;
       REG(edx) = handles[h].size / 1024;
-      LWORD(eax) = 1;
       x_printf("XMS Get EMB info(new) %d\n", h);
     }
+    HI(bx) = handles[h].lockcount;
+    return 0;
   }
   else {
-    LWORD(eax) = 0;
-    LO(bx) = 0xa2;		/* invalid handle */
+    return 0xa2;		/* invalid handle */
   }
 }
 
 /* untested! if you test it, please tell me! */
-static void
+static unsigned char
 xms_realloc_EMB(int api)
 {
   int h;
@@ -914,18 +872,12 @@ xms_realloc_EMB(int api)
 
   if (!ValidHandle(h)) {
     x_printf("XMS: invalid handle %d, cannot realloc\n", h);
-
-    LO(bx) = 0xa2;
-    LWORD(eax) = 0;
-    return;
+    return 0xa2;
   }
   else if (handles[h].lockcount) {
     x_printf("XMS: handle %d locked (%d), cannot realloc\n",
 	     h, handles[h].lockcount);
-
-    LO(bx) = 0xab;
-    LWORD(eax) = 0;
-    return;
+    return 0xab;
   }
 
   oldsize = handles[h].size;
@@ -943,7 +895,7 @@ xms_realloc_EMB(int api)
 	   "XMS realloc EMB(new) %d to size 0x%08lx\n",
 	   h, handles[h].size);
 
-  handles[h].addr = malloc(handles[h].size);
+  handles[h].addr = smalloc(&mp, handles[h].size);
 
   /* memcpy() the old data into the new block, but only
    * min(new,old) bytes.
@@ -952,31 +904,8 @@ xms_realloc_EMB(int api)
 	 (oldsize <= handles[h].size) ? oldsize : handles[h].size);
 
   /* free the EMB's old Linux memory */
-  free(oldaddr);
-
-  LWORD(eax) = 1;		/* success */
-}
-
-void
-xms_int15(void)
-{
-  if (HI(ax) == 0x88) {
-    NOCARRY;
-    if (xms_grab_int15) {
-      x_printf("XMS int 15 free mem returning 0 to protect HMA\n");
-      LWORD(eax) &= ~0xffff;
-    }
-    else {
-      x_printf("XMS early int15 call...%dK free\n", config.xms_size);
-      LWORD(eax) = config.xms_size;
-    }
-  }
-  else {			/* AH = 0x87 */
-    x_printf("XMS int 15 block move failed AX=0x%04x\n", (u_short)REG(eax));
-    LWORD(eax) &= 0xFF;
-    HI(ax) = 3;			/* say A20 gate failed */
-    CARRY;
-  }
+  smfree(&mp, oldaddr);
+  return 0;
 }
 
 struct EMM
@@ -985,19 +914,8 @@ get_emm(unsigned int seg, unsigned int off)
   struct EMM e;
   unsigned char *p;
 
-  p = (char *) ((seg << 4) + off);
+  p = MK_FP32(seg, off);
 
-#if 0
-  e.Length = *(unsigned long *) p;
-  p += 4;
-  e.SourceHandle = *(unsigned short *) p;
-  p += 2;
-  e.SourceOffset = *(unsigned long *) p;
-  p += 4;
-  e.DestHandle = *(unsigned short *) p;
-  p += 2;
-  e.DestOffset = *(unsigned long *) p;
-#else
   e.Length = *(unsigned int *) p;
   p += 4;
   e.SourceHandle = *(unsigned short *) p;
@@ -1007,7 +925,6 @@ get_emm(unsigned int seg, unsigned int off)
   e.DestHandle = *(unsigned short *) p;
   p += 2;
   e.DestOffset = *(unsigned int *) p;
-#endif
 
   return e;
 }
