@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -30,10 +30,9 @@
  *
  */
 
-#include <features.h>
+#include <unistd.h>
 #include "emu.h"
 #include "pktdrvr.h"
-#include "dosio.h"
 #include "memory.h"
 #include "int.h"
 #include "bios.h"
@@ -43,32 +42,28 @@
 #include <errno.h>
 #include <sys/time.h>
 #include <netinet/in.h>
-#if GLIBC_VERSION_CODE >= 2000
-  #include <netinet/if_ether.h>
-#else
-  #include "Linux/if_ether.h"
-#endif
+#include <netinet/if_ether.h>
 #include "libpacket.h"
 #include "dosnet.h"
-#include "dosemu_select.h"
-
-/* flag to activate use of pic by packet driver */
-#ifdef PICPKT
 #include "pic.h"
 #include "dpmi.h"
-#endif
 
 #define min(a,b)	((a) < (b)? (a) : (b))
 
-static int Open_sockets(void);
+static int Open_sockets(char *name);
 static int Insert_Type(int, int, char *);
 static int Remove_Type(int);
 int Find_Handle(u_char *buf);
 static void printbuf(char *, struct ethhdr *);
+static int pkt_check_receive(int ilevel);
+static void pkt_receiver_callback(void);
 
 int pkt_fd=-1, pkt_broadcast_fd=-1, max_pkt_fd;
 static int pktdrvr_installed;
 
+unsigned short receive_mode;
+static unsigned short local_receive_mode;
+
 /* array used by virtual net to keep track of packet types */
 #define MAX_PKT_TYPE_SIZE 10
 struct pkt_type {
@@ -92,7 +87,6 @@ struct pkt_globs
     unsigned char hw_address[8];	/* our hardware address */
     short type;				/* our type */
     int flags;				/* configuration flags */
-    int helpvec;			/* vector number for helper */
     int nfds;				/* number of fd's for select() */
     fd_set sockset;			/* set of sockets for select() */
 
@@ -108,65 +102,95 @@ struct pkt_globs
     } handle[MAX_HANDLE];          
 } pg;
 
-static char devname[10];		/* linux device name */
-
-/* various declarations for interfacing with the packet driver code in
-   bios.S */
-
-extern void PKTDRV_size();
-extern void PKTDRV_handle();
-extern void PKTDRV_receiver();
-extern void PKTDRV_buf();
-extern void PKTDRV_helper();
-extern void PKTDRV_signature();
-extern void PKTDRV_driver_name();
-extern void PKTDRV_param();
-extern void PKTDRV_stats();
-extern void PKTDRV_start();
-
 /* creates a pointer into the BIOS from the asm exported labels */
 #define MK_PTR(ofs) ( (void *)((long)(ofs)-(long)bios_f000+(BIOSSEG << 4)) )
 
 /* calculates offset of a label from the start of the packet driver */
 #define MK_PKT_OFS(ofs) ((long)(ofs)-(long)PKTDRV_start)
 
-#define pkt_buf ((char *)MK_PTR(PKTDRV_buf))
+char pkt_buf[PKT_BUF_SIZE];
 
-short *p_helper_size;
-long *p_helper_receiver;
-short *p_helper_handle;
+short p_helper_size;
+long p_helper_receiver;
+short p_helper_handle;
 struct pkt_param *p_param;
 struct pkt_statistics *p_stats;
 
 /************************************************************************/
 
 /* initialize the packet driver interface (called at startup) */
-void
-pkt_init(int vec)
+void pkt_priv_init(void)
 {
+    int ret = 0;
     /* initialize the globals */
+    if (!config.pktdrv)
+      return;
 
     pktdrvr_installed = 1; /* Will be cleared if some error occurs */
 
-    p_helper_size = MK_PTR(PKTDRV_size);
-    p_helper_receiver = MK_PTR(PKTDRV_receiver);
-    p_helper_handle = MK_PTR(PKTDRV_handle);
+    switch (config.vnet) {
+      case VNET_TYPE_ETH:
+        ret = Open_sockets(config.netdev);
+	break;
+      case VNET_TYPE_DSN:
+        ret = Open_sockets(DOSNET_DEVICE);
+	break;
+    }
+
+    if (ret < 0) {
+      warn("PKT: Cannot open raw sockets: %s\n", strerror(errno));
+      pktdrvr_installed = -1;
+    }
+}
+
+void
+pkt_init(void)
+{
+    char devname[10];
+    if (!config.pktdrv)
+      return;
+    if (pktdrvr_installed == -1)
+      goto fail;
     p_param = MK_PTR(PKTDRV_param);
     p_stats = MK_PTR(PKTDRV_stats);
 
     /* use dosnet device (dsn0) for virtual net */
-    if (config.vnet) {
-	strcpy(devname, DOSNET_DEVICE);
-	if (Open_sockets() < 0)
-	    goto fail;
-    } else {
-      strncpy(devname, config.netdev, sizeof(devname) - 1);
-      devname[sizeof(devname) - 1] = 0;
-    }
+    pd_printf("PKT: VNET mode is %i\n", config.vnet);
+    switch (config.vnet) {
+      case VNET_TYPE_ETH:
+	strncpy(devname, config.netdev, sizeof(devname) - 1);
+        devname[sizeof(devname) - 1] = 0;
+	add_to_io_select(pkt_fd, 1, pkt_receive_async);
+	break;
 
-    /* hook the interrupt vector by pointing it into the magic table */
-    SETIVEC(vec, PKTDRV_SEG, PKTDRV_OFF);
-    strcpy(MK_PTR(PKTDRV_signature), "PKT DRVR");
+      case VNET_TYPE_DSN:
+	strcpy(devname, DOSNET_DEVICE);
+	add_to_io_select(pkt_fd, 1, pkt_receive_async);
+	add_to_io_select(pkt_broadcast_fd, 1, pkt_receive_async);
+	break;
+
+      case VNET_TYPE_TAP:
+        strcpy(devname, TAP_DEVICE);
+        if (strncmp(config.netdev, TAP_DEVICE, 3) == 0) {
+          pd_printf("PKT: trying to bind to device %s\n", config.netdev);
+          strcpy(devname, config.netdev);
+        }
+        if (Open_sockets(devname) < 0) {
+          warn("PKT: Cannot allocate TAP device %s: %s\n",
+	    strcmp(devname, TAP_DEVICE) ? devname : "(dynamic)",
+	    strerror(errno));
+          goto fail;
+        }
+	add_to_io_select(pkt_fd, 1, pkt_receive_async);
+	break;
+
+      default:
+        error("Unknown vnet mode %i\n", config.vnet);
+	goto fail;
+    }
+    pd_printf("PKT: Using device %s\n", devname);
+
+    pic_seti(PIC_NET, pkt_check_receive, 0, pkt_receiver_callback);
     
     /* fill other global data */
 
@@ -184,14 +208,21 @@ pkt_init(int vec)
     p_param->rcv_bufs = 8 - 1;		/* a guess */
     p_param->xmt_bufs = 2 - 1;
 
-    pg.helpvec = vec + 1;		/* use this for helper vec */
-    SETIVEC(pg.helpvec, BIOSSEG, (long)PKTDRV_helper-(long)bios_f000);
     return;
 
 fail:
     pktdrvr_installed = 0;
 }
 
+void
+pkt_reset(void)
+{
+    if (!config.pktdrv)
+      return;
+    /* hook the interrupt vector by pointing it into the magic table */
+    SETIVEC(0x60, PKTDRV_SEG, PKTDRV_OFF);
+}
+
 /* this is the handler for INT calls from DOS to the packet driver */
 int
 pkt_int (void)
@@ -351,23 +382,8 @@ pkt_int (void)
 		    }
 		}
 	    }
-	    if (config.vnet)
-		Insert_Type(free_handle, hdlp->packet_type_len, 
+	    Insert_Type(free_handle, hdlp->packet_type_len, 
 			    hdlp->packet_type);
-	    else {
-		hdlp->sock = OpenNetworkType(type); /* open the socket */
-
-		if (hdlp->sock < 0) {
-		    hdlp->in_use = 0;		/* forget about this handle */
-		    HI(dx) = E_BAD_TYPE;
-		    break;
-		}
-
-		FD_SET(hdlp->sock, &pg.sockset); /* keep track of sockets */
-		add_to_io_select(hdlp->sock, 1);
-		if (hdlp->sock >= pg.nfds)
-		    pg.nfds = hdlp->sock + 1;
-	    }
 	    REG(eax) = free_handle;		/* return the handle */
 	}
 	return 1;
@@ -378,44 +394,18 @@ pkt_int (void)
 	    break;
 	}
 
-	if (config.vnet)
-	    Remove_Type(hdlp_handle);
-	else {
-	    if (hdlp->in_use) {
-		int i, n;
-		
-		CloseNetworkLink(hdlp->sock);      /* close the socket */
-		FD_CLR(hdlp->sock, &pg.sockset);  /* keep track of sockets */
-		remove_from_io_select(hdlp->sock, 1);
-		n = pg.nfds;
-		pg.nfds = 0;
-
-		for (i = 0; i < n; i++)
-		    if (FD_ISSET(i,&pg.sockset))
-			pg.nfds = i + 1;
-	    }
-	}
-	if (hdlp->in_use)
-	    hdlp->in_use = 0;	/* no longer in use */
+	Remove_Type(hdlp_handle);
+	hdlp->in_use = 0;	/* no longer in use */
 	return 1;
 
     case F_SEND_PKT: {
-	/* unfortunately, a handle is not passed as a parameter for the */
-	/* SEND_PKT call.  so, we will have to scan the handle table to */
-	/* find a handle which is in use, and use its socket for the send */
-	int handle, socket;
-
 	p_stats->packets_out++;
 	p_stats->bytes_out += LWORD(ecx);
 
 	pd_printf("========Sending packet======\n");
 	printbuf("packet to send:", SEG_ADR((struct ethhdr *), ds, si)); 
-	for (handle = 0; handle < MAX_HANDLE; handle++) {
-	    hdlp = &pg.handle[handle];
-
-	    if (hdlp->in_use) {
-		if (pg.flags & FLAG_NOVELL)	/* Novell hack? */
-		{
+	if (pg.flags & FLAG_NOVELL)	/* Novell hack? */
+	{
 		    char *p;
 		    short len;
 		    
@@ -434,31 +424,15 @@ pkt_int (void)
 			p[0] = len >> 8;
 			p[1] = (char)len;
 		    }
-		}
-		if (config.vnet)
-		    socket = pkt_fd;
-		else
-		    socket = hdlp->sock;
+	}
 
-		if (WriteToNetwork(socket, devname, 
-				   SEG_ADR((char *), ds, si),
-				   LWORD(ecx)) >= 0) {
+	if (write(pkt_fd, SEG_ADR((char *), ds, si), LWORD(ecx)) >= 0) {
 		    pd_printf("Write to net was ok\n");
-		    /* check for a reply to speedup things 
-		       (background poll is quite slow) */
-#ifdef PICPKT
-		    pic_request(PIC_NET);
-#else
-		    pkt_check_receive(50000);
-#endif
 		    return 1;
-		} 
-		else {
-		    warn("WriteToNetwork(%d,\"%s\",buffer,%u): error %d\n",
-			 socket, devname, LWORD(ecx), errno);
+	} else {
+		    warn("WriteToNetwork(%d,buffer,%u): error %d\n",
+			 pkt_fd, LWORD(ecx), errno);
 		    break;
-		}
-	    }
 	}
 	
 	p_stats->errors_out++;
@@ -501,10 +475,11 @@ pkt_int (void)
 	    HI(dx) = E_BAD_HANDLE;
 	    break;
 	}
-	if (LWORD(ecx) != 3) {			/* only mode 3 supported */
+	if (LWORD(ecx) != receive_mode && LWORD(ecx) != 1) {
 	    HI(dx) = E_BAD_MODE;
 	    break;
 	}
+	local_receive_mode = LWORD(ecx);
 	return 1;
 
     case F_GET_RCV_MODE:
@@ -512,7 +487,7 @@ pkt_int (void)
 	    HI(dx) = E_BAD_HANDLE;
 	    break;
 	}
-	REG(eax) = 3;
+	REG(eax) = local_receive_mode;
 	return 1;
 
     case F_GET_STATS:
@@ -546,24 +521,33 @@ pkt_int (void)
 /* Open two sockets for the virtual network, one for normal packets
    and one for broadcasts. */
 static int
-Open_sockets(void)
+Open_sockets(char *name)
 {
+    GenerateDosnetID();
+
    /* This handle is inserted to receive packets of type "dosnet broadcast". 
       These are really speaking broadcast packets meant to be received by
       all dosemus. Their destination addresses are changed by dosemu, and
       are changed back to 'ffffff..' when these packets are received by dosemu.    */
-    pkt_broadcast_fd = OpenBroadcastNetworkType();
-    if (pkt_broadcast_fd < 0)
+    if (config.vnet == VNET_TYPE_DSN) {
+      pkt_broadcast_fd = OpenNetworkLink(name, DOSNET_BROADCAST_TYPE);
+      if (pkt_broadcast_fd < 0)
 	return pkt_broadcast_fd;
+    }
 
     /* The socket for normal packets */
-    pkt_fd = OpenNetworkType((u_short)0);
+    pkt_fd = OpenNetworkLink(name,
+      config.vnet == VNET_TYPE_ETH ? ETH_P_ALL : GetDosnetID());
     if (pkt_fd < 0)
 	return pkt_fd;
 
     max_pkt_fd = pkt_fd + 1;
     if (max_pkt_fd <= pkt_broadcast_fd) 
 	max_pkt_fd = pkt_broadcast_fd + 1; 
+
+    local_receive_mode = receive_mode;
+    pd_printf("PKT: detected receive mode %i\n", receive_mode);
+
     return 0;    
 }
 
@@ -613,97 +597,91 @@ Remove_Type(int handle)
     return 0;
 }
 
-void pkt_receiver_callback(void)
+static void pkt_receiver_callback(void)
 {
+    struct vm86_regs saved_regs;
+
+    if (p_helper_size == 0)
+      return;
+
     if(in_dpmi && !in_dpmi_dos_int)
 	fake_pm_int();
-    run_int(pg.helpvec);
-   /* push iret frame. At F000:20F7 (bios.S) we get an
-    * iret and return to F000:2140 for EOI */
-    fake_int(BIOSSEG, EOI_OFF);
+    fake_int_to(BIOSSEG, EOI_OFF);
+
+    saved_regs = REGS;
+
+    _AX = 0;
+    _BX = p_helper_handle;
+    _CX = p_helper_size;
+    do_call_back(p_helper_receiver);
+    if (_ES == 0 && _DI == 0)
+      goto out;
+    MEMCPY_2DOS(SEGOFF2LINEAR(_ES, _DI), pkt_buf, p_helper_size);
+    _DS = _ES;
+    _SI = _DI;
+    _AX = 1;
+    _BX = p_helper_handle;
+    do_call_back(p_helper_receiver);
+
+out:
+    p_helper_size = 0;
+    REGS = saved_regs;
 }
 
-void
-pkt_check_receive_quick(void)
+void pkt_receive_async(void)
 {
-    pkt_check_receive(0);
+  if (local_receive_mode == 1)
+    return;
+  pic_request(PIC_NET);
 }
 
-int
-pkt_check_receive(int timeout)
+static int pkt_receive(void)
 {
     int size,handle, fd;
     struct per_handle *hdlp;
-    char *p;
     struct timeval tv;
     fd_set readset;
-    char device[32];
 
     if (!pktdrvr_installed) {
         pd_printf("Driver not initialized ...\n");
-	return -1;
-    }
-
-    /*
-    ** sometimes, dosemu enters this routine with the previous packet 
-    ** not sent by helper, and the packet driver hangs. 
-    ** This patch tryes to resend the packet, and make dosemu happy :)
-    **
-    ** Gloriano 
-    */
-    if (*p_helper_size != 0)			/* transfer area busy? */
-    {
-#ifdef PICPKT
-	do_irq();
-#else
-	run_int(pg.helpvec);
-#endif
-	pd_printf("Called the helpvector... (after failure)\n");
-	return 1;
+	return 0;
     }
 
     tv.tv_sec = 0;				/* set a (small) timeout */
-    tv.tv_usec = timeout;
+    tv.tv_usec = 0;
 
     /* anything ready? */
-    if (config.vnet) {
-	FD_ZERO(&readset);
-	FD_SET(pkt_fd, &readset);
-	FD_SET(pkt_broadcast_fd, &readset);
-	add_to_io_select(pkt_broadcast_fd, 1);
-	add_to_io_select(pkt_fd, 1);
+    FD_ZERO(&readset);
+    FD_SET(pkt_fd, &readset);
+    if (config.vnet == VNET_TYPE_DSN) {
+      FD_SET(pkt_broadcast_fd, &readset);
+    }
+    /* anything ready? */
+    if (select(max_pkt_fd,&readset,NULL,NULL,&tv) <= 0)
+        return 0;
 
-	/* anything ready? */
-	if (select(max_pkt_fd,&readset,NULL,NULL,&tv) <= 0)
-	    return 0;
-	if(FD_ISSET(pkt_fd, &readset)) 
-	    fd = pkt_fd;
-	else if(FD_ISSET(pkt_broadcast_fd, &readset)) 
-	    fd = pkt_broadcast_fd;
-	else return 0;
+    if(FD_ISSET(pkt_fd, &readset)) 
+        fd = pkt_fd;
+    else if(config.vnet == VNET_TYPE_DSN && FD_ISSET(pkt_broadcast_fd, &readset)) 
+        fd = pkt_broadcast_fd;
+    else return 0;
 
-	size = ReadFromNetwork(fd, device, pkt_buf, PKT_BUF_SIZE);
-	if (size < 0) {
-	    p_stats->errors_in++;		/* select() somehow lied */
-	    return 0;
-	}
-	if (strcmp(device, devname)) {
-	    pd_printf("strcmp(device != devname) ...\n");
-	    return 0;
-	}
+    size = read(fd, pkt_buf, PKT_BUF_SIZE);
+    if (size < 0) {
+        p_stats->errors_in++;		/* select() somehow lied */
+        return 0;
+    }
    
-	pd_printf("========Processing New packet======\n");
-	handle = Find_Handle(pkt_buf);
-	if (handle == -1) 
-	    return -1;
-	pd_printf("Found handle %d\n", handle);
-
-	hdlp = &pg.handle[handle];
-	if (hdlp->in_use) {
-	    printbuf("received packet:", (struct ethhdr *)pkt_buf); 
+    pd_printf("========Processing New packet======\n");
+    handle = Find_Handle(pkt_buf);
+    if (handle == -1) 
+        return 0;
+    pd_printf("Found handle %d\n", handle);
 
+    hdlp = &pg.handle[handle];
+    if (hdlp->in_use) {
             /* VINOD: If it is broadcast type, translate it back ... */
-	    if (memcmp(pkt_buf, DOSNET_BROADCAST_ADDRESS, 4) == 0) {
+	    if (config.vnet == VNET_TYPE_DSN && memcmp(pkt_buf, DOSNET_BROADCAST_ADDRESS, 4) == 0) {
 		pd_printf("It is a broadcast packet\n");
 		if(memcmp(pkt_buf + ETH_ALEN, pg.hw_address, ETH_ALEN) == 0) {
 		    /* Ignore our own ethernet broadcast. */
@@ -714,9 +692,10 @@ pkt_check_receive(int timeout)
 		printbuf("Translated:", (struct ethhdr *)pkt_buf); 
 	    }
 
-	    /* No need to check the type again. Also, for now, NOVELL_HACK
-	       is diabled... */
-	    if (0) {
+	    /* No need to hack the incoming packets it seems. */
+#if 0
+	    if (pg.flags & FLAG_NOVELL)	{ /* Novell hack? */
+		char *p;
 		/* check if the packet's type matches the specified type */
 		/* in the ACCESS_TYPE call.  the position depends on the */
 		/* driver class! */
@@ -726,102 +705,37 @@ pkt_check_receive(int timeout)
 		else
 		    p = pkt_buf + 2 * ETH_ALEN + 2;	/* IEEE 802.3 */
 
-		if (size >= ((p - pkt_buf) + hdlp->packet_type_len) &&
-		    !memcmp(p,hdlp->packet_type,hdlp->packet_type_len)) {
-                }
-		if (hdlp->flags & FLAG_NOVELL) {	/* Novell hack? */
-		    *--p = (char)ETH_P_IPX; /* overwrite length with type */
-		    *--p = (char)(ETH_P_IPX >> 8);
-		}
+		*--p = (char)ETH_P_IPX; /* overwrite length with type */
+		*--p = (char)(ETH_P_IPX >> 8);
 	    }
+#endif
 	    p_stats->packets_in++;
 	    p_stats->bytes_in += size;
 
+	    printbuf("received packet:", (struct ethhdr *)pkt_buf); 
 	    /* stuff things in global vars and queue a hardware */
 	    /* interrupt which will perform the upcall */
-
-	    *p_helper_size = size;
-	    *p_helper_receiver = hdlp->receiver;
-	    *p_helper_handle = handle;
-#ifdef PICPKT
-	    do_irq();
-#else
-	    run_int(pg.helpvec);
-#endif
+	    if (p_helper_size)
+		error("PKT: Receiver is not ready, packet dropped (size=%i)\n",
+		  p_helper_size);
+	    p_helper_size = size;
+	    p_helper_receiver = hdlp->receiver;
+	    p_helper_handle = handle;
 	    pd_printf("Called the helpvector ... \n");
 	    return 1;
-	} else {
-	    p_stats->packets_lost++;	/* not really lost... */
-	    pd_printf("Handle not in use, ignored this packet\n");
-	    return 0;
-	}
+    } else {
+        p_stats->packets_lost++;	/* not really lost... */
+        pd_printf("Handle not in use, ignored this packet\n");
+        return 0;
     }
-    else { /* !config.vnet */
-	readset = pg.sockset;
-	if (select(pg.nfds, &readset, NULL, NULL, &tv) <= 0)
-	    return 0;
+    return 0;
+}
 
-	for (handle = 0; handle < MAX_HANDLE; handle++) {
-	    hdlp = &pg.handle[handle];
-	    
-	    if (hdlp->in_use && FD_ISSET(hdlp->sock,&readset)) {
-		/* something is available on this handle's socket */
-		/* attempt to read it. Should always succeed. */
-
-		size = ReadFromNetwork(hdlp->sock, device, 
-				       pkt_buf, PKT_BUF_SIZE);
-
-		/* check if something has been received */
-		/* there was a check for "not multicast" here, but I think */
-		/* it should not be there... */
-
-		if (size >= 0) {
-		    /* verify the source of the packet. when not from eth0, */
-		    /* discard it for now (multiple network cards are next) */
-
-		    if (strcmp(device,devname))
-			continue;
-
-		    /* check if the packet's type matches the specified type */
-		    /* in the ACCESS_TYPE call.  the position depends on the */
-		    /* driver class! */
-		    if (hdlp->class == ETHER_CLASS)
-			p = pkt_buf + 2 * ETH_ALEN;	/* Ethernet-II */
-		    else
-			p = pkt_buf + 2 * ETH_ALEN + 2;	/* IEEE 802.3 */
-		    
-		    if (size >= ((p - pkt_buf) + hdlp->packet_type_len) &&
-			!memcmp(p,hdlp->packet_type,hdlp->packet_type_len))
-		    {
-			p_stats->packets_in++;
-			p_stats->bytes_in += size;
-			
-			if (hdlp->flags & FLAG_NOVELL) {
-			    /* overwrite length with type */
-			    *--p = (char)ETH_P_IPX;
-			    *--p = (char)(ETH_P_IPX >> 8);
-			}
-			
-		    /* stuff things in global vars and queue a hardware */
-		    /* interrupt which will perform the upcall */
-			*p_helper_size = size;
-			*p_helper_receiver = hdlp->receiver;
-			*p_helper_handle = handle;
-#ifdef PICPKT
-			do_irq();
-#else
-			run_int(pg.helpvec);
-#endif
-			return 1;
-		    } else
-			p_stats->packets_lost++; /* not really lost... */
-		} else
-		    p_stats->errors_in++; /* select() somehow lied */
-	    }
-	}
-	
-	return 0;
-    }
+static int pkt_check_receive(int ilevel)
+{
+  if (pkt_receive())
+    return 1;	/* run IRQ */
+  return 0;
 }
 
 /*  Find_Handle does type demultiplexing. 
@@ -854,10 +768,19 @@ Find_Handle(u_char *buf)
 
 static void
 printbuf(char *mesg, struct ethhdr *buf)
-{ int i;
+{
+int i;
+u_char *p;
   pd_printf( "%s :\n Dest.=", mesg);
   for (i=0;i<6;i++) pd_printf("%x:",buf->h_dest[i]);
   pd_printf( " Source=");
   for (i=0;i<6;i++) pd_printf("%x:",buf->h_source[i]);
-  pd_printf( " Type= 0x%x \n", ntohs(buf->h_proto));
+  if (ntohs(buf->h_proto) >= 1536) {
+    p = (u_char *)buf + 2 * ETH_ALEN;		/* Ethernet-II */
+    pd_printf(" Ethernet-II;");
+  } else {
+    p = (u_char *)buf + 2 * ETH_ALEN + 2;     /* All the rest frame types. */
+    pd_printf(" 802.3;");
+  }
+  pd_printf( " Type= 0x%x \n", ntohs(*(u_short *)p));
 }
