@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* 
@@ -25,10 +25,12 @@
 #include <unistd.h>
 #include <string.h>
 #include <fcntl.h>
+#include <limits.h>
 #include <sys/time.h>
 #include <errno.h>
 
 #include "config.h"
+#include "dosemu_config.h"
 #include "memory.h"
 #include "termio.h"
 #include "emu.h"
@@ -40,10 +42,9 @@
 #include "dpmi.h"
 #include "priv.h"
 
-extern void xms_control(void);
-
 #ifdef X86_EMULATOR
 #include "simx86/syncpu.h"
+#include "cpu-emu.h"
 #define CRs	TheCPU.cr
 #define DRs	TheCPU.dr
 #define TRs	TheCPU.tr
@@ -51,7 +52,7 @@ extern void xms_control(void);
 /* Extra CPU registers. Note that GDTR,LDTR,IDTR are internal
  * to the cpuemu and are defined in emu-ldt.c:
  */
-unsigned long CRs[5] =
+unsigned int CRs[5] =
 {
 	0x00000013,	/* valid bits: 0xe005003f */
 	0x00000000,	/* invalid */
@@ -78,7 +79,7 @@ unsigned long CRs[5] =
  *	b24-27= LLRW bp#2
  *	b28-31= LLRW bp#3
  */
-unsigned long DRs[8] =
+unsigned int DRs[8] =
 {
 	0x00000000,
 	0x00000000,
@@ -90,13 +91,16 @@ unsigned long DRs[8] =
 	0x00000400
 };
 
-unsigned long TRs[2] =
+unsigned int TRs[2] =
 {
 	0x00000000,
 	0x00000000
 };
 #endif
 
+/* fpu_state needs to be paragraph aligned for fxrstor/fxsave */
+struct _fpstate vm86_fpu_state __attribute__ ((aligned(16)));
+
 /* 
  * DANG_BEGIN_FUNCTION cpu_trap_0f
  *
@@ -110,15 +114,12 @@ unsigned long TRs[2] =
  */
 int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 {
-  	if (in_dpmi && (scp==NULL)) return 0; /* for safety */
-	/* moved access to "csp" after safety check,
-	 *                          -- 980627 Andreas Kirschbaum
-	 */
+	int increment_ip = 0;
 	g_printf("CPU: TRAP op 0F %02x %02x\n",csp[1],csp[2]);
 
 	if (csp[1] == 0x06) {
-		(in_dpmi? scp->eip:LWORD(eip)) += 2;  /* CLTS - ignore */
-		return 1;
+		/* CLTS - ignore */
+		increment_ip = 2;
 	}
 	else if (csp[1] == 0x31) {
 		/* ref: Van Gilluwe, "The Undocumented PC". The program
@@ -143,15 +144,15 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 		  gettimeofday(&tv, NULL);
 		  t = (unsigned long long)tv.tv_sec*100000000 +
 		      (unsigned long long)tv.tv_usec*100;
-		  REG(eax)=((int *)&t)[0];
-		  REG(edx)=((int *)&t)[1];
+		  REG(eax)=t & 0xffffffff;
+		  REG(edx)=t >> 32;
 		}
-		(in_dpmi? scp->eip:LWORD(eip)) += 2;
-		return 1;
+		increment_ip = 2;
 	}
 	else if ((((csp[1] & 0xfc)==0x20)||(csp[1]==0x24)||(csp[1]==0x26)) &&
 		((csp[2] & 0xc0) == 0xc0)) {
-		unsigned long *cdt, *srg;
+		unsigned int *cdt;
+		unsigned int *srg;
 		int idx;
 		boolean rnotw;
 
@@ -170,26 +171,26 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 		}
 
 		switch (csp[2]&7) {
-			case 0: srg = (in_dpmi? &(scp->eax):(unsigned long *)&(REGS.eax));
+			case 0: srg = (scp ? &_eax:&_EAX);
 				break;
-			case 1: srg = (in_dpmi? &(scp->ecx):(unsigned long *)&(REGS.ecx));
+			case 1: srg = (scp ? &_ecx:&_ECX);
 				break;
-			case 2: srg = (in_dpmi? &(scp->edx):(unsigned long *)&(REGS.edx));
+			case 2: srg = (scp ? &_edx:&_EDX);
 				break;
-			case 3: srg = (in_dpmi? &(scp->ebx):(unsigned long *)&(REGS.ebx));
+			case 3: srg = (scp ? &_ebx:&_EBX);
 				break;
-			case 6: srg = (in_dpmi? &(scp->esi):(unsigned long *)&(REGS.esi));
+			case 6: srg = (scp ? &_esi:&_ESI);
 				break;
-			case 7: srg = (in_dpmi? &(scp->edi):(unsigned long *)&(REGS.edi));
+			case 7: srg = (scp ? &_edi:&_EDI);
 				break;
 			default:	/* ESP(4),EBP(5) */
 				return 0;
 		}
 		if (rnotw) {
 		  *srg = cdt[idx];
-		  g_printf("CPU: read =%08lx\n",*srg);
+		  g_printf("CPU: read =%08x\n",*srg);
 		} else {
-		  g_printf("CPU: write=%08lx\n",*srg);
+		  g_printf("CPU: write=%08x\n",*srg);
 		  if (cdt==CRs) {
 		    /* special cases... they are too many, I hope this
 		     * will suffice for all */
@@ -198,7 +199,13 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 		  }
 		  else cdt[idx] = *srg;
 		}
-		(in_dpmi? scp->eip:LWORD(eip)) += 3;
+		increment_ip = 3;
+	}
+	if (increment_ip) {
+		if (scp)
+			_eip += increment_ip;
+		else
+			LWORD(eip) += increment_ip;
 		return 1;
 	}
 	/* all other trapped combinations:
@@ -207,6 +214,27 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
 	return 0;
 }
 
+void cpu_reset(void)
+{
+  /* ax,bx,cx,dx,si,di,bp,fs,gs can probably can be anything */
+  REG(eax) = 0;
+  REG(ebx) = 0;
+  REG(ecx) = 0;
+  REG(edx) = 0;
+  REG(esi) = 0;
+  REG(edi) = 0;
+  REG(ebp) = 0;
+  REG(eip) = 0;
+  REG(cs) = 0xffff;
+  REG(esp) = 0x100;
+  REG(ss) = 0x30;		/* This is the standard pc bios stack */
+  REG(es) = 0;			/* standard pc es */
+  REG(ds) = 0x40;		/* standard pc ds */
+  REG(fs) = 0;
+  REG(gs) = 0;
+  REG(eflags) = 0;
+}
+
 /* 
  * DANG_BEGIN_FUNCTION cpu_setup
  *
@@ -219,33 +247,35 @@ int cpu_trap_0f (unsigned char *csp, struct sigcontext_struct *scp)
  */
 void cpu_setup(void)
 {
-  extern void int_vector_setup(void);
-  extern void init_emu_cpu (void);
+  unsigned long int stk_ptr, stk_beg, stk_end;
+  FILE *fp;
+  int fd;
 
   int_vector_setup();
 
-  /* ax,bx,cx,dx,si,di,bp,fs,gs can probably can be anything */
-  REG(eax) = 0;
-  REG(ebx) = 0;
-  REG(ecx) = 0;
-  REG(edx) = 0;
-  REG(esi) = 0;
-  REG(edi) = 0;
-  REG(ebp) = 0;
-  REG(eip) = 0x7c00;
-  REG(cs) = 0;			/* Some boot sectors require cs=0 */
-  REG(esp) = 0x100;
-  REG(ss) = 0x30;		/* This is the standard pc bios stack */
-  REG(es) = 0;			/* standard pc es */
-  REG(ds) = 0x40;		/* standard pc ds */
-  REG(fs) = 0;
-  REG(gs) = 0;
-#if 0
-  REG(eflags) |= (IF | VIF | VIP);
+  cpu_reset();
+
+  savefpstate(vm86_fpu_state);
+#ifdef __x86_64__
+  stk_ptr = getregister(rsp);
 #else
-  REG(eflags) |= (VIF | VIP);
+  stk_ptr = getregister(esp);
 #endif
 
+  fd = dup(dosemu_proc_self_maps_fd);
+  if ((fp = fdopen(fd, "r"))) {
+    while(fscanf(fp, "%lx-%lx%*[^\n]", &stk_beg, &stk_end) == 2) {
+      if (stk_ptr >= stk_beg && stk_ptr < stk_end) {
+        stack_init_top = stk_end;
+        stack_init_bot = stk_beg;
+        c_printf("CPU: Stack bottom %#lx, top %#lx, esp=%#lx\n",
+	  stack_init_bot, stack_init_top, stk_ptr);
+	break;
+      }
+    }
+    fclose(fp);
+  }
+
 #ifdef X86_EMULATOR
   if (config.cpuemu) {
     init_emu_cpu();
@@ -253,9 +283,6 @@ void cpu_setup(void)
 #endif
 }
 
-
-
-
 int
 do_soft_int(int intno)
 {
