@@ -1,9 +1,9 @@
 /***************************************************************************
  * 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  *
  *
  *  SIMX86 a Intel 80x86 cpu emulator
@@ -38,13 +38,16 @@
 #include <stdlib.h>
 #include <string.h>		/* for memset */
 #include <setjmp.h>
+#include <sys/time.h>
 #include "emu.h"
+#include "vm86plus.h"
 #include "timers.h"
 #include "pic.h"
 #include "cpu-emu.h"
 #include "emu86.h"
 #include "codegen-arch.h"
 #include "dpmi.h"
+#include "dis8086.h"
 
 #define e_set_flags(X,new,mask) \
 	((X) = ((X) & ~(mask)) | ((new) & (mask)))
@@ -62,23 +65,16 @@
 #ifdef PROFILE
 hitimer_t AddTime, SearchTime, ExecTime, CleanupTime; // for debug
 hitimer_t GenTime, LinkTime;
-extern int MaxDepth, MaxNodes, MaxNodeSize, TotalNodesParsed,
-	TotalNodesExecd, PageFaults, EmuSignals;
-extern int NodesFound, NodesFastFound, NodesNotFound, TreeCleanups;
 #endif
 
 hitimer_t TotalTime;
 static int iniflag = 0;
-extern TNode *LastXNode;
+static int vm86only = 0;
 
 hitimer_t sigEMUtime = 0;
 static hitimer_t lastEMUsig = 0;
 static unsigned long sigEMUdelta = 0;
-#ifdef HOST_ARCH_SIM
-int eTimeCorrect = 0;		// full backtime stretch
-#else
-int eTimeCorrect = 1;		// 1/2 backtime stretch
-#endif
+int eTimeCorrect;
 
 /* This needs to be merged someday with 'mode' */
 volatile int CEmuStat = 0;
@@ -91,14 +87,13 @@ int IsDpmiEmu = 1;
  * which should be considered in time stretching */
 int e_sigpa_count;
 
-int emu_dpmi_retcode = -1;
 int in_vm86_emu = 0;
 int in_dpmi_emu = 0;
 
 SynCPU	TheCPU;
 
 int Running = 0;
-int InCompiledCode = 0;
+volatile int InCompiledCode = 0;
 
 unsigned long trans_addr, return_addr;	// PC
 
@@ -159,15 +154,7 @@ FILE *aLog = NULL;
  * 20	unsigned short gs, __gsh;
  * --------------------------------------------------------------
  */
-struct sigcontext_struct e_scp = {
-	0,0,0,0,0,0,0,0,
-	0,0,0,0,0,0,0,0,
-	0,0,0,
-	0,0,0,0,0,0,0,0,0
-};
-
-extern void dosemu_fault1(int signal, struct sigcontext_struct *scp);
-extern void e_sigalrm(struct sigcontext_struct *scp);
+struct sigcontext_struct e_scp; /* initialized to 0 */
 
 unsigned long io_bitmap[IO_BITMAP_SIZE+1];
 
@@ -176,87 +163,12 @@ unsigned long io_bitmap[IO_BITMAP_SIZE+1];
 unsigned long eTSSMASK = 0;
 #define VFLAGS	(*(unsigned short *)&TheCPU.veflags)
 
-#undef pushb
-#undef pushw
-#undef pushl
-#define pushb(base, ptr, val) \
-__asm__ __volatile__( \
-	"decw %w0\n\t" \
-	"movb %2,0(%1,%0)" \
-	: "=r" (ptr) \
-	: "r" (base), "q" (val), "0" (ptr))
-
-#define pushw(base, ptr, val) \
-__asm__ __volatile__( \
-	"decw %w0\n\t" \
-	"movb %h2,0(%1,%0)\n\t" \
-	"decw %w0\n\t" \
-	"movb %b2,0(%1,%0)" \
-	: "=r" (ptr) \
-	: "r" (base), "q" (val), "0" (ptr))
-
-#define pushl(base, ptr, val) \
-__asm__ __volatile__( \
-	"decw %w0\n\t" \
-	"rorl $16,%2\n\t" \
-	"movb %h2,0(%1,%0)\n\t" \
-	"decw %w0\n\t" \
-	"movb %b2,0(%1,%0)\n\t" \
-	"decw %w0\n\t" \
-	"rorl $16,%2\n\t" \
-	"movb %h2,0(%1,%0)\n\t" \
-	"decw %w0\n\t" \
-	"movb %b2,0(%1,%0)" \
-	: "=r" (ptr) \
-	: "r" (base), "q" (val), "0" (ptr))
-
-#undef popb
-#undef popw
-#undef popl
-#define popb(base, ptr) \
-({ unsigned long __res; \
-__asm__ __volatile__( \
-	"movb 0(%1,%0),%b2\n\t" \
-	"incw %w0" \
-	: "=r" (ptr), "=r" (base), "=q" (__res) \
-	: "0" (ptr), "1" (base), "2" (0)); \
-__res; })
-
-#define popw(base, ptr) \
-({ unsigned long __res; \
-__asm__ __volatile__( \
-	"movb 0(%1,%0),%b2\n\t" \
-	"incw %w0\n\t" \
-	"movb 0(%1,%0),%h2\n\t" \
-	"incw %w0" \
-	: "=r" (ptr), "=r" (base), "=q" (__res) \
-	: "0" (ptr), "1" (base), "2" (0)); \
-__res; })
-
-#define popl(base, ptr) \
-({ unsigned long __res; \
-__asm__ __volatile__( \
-	"movb 0(%1,%0),%b2\n\t" \
-	"incw %w0\n\t" \
-	"movb 0(%1,%0),%h2\n\t" \
-	"incw %w0\n\t" \
-	"rorl $16,%2\n\t" \
-	"movb 0(%1,%0),%b2\n\t" \
-	"incw %w0\n\t" \
-	"movb 0(%1,%0),%h2\n\t" \
-	"incw %w0\n\t" \
-	"rorl $16,%2" \
-	: "=r" (ptr), "=r" (base), "=q" (__res) \
-	: "0" (ptr), "1" (base)); \
-__res; })
-
-
 void e_priv_iopl(int pl)
 {
     pl &= 3;
     TheCPU.eflags = (TheCPU.eflags & ~EFLAGS_IOPL) |
 	(pl << 12);
-    e_printf("eIOPL: set IOPL to %d, flags=%#lx\n",pl,TheCPU.eflags);
+    e_printf("eIOPL: set IOPL to %d, flags=%#x\n",pl,TheCPU.eflags);
 }
 
 void InvalidateSegs(void)
@@ -275,7 +187,7 @@ void InvalidateSegs(void)
 static char ehextab[] = "0123456789abcdef";
 
 /* WARNING - do not convert spaces to tabs! */
-#define ERB_LLEN	0x35
+#define ERB_LLEN	0x39
 #define ERB_LEFTM	5
 #define ERB_L1		0x00
 #define ERB_L2		(ERB_L1+ERB_LLEN)
@@ -284,17 +196,21 @@ static char ehextab[] = "0123456789abcdef";
 #define ERB_L5		(ERB_L1+4*ERB_LLEN)
 #define ERB_L6		(ERB_L5+0x37)
 static char eregbuf[] =
-/*00*/	"\teax=00000000 ebx=00000000 ecx=00000000 edx=00000000\n"
-/*35*/	"\tesi=00000000 edi=00000000 ebp=00000000 esp=00000000\n"
-/*6a*/	"\t vf=00000000  cs=0000      ds=0000      es=0000    \n"
-/*9f*/	"\t fs=0000      gs=0000      ss=0000     flg=00000000\n"
-/*d4*/  "\tstk=0000 0000 0000 0000 0000 0000 0000 0000 0000 0000\n";
-/*10b*/
+/*00*/	"\teax=00000000 ebx=00000000 ecx=00000000 edx=00000000    \n"
+/*39*/	"\tesi=00000000 edi=00000000 ebp=00000000 esp=00000000    \n"
+/*72*/	"\t vf=00000000  cs=0000      ds=0000      es=0000        \n"
+/*ab*/	"\t fs=0000      gs=0000      ss=0000     flg=00000000    \n"
+/*e4*/  "\tstk=0000 0000 0000 0000 0000 0000 0000 0000 0000 0000\n";
+/*11b*/
 
 static inline void exprintl(unsigned long val,char *bf,unsigned int pos)
 {
 	char *p=bf+pos+7;
 	unsigned long v	= val;
+#ifdef __x86_64__
+	v &= 0xffffffffffff;
+	if (v > 0xffffffff) p += 4;
+#endif
 	while (v) { *p-- = ehextab[v&15]; v>>=4; }
 }
 
@@ -323,7 +239,7 @@ char *e_print_regs(void)
 	if (TheCPU.eflags&EFLAGS_VM)
 		exprintl(TheCPU.veflags,buf,(ERB_L3+ERB_LEFTM));
 	else
-		exprintl(dpmi_eflags,buf,(ERB_L3+ERB_LEFTM));
+		exprintl(get_vFLAGS(TheCPU.eflags),buf,(ERB_L3+ERB_LEFTM));
 	exprintw(TheCPU.cs,buf,(ERB_L3+ERB_LEFTM)+13);
 	exprintw(TheCPU.ds,buf,(ERB_L3+ERB_LEFTM)+26);
 	exprintw(TheCPU.es,buf,(ERB_L3+ERB_LEFTM)+39);
@@ -333,9 +249,14 @@ char *e_print_regs(void)
 	exprintl(TheCPU.eflags,buf,(ERB_L4+ERB_LEFTM)+39);
 	if (debug_level('e')>4) {
 		int i;
-		unsigned short *stk = (unsigned short *)(LONG_SS+TheCPU.esp);
-		for (i=(ERB_L5+ERB_LEFTM); i<(ERB_L6-2); i+=5) {
-		   exprintw(*stk++,buf,i);
+		unsigned st = LONG_SS+TheCPU.esp;
+		if (st < 0x1100000 ||
+		    (st > config.dpmi_base &&
+		     st <= config.dpmi_base + config.dpmi * 1024)) {
+			unsigned short *stk = (unsigned short *)(uintptr_t)st;
+			for (i=(ERB_L5+ERB_LEFTM); i<(ERB_L6-2); i+=5) {
+			   exprintw(*stk++,buf,i);
+			}
 		}
 	}
 	else
@@ -343,10 +264,12 @@ char *e_print_regs(void)
 	return buf;
 }
 
-#define GetSegmentBaseAddress(s)	(((s) >= (MAX_SELECTORS << 3))? 0 :\
-					Segments[(s) >> 3].base_addr)
-#define IsSegment32(s)			(((s) >= (MAX_SELECTORS << 3))? 0 :\
-					Segments[(s) >> 3].is_32)
+#if MAX_SELECTORS != 8192
+#error MAX_SELECTORS needs to be 8192
+#endif
+
+#define GetSegmentBaseAddress(s)	((unsigned long)Segments[(s) >> 3].base_addr)
+#define IsSegment32(s)			(Segments[(s) >> 3].is_32)
 
 char *e_print_scp_regs(struct sigcontext_struct *scp, int pmode)
 {
@@ -357,16 +280,16 @@ char *e_print_scp_regs(struct sigcontext_struct *scp, int pmode)
 	int i;
 
 	while (*q) *p++ = *q++; *p=0;
-	exprintl(_eax,buf,(ERB_L1+ERB_LEFTM));
-	exprintl(_ebx,buf,(ERB_L1+ERB_LEFTM)+13);
-	exprintl(_ecx,buf,(ERB_L1+ERB_LEFTM)+26);
-	exprintl(_edx,buf,(ERB_L1+ERB_LEFTM)+39);
-	exprintl(_esi,buf,(ERB_L2+ERB_LEFTM));
-	exprintl(_edi,buf,(ERB_L2+ERB_LEFTM)+13);
-	exprintl(_ebp,buf,(ERB_L2+ERB_LEFTM)+26);
-	exprintl(_esp,buf,(ERB_L2+ERB_LEFTM)+39);
+	exprintl(_rax,buf,(ERB_L1+ERB_LEFTM));
+	exprintl(_rbx,buf,(ERB_L1+ERB_LEFTM)+13);
+	exprintl(_rcx,buf,(ERB_L1+ERB_LEFTM)+26);
+	exprintl(_rdx,buf,(ERB_L1+ERB_LEFTM)+39);
+	exprintl(_rsi,buf,(ERB_L2+ERB_LEFTM));
+	exprintl(_rdi,buf,(ERB_L2+ERB_LEFTM)+13);
+	exprintl(_rbp,buf,(ERB_L2+ERB_LEFTM)+26);
+	exprintl(_rsp,buf,(ERB_L2+ERB_LEFTM)+39);
 	if (pmode & 1)
-		exprintl(dpmi_eflags,buf,(ERB_L3+ERB_LEFTM));
+		exprintl(get_vFLAGS(TheCPU.eflags),buf,(ERB_L3+ERB_LEFTM));
 	else
 		exprintl(TheCPU.veflags,buf,(ERB_L3+ERB_LEFTM));
 	exprintw(_cs,buf,(ERB_L3+ERB_LEFTM)+13);
@@ -387,7 +310,7 @@ char *e_print_scp_regs(struct sigcontext_struct *scp, int pmode)
 		    stk = (unsigned short *)(GetSegmentBaseAddress(_ss)+_LWORD(esp));
 	    }
 	    else
-		stk = (unsigned short *)((_ss<<4)+_LWORD(esp));
+		stk = MK_FP32(_ss,_LWORD(esp));
 	    for (i=(ERB_L5+ERB_LEFTM); i<(ERB_L6-2); i+=5) {
 		exprintw(*stk++,buf,i);
 	    }
@@ -396,10 +319,6 @@ char *e_print_scp_regs(struct sigcontext_struct *scp, int pmode)
 }
 
 
-extern int  dis_8086(unsigned int, const unsigned char *,
-                     unsigned char *, int, unsigned int *, unsigned int *,
-                     unsigned int, int);
-
 char *e_emu_disasm(unsigned char *org, int is32)
 {
    static char buf[512];
@@ -407,15 +326,16 @@ char *e_emu_disasm(unsigned char *org, int is32)
    int rc;
    int i;
    char *p, *p1;
-   unsigned char *code, *org2;
+   unsigned char *code;
+   unsigned int org2;
    unsigned int refseg;
    unsigned int ref;
 
    refseg = TheCPU.cs;
-   org2 = org-LONG_CS;
+   org2 = org-(unsigned char *)(uintptr_t)LONG_CS;
    code = org;
 
-   rc = dis_8086((unsigned long)org, code, frmtbuf, is32, &refseg, &ref, 0, 1);
+   rc = dis_8086(code, frmtbuf, is32, &ref, refseg * 16);
 
    p = buf + sprintf(buf,"%08lx: ",(long)org);
    for (i=0; i<rc && i<8; i++) {
@@ -423,7 +343,7 @@ char *e_emu_disasm(unsigned char *org, int is32)
    }
    sprintf(p,"%20s", " ");
    p1 = buf + 28;
-   p = p1 + sprintf(p1, "%04x:%04x %s", TheCPU.cs, (int)org2, frmtbuf);
+   p = p1 + sprintf(p1, "%04x:%04x %s", TheCPU.cs, org2, frmtbuf);
 
    return buf;
 }
@@ -449,14 +369,16 @@ char *e_scp_disasm(struct sigcontext_struct *scp, int pmode)
       csp2 = org = (unsigned char *)scp->eip;
    }
    else {
-      csp2 = (unsigned char *)(GetSegmentBaseAddress(scp->cs));
+      csp2 = NULL;
+      if (scp->cs <= 0xffff)
+         csp2 = (unsigned char *)(GetSegmentBaseAddress(scp->cs));
       org  = (unsigned char *)(csp2 + scp->eip);
    }
    if ((long)org==lasta) { insrep=1; return buf; } /* skip 'rep xxx' steps */
    lasta = (long)org; insrep = 0;
 
-   rc = dis_8086((unsigned long)org, org, frmtbuf, (pmode&&IsSegment32(scp->cs)? 3:0),
-   	&refseg, &ref, (pmode? (int)csp2 : 0), 1);
+   rc = dis_8086(org, frmtbuf, pmode&&IsSegment32(scp->cs),
+   	&ref, (pmode? (int)csp2 : refseg * 16));
 
    pb = buf;
    org2 = org;
@@ -517,7 +439,7 @@ void GCPrint(unsigned char *cp, unsigned char *cbase, int len)
 {
 	int i;
 	while (len) {
-		dbug_printf(">>> %08x:",cp-cbase);
+		dbug_printf(">>> %08tx:",cp-cbase);
 		for (i=0; (i<16) && len; i++,len--) dbug_printf(" %02x",*cp++);
 		dbug_printf("\n");
 	}
@@ -543,7 +465,7 @@ char *showreg(signed char r)
 	i = (i+8) & 0x18; p = m1+i;	// for side effects in printf
 	if ((ix&0xff)==0xff) { *p=0; return p; }
 	ix = (ix>>2)&0x3f;
-	*((long *)p) = *((long *)(s4[ix]));
+	*((int *)p) = *((int *)(s4[ix]));
 	if ((r&3)==0) p[4]=0; else {
 		p[4]='+'; p[5]=(r&3)+'0'; p[6]=0;
 	}
@@ -587,21 +509,18 @@ static void Reg2Cpu (int mode)
   /* get the protected mode flags. Note that RF and VM are cleared
    * by pushfd (but not by ints and traps). Equivalent to regs32->eflags
    * in vm86.c */
-  __asm__ __volatile__ (" \
-	pushfl\n \
-	popl	%0" \
-	: "=m"(flg) : : "memory");
+  flg = getflags();
   TheCPU.eflags |= (flg & notSAFE_MASK); // which VIP do we get here?
   TheCPU.eflags |= (VM | RF);	// RF is cosmetic...
 
-  if (config.cpuemu>2) {
+  if (config.cpuemu==2) {
     /* a vm86 call switch has been detected.
        Setup flags for the 1st time. */
-    config.cpuemu=2;
+    config.cpuemu=4-vm86only;
   }
 
-  if (debug_level('e')>1) e_printf("Reg2Cpu> vm86=%08lx dpm=%08x emu=%08lx evf=%08lx\n",
-	vm86s.regs.eflags,dpmi_eflags,TheCPU.eflags,TheCPU.veflags);
+  if (debug_level('e')>1) e_printf("Reg2Cpu> vm86=%08lx dpm=%08x emu=%08x evf=%08x\n",
+	vm86s.regs.eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,TheCPU.veflags);
   TheCPU.eax     = vm86s.regs.eax;	/* 2c -> 18 */
   TheCPU.ebx     = vm86s.regs.ebx;	/* 20 -> 00 */
   TheCPU.ecx     = vm86s.regs.ecx;	/* 28 -> 04 */
@@ -622,11 +541,11 @@ static void Reg2Cpu (int mode)
   trans_addr     = LONG_CS + TheCPU.eip;
 
   if (debug_level('e')>1) {
-	if (debug_level('e')==3) e_printf("Reg2Cpu< vm86=%08lx dpm=%08x emu=%08lx evf=%08lx\n%s\n",
-		vm86s.regs.eflags,dpmi_eflags,TheCPU.eflags,TheCPU.veflags,
+	if (debug_level('e')==3) e_printf("Reg2Cpu< vm86=%08lx dpm=%08x emu=%08x evf=%08x\n%s\n",
+		vm86s.regs.eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,TheCPU.veflags,
 		e_print_regs());
-	else e_printf("Reg2Cpu< vm86=%08lx dpm=%08x emu=%08lx evf=%08lx\n",
-		vm86s.regs.eflags,dpmi_eflags,TheCPU.eflags,TheCPU.veflags);
+	else e_printf("Reg2Cpu< vm86=%08lx dpm=%08x emu=%08x evf=%08x\n",
+		vm86s.regs.eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,TheCPU.veflags);
   }
 }
 
@@ -636,8 +555,8 @@ static void Reg2Cpu (int mode)
 static void Cpu2Reg (void)
 {
   int mask;
-  if (debug_level('e')>1) e_printf("Cpu2Reg> vm86=%08lx dpm=%08x emu=%08lx evf=%08lx\n",
-	vm86s.regs.eflags,dpmi_eflags,TheCPU.eflags,TheCPU.veflags);
+  if (debug_level('e')>1) e_printf("Cpu2Reg> vm86=%08lx dpm=%08x emu=%08x evf=%08x\n",
+	vm86s.regs.eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,TheCPU.veflags);
   vm86s.regs.eax = TheCPU.eax;
   vm86s.regs.ebx = TheCPU.ebx;
   vm86s.regs.ecx = TheCPU.ecx;
@@ -662,31 +581,61 @@ static void Cpu2Reg (void)
   vm86s.regs.eflags = (vm86s.regs.eflags & VIP) | 
   			(eVEFLAGS & mask) | (TheCPU.eflags & ~(mask|VIP));
 
-  if (debug_level('e')>1) e_printf("Cpu2Reg< vm86=%08lx dpm=%08x emu=%08lx evf=%08lx\n",
-	vm86s.regs.eflags,dpmi_eflags,TheCPU.eflags,TheCPU.veflags);
+  if (debug_level('e')>1) e_printf("Cpu2Reg< vm86=%08lx dpm=%08x emu=%08x evf=%08x\n",
+	vm86s.regs.eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,TheCPU.veflags);
 }
 
 
 /* ======================================================================= */
 
+static void Scp2Cpu (struct sigcontext_struct *scp)
+{
+#ifdef __x86_64__
+  TheCPU.eax = _eax;
+  TheCPU.ebx = _ebx;
+  TheCPU.ecx = _ecx;
+  TheCPU.edx = _edx;
+  TheCPU.esi = _esi;
+  TheCPU.edi = _edi;
+  TheCPU.ebp = _ebp;
+  TheCPU.esp = _esp;
+
+  TheCPU.eip = _eip;
+  TheCPU.eflags = _eflags;
+
+  TheCPU.cs = _cs;
+  TheCPU.fs = _fs;
+  TheCPU.gs = _gs;
+
+  TheCPU.ds = _ds;
+  TheCPU.es = _es;
+
+  TheCPU.scp_err = _err;
+#else
+  memcpy(&TheCPU.gs,scp,offsetof(struct sigcontext_struct,esp_at_signal));
+#endif
+  TheCPU.ss = _ss;
+  TheCPU.cr2 = _cr2;
+}
+
 /*
  * Return back from fault handling to VM86
  */
 static void Scp2CpuR (struct sigcontext_struct *scp)
 {
-  if (debug_level('e')>1) e_printf("Scp2CpuR> scp=%08lx dpm=%08x fl=%08lx vf=%08lx\n",
-	scp->eflags,dpmi_eflags,TheCPU.eflags,eVEFLAGS);
-  __memcpy(&TheCPU.gs,scp,sizeof(struct sigcontext_struct));
+  if (debug_level('e')>1) e_printf("Scp2CpuR> scp=%08lx dpm=%08x fl=%08x vf=%08x\n",
+	_eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,eVEFLAGS);
+  Scp2Cpu(scp);
   TheCPU.err = 0;
 
   if (in_dpmi) {	// vm86 during dpmi active
 	e_printf("** Scp2Cpu VM in_dpmi\n");
   }
-  TheCPU.eflags = (scp->eflags&(eTSSMASK|0x10ed5)) | 0x20002;
-  trans_addr = ((scp->cs<<4) + scp->eip);
+  TheCPU.eflags = (_eflags&(eTSSMASK|0x10ed5)) | 0x20002;
+  trans_addr = ((_cs<<4) + _eip);
 
-  if (debug_level('e')>1) e_printf("Scp2CpuR< scp=%08lx dpm=%08x fl=%08lx vf=%08lx\n",
-	scp->eflags,dpmi_eflags,TheCPU.eflags,eVEFLAGS);
+  if (debug_level('e')>1) e_printf("Scp2CpuR< scp=%08lx dpm=%08x fl=%08x vf=%08x\n",
+	_eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,eVEFLAGS);
 }
 
 /*
@@ -694,38 +643,62 @@ static void Scp2CpuR (struct sigcontext_struct *scp)
  */
 static void Cpu2Scp (struct sigcontext_struct *scp, int trapno)
 {
-  if (debug_level('e')>1) e_printf("Cpu2Scp> scp=%08lx dpm=%08x fl=%08lx vf=%08lx\n",
-	scp->eflags,dpmi_eflags,TheCPU.eflags,eVEFLAGS);
+  if (debug_level('e')>1) e_printf("Cpu2Scp> scp=%08lx dpm=%08x fl=%08x vf=%08x\n",
+	_eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,eVEFLAGS);
 
   /* setup stack context from cpu registers */
-  __memcpy(scp,&TheCPU.gs,sizeof(struct sigcontext_struct));
+#ifdef __x86_64__
+  _eax = TheCPU.eax;
+  _ebx = TheCPU.ebx;
+  _ecx = TheCPU.ecx;
+  _edx = TheCPU.edx;
+  _esi = TheCPU.esi;
+  _edi = TheCPU.edi;
+  _ebp = TheCPU.ebp;
+  _esp = TheCPU.esp;
 
-  scp->trapno = trapno;
+  _eip = TheCPU.eip;
+  _eflags = TheCPU.eflags;
+
+  _cs = TheCPU.cs;
+  _fs = TheCPU.fs;
+  _gs = TheCPU.gs;
+
+  _ds = TheCPU.ds;
+  _es = TheCPU.es;
+
+  _err = TheCPU.scp_err;
+#else
+  memcpy(scp,&TheCPU.gs,offsetof(struct sigcontext_struct,esp_at_signal));
+#endif
+  _ss = TheCPU.ss;
+  _cr2 = TheCPU.cr2;
+  _trapno = trapno;
   /* Error code format:
    * b31-b16 = 0 (undef)
    * b15-b0  = selector, where b2=LDT/GDT, b1=IDT, b0=EXT
    * (b0-b1 are currently unimplemented here)
    */
-  if (!TheCPU.err) scp->err = 0;		//???
+  if (!TheCPU.err) _err = 0;		//???
 
   if (in_dpmi) {
-    scp->cs = TheCPU.cs;
-    scp->eip = return_addr - LONG_CS;
+    _cs = TheCPU.cs;
+    _eip = return_addr - LONG_CS;
     /* push running flags - same as eflags, RF is cosmetic */
-    scp->eflags = (TheCPU.eflags & (eTSSMASK|0xfd5)) | 0x10002;
+    _eflags = (TheCPU.eflags & (eTSSMASK|0xfd5)) | 0x10002;
   }
   else {
     unsigned long mask;
-    scp->cs  = return_addr >> 16;
-    scp->eip = return_addr & 0xffff;
+    _cs  = return_addr >> 16;
+    _eip = return_addr & 0xffff;
     /* rebuild running flags */
     mask = VIF | eTSSMASK;
     vm86s.regs.eflags = (vm86s.regs.eflags & VIP) | 
   			(eVEFLAGS & mask) | (TheCPU.eflags & ~(mask|VIP));
-    scp->eflags = vm86s.regs.eflags & ~VM;
+    _eflags = vm86s.regs.eflags & ~VM;
   }
-  if (debug_level('e')>1) e_printf("Cpu2Scp< scp=%08lx vm86=%08lx dpm=%08x fl=%08lx vf=%08lx\n",
-	scp->eflags,vm86s.regs.eflags,dpmi_eflags,TheCPU.eflags,eVEFLAGS);
+  if (debug_level('e')>1) e_printf("Cpu2Scp< scp=%08lx vm86=%08lx dpm=%08x fl=%08x vf=%08x\n",
+	_eflags,vm86s.regs.eflags,get_vFLAGS(TheCPU.eflags),TheCPU.eflags,eVEFLAGS);
 }
 
 
@@ -740,10 +713,11 @@ static void Cpu2Scp (struct sigcontext_struct *scp, int trapno)
  */
 static int Scp2CpuD (struct sigcontext_struct *scp)
 {
-  char big; int mode=0;
+  char big; int mode=0, amask, oldfl;
 
   /* copy registers from current dpmi client to our cpu */
-  __memcpy(&TheCPU.gs,scp,sizeof(struct sigcontext_struct));
+  oldfl = TheCPU.eflags & ~(EFLAGS_VM|EFLAGS_RF); /* to preserve IOPL */
+  Scp2Cpu(scp);
 
   mode |= ADDR16;
   TheCPU.err = SetSegProt(mode&ADDR16,Ofs_CS,&big,TheCPU.cs);
@@ -762,15 +736,18 @@ static int Scp2CpuD (struct sigcontext_struct *scp)
   TheCPU.err = SetSegProt(mode&ADDR16,Ofs_GS,&big,TheCPU.gs);
 erseg:
   /* push scp flags, pop eflags - this clears RF,VM */
-  TheCPU.eflags = (scp->eflags & (eTSSMASK|0xed5)) | 2;
-  trans_addr = LONG_CS + scp->eip;
+  amask = (CPL==0? 0:EFLAGS_IOPL_MASK) | (CPL<=IOPL? 0:EFLAGS_IF) |
+    (EFLAGS_VM|EFLAGS_RF) | 2;
+  TheCPU.eflags = (oldfl & amask) | ((_eflags&(eTSSMASK|0xfd7))&~amask);
+
+  trans_addr = LONG_CS + _eip;
   if (debug_level('e')>1) {
-	if (debug_level('e')==3) e_printf("Scp2CpuD%s: %08lx -> %08lx\n\tIP=%08lx:%08lx\n%s\n",
+	if (debug_level('e')==3) e_printf("Scp2CpuD%s: %08lx -> %08x\n\tIP=%08x:%08x\n%s\n",
 			(TheCPU.err? " ERR":""),
-			scp->eflags, TheCPU.eflags, LONG_CS, scp->eip,
+			_eflags, TheCPU.eflags, LONG_CS, _eip,
 			e_print_regs());
-	else e_printf("Scp2CpuD%s: %08lx -> %08lx\n",
-			(TheCPU.err? " ERR":""), scp->eflags, TheCPU.eflags);
+	else e_printf("Scp2CpuD%s: %08lx -> %08x\n",
+			(TheCPU.err? " ERR":""), _eflags, TheCPU.eflags);
   }
   return mode;
 }
@@ -781,8 +758,13 @@ erseg:
 
 void init_emu_cpu (void)
 {
-  extern void init_emu_npu();
-
+  eTimeCorrect = 0;		// full backtime stretch
+#ifdef HOST_ARCH_X86
+  if (!CONFIG_CPUSIM)
+    eTimeCorrect = 1;		// 1/2 backtime stretch
+#endif
+  if (config.cpuemu == 3)
+    vm86only = 1;
   memset(&TheCPU, 0, sizeof(SynCPU));
   TheCPU.cr[0] = 0x13;	/* valid bits: 0xe005003f */
   TheCPU.dr[4] = 0xffff1ff0;
@@ -831,8 +813,11 @@ void init_emu_cpu (void)
  * asynchronous signals because without it any badly-behaved pgm
  * can stop us forever.
  */
-static void e_gen_sigalrm(int sig, struct sigcontext_struct context)
+int e_gen_sigalrm(struct sigcontext_struct *scp)
 {
+	if(config.cpuemu < 2)
+	    return 1;
+
 	/* here we come from the kernel with cs==UCODESEL, as
 	 * the passed context is that of dosemu, NOT that of the
 	 * emulated CPU! */
@@ -850,35 +835,33 @@ static void e_gen_sigalrm(int sig, struct sigcontext_struct context)
 		sigEMUtime += sigEMUdelta;
 		/* we can't call sigalrm() because of the way the
 		 * context parameter is passed. */
-		e_sigalrm(&context);	/* -> signal_save */
+		return 1;	/* -> signal_save */
 	}
 	/* here we return back to dosemu */
+	return 0;
 }
 
-static void e_gen_sigprof(int sig, struct sigcontext_struct context)
+static void e_gen_sigprof(struct sigcontext_struct *scp)
 {
 	e_sigpa_count -= sigEMUdelta;
 	TheCPU.sigprof_pending += 1;
 }
 
 
-Descriptor *alloc_LDT = NULL;
-
 void enter_cpu_emu(void)
 {
 	struct itimerval itv;
-	struct sigaction sa;
+	unsigned int realdelta = config.update / TIMER_DIVISOR;
 
-	if (config.rdtsc==0) {
+	if (config.realcpu < CPU_586) {
 	  fprintf(stderr,"Cannot execute CPUEMU without TSC counter\n");
 	  leavedos(0);
 	}
-	config.cpuemu=3;	/* for saving CPU flags */
-	emu_dpmi_retcode = -1;
+	config.cpuemu=2;	/* for saving CPU flags */
 	GDT = NULL; IDT = NULL;
-	/* allocate the LDT used by dpmi (w/o GDT) */
+	/* use the cached LDT used by dpmi (w/o GDT) */
 	if (LDT==NULL) {
-		alloc_LDT = LDT = (Descriptor *)calloc(LGDT_ENTRIES, sizeof(Descriptor));
+		LDT = (Descriptor *)ldt_buffer;
 		e_printf("LDT allocated at %08lx\n",(long)LDT);
 		TheCPU.LDTR.Base = (long)LDT;
 		TheCPU.LDTR.Limit = 0xffff;
@@ -887,28 +870,39 @@ void enter_cpu_emu(void)
 	if ((ISEG(0x10)==INT10_WATCHER_SEG)&&(IOFF(0x10)==INT10_WATCHER_OFF))
 		IOFF(0x10)=CPUEMU_WATCHER_OFF;
 #endif
-	e_printf("EMU86: switching SIGALRMs\n");
+#ifdef HOST_ARCH_X86
+	TheCPU.unprotect_stub = stub_rep;
+	TheCPU.stub_wri_8 = stub_wri_8;
+	TheCPU.stub_wri_16 = stub_wri_16;
+	TheCPU.stub_wri_32 = stub_wri_32;
+	TheCPU.stub_stk_16 = stub_stk_16;
+	TheCPU.stub_stk_32 = stub_stk_32;
+	TheCPU.stub_movsb = stub_movsb;
+	TheCPU.stub_movsw = stub_movsw;
+	TheCPU.stub_movsl = stub_movsl;
+	TheCPU.stub_stosb = stub_stosb;
+	TheCPU.stub_stosw = stub_stosw;
+	TheCPU.stub_stosl = stub_stosl;
+#endif
+
 	TheCPU.EMUtime = GETTSC();
-	sigEMUdelta = config.realdelta*config.CPUSpeedInMhz;
+	sigEMUdelta = realdelta*config.CPUSpeedInMhz;
 	sigEMUtime = TheCPU.EMUtime + sigEMUdelta;
 	TotalTime = 0;
 #ifdef PROFILE
 	SearchTime = AddTime = ExecTime = CleanupTime =
 	GenTime = LinkTime = 0;
 #endif
-	e_printf("EMU86: delta alrm=%d speed=%d\n",config.realdelta,config.CPUSpeedInMhz);
+	e_printf("EMU86: delta alrm=%d speed=%d\n",realdelta,config.CPUSpeedInMhz);
 	e_sigpa_count = 0;
-	SETSIG(SIGALRM, e_gen_sigalrm);
 
 	itv.it_interval.tv_sec = 0;
-	itv.it_interval.tv_usec = config.realdelta;
+	itv.it_interval.tv_usec = realdelta;
 	itv.it_value.tv_sec = 0;
-	itv.it_value.tv_usec = config.realdelta;
-	e_printf("TIME: using %d usec for updating PROF timer\n", config.realdelta);
+	itv.it_value.tv_usec = realdelta;
+	e_printf("TIME: using %d usec for updating PROF timer\n", realdelta);
 	setitimer(ITIMER_PROF, &itv, NULL);
-	SETSIG(SIGPROF, e_gen_sigprof);
-	NEWSETSIG(SIGFPE, e_emu_fault);
-	NEWSETSIG(SIGSEGV, e_emu_fault);
+	registersig(SIGPROF, e_gen_sigprof);
 
 #ifdef DEBUG_TREE
 	tLog = fopen(DEBUG_TREE_FILE,"w");
@@ -924,26 +918,21 @@ void enter_cpu_emu(void)
 void leave_cpu_emu(void)
 {
 	struct itimerval itv;
-	struct sigaction sa;
 
 	if (config.cpuemu > 1) {
+		iniflag = 0;
 		config.cpuemu=1;
 #ifdef SKIP_EMU_VBIOS
 		if (IOFF(0x10)==CPUEMU_WATCHER_OFF)
 			IOFF(0x10)=INT10_WATCHER_OFF;
 #endif
-		NEWSETSIG(SIGFPE, dosemu_fault);
-		NEWSETSIG(SIGSEGV, dosemu_fault);
-		e_printf("EMU86: switching SIGALRMs\n");
-		NEWSETQSIG(SIGALRM, sigalrm);
-
 		itv.it_interval.tv_sec = 0;
 		itv.it_interval.tv_usec = 0;
 		itv.it_value.tv_sec = 0;
 		itv.it_value.tv_usec = 0;
 		e_printf("TIME: disabling PROF timer\n");
 		setitimer(ITIMER_PROF, &itv, NULL);
-		SETSIG(SIGPROF, SIG_IGN);
+		registersig(SIGPROF, NULL);
 
 		EndGen();
 #ifdef DEBUG_TREE
@@ -954,17 +943,16 @@ void leave_cpu_emu(void)
 #endif
 		mprot_end();
 
-		if (alloc_LDT!=NULL) { free(alloc_LDT); alloc_LDT=LDT=NULL; }
-		GDT = NULL; IDT = NULL;
+		LDT = NULL; GDT = NULL; IDT = NULL;
 		dbug_printf("======================= LEAVE CPU-EMU ===============\n");
 #ifdef PROFILE
-		dbug_printf("Total cpuemu time %16Ld us (incl.trace)\n",TotalTime/config.CPUSpeedInMhz);
-		dbug_printf("Total codgen time %16Ld us\n",GenTime/config.CPUSpeedInMhz);
-		dbug_printf("Total linker time %16Ld us\n",LinkTime/config.CPUSpeedInMhz);
-		dbug_printf("Total exec   time %16Ld us (incl.faults)\n",ExecTime/config.CPUSpeedInMhz);
-		dbug_printf("Total insert time %16Ld us\n",AddTime/config.CPUSpeedInMhz);
-		dbug_printf("Total search time %16Ld us\n",SearchTime/config.CPUSpeedInMhz);
-		dbug_printf("Total clean  time %16Ld us\n",CleanupTime/config.CPUSpeedInMhz);
+		dbug_printf("Total cpuemu time %16lld us (incl.trace)\n",(long long)TotalTime/config.CPUSpeedInMhz);
+		dbug_printf("Total codgen time %16lld us\n",(long long)GenTime/config.CPUSpeedInMhz);
+		dbug_printf("Total linker time %16lld us\n",(long long)LinkTime/config.CPUSpeedInMhz);
+		dbug_printf("Total exec   time %16lld us (incl.faults)\n",(long long)ExecTime/config.CPUSpeedInMhz);
+		dbug_printf("Total insert time %16lld us\n",(long long)AddTime/config.CPUSpeedInMhz);
+		dbug_printf("Total search time %16lld us\n",(long long)SearchTime/config.CPUSpeedInMhz);
+		dbug_printf("Total clean  time %16lld us\n",(long long)CleanupTime/config.CPUSpeedInMhz);
 		dbug_printf("Max tree nodes    %16d\n",MaxNodes);
 		dbug_printf("Max node size     %16d\n",MaxNodeSize);
 		dbug_printf("Max tree depth    %16d\n",MaxDepth);
@@ -974,9 +962,9 @@ void leave_cpu_emu(void)
 		if (TotalNodesExecd) {
 		  unsigned long long k;
 		  k = ((long long)NodesFound * 100UL) / (long long)TotalNodesExecd;
-		  dbug_printf("Find hits         %16d (%Ld%%)\n",NodesFound,k);
+		  dbug_printf("Find hits         %16d (%lld%%)\n",NodesFound,k);
 		  k = ((long long)NodesFastFound * 100UL) / (long long)TotalNodesExecd;
-		  dbug_printf("Find last hits    %16d (%Ld%%)\n",NodesFastFound,k);
+		  dbug_printf("Find last hits    %16d (%lld%%)\n",NodesFastFound,k);
 		}
 		dbug_printf("Page faults       %16d\n",PageFaults);
 		dbug_printf("Signals received  %16d\n",EmuSignals);
@@ -1013,14 +1001,17 @@ static inline int e_revectored(int nr, struct revectored_struct * bitmap)
 
 static void e_should_clean_tree(int i)
 {
-	if ((i==0x21) && ((HI(ax)==0x4b)||(HI(ax)==0x4c))) {
+#ifdef HOST_ARCH_X86
+	if ((i==0x21) && ((HI(ax)==0x4b)||(HI(ax)==0x4c)) &&
+	    !CONFIG_CPUSIM) {
 	    FLUSH_TREE;
 	}
+#endif
 }
 
 static int e_do_int(int i, unsigned char * ssp, unsigned long sp)
 {
-	unsigned long *intr_ptr, segoffs;
+	unsigned int *intr_ptr, segoffs;
 
 	if (_CS == BIOSSEG)
 		goto cannot_handle;
@@ -1028,7 +1019,7 @@ static int e_do_int(int i, unsigned char * ssp, unsigned long sp)
 		goto cannot_handle;
 	if (i==0x21 && e_revectored(_AH,&vm86s.int21_revectored))
 		goto cannot_handle;
-	intr_ptr = (unsigned long *) (i << 2);
+	intr_ptr = MK_FP32(0, i << 2);
 	segoffs = *intr_ptr;
 	if ((segoffs >> 16) == BIOSSEG)
 		goto cannot_handle;
@@ -1056,22 +1047,22 @@ cannot_handle:
 }
 
 
-static int handle_vm86_trap(long *error_code, int trapno)
+static int handle_vm86_trap(int *error_code, int trapno)
 {
 	if ( (trapno==3) || (trapno==1) )
 		return (VM86_TRAP + (trapno << 8));
-	e_do_int(trapno, (unsigned char *) (_SS << 4), _SP);
+	e_do_int(trapno, SEG2LINEAR(_SS), _SP);
 	return -1;
 }
 
 
-static int handle_vm86_fault(long *error_code)
+static int handle_vm86_fault(int *error_code)
 {
 	unsigned char *csp, *ssp, op;
 	unsigned long ip, sp;
 
-	csp = (unsigned char *) (_CS << 4);
-	ssp = (unsigned char *) (_SS << 4);
+	csp = SEG2LINEAR(_CS);
+	ssp = SEG2LINEAR(_SS);
 	sp = _SP;
 	ip = _IP;
 	op = popb(csp, ip);
@@ -1109,8 +1100,9 @@ int e_vm86(void)
 #ifdef SKIP_VM86_TRACE
   int demusav;
 #endif
-  long errcode;
+  int errcode;
 
+#ifdef __i386__
   /* skip emulation of video BIOS, as it is too much timing-dependent */
   if ((!IsV86Emu) || (config.cpuemu<2)
 #ifdef SKIP_EMU_VBIOS
@@ -1119,8 +1111,9 @@ int e_vm86(void)
    ) {
 	s_munprotect(0);
 	InvalidateSegs();
-	return TRUE_VM86(&vm86s);
+	return true_vm86(&vm86s);
   }
+#endif
   if (iniflag==0) enter_cpu_emu();
 
   tt0 = GETTSC();
@@ -1131,38 +1124,37 @@ int e_vm86(void)
   demusav=debug_level('e'); if (debug_level('e')) set_debug_level('e', 1);
 #endif
 //  if (lastEMUsig && (debug_level('e')>1))
-//    e_printf("EMU86: last sig at %Ld, curr=%Ld, next=%Ld\n",lastEMUsig>>16,
+//    e_printf("EMU86: last sig at %lld, curr=%lld, next=%lld\n",lastEMUsig>>16,
 //    	TheCPU.EMUtime>>16,sigEMUtime>>16);
 
  /* This emulates VM86_ENTER */
   /* ------ OUTER LOOP: exit for code >=0 and return to dosemu code */
   do {
     Reg2Cpu(mode);
-#ifdef HOST_ARCH_SIM
-    RFL.valid = V_INVALID;
-#endif
+    if (CONFIG_CPUSIM) {
+      RFL.valid = V_INVALID;
+    }
     /* ---- INNER LOOP: exit with error or code>0 (vm86 fault) ---- */
     do {
       /* enter VM86 mode */
       in_vm86_emu = 1;
       e_printf("INTERP: enter=%08lx\n",trans_addr);
       return_addr = (long)Interp86((char *)trans_addr, mode);
-      e_printf("INTERP: exit=%08lx err=%ld\n",return_addr,TheCPU.err-1);
+      e_printf("INTERP: exit=%08lx err=%d\n",return_addr,TheCPU.err-1);
       xval = TheCPU.err;
       in_vm86_emu = 0;
       /* 0 if ok, else exception code+1 or negative if dosemu err */
       if (xval < 0) {
         error("EMU86: error %d\n", -xval);
         in_vm86=0;
-        leavedos(0);
+        leavedos(1);
       }
       trans_addr = return_addr;
     }
     while (xval==0);
     /* ---- INNER LOOP -- exit for exception ---------------------- */
-#ifdef HOST_ARCH_SIM
-    FlagSync_All();
-#endif
+    if (CONFIG_CPUSIM)
+      FlagSync_All();
 
     Cpu2Reg();
     if (debug_level('e')>1) e_printf("---------------------\n\t   EMU86: EXCP %#x\n", xval-1);
@@ -1195,6 +1187,7 @@ int e_vm86(void)
 #endif
 	      }
 	      break;
+	    case EXCP01_SSTP:
 	    case EXCP03_INT3: {	/* to kernel vm86 */
 		retval=handle_vm86_trap(&errcode,xval-1); /* kernel level */
 	      }
@@ -1252,16 +1245,15 @@ int e_dpmi(struct sigcontext_struct *scp)
 	D_printf("EMU86: DPMI enter at %08lx\n",swa);
   }
 //  if (lastEMUsig)
-//    e_printf("DPM86: last sig at %Ld, curr=%Ld, next=%Ld\n",lastEMUsig>>16,
+//    e_printf("DPM86: last sig at %lld, curr=%lld, next=%lld\n",lastEMUsig>>16,
 //    	TheCPU.EMUtime>>16,sigEMUtime>>16);
 
   /* ------ OUTER LOOP: exit for code >=0 and return to dosemu code */
   do {
     TheCPU.err = 0;
     mode = Scp2CpuD (scp);
-#ifdef HOST_ARCH_SIM
-    RFL.valid = V_INVALID;
-#endif
+    if (CONFIG_CPUSIM)
+      RFL.valid = V_INVALID;
     if (TheCPU.err) {
         error("DPM86: segment error %d\n", TheCPU.err);
         leavedos(0);
@@ -1274,7 +1266,7 @@ int e_dpmi(struct sigcontext_struct *scp)
       in_dpmi_emu = 1;
       e_printf("INTERP: enter=%08lx mode=%04x\n",trans_addr,mode);
       return_addr = (long)Interp86((char *)trans_addr, mode);
-      e_printf("INTERP: exit=%08lx err=%ld\n",return_addr,TheCPU.err-1);
+      e_printf("INTERP: exit=%08lx err=%d\n",return_addr,TheCPU.err-1);
       xval = TheCPU.err;
       in_dpmi_emu = 0;
       /* 0 if ok, else exception code+1 or negative if dosemu err */
@@ -1286,9 +1278,8 @@ int e_dpmi(struct sigcontext_struct *scp)
     }
     while (xval==0);
     /* ---- INNER LOOP -- exit for exception ---------------------- */
-#ifdef HOST_ARCH_SIM
-    FlagSync_All();
-#endif
+    if (CONFIG_CPUSIM)
+      FlagSync_All();
 
     if (debug_level('e')>1) e_printf("DPM86: EXCP %#x eflags=%08lx\n",
 	xval-1, REG(eflags));
@@ -1300,31 +1291,24 @@ int e_dpmi(struct sigcontext_struct *scp)
     E_TIME_STRETCH;
 
     if ((xval==EXCP_SIGNAL) || (xval==EXCP_PICSIGNAL) || (xval==EXCP_STISIGNAL)) {
-	if (debug_level('e')>2) e_printf("DPMI retcode = %d\n",emu_dpmi_retcode);
-	if (emu_dpmi_retcode >= 0) {
-	    retval=emu_dpmi_retcode; emu_dpmi_retcode = -1;
+	if (debug_level('e')>2) e_printf("DPMI sigpending = %d\n",signal_pending);
+	if (signal_pending) {
+	    retval=0;
 	}
     }
     else if (xval==EXCP_GOBACK) {
         retval = 0;
     }
     else {
+	int emu_dpmi_retcode;
 	TotalTime += (GETTSC() - tt0);
-	dpmi_fault(scp);
+	emu_dpmi_retcode = dpmi_fault(scp);
 	tt0 = GETTSC();
-	if (emu_dpmi_retcode >= 0) {
-		retval=emu_dpmi_retcode; emu_dpmi_retcode = -1;
+	if (emu_dpmi_retcode != 0) {
+	    retval=emu_dpmi_retcode; emu_dpmi_retcode = 0;
+	    if (retval == -1)
+		retval = 0;
 	}
-	else {
-	  switch (scp->trapno) {
-	    case 0x6c: case 0x6d: case 0x6e: case 0x6f:
-	    case 0xe4: case 0xe5: case 0xe6: case 0xe7:
-	    case 0xec: case 0xed: case 0xee: case 0xef:
-	    case 0xfa: case 0xfb: retval = -1; break;
-	    default: retval = 0; break;
-	  }
-	}
-	scp->trapno = 0;
     }
   }
   while (retval < 0);
