@@ -1,9 +1,9 @@
 /***************************************************************************
  * 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  *
  *
  *  SIMX86 a Intel 80x86 cpu emulator
@@ -66,16 +66,16 @@ static unsigned short sysxfer[] = {
 	DT_XFER_CG32, DT_NO_XFER, DT_XFER_IG32, DT_XFER_TRP32
 };
 
-static char ofsnam[] =	"??? GS: FS: ES: DS: ??? ??? ??? "
+static char ofsnam[] =	"??? ??? ??? GS: FS: ES: DS: ??? "
 			"??? ??? ??? ??? ??? ??? ??? ??? "
-			"CS: ??? ??? SS: ??? ??? ??? ??? ";
+			"??? ??? CS: ??? SS: ??? ??? ??? ";
 
-static SDTR *ofsseg[] = {
-		NULL,   &GS_DTR,&FS_DTR,&ES_DTR,&DS_DTR,NULL,NULL,NULL,
-		NULL,   NULL,   NULL,   NULL,   NULL,   NULL,NULL,NULL,
-		&CS_DTR,NULL,   NULL,   &SS_DTR,NULL,   NULL,NULL,NULL };
+signed char e_ofsseg[] = {
+	0,   0,       0, Ofs_XGS, Ofs_XFS, Ofs_XES, Ofs_XDS, 0,
+	0,   0,       0,       0,       0,       0,       0, 0,
+	0,   0, Ofs_XCS,       0, Ofs_XSS,       0,       0, 0 };
 
-#define MKOFSNAM(o,b)   ((*((long *)(b))=*((long *)(ofsnam+(o)))), \
+#define MKOFSNAM(o,b)   ((*((int *)(b))=*((int *)(ofsnam+(o)))), \
                           ((b)[3]=0), (b))
 
 int SetSegReal(unsigned short sel, int ofs)
@@ -83,18 +83,13 @@ int SetSegReal(unsigned short sel, int ofs)
 	static char buf[4];
 	SDTR *sd;
 
-	sd = ofsseg[(ofs>>2)];
+	sd = (SDTR *)CPUOFFS(e_ofsseg[(ofs>>2)]);
 
 	CPUWORD(ofs) = sel;
-	if ((sd->Oldsel==sel)&&(sd->Attrib==2)) {
-	    e_printf("SetSeg REAL %s%04x cached\n",MKOFSNAM(ofs,buf),sel);
-	}
-	else {
-	    sd->BoundL = (sel<<4);
-	    sd->BoundH = sd->BoundL + 0xffff;
-	    sd->Oldsel = sel; sd->Attrib = 2;
-	    e_printf("SetSeg REAL %s%04x\n",MKOFSNAM(ofs,buf),sel);
-	}
+	sd->BoundL = (sel<<4);
+	sd->BoundH = sd->BoundL + 0xffff;
+	sd->Attrib = 2;
+	e_printf("SetSeg REAL %s%04x\n",MKOFSNAM(ofs,buf),sel);
 	return 0;
 }
 
@@ -107,7 +102,7 @@ int SetSegProt(int a16, int ofs, unsigned char *big, unsigned long sel)
 	Descriptor *dt;
 	SDTR *sd;
 
-	sd = ofsseg[(ofs>>2)];
+	sd = (SDTR *)CPUOFFS(e_ofsseg[(ofs>>2)]);
 
 	if ((sd->Oldsel==sel)&&((sd->Attrib&3)==1)) {
 	    e_printf("SetSeg PROT %s%04lx cached\n",MKOFSNAM(ofs,buf),sel);
@@ -199,12 +194,13 @@ int SetSegProt(int a16, int ofs, unsigned char *big, unsigned long sel)
 
 	if (big) *big = lbig;
 	if (debug_level('e')>2) {
-		e_printf("PMSEL %#04lx bounds=%08lx:%08lx flg=%04x big=%d\n",
+		e_printf("PMSEL %#04lx bounds=%08x:%08x flg=%04x big=%d\n",
 			sel, sd->BoundL, sd->BoundH, wFlags, lbig&1);
 	}
 	return 0;
 }
 
+#if 0
 /* not (yet) used */
 void ValidateAddr(unsigned char *addr, unsigned short sel)
 {
@@ -227,7 +223,7 @@ void ValidateAddr(unsigned char *addr, unsigned short sel)
 	    dbug_printf("selector %#x not present\n",sel);
 	leavedos(0);
 }
-
+#endif
 
 /* flags with 0x409e mask:
  *	0090	16 RO data
@@ -304,132 +300,6 @@ int hsw_verw(unsigned short sel)
 }
 
 
-/* ======================================================================= */
-
-
-static int emu_read_ldt(char *ptr, unsigned long bytecount)
-{
-	__u32 *lp = (__u32 *)LDT;
-	int i, size=0;
-
-	for (i = 0; (i < LGDT_ENTRIES) && (size < bytecount); i++) {
-		if (*lp) {
-			D_printf("EMU86: read LDT entry %04x: %08x%08x\n",i,
-				lp[1], lp[0]);
-		}
-		if (ptr) {
-		  __memcpy(ptr, lp, LGDT_ENTRY_SIZE);
-		  ptr += LGDT_ENTRY_SIZE;
-		}
-		size += LGDT_ENTRY_SIZE;
-		lp+=2;
-	}
-	e_printf("EMU86: %d LDT entries read\n", i);
-	return size;
-}
-
-
-
-static int emu_update_LDT (struct modify_ldt_ldt_s *ldt_info, int oldmode)
-{
-	static char *xftab[] = { "D16","D32","C16","C32" };
-	Descriptor *lp;
-	int bSelType;
-
-	/* Install the new entry ...  */
-	lp = &LDT[ldt_info->entry_number];
-
-	MKBASE(lp,(long)ldt_info->base_addr);
-	MKLIMIT(lp,(long)ldt_info->limit);
-	/*
-	 * LDT bits 40-43 (5th byte):
-	 *  3  2  1  0
-	 * C/D E  W  A
-	 *  0  0  0		R/O data
-	 *  0  0  1		R/W data, expand-up stack
-	 *  0  1  0		R/O data
-	 *  0  1  1		R/W data, expand-down stack
-	 * C/D C  R  A
-	 *  1  0  0		F/O code, non-conforming
-	 *  1  0  1		F/R code, non-conforming
-	 *  1  1  0		F/O code, conforming
-	 *  1  1  1		F/R code, conforming
-	 *
-	 */
-	lp->type = ((ldt_info->read_exec_only ^ 1) << 1) |
-		   (ldt_info->contents << 2);
-	lp->S = 1;	/* not SYS */
-	lp->DPL = 3;
-	lp->present = (ldt_info->seg_not_present ^ 1);
-	lp->AVL = (oldmode? 0:ldt_info->useable);
-	lp->DB = ldt_info->seg_32bit;
-	lp->gran = ldt_info->limit_in_pages;
-
-	bSelType = ((lp->type>>2)&2) | lp->DB;
-
-	D_printf("EMU86: LDT entry %#x type %s: b=%08x l=%x%s fl=%04x\n",
-		ldt_info->entry_number, xftab[bSelType], DT_BASE(lp),
-		DT_LIMIT(lp), (lp->gran? "fff":""), DT_FLAGS(lp));
-	return 0;
-}
-
-
-static int emu_write_ldt(void *ptr, unsigned long bytecount, int oldmode)
-{
-	int error;
-	struct modify_ldt_ldt_s ldt_info;
-
-	error = -EINVAL;
-	if (bytecount != sizeof(ldt_info)) {
-		dbug_printf("EMU86: write_ldt: bytecount=%ld\n",bytecount);
-		goto out;
-	}
-	__memcpy(&ldt_info, ptr, sizeof(ldt_info));
-
-	error = -EINVAL;
-	if (ldt_info.entry_number >= LGDT_ENTRIES) {
-		dbug_printf("EMU86: write_ldt: entry=%d\n",ldt_info.entry_number);
-		goto out;
-	}
-	if (ldt_info.contents == 3) {
-		if (oldmode) {
-			dbug_printf("EMU86: write_ldt: oldmode\n");
-			goto out;
-		}
-		if (ldt_info.seg_not_present == 0) {
-			dbug_printf("EMU86: write_ldt: seg_not_present\n");
-			goto out;
-		}
-	}
-	error = emu_update_LDT(&ldt_info, oldmode);
-out:
-	return error;
-}
-
-int emu_modify_ldt(int func, void *ptr, unsigned long bytecount)
-{
-	int ret = -ENOSYS;
-
-#if 1
-	{ long *lptr = (long *)ptr;
-	  D_printf("EMU86: modify_ldt %02x %ld [%08lx %08lx %08lx %08lx]\n",
-		func, bytecount, lptr[0], lptr[1], lptr[2], lptr[3] ); }
-#endif
-	switch (func) {
-	case 0:
-		ret = emu_read_ldt((char *)ptr, bytecount);
-		break;
-	case 1:
-		ret = emu_write_ldt(ptr, bytecount, 1);
-		break;
-	case 0x11:
-		ret = emu_write_ldt(ptr, bytecount, 0);
-		break;
-	}
-	return ret;
-}
-
-
 /* ======================================================================= */
 
 int e_larlsl(int mode, unsigned short sel)
@@ -454,7 +324,7 @@ int e_larlsl(int mode, unsigned short sel)
 	/* "if the selector is visible at the current privilege level
 	 *  (modified by the selector's RPL)" */
 	pl = CPL; if ((sel&3)>pl) pl=(sel&3);
-	if (pl > dt->DPL) {
+	if (pl > dt[sel>>3].DPL) {
 		return 0;
 	}
 	/* Intel manual says nothing about present bit */
