@@ -1,9 +1,9 @@
 /***************************************************************************
  * 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  *
  *
  *  SIMX86 a Intel 80x86 cpu emulator
@@ -39,61 +39,80 @@
 
 #include "host.h"
 #include "protmode.h"
+#include <limits.h>
+
+#if ULONG_MAX > 0xffffffffUL
+#define PADDING32BIT(n)
+#else
+#define PADDING32BIT(n) unsigned int padding##n;
+#endif
 
 typedef struct {
 /* offsets are 8-bit signed */
-#define FIELD0		rzero	/* field of SynCPU at offset 00 */
+#define FIELD0		unprotect_stub	/* field of SynCPU at offset 00 */
 /* ------------------------------------------------ */
-/*80	... */
-/*84*/  unsigned long cr2smc;
-/*88*/	unsigned long cr[5];
-/*9c*/	unsigned int  mode;
-/*a0*/	SDTR gs_cache;
-/*ac*/	SDTR fs_cache;
-/*b8*/	SDTR es_cache;
-/*c4*/	SDTR ds_cache;
-/*d0*/	SDTR cs_cache;
-/*dc*/	SDTR ss_cache;
+/*80*/  double   *fpregs;
+/*84*/  PADDING32BIT(1)
+/*88*/	unsigned long long EMUtime;
+/*90*/	SDTR gs_cache;
+/*9c*/	SDTR fs_cache;
+/*a8*/	SDTR es_cache;
+/*b4*/	SDTR ds_cache;
+/*c0*/	SDTR cs_cache;
+/*cc*/	SDTR ss_cache;
 /* ------------------------------------------------ */
-/*e8*/	double   *fpregs;
-/*ec*/	unsigned short fpuc, fpus;
-/*f0*/	unsigned short fpstt, fptag;
-/*f4*/	unsigned long _fni[3];
+/*d8*/  void (*stub_stk_16)(void);
+/*dc*/  PADDING32BIT(2)
+/*e0*/  void (*stub_stk_32)(void);
+/*e4*/  PADDING32BIT(3)
+/*e8*/  void (*stub_wri_8)(void);
+/*ec*/  PADDING32BIT(4)
+/*f0*/  void (*stub_wri_16)(void);
+/*f4*/  PADDING32BIT(5)
+/*f8*/  void (*stub_wri_32)(void);
+/*fc*/  PADDING32BIT(6)
 /* ------------------------------------------------ */
-/*00*/	unsigned long rzero;
-/*04*/	unsigned short gs, __gsh;
-/*08*/	unsigned short fs, __fsh;
-/*0c*/	unsigned short es, __esh;
-/*10*/	unsigned short ds, __dsh;
-/*14*/	unsigned long edi;
-/*18*/	unsigned long esi;
-/*1c*/	unsigned long ebp;
-/*20*/	unsigned long esp;
-/*24*/	unsigned long ebx;
-/*28*/	unsigned long edx;
-/*2c*/	unsigned long ecx;
-/*30*/	unsigned long eax;
-/*34*/	unsigned long trapno;
-/*38*/	unsigned long scp_err;
-/*3c*/	unsigned long eip;
-/*40*/	unsigned short cs, __csh;
-/*44*/	unsigned long eflags;
-/*48*/	unsigned long esp_at_signal;
-/*4c*/	unsigned short ss, __ssh;
-/*50*/	struct _fpstate *fpstate;
-/*54*/	unsigned long oldmask;
-/*58*/	unsigned long cr2;
+/*00*/  void (*unprotect_stub)(void); /* must be at 0 for call (%ebx) */
+/*04*/  PADDING32BIT(7)
+/*08*/	unsigned int rzero;
+/*0c*/	unsigned short gs, __gsh;
+/*10*/	unsigned short fs, __fsh;
+/*14*/	unsigned short es, __esh;
+/*18*/	unsigned short ds, __dsh;
+/*1c*/	unsigned int edi;
+/*20*/	unsigned int esi;
+/*24*/	unsigned int ebp;
+/*28*/	unsigned int esp;
+/*2c*/	unsigned int ebx;
+/*30*/	unsigned int edx;
+/*34*/	unsigned int ecx;
+/*38*/	unsigned int eax;
+/*3c*/	unsigned int trapno;
+/*40*/	unsigned int scp_err;
+/*44*/	unsigned int eip;
+/*48*/	unsigned short cs, __csh;
+/*4c*/	unsigned int eflags;
+/* ----end of i386 sigcontext 1:1 correspondence--- */
+/*50*/	unsigned short ss, __ssh;
+/*54*/	unsigned int cr2;
 /* ------------------------------------------------ */
-/*5c*/	unsigned long sreg1;
-/*60*/  unsigned long dreg1;
-/*64*/	unsigned long xreg1;
-/*68*/	unsigned short sigalrm_pending, sigprof_pending;
-/*6c*/	unsigned long veflags;
-/*70*/		 long err;
-/*74*/	unsigned long long EMUtime;
-/*7c*/	unsigned long StackMask;
+/*58*/	unsigned short fpuc, fpus;
+/*5c*/	unsigned short fpstt, fptag;
+/*60*/	unsigned int _fni[3];
 /* ------------------------------------------------ */
-/*80*/	unsigned long tr[2];
+/*6c*/	unsigned short sigalrm_pending, sigprof_pending;
+/*70*/		 int err;
+/*74*/	unsigned int StackMask;
+/*78*/	unsigned int  mode;
+/*7c*/	unsigned int cr[5]; /* only cr[0] is used in compiled code */
+/* ------------------------------------------------ */
+/*90*/	unsigned int tr[2];
+
+	unsigned int sreg1;
+	unsigned int dreg1;
+	unsigned int xreg1;
+	unsigned int veflags;
+
 /*
  * DR0-3 = linear address of breakpoint 0-3
  * DR4=5 = reserved
@@ -112,7 +131,7 @@ typedef struct {
  *	b24-27= LLRW bp#2
  *	b28-31= LLRW bp#3
  */
-	unsigned long dr[8];
+	unsigned int dr[8];
 	unsigned long mem_ref;
 /* CPU register: base(32) limit(16) */
 	DTR  GDTR;
@@ -124,6 +143,13 @@ typedef struct {
 /* CPU register: sel(16) base(32) limit(16) attr(8) */
 	unsigned short TR_SEL;
 	DTR  TR;
+
+	void (*stub_stosb)(void);
+	void (*stub_stosw)(void);
+	void (*stub_stosl)(void);
+	void (*stub_movsb)(void);
+	void (*stub_movsw)(void);
+	void (*stub_movsl)(void);
 } SynCPU;
 
 extern SynCPU TheCPU;
@@ -138,7 +164,7 @@ extern SynCPU TheCPU;
 
 #define CPUBYTE(o)	*((unsigned char *)CPUOFFS(o))
 #define CPUWORD(o)	*((unsigned short *)CPUOFFS(o))
-#define CPULONG(o)	*((unsigned long *)CPUOFFS(o))
+#define CPULONG(o)	*((unsigned int *)CPUOFFS(o))
 
 #define rEAX		TheCPU.eax
 #define Ofs_EAX		(char)(offsetof(SynCPU,eax)-SCBASE)
@@ -165,7 +191,6 @@ extern SynCPU TheCPU;
 #define Ofs_GS		(char)(offsetof(SynCPU,gs)-SCBASE)
 #define Ofs_EFLAGS	(char)(offsetof(SynCPU,eflags)-SCBASE)
 #define Ofs_ERR		(char)(offsetof(SynCPU,err)-SCBASE)
-#define Ofs_REG1	(char)(offsetof(SynCPU,sreg1)-SCBASE)
 #define Ofs_CR0		(char)(offsetof(SynCPU,cr[0])-SCBASE)
 #define Ofs_CR2		(char)(offsetof(SynCPU,cr2)-SCBASE)
 #define Ofs_STACKM	(char)(offsetof(SynCPU,StackMask)-SCBASE)
@@ -188,6 +213,18 @@ extern SynCPU TheCPU;
 #define Ofs_XFS		(char)(offsetof(SynCPU,fs_cache)-SCBASE)
 #define Ofs_XGS		(char)(offsetof(SynCPU,gs_cache)-SCBASE)
 
+#define Ofs_stub_wri_8	(char)(offsetof(SynCPU,stub_wri_8)-SCBASE)
+#define Ofs_stub_wri_16	(char)(offsetof(SynCPU,stub_wri_16)-SCBASE)
+#define Ofs_stub_wri_32	(char)(offsetof(SynCPU,stub_wri_32)-SCBASE)
+#define Ofs_stub_stk_16	(char)(offsetof(SynCPU,stub_stk_16)-SCBASE)
+#define Ofs_stub_stk_32	(char)(offsetof(SynCPU,stub_stk_32)-SCBASE)
+#define Ofs_stub_movsb	(int)(offsetof(SynCPU,stub_movsb)-SCBASE)
+#define Ofs_stub_movsw	(int)(offsetof(SynCPU,stub_movsw)-SCBASE)
+#define Ofs_stub_movsl	(int)(offsetof(SynCPU,stub_movsl)-SCBASE)
+#define Ofs_stub_stosb	(int)(offsetof(SynCPU,stub_stosb)-SCBASE)
+#define Ofs_stub_stosw	(int)(offsetof(SynCPU,stub_stosw)-SCBASE)
+#define Ofs_stub_stosl	(int)(offsetof(SynCPU,stub_stosl)-SCBASE)
+
 #define rAX		*((unsigned short *)&rEAX)
 #define Ofs_AX		(Ofs_EAX)
 #define Ofs_AXH		(Ofs_EAX+2)
