@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -40,6 +40,7 @@
  * DANG_END_REMARK
  */
 
+#include "config.h"
 
 #ifndef __ELF__
 /*
@@ -55,7 +56,6 @@
 __asm__("___START___: jmp _emulate\n");
 #endif
 
-#include <features.h>
 #include <stdio.h>
 #include <termios.h>
 #include <unistd.h>
@@ -78,26 +78,22 @@ __asm__("___START___: jmp _emulate\n");
 #include <locale.h>
 
 #ifdef __linux__
-#if GLIBC_VERSION_CODE >= 2000
 #include <sys/vt.h>
-#else
-#include "Linux/vt.h"
-#endif
 #include <sys/kd.h>
 #include "Linux/fd.h"
 #include "Linux/hdreg.h"
-#include <sys/vm86.h>
 #include <syscall.h>
 #endif
 
-#include "config.h"
 #include "memory.h"
 
 #ifdef USE_MHPDBG
 #include "mhpdbg.h"
 #endif
+#include "debug.h"
 
 #include "emu.h"
+#include "vm86plus.h"
 
 #include "bios.h"
 #include "termio.h"
@@ -105,15 +101,12 @@ __asm__("___START___: jmp _emulate\n");
 #include "timers.h"
 #include "cmos.h"
 #include "mouse.h"
-#include "dosio.h"
 #include "disks.h"
 #include "xms.h"
-#include "hgc.h"
 #include "ipx.h"		/* TRB - add support for ipx */
 #include "serial.h"
 #include "int.h"
 #include "bitops.h"
-#include "port.h"
 #include "pic.h"
 #include "dpmi.h"
 #include "priv.h"   /* for priv_init */
@@ -123,8 +116,12 @@ __asm__("___START___: jmp _emulate\n");
 #include "utilities.h"
 #include "dos2linux.h"
 #include "iodev.h"
+#include "mapping.h"
 #include "dosemu_config.h"
-
+#include "userhook.h"
+#include "pktdrvr.h"
+#include "dma.h"
+#include "keyb_server.h"
 #include "keyb_clients.h"
 
 #ifdef USE_SBEMU
@@ -134,25 +131,7 @@ __asm__("___START___: jmp _emulate\n");
 #include "cpu-emu.h"
 #endif
 
-extern void     stdio_init(void);
-extern void     time_setting_init(void);
-extern void     low_mem_init(void);
-extern void	mapping_init(void);
-extern void	mapping_close(void);
-extern void	pcibios_init(void);
-extern void     shared_memory_exit(void);
-extern void     restore_vt(u_short);
-extern void     disallocate_vt(void);
-extern void     vm86_GP_fault();
-extern void     config_init(int argc, char **argv);
-extern void	timer_int_engine(void);
-extern void	disk_open(struct disk *dp);
-extern void     print_version(void);
-
-extern void io_select_init(void);
-
-
-jmp_buf NotJEnv;
+sigjmp_buf NotJEnv;
 
 void 
 boot(void)
@@ -181,7 +160,7 @@ boot(void)
     disk_close();
     disk_open(dp);
 
-    buffer = (char *) 0x7c00;
+    buffer = LOWMEM(0x7c00);
 
     if (dp->boot_name) {/* Boot from the specified file */
         int bfd;
@@ -219,19 +198,27 @@ boot(void)
 static inline void 
 vm86plus_init(void)
 {
-    static inline int vm86_old(struct vm86_struct* v86)
-    {
-        int __res;
-        __asm__ __volatile__("int $0x80\n"
-        :"=a" (__res):"a" ((int)113), "b" ((int)v86));
-        return __res;
-    }
+#ifdef __i386__
     if (!vm86_plus(VM86_PLUS_INSTALL_CHECK,0)) return;
-    if (!vm86_old((void *)0xffffff01)) {
+    if (!syscall(SYS_vm86old, (void *)0xffffff01)) {
       fprintf(stderr, "your kernel contains an older (interim) vm86plus version\n\r"
       		      "please upgrade to an newer one\n\r");
     }
-    else fprintf(stderr, "vm86plus service not available in your kernel\n\r");
+    else
+#endif
+    {
+#ifdef X86_EMULATOR
+      warn("WARN: vm86plus service not available in your kernel\n");
+      warn("WARN: using CPU emulation for vm86()\n");
+      if (config.cpuemu < 3) {
+	config.cpuemu = 3;
+      }
+      if (config.cpuemu < 3) config.cpuemu = 3;
+      return;
+#else
+      fprintf(stderr, "vm86plus service not available in your kernel\n\r");
+#endif
+    }
     fflush(stdout);
     fflush(stderr);
     _exit(1);
@@ -245,30 +232,51 @@ module_init(void)
 }
 
 
-static void do_liability_disclaimer_prompt(void)
+void do_liability_disclaimer_prompt(int dosboot, int prompt)
 {
   FILE *f;
   char buf[32];
-  char disclaimer_file_name[256];
+  char *disclaimer_file_name;
   static char text[] =
-  "The Linux DOSEMU, Copyright (C) 2002 the 'DOSEMU-Development-Team'.\n"
+  "\nWelcome to DOSEMU "VERSTR", a DOS emulator"
+#ifdef __linux__
+    " for Linux"
+#endif
+  ".\nCopyright (C) 1992-2006 the 'DOSEMU-Development-Team'.\n"
   "This program is  distributed  in  the  hope that it will be useful,\n"
   "but  WITHOUT  ANY  WARRANTY;   without even the implied warranty of\n"
-  "MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See the file\n"
-  "COPYING for more details.  Use  this  program  at  your  own  risk!\n\n"
-  "By continuing execution of this program,  you  are stating that you\n"
-  "have read the file  COPYING  and the above liability disclaimer and\n"
-  "that you accept these conditions.\n\n"
-  "Enter 'yes' to confirm/continue: ";
+  "MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See the files\n"
+  "COPYING.DOSEMU and COPYING for more details.\n"
+  "Use  this  program  at  your  own  risk!\n\n";
 
+  static char text2[] =
+  "Press ENTER to confirm, and boot DOSEMU, or [Ctrl-C] to abort\n";
 
-  snprintf(disclaimer_file_name, 256, "%s/disclaimer", LOCALDIR);
-  if (exists_file(disclaimer_file_name)) return;
+  disclaimer_file_name = assemble_path(LOCALDIR, "disclaimer", 0);
+  if (exists_file(disclaimer_file_name)) {
+    free(disclaimer_file_name);
+    return;
+  }
 
-  fputs(text, stdout);
-  fgets(buf, sizeof(buf), stdin);
-  if (strncmp(buf, "yes", 3)) {
-    exit(1);
+  if (dosboot) {
+    p_dos_str(text);
+  } else {
+    fputs(text, stdout);
+  }
+
+  buf[0] = '\0';
+  if (prompt) {
+    int size = 0;
+    if (dosboot) {
+      p_dos_str(text2);
+      size = com_biosread(buf, sizeof(buf)-2);
+    } else {
+      fputs(text2, stdout);
+      fgets(buf, sizeof(buf), stdin);
+    }
+    if (buf[0] == 3) {
+      leavedos(1);
+    }
   }
 
   /*
@@ -278,10 +286,11 @@ static void do_liability_disclaimer_prompt(void)
    f = fopen(disclaimer_file_name, "w");
    if (!f) {
      fprintf(stderr, "cannot create %s\n", disclaimer_file_name);
-     exit(1);
+     leavedos(1);
    }
    fprintf(f, "%s%s\n", text, buf);
    fclose(f);
+   free(disclaimer_file_name);
 }
 
 
@@ -310,12 +319,14 @@ int
 emulate(int argc, char **argv)
 #endif
 {
+    char *signalstack[16384];
     int e;
 
     srand(time(NULL));
     memset(&config, 0, sizeof(config));
+    cstack = &signalstack;
 
-    if ((e=setjmp(NotJEnv))) {
+    if ((e=sigsetjmp(NotJEnv, 1))) {
         flush_log();
     	fprintf(stderr,"EMERGENCY JUMP %x!!!\n",e);
     	/* there's no other way to stop a signal 11 from hanging dosemu
@@ -343,12 +354,6 @@ emulate(int argc, char **argv)
      */
     parse_dosemu_users();
 
-    /*
-     * Do the 'liability disclaimer' stuff we need to avoid problems
-     * with laws in some countries.
-     */
-    do_liability_disclaimer_prompt();
-
     /* the transposal of (config_|stdio_)init allows the addition of -o */
     /* to specify a debug out filename, if you're wondering */
 
@@ -368,19 +373,15 @@ emulate(int argc, char **argv)
 #endif
     get_time_init();
     stdio_init();		/* initialize stdio & open debug file */
-    mapping_init();		/* initialize mapping drivers */
     print_version();            /* log version information */
     module_init();
-    low_mem_init();		/* initialize the lower 1Meg */
     time_setting_init();	/* get the startup time */
-    signal_init();		/* initialize sig's & sig handlers */
-    device_init();		/* initialize keyboard, disk, video, etc. */
     cpu_setup();		/* setup the CPU */
-    pcibios_init();
     pci_setup();
-    hardware_setup();		/* setup any hardware */
+    device_init();		/* priv initialization of video etc. */
     extra_port_init();		/* setup ports dependent on config */
-    memory_init();		/* initialize the memory contents */
+    SIG_init();			/* priv part of the signal init */
+    pkt_priv_init();
 
     /* here we include the hooks to possible plug-ins */
     #include "plugin_init.h"
@@ -390,7 +391,40 @@ emulate(int argc, char **argv)
       leavedos(0);
     }
 
-    boot();			/* read the boot sector & get moving */
+    if (can_do_root_stuff && !under_root_login) {
+        g_printf("dropping root privileges\n");
+	open_kmem();
+    }
+    priv_drop();
+    
+    mapping_init();		/* initialize mapping drivers */
+    low_mem_init();		/* initialize the lower 1Meg */
+    map_hardware_ram();         /* map the direct hardware ram */
+    map_video_bios();           /* map (really: copy) the video bios */
+    close_kmem();
+
+    /*
+     * Do the 'liability disclaimer' stuff we need to avoid problems
+     * with laws in some countries.
+     *
+     * show it at the beginning only for terminals and direct console
+     * else the banner code in int.c does it.
+     */
+    if (!config.X)
+	install_dos(0);
+
+    HMA_init();			/* HMA can only be done now after mapping
+                                   is initialized*/
+    memory_init();		/* initialize the memory contents */
+    /* iodev_init() can load plugins, like SDL, that can spawn a thread.
+     * This must be done before initializing signals, or problems ensue.
+     * This also must be done when the signals are blocked, so after
+     * the io_select_init(), which right now blocks the signals. */
+    iodev_init();		/* initialize devices */
+    signal_init();              /* initialize sig's & sig handlers */
+    ems_init();			/* initialize ems */
+    xms_init();			/* initialize xms */
+    dpmi_setup();
 
     if (not_use_sigio)
 	k_printf("Atleast 1 NON-SIGIO file handle in use.\n");
@@ -405,7 +439,11 @@ emulate(int argc, char **argv)
 #endif
     timer_interrupt_init();	/* start sending int 8h int signals */
 
-    while (!fatalerr) {
+    /* remap conventional memory just before booting */
+    mmap_mapping(MAPPING_LOWMEM, 0, config.mem_size * 1024,
+		 PROT_READ | PROT_WRITE | PROT_EXEC, 0);
+
+    while (!fatalerr && !config.exitearly) {
 	loopstep_run_vm86();
     }
 
@@ -415,81 +453,31 @@ emulate(int argc, char **argv)
     return 0;  /* just to make gcc happy */
 }
 
-#if 0    /* disable C-A-D until someone will fix it (if really needed) */
-extern void HMA_MAP(int);
-static int      special_nowait = 0;
-
 void
 dos_ctrl_alt_del(void)
 {
     dbug_printf("DOS ctrl-alt-del requested.  Rebooting!\n");
-    HMA_MAP(1);
-    time_setting_init();
-#ifdef HAVE_KEYBOARD_V1
-    keyb_server_reset();
-#endif
-    iodev_reset();
-    video_config_init();
-    serial_init();
-    dosemu_mouse_init();
-    printer_init();
-    disk_close();
-    disk_init();
-    scr_state_init();
-    clear_screen(READ_BYTE(BIOS_CURRENT_SCREEN_PAGE), 7);
-    special_nowait = 0;
-    p_dos_str("Rebooting DOS.  Be careful...this is partially implemented\r\n");
-    disk_init();
-    cpu_setup();
-    hardware_setup();
-    memcheck_init();		/* lower 1M memory map support */
-    memory_init();
-    boot();
-    timer_interrupt_init();	/* start sending int 8h int signals */
-}
-#endif
-
-static void
-ign_sigs(int sig)
-{
-    static int      timerints = 0;
-    static int      otherints = 0;
-
-    error("signal %d received in leavedos()\n", sig);
-    show_regs(__FILE__, __LINE__);
-    flush_log();
-    if (sig == SIGALRM)
-	timerints++;
-    else
-	otherints++;
-
-#define LEAVEDOS_TIMEOUT (3 * config.freq)
-#define LEAVEDOS_SIGOUT  5
-    if ((sig==11) || (timerints >= LEAVEDOS_TIMEOUT) || (otherints >= LEAVEDOS_SIGOUT)) {
-	error("timed/signalled out in leavedos()\n");
-	longjmp(NotJEnv, 0x11);
+    while (in_dpmi) {
+	in_dpmi_dos_int = 1;
+	dpmi_cleanup();
     }
+    cpu_reset();
 }
 
-int leavedos_recurse_check = 0;
-
-
 /* "graceful" shutdown */
 void
 leavedos(int sig)
 {
-    struct sigaction sa;
     struct itimerval itv;
-    extern void SIG_close(); /* obsolete now; integrated into the kernel */
-    extern void do_r3da_pending (void);	/* emuretrace stuff */
-   
-    if (leavedos_recurse_check)
+
+    if (in_leavedos)
       {
        error("leavedos called recursively, forgetting the graceful exit!\n");
+       gdb_debug();
        flush_log();
-       longjmp(NotJEnv, sig);
+       siglongjmp(NotJEnv, sig);
       }
-    leavedos_recurse_check = 1;
+    in_leavedos++;
     dbug_printf("leavedos(%d|0x%x) called - shutting down\n", sig, sig);
 #if 1 /* BUG CATCHER */
     if (in_vm86) {
@@ -500,25 +488,21 @@ leavedos(int sig)
     in_vm86 = 0;
 
     /* try to notify dosdebug */
-    {
-      extern void mhp_exit_intercept(int errcode);
+#ifdef USE_MHPDBG
+    if (fault_cnt > 0)
+      dbug_printf("leavedos() called from within a signal context!\n");
+    else
       mhp_exit_intercept(sig);
-    }
-
+    g_printf("closing debugger pipes\n");
+    mhp_close();
+#endif
+    
     itv.it_interval.tv_sec = itv.it_interval.tv_usec = 0;
     itv.it_value = itv.it_interval;
     if (setitimer(ITIMER_REAL, &itv, NULL) == -1) {
 	g_printf("can't turn off timer at shutdown: %s\n", strerror(errno));
     }
-    SETSIG(SIGALRM, ign_sigs);
-#ifdef X86_EMULATOR
-    setitimer(ITIMER_PROF, &itv, NULL);
-    SETSIG(SIGPROF, ign_sigs);
-#endif
-    SETSIG(SIGSEGV, ign_sigs);
-    SETSIG(SIGILL, ign_sigs);
-    SETSIG(SIGFPE, ign_sigs);
-    SETSIG(SIGTRAP, ign_sigs);
+    registersig(SIGALRM, NULL);
 
     /* here we include the hooks to possible plug-ins */
     #include "plugin_close.h"
@@ -535,7 +519,8 @@ leavedos(int sig)
 	* My port logic is actually stolen from kd_nosound in the kernel.
 	* 		--EB 21 September 1997
 	*/
-       port_safe_outb(0x61, port_safe_inb(0x61)&0xFC); /* turn off any sound */
+       if (portserver_pid)
+          port_safe_outb(0x61, port_safe_inb(0x61)&0xFC); /* turn off any sound */
     }
 
 #ifdef SIG
@@ -545,10 +530,6 @@ leavedos(int sig)
 
     /* try to regain control of keyboard and video first */
     g_printf("calling keyboard_close\n");
-#ifdef HAVE_KEYBOARD_V1
-    keyb_server_close();
-    keyb_client_close();
-#endif
     iodev_term();
 
 #if defined(X86_EMULATOR)
@@ -560,8 +541,6 @@ leavedos(int sig)
     show_ints(0, 0x33);
     g_printf("calling disk_close_all\n");
     disk_close_all();
-    g_printf("calling video_close\n");
-    video_close();
    
     if (config.emuretrace) {
       do_r3da_pending ();
@@ -570,21 +549,16 @@ leavedos(int sig)
       config.emuretrace = 0;
     }
 
+    /* terminate port server */
+    port_exit();
+
     g_printf("releasing ports and blocked devices\n");
     release_ports();
 
     g_printf("calling shared memory exit\n");
-    shared_memory_exit();
     g_printf("calling HMA exit\n");
     hma_exit();
-    {
-	extern void close_uhook(void);
-	close_uhook();
-    }
-#ifdef USE_MHPDBG
-    g_printf("closing debugger pipes\n");
-    mhp_close();
-#endif
+    close_uhook();
 
     g_printf("calling mapping_close()\n");
     mapping_close();
@@ -595,31 +569,14 @@ leavedos(int sig)
     }
 
 #ifdef IPX
-    {
-      extern void ipx_close(void);
-      ipx_close();
-    }
+    ipx_close();
 #endif
 
     g_printf("calling close_all_printers\n");
     close_all_printers();
 
-    g_printf("calling mouse_close\n");
-    dosemu_mouse_close();
-
     flush_log();
 
-    /* remove per process tmpdir and its contents */
-    priv_drop(); /* drop any priviledges before running system() !! */
-    /* first try rmdir; if the directory is not empty, try system */
-    if (rmdir(TMPDIR_PROCESS) != 0) {
-       char *command = strcatdup("/bin/rm -rf >/dev/null 2>&1 ", TMPDIR_PROCESS);
-       if (command == NULL || system(command) != 0) {
-          g_printf("Failed to remove %s, you'll have to clean it up yourself.\n", 
-         TMPDIR_PROCESS);
-       }
-    }
-
     _exit(sig);
 }
 
@@ -650,15 +607,20 @@ d_ready(int fd)
     } else
 	return (0);
 }
+
+void activate(int con_num)
+{
+    ioctl(kbd_fd, VT_ACTIVATE, con_num);
+}
 #endif
 
-
-void
-activate(int con_num)
+void hardware_run(void)
 {
-    if (in_ioctl) {
-	k_printf("KBD: can't ioctl for activate, in a signal handler\n");
-	do_ioctl(kbd_fd, VT_ACTIVATE, con_num);
-    } else
-	do_ioctl(kbd_fd, VT_ACTIVATE, con_num);
+	dma_run ();
+#ifdef USE_SBEMU
+	run_sb(); /* Beat Karcher to this one .. 8-) - AM */
+#endif
+	keyb_server_run();
+	rtc_run();
 }
+
