@@ -1,28 +1,23 @@
 /* 
+ * (C) 1994 Martin Ludwig  (original code)
+ *          Martin.Ludwig@ruba.rz.ruhr-uni-bochum.de
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
  * hgc.c
  *
  * This file contains support for HGC-cards
- *
- * (C) 1994 Martin Ludwig
- *          Martin.Ludwig@ruba.rz.ruhr-uni-bochum.de
- *
  */
 
-#define HGC_C
-
-
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
 #include <errno.h>
-#include <malloc.h>
+#include <stdlib.h>
 
 #include "config.h"
 #include "emu.h"
@@ -30,80 +25,52 @@
 #include "video.h"
 #include "vc.h"
 #include "vga.h"
-#include "hgc.h"
 #include "mapping.h"
+#include "vgaemu.h"
 
-void set_hgc_page(int page);
-void map_hgc_page( int fullmode );
+#define HGC_BASE0 0xb0000
+#define HGC_BASE1 0xb8000
+#define HGC_PLEN (32 * 1024)
+
+static void hgc_meminit(void);
+static void mda_initialize(void);
+static void mda_reinitialize(void);
+static void set_hgc_page(int page);
+
+static char hgc_Mode;
+static char hgc_Konv;
+static int hgc_Page;
+static int hgc_ctrl;
 
-char hgc_Mode = 0;
-char hgc_Konv = 0;
-int hgc_Page = 0;
-int hgc_ctrl = 0;
 static char * phgcp0;
 static char * phgcp1;
 static char * syncadr;
 
-extern int dos_has_vt;
-
 static void hga_restore_cursor(void)
 {
-  get_perm();
   port_out(10,0x03b4); port_out( 11,0x03b5);
   port_out(11,0x03b4); port_out( 12,0x03b5);
-  release_perm();
 }
 
-
-void poshgacur(int x, int  y)
-{
-  if ( !dos_has_vt )
-    {
-      hga_restore_cursor();
-      return;
-    }
-
-  get_perm();
-
-  /* cursor address high */
-  port_out(14,0x03b4); port_out( ( ( y*80+x ) & 0xFF00 )>>8 ,0x03b5);
-
-  /* cursor address low */
-  port_out(15,0x03b4); port_out( ( y*80+x ) & 0x0FF ,0x03b5);
-
-  /* cursor start */
-  port_out(10,0x03b4); port_out( CURSOR_START(cursor_shape),0x03b5);
-
-  /* cursor end */
-  port_out(11,0x03b4); port_out( CURSOR_END(cursor_shape),0x03b5);
-
-  release_perm();
-
-  return;
-}
-
-void hgc_meminit(void)
+static void hgc_meminit(void)
 {
   char *maperr;
 
-  open_mapping(MAPPING_HGC);
   hgc_ctrl = 1;
 
-  if ( ( phgcp0 = valloc( 32*1024 ) ) == NULL ) /* memory for page 0 */
+  if ( ( phgcp0 = malloc( 32*1024 ) ) == NULL ) /* memory for page 0 */
     hgc_ctrl = 0;
 
-  if ( ( phgcp1 = valloc( 32*1024 ) ) == NULL ) /* memory for page 1 */
-    hgc_ctrl = 0;
-
-  if ( ( syncadr = valloc( 32*1024 ) ) == NULL ) /* memory for sync */
+  if ( ( phgcp1 = malloc( 32*1024 ) ) == NULL ) /* memory for page 1 */
     hgc_ctrl = 0;
 
   /* map real HGC-mem (page 0) */
+  alloc_mapping(MAPPING_HGC | MAPPING_KMEM, HGC_PLEN, HGC_BASE0);
   maperr = (caddr_t) mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
-     HGC_BASE0,
+     (void *) HGC_BASE0,
      HGC_PLEN,
-     PROT_READ | PROT_WRITE,
-     (void *) HGC_BASE0);
+     PROT_READ | PROT_WRITE | PROT_EXEC,
+     HGC_BASE0);
   if ( maperr == (caddr_t) -1 ){
     error("can't map HGC-mem: errno=%d, %s \n",
    errno, strerror(errno));
@@ -112,11 +79,9 @@ void hgc_meminit(void)
   }
 }
 
-void mda_initialize(void)
+static void mda_initialize(void)
 {
-  cursor_shape = 0x0b0c;
-
-  get_perm();
+  unsigned cursor_shape = 0x0b0c;
 
   /* init 6845 Video-Controller */
   /* Values from c't 10/88 page 216 */
@@ -179,15 +144,15 @@ void mda_initialize(void)
   hgc_Mode = 0x28;
   hgc_Page = 0;
   video_initialized = 1;
-
-  /*  permissions _must_ be released, because we had to manage access
-   *  ourself, so we need an exception in vm86!
-   */
-  release_perm();
 }
 
-void mda_reinitialize(void)
+static void map_hgc_page( int fullmode )
 {
+}
+
+static void mda_reinitialize(void)
+{
+  unsigned cursor_shape = 0x0b0c;
 
 #if 0
   /* init 6845 Video-Controller */
@@ -265,8 +230,6 @@ void mda_reinitialize(void)
 #endif
 
 
-  get_perm();
-
   /* init 6845 Video-Controller */
   /* Values from c't 10/88 page 216 */
 
@@ -330,16 +293,10 @@ void mda_reinitialize(void)
   hgc_Mode = 0x28;
   hgc_Page = 0;
   video_initialized = 1;
-
-  /*  permissions _must_ be released, because we had to manage access
-   *  ourself, so we need an exception in vm86!
-   */
-  release_perm();
-
 }
 
 
-void set_hgc_page(int page)
+static void set_hgc_page(int page)
 {
   caddr_t Test;
 
@@ -352,19 +309,15 @@ void set_hgc_page(int page)
       if ( page == 0 ){
 
  /* unmap old visible pages */
- munmap_mapping(MAPPING_HGC, HGC_BASE1, HGC_PLEN);
+ munmap_mapping(MAPPING_HGC, (void *) HGC_BASE1, HGC_PLEN);
 
  /* map real HGC-mem to a place from where we can sync */
- Test = mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
-      syncadr,
+ alloc_mapping(MAPPING_HGC | MAPPING_KMEM, HGC_PLEN, HGC_BASE0);
+ syncadr = mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
+      (void *)-1,
       HGC_PLEN,
-      PROT_READ | PROT_WRITE,
-      (void *) HGC_BASE0);
-
-
-   v_printf("Syncadr: soll %u ist %u\n",
-     (unsigned int) syncadr,
-     (unsigned int) Test);
+      PROT_READ | PROT_WRITE | PROT_EXEC,
+      HGC_BASE0);
 
  /* save old visible page (page 1) */
  memcpy( phgcp1, syncadr, HGC_PLEN );
@@ -376,15 +329,14 @@ void set_hgc_page(int page)
  munmap_mapping(MAPPING_HGC, syncadr, HGC_PLEN );
 
  /* map real HGC-mem to page 0 */
+ alloc_mapping(MAPPING_HGC | MAPPING_KMEM, HGC_PLEN, HGC_BASE0);
  Test = mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
-      HGC_BASE0,
+      (void *) HGC_BASE0,
       HGC_PLEN,
-      PROT_READ | PROT_WRITE,
-      (void *) HGC_BASE0);
+      PROT_READ | PROT_WRITE | PROT_EXEC,
+      HGC_BASE0);
 
-   v_printf("MEM: soll %u ist %u\n",
-     (unsigned int) HGC_BASE1,
-     (unsigned int) Test);
+   v_printf("MEM: soll %#x ist %p\n", HGC_BASE0, Test);
 
  hgc_Page = 0;
  hgc_Mode = hgc_Mode & 0x7f;
@@ -395,19 +347,15 @@ void set_hgc_page(int page)
    v_printf("HGC Map Page 1 allowed\n" );
 
    /* unmap old visible pages */
-   munmap_mapping(MAPPING_HGC, HGC_BASE0, HGC_PLEN);
+   munmap_mapping(MAPPING_HGC, (void *) HGC_BASE0, HGC_PLEN);
 
    /* map real HGC-mem to a place from where we can sync */
-   Test = mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
-    syncadr,
+   alloc_mapping(MAPPING_HGC | MAPPING_KMEM, HGC_PLEN, HGC_BASE0);
+   syncadr = mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
+    (void *)-1,
     HGC_PLEN,
-    PROT_READ | PROT_WRITE,
-    (void *) HGC_BASE0);
-
-
-   v_printf("Syncadr: soll %u ist %u\n",
-     (unsigned int) syncadr,
-     (unsigned int) Test);
+    PROT_READ | PROT_WRITE | PROT_EXEC,
+    HGC_BASE0);
 
    /* save old visible page (page 0) */
    memcpy( phgcp0, syncadr, HGC_PLEN );
@@ -419,14 +367,13 @@ void set_hgc_page(int page)
    munmap_mapping(MAPPING_HGC, syncadr, HGC_PLEN );
 
    /* map real HGC-mem to page 1 */
+   alloc_mapping(MAPPING_HGC | MAPPING_KMEM, HGC_PLEN, HGC_BASE0);
    Test = mmap_mapping(MAPPING_HGC | MAPPING_KMEM,
-        HGC_BASE1,
+        (void *) HGC_BASE1,
         HGC_PLEN,
-        PROT_READ | PROT_WRITE,
-        (void *) HGC_BASE0);
-   v_printf("MEM: soll %u ist %u\n",
-     (unsigned int) HGC_BASE1,
-     (unsigned int) Test);
+        PROT_READ | PROT_WRITE | PROT_EXEC,
+        HGC_BASE0);
+   v_printf("MEM: soll %#x ist %p\n", HGC_BASE1, Test);
 
 
    hgc_Page = 1;
@@ -438,11 +385,7 @@ void set_hgc_page(int page)
   return;
 }
 
-void map_hgc_page( int fullmode )
-{
-}
-
-int hgc_init(void)
+static int hgc_init(void)
 {
   hgc_meminit();
   mda_initialize();
@@ -450,7 +393,7 @@ int hgc_init(void)
   {
 #include "memory.h"
     int i;
-    unsigned short blank = ' ' | (7 << 8), *p = (unsigned short *)SCREEN_ADR(0);
+    unsigned short blank = ' ' | (7 << 8), *p = (unsigned short *)MDA_PHYS_TEXT_BASE;
     for ( i = 0; i < 2000; i++ )
       *p++ = blank;
   }
@@ -458,20 +401,19 @@ int hgc_init(void)
   return 0;
 }
 
-int hgc_setmode(int type, int xsize,int ysize)
+static int hgc_post_init(void)
+{
+  Video_console.init();
+  return 0;
+}
+
+static int hgc_setmode(int type, int xsize,int ysize)
 {
   mda_reinitialize();
     /* port_real_outb(0x03b8, 0x28); *//* 6845: text, visible, page0, cursor */
   return 0;
 }
 
-void do_hgc_update_cursor(void)
-{
-  poshgacur(cursor_col,cursor_row);
-  return;
-}
-
-
 static void hga_close(void)
 {
   /* Later modifications may make it necessary to put the
@@ -481,19 +423,19 @@ static void hga_close(void)
 
   /* The display is in text mode, use the kernel to put the cursor
      at the end of the screen, then scroll up. */
-  fputs("\033[25;80H\n", stdout);
+  fputs("\033[?25h\033[25;80H\n", stdout);
+
+  clear_console_video();
 }
 
 
 struct video_system Video_hgc = {
-   1,                /* is_mapped */
    hgc_init,
+   hgc_post_init,
    hga_close,
    hgc_setmode,
    NULL,             /* update_screen */
-   do_hgc_update_cursor
+   NULL,
+   NULL,
+   NULL
 };
-
-
-#undef HGC_C
-
