@@ -1,8 +1,7 @@
 /* 
- * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -77,12 +76,6 @@ hitimer_t t_vretrace = 0;		/* cf. base/dev/misc/timers.c */
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
-/*
- * Misc_get_input_status_1() is used not only by VGAEmu (cf. emu-i386/ports.c);
- * hence the following #ifdef X_SUPPORT.
- */
-
-#ifdef X_SUPPORT
 /*
  * DANG_BEGIN_FUNCTION Misc_init
  *
@@ -134,6 +127,11 @@ void Misc_init()
 
   vga.config.mono_port = (vga.misc.misc_output & 1) ^ 1;
 
+  if (vga.VGA_mode == 0x6)
+    Misc_set_color_select(0x3f);
+  else if (vga.VGA_mode <= 0x7)
+    Misc_set_color_select(0x30);
+
   misc_msg("Misc_init done\n");
 }
 
@@ -163,10 +161,56 @@ void Misc_set_misc_output(unsigned char data)
   if(u != vga.config.mono_port) {
     vga.config.mono_port = u;
     misc_msg("Misc_set_misc_output: VGA changed to %s mode\n", u ? "mono" : "color");
+    vgaemu_adj_cfg(CFG_MODE_CONTROL, 0);
   }
 }
 
 
+/*
+ * DANG_BEGIN_FUNCTION Misc_set_color_select
+ *
+ * Emulate CGA color select Register 0x3d9.
+ * This is a hardware emulation function.
+ * Don't do background colors for now.
+ *
+ * DANG_END_FUNCTION
+ *
+ */
+void Misc_set_color_select(unsigned char data)
+{
+  int i;
+  int colors = 1 << vga.color_bits;
+
+  misc_deb2("Misc_set_color_select: 0x%02x\n", (unsigned) data);
+  if (vga.mode_class == TEXT) {
+    /* border colour */
+    vga.attr.data[0x11] = data & 0xf;
+    vga.attr.dirty[0x11] = True;
+  } else {
+    if (colors == 2) {
+      vga.attr.data[1] = data & 0xf;
+    } else if (colors == 4) {
+      if (data & 0x20) { /* cyan, magenta and white */
+	vga.attr.data[1] = 3;
+	vga.attr.data[2] = 5;
+	vga.attr.data[3] = 7;
+      } else { /* green, red and brown (yellow) */
+	vga.attr.data[1] = 2;
+	vga.attr.data[2] = 4;
+	vga.attr.data[3] = 6;
+      }
+    } else {
+      return;
+    }
+    vga.attr.data[0] = 0;
+    for (i = 0; i < colors; i++) {
+      vga.attr.dirty[i] = True;
+      if ((data & 0x10) && i > 0) /* bright colors */
+	vga.attr.data[i] |= 0x10;
+    }
+  }
+}
+
 /*
  * DANG_BEGIN_FUNCTION Misc_get_misc_output
  *
@@ -243,8 +287,6 @@ unsigned char Misc_get_input_status_0()
 
   return u;
 }
-#endif	/* X_SUPPORT */
-
 
 /*
  * DANG_BEGIN_FUNCTION Misc_get_input_status_1
@@ -283,6 +325,7 @@ unsigned char Misc_get_input_status_1()
    */
   static unsigned char hretrace = 0, vretrace = 0, first = 1;
   static hitimer_t t_vretrace = 0;
+  static int flip = 0;
   /* Timings are 'ballpark' guesses and may vary from mode to mode, but
      such accuracy is probably not important... I hope. (--adm) */
   static int vvfreq = 17000;	/* 70 Hz - but the best we'll get with
@@ -290,9 +333,7 @@ unsigned char Misc_get_input_status_1()
   hitimer_t t, tdiff;
   unsigned char retval;
 
-#ifdef X_SUPPORT
   vga.attr.flipflop = 0;	 	/* ATTR_INDEX_FLIPFLOP */
-#endif
 
 #ifdef OLD_CGA_SNOW_CODE
   /* old 'cga snow' code with the new variables - looks terrible,
@@ -300,7 +341,7 @@ unsigned char Misc_get_input_status_1()
   hretrace ^= 0x01; vretrace++;
   retval = 0xc6 | hretrace | (vretrace & 0xfc ? 0 : 0x09);
 #else
-  t = pic_sys_time = GETtickTIME(0);
+  t = GETusTIME(0);
 
   if(first) { t_vretrace = t; first = 0; }
   tdiff = t - t_vretrace;
@@ -321,20 +362,26 @@ unsigned char Misc_get_input_status_1()
       vretrace = 0x09; t_vretrace = t;
     }
     else {
+      /* The timer can't be relied upon for the very short intervals necessary
+	 for horizontal retrace emulation
+	 (such as the old "hretrace = (tdiff%49) > 35;").
+	 The following is a crude switch-on switch-off approach after
+	 10 reads taken from DosBox. It seems to work in most cases,
+         (in particular Commander Keen 4) */
+      flip++;
+      if (flip > 20) flip = 0;
       /* We're in horizontal retrace?  If so, just set DE flag, 0 in VR */
-      hretrace = (tdiff%49) > 35;
+      hretrace = flip > 10;
     }
   }
 
   retval = 0xc4 | hretrace | vretrace;
 
-#ifdef X_SUPPORT
   /*
    * Hercules cards use bit 7 for vertical retrace. This bit is not used
    * by VGA cards, so there should be no conflict here. -- sw
    */
   if(vga.config.mono_port && vretrace) retval &= ~0x80;
-#endif
 
 #endif	/* OLD_CGA_SNOW_CODE */
 
