@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -101,23 +101,20 @@
 #define True 1
 #endif
 
-#define seq_msg(x...) v_printf("VGAEmu: " x)
+#define seq_msg(x...) v_printf("VGAEmu: Seq: " x)
 
 #if DEBUG_SEQ >= 1
-#define seq_deb(x...) v_printf("VGAEmu: " x)
+#define seq_deb(x...) v_printf("VGAEmu: Seq: " x)
 #else
 #define seq_deb(x...)
 #endif
 
 #if DEBUG_SEQ >= 2
-#define seq_deb2(x...) v_printf("VGAEmu: " x)
+#define seq_deb2(x...) v_printf("VGAEmu: Seq: " x)
 #else
 #define seq_deb2(x...)
 #endif
 
-#define NEWBITS(a) ((vga.seq.data[ind] ^ data) & (a))
-
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 #include "config.h"
 #include "emu.h"
@@ -149,6 +146,23 @@ static unsigned char seq_ival[16][5] = {
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
+/* calculate font bank offsets from map select value */
+static void seq_map_select(unsigned data)
+{
+  /* map select processing: prepare font mem offsets */
+  /* actually, the single "8k" bit should be ignored on EGA! */
+  vga.seq.fontofs[0] = ((data & 3) << 1) + /* 16k bits */
+    ((data & 0x10) >> 4); /* 8k bit */
+  vga.seq.fontofs[1] = ((data & 0x0c) >> 1) + /* 16k bits */
+    ((data & 0x20) >> 5); /* 8k bit */
+  vga.seq.fontofs[0] <<= 13; /* factor is 8k */
+  vga.seq.fontofs[1] <<= 13; /* factor is 8k */
+  seq_msg("Setting font offsets to 0x%04x/0x%04x (0x%02x)\n",
+          vga.seq.fontofs[0], vga.seq.fontofs[1], data);
+  vga.reconfig.display = 1;
+  vga.reconfig.re_init = 1;
+}
+
 void Seq_init()
 {
   unsigned i, j = 15;
@@ -161,6 +175,13 @@ void Seq_init()
   for(i = 0; i < 5; i++) {
     vga.seq.data[i] = seq_ival[j][i];
   }
+  if(j == 15 ) {
+    /* adjust reg 4 for vesa modes */
+    vga.seq.data[4] = vga.mode_class == TEXT ? 0x02 :
+      vga.pixel_size == 4 ? 0x06 : 0x0a;
+  }
+  /* font offset processing, see below: */
+  seq_map_select(vga.seq.data[3]);
   while(i <= SEQ_MAX_INDEX) vga.seq.data[i++] = 0;
   vga.seq.data[0x0b] = CHIP_ID;
 
@@ -207,9 +228,9 @@ unsigned char Seq_get_index()
 
 void Seq_write_value(unsigned char data)
 {
-  unsigned u, u1, ind = vga.seq.index;
+#define NEWBITS(a) (delta & (a))
+  unsigned u, u1, delta, ind = vga.seq.index;
   unsigned char uc1;
-  unsigned todo_ind, todo[4];
 
   if(ind > SEQ_MAX_INDEX) {
     seq_deb("Seq_write_value: data (0x%02x) ignored\n", (unsigned) data);
@@ -239,8 +260,8 @@ void Seq_write_value(unsigned char data)
     if(vga.seq.data[ind] == data) return;
   }
 
-  for(todo_ind = 0; todo_ind < sizeof todo / sizeof *todo; todo_ind++) todo[todo_ind] = 0;
-  todo_ind = 0;
+  delta = vga.seq.data[ind] ^ data;
+  vga.seq.data[ind] = data;
 
   switch(ind) {
     case 0x00:		/* Reset */
@@ -256,10 +277,16 @@ void Seq_write_value(unsigned char data)
           seq_deb("Seq_write_value: video signal turned %s\n", vga.config.video_off ? "off" : "on");
         }
       }
+      if(NEWBITS(0x9)) {
+        vgaemu_adj_cfg(CFG_SEQ_ADDR_MODE, 0);
+      }
       break;
 
     case 0x02:		/* Map Mask */
-      u = vga.seq.map_mask = data & 0x0f;
+      u = data;
+      if (vga.color_bits != 8)
+        u &= 0xf;
+      vga.seq.map_mask = u;
       u1 = 0;
       if(u) while(!(u & 1)) u >>= 1, u1++;
       seq_deb("Seq_write_value: map mask = 0x%x, write plane = %u\n",
@@ -289,12 +316,14 @@ void Seq_write_value(unsigned char data)
       break;
 
     case 0x03:		/* Character Map Select */
+      if (NEWBITS(0xff))
+        seq_map_select(data);
       break;
 
 
     case 0x04:		/* Memory Mode */
       if(NEWBITS(0x0c)) {
-        todo[todo_ind++] = CFG_SEQ_ADDR_MODE;
+        vgaemu_adj_cfg(CFG_SEQ_ADDR_MODE, 0);
       }
       break;
 
@@ -311,8 +340,9 @@ void Seq_write_value(unsigned char data)
 
     case 0x0e:		/* Trident: Old/New Mode Control 1 */
       if(vga.seq.mode == NEW_MODE) {
-        u = (data ^= 0x02) & 0x0f;	/* XOR 0x02, used for Trident detection */
+        u = (data ^ 0x02) & 0x0f;	/* XOR 0x02, used for Trident detection */
         vga_emu_switch_bank(u);
+        vga.seq.data[0x0e] = data ^ 0x02;
       }
       else {	/* Seq_mode == OLD_MODE */
         /* don't know what to do, we don't support 128K pages -- Erik */
@@ -322,12 +352,6 @@ void Seq_write_value(unsigned char data)
     case 0x0f:		/* Trident: Power Up Mode 2 */
       break;
   }
-
-  vga.seq.data[ind] = data;
-
-  for(todo_ind = 0; todo_ind < sizeof todo / sizeof *todo; todo_ind++) {
-    if(todo[todo_ind]) vgaemu_adj_cfg(todo[todo_ind], 0);
-  }
 }
 
 
