@@ -1,8 +1,9 @@
 /* 
+ * Copyright (c) 1997 Steffen Winterfeldt (original code)
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -17,8 +18,6 @@
  * /REMARK
  * DANG_END_MODULE
  *
- * Copyright (c) 1997 Steffen Winterfeldt
- *
  * DANG_BEGIN_CHANGELOG
  *
  * 1995/1996: Initial version by Erik Mouw and Arjan Filius
@@ -53,16 +52,17 @@
 #define	DEBUG_VBE		/* general info */
 
 
-#include <features.h>
 #include <string.h>
+#include <unistd.h>
+#include <fcntl.h>
 #include "cpu.h"
 #include "emu.h"
 #include "video.h"
 #include "remap.h"
 #include "vgaemu.h"
-
-/* from dosext/dpmi/dpmi.c */
-unsigned long dpmi_GetSegmentBaseAddress(unsigned short);
+#include "dpmi.h"
+#include "vesa.h"
+#include "int.h"
 
 #define VBE_BIOS_MAXPAGES	2	/* max. 8k BIOS size, more than enough */
 
@@ -72,72 +72,6 @@ unsigned long dpmi_GetSegmentBaseAddress(unsigned short);
 #define VBE_OEMSoftRev		0x100		/* 1.0 */
 #define VBE_OEMProductRev	"1.0"
 
-/* VBE error codes */
-#define VBE_ERROR_UNSUP			-1		/* unsupported VBE function */
-#define VBE_ERROR_GENERAL_FAIL		1		/* VBE call failed */
-#define VBE_ERROR_HARDWARE_FAIL		2		/* invalid on current hardware */
-#define VBE_ERROR_MODE_FAIL		3		/* invalid in current mode */
-
-/*
- * macros for easier access to some VBE data structures
- */
-
-/* VBE info block */
-#define VBE_viVBESig		VBE_W(0x00)
-#define VBE_viVESAVersion	VBE_S(0x04)
-#define VBE_viOEMID		VBE_W(0x06)
-#define VBE_viCapabilities	VBE_W(0x0a)
-#define VBE_viModeList		VBE_W(0x0e)
-#define VBE_viMemory		VBE_S(0x12)
-/* VBE 2.0+ */
-#define VBE_viOEMSoftRev	VBE_S(0x14)
-#define VBE_viOEMVendorName	VBE_W(0x16)
-#define VBE_viOEMProdName	VBE_W(0x1a)
-#define VBE_viOEMProductRev	VBE_W(0x1e)
-#define VBE_OEMData		0x100		/* offset to OEM data block */
-
-/* VBE mode info block */
-#define VBE_vmModeAttrib	VBE_S(0x00)
-#define VBE_vmWinAAttrib	VBE_B(0x02)
-#define VBE_vmWinBAttrib	VBE_B(0x03)
-#define VBE_vmWinGran		VBE_S(0x04)
-#define VBE_vmWinSize		VBE_S(0x06)
-#define VBE_vmWinASeg		VBE_S(0x08)
-#define VBE_vmWinBSeg		VBE_S(0x0a)
-#define VBE_vmWinFuncPtr	VBE_W(0x0c)
-#define VBE_vmBytesPLine	VBE_S(0x10)
-#define VBE_vmXRes		VBE_S(0x12)
-#define VBE_vmYRes		VBE_S(0x14)
-#define VBE_vmXCharSize		VBE_B(0x16)
-#define VBE_vmYCharSize		VBE_B(0x17)
-#define VBE_vmNumPlanes		VBE_B(0x18)
-#define VBE_vmBitsPPixel	VBE_B(0x19)
-#define VBE_vmBanks		VBE_B(0x1a)
-#define VBE_vmMemModel		VBE_B(0x1b)
-#define VBE_vmBankSize		VBE_B(0x1c)
-#define VBE_vmPages		VBE_B(0x1d)
-#define VBE_vmReserved1		VBE_B(0x1e)
-#define VBE_vmRedMaskSize	VBE_B(0x1f)
-#define VBE_vmRedFieldPos	VBE_B(0x20)
-#define VBE_vmGreenMaskSize	VBE_B(0x21)
-#define VBE_vmGreenFieldPos	VBE_B(0x22)
-#define VBE_vmBlueMaskSize	VBE_B(0x23)
-#define VBE_vmBlueFieldPos	VBE_B(0x24)
-#define VBE_vmRsvdMaskSize	VBE_B(0x25)
-#define VBE_vmRsvdFieldPos	VBE_B(0x26)
-#define VBE_vmDirectColor	VBE_B(0x27)
-#define VBE_vmPhysBasePtr	VBE_W(0x28)
-#define VBE_vmOffScreenOfs	VBE_W(0x2c)
-#define VBE_vmOffScreenMem	VBE_S(0x30)
-
-/* some useful macros */
-#define VBE_B(a) VBE_BUFFER[a]
-#define VBE_S(a) *((unsigned short *) (VBE_BUFFER + a))
-#define VBE_W(a) *((unsigned *) (VBE_BUFFER + a))
-
-#define VBE_SEG_OFS(a, b)	(((a) << 16) + ((b) & 0xffff))
-
-
 /*
  * Type of display we are attached to. This info is needed to report
  * the color masks of hi/true-color displays back to the DOS app.
@@ -164,21 +98,6 @@ static int vbe_pm_interface(unsigned);
 static int vbe_power_state(unsigned, unsigned);
 
 
-/*
- * The actual VGA/VBE BIOS is in vesabios.S and vesabios_pm.S. These are
- * some labels pointing to interesting code portions. They are used in
- * vbe_init() to setup the struct vgaemu_bios which holds all relevant
- * info about our BIOS.
- */
-
-extern void vgaemu_bios_start();
-extern void vgaemu_bios_prod_name();
-extern void vgaemu_bios_win_func();
-extern void vgaemu_bios_end();
-extern void vgaemu_bios_pm_interface();
-extern void vgaemu_bios_pm_interface_end();
-
-
 /*
  * DANG_BEGIN_FUNCTION vbe_init
  *
@@ -219,63 +138,67 @@ void vbe_init(vgaemu_display_type *vedt)
     save_function_flags:0x3f 
   };
 
-  vbe_screen = *vedt;
-
   memset(dos_vga_bios, 0, VBE_BIOS_MAXPAGES << 12);	/* one page */
   memcpy(dos_vga_bios, vgaemu_bios_start, bios_ptr);
 
   vgaemu_bios.prod_name = (char *) vgaemu_bios_prod_name - (char *) vgaemu_bios_start;
 
-  i = (char *) vgaemu_bios_pm_interface_end - (char *) vgaemu_bios_pm_interface;
+  if (Video->setmode) {
+    i = (char *) vgaemu_bios_pm_interface_end - (char *) vgaemu_bios_pm_interface;
 
-  if(i + bios_ptr > (VBE_BIOS_MAXPAGES << 12) - 8) {
-    error("VBE: vbe_init: protected mode interface to large, disabling\n");
-    vgaemu_bios.vbe_pm_interface_len =
-    vgaemu_bios.vbe_pm_interface = 0;
-  }
+    if(i + bios_ptr > (VBE_BIOS_MAXPAGES << 12) - 8) {
+      error("VBE: vbe_init: protected mode interface to large, disabling\n");
+      vgaemu_bios.vbe_pm_interface_len =
+	vgaemu_bios.vbe_pm_interface = 0;
+    }
 
-  vgaemu_bios.vbe_pm_interface_len = i;
-  vgaemu_bios.vbe_pm_interface = bios_ptr;
-  memcpy(dos_vga_bios + bios_ptr, vgaemu_bios_pm_interface, i);
-  bios_ptr += i;
+    vgaemu_bios.vbe_pm_interface_len = i;
+    vgaemu_bios.vbe_pm_interface = bios_ptr;
+    memcpy(dos_vga_bios + bios_ptr, vgaemu_bios_pm_interface, i);
+    bios_ptr += i;
 
-  bios_ptr = (bios_ptr + 3) & ~3;
-  vgaemu_bios.vbe_mode_list = bios_ptr;
+    bios_ptr = (bios_ptr + 3) & ~3;
+    vgaemu_bios.vbe_mode_list = bios_ptr;
 
-  /* set up video mode list */
-  for(i = 0x100; i <= vgaemu_bios.vbe_last_mode; i++) {
-    if((vmi = vga_emu_find_mode(i, NULL))) {
-      if(vmi->VESA_mode != -1 && bios_ptr < ((VBE_BIOS_MAXPAGES << 12) - 4)) {
-        *((unsigned short *) (dos_vga_bios + bios_ptr)) = vmi->VESA_mode;
-        bios_ptr += 2;
+    /* set up video mode list */
+    for(i = 0x100; i <= vgaemu_bios.vbe_last_mode; i++) {
+      if((vmi = vga_emu_find_mode(i, NULL))) {
+	if(vmi->VESA_mode != -1 && bios_ptr < ((VBE_BIOS_MAXPAGES << 12) - 4)) {
+	  *((unsigned short *) (dos_vga_bios + bios_ptr)) = vmi->VESA_mode;
+	  bios_ptr += 2;
+	}
       }
     }
+
+    *((short *) (dos_vga_bios + bios_ptr)) = -1;
+    bios_ptr += 2;
+
+    /* add fonts */
+    vgaemu_bios.font_8 = bios_ptr;
+    memcpy(dos_vga_bios + bios_ptr, vga_rom_08, sizeof vga_rom_08);
+    bios_ptr += sizeof vga_rom_08;
+
+    vgaemu_bios.font_14 = bios_ptr;
+    memcpy(dos_vga_bios + bios_ptr, vga_rom_14, sizeof vga_rom_14);
+    bios_ptr += sizeof vga_rom_14;
+
+    vgaemu_bios.font_16 = bios_ptr;
+    memcpy(dos_vga_bios + bios_ptr, vga_rom_16, sizeof vga_rom_16);
+    bios_ptr += sizeof vga_rom_16;
+
+    vgaemu_bios.font_14_alt = bios_ptr;
+    memcpy(dos_vga_bios + bios_ptr, vga_rom_14_alt, sizeof vga_rom_14_alt);
+    bios_ptr += sizeof vga_rom_14_alt;
+
+    vgaemu_bios.font_16_alt = bios_ptr;
+    memcpy(dos_vga_bios + bios_ptr, vga_rom_16_alt, sizeof vga_rom_16_alt);
+    bios_ptr += sizeof vga_rom_16_alt;
+  } else {
+    /* only support the initial video mode, can't change it */
+    vgaemu_bios_functionality_table.modes[0] = 1 << video_mode;
+    vgaemu_bios_functionality_table.modes[1] =
+      vgaemu_bios_functionality_table.modes[2] = 0;
   }
-
-  *((short *) (dos_vga_bios + bios_ptr)) = -1;
-  bios_ptr += 2;
-
-  /* add fonts */
-  vgaemu_bios.font_8 = bios_ptr;
-  memcpy(dos_vga_bios + bios_ptr, vga_rom_08, sizeof vga_rom_08);
-  bios_ptr += sizeof vga_rom_08;
-
-  vgaemu_bios.font_14 = bios_ptr;
-  memcpy(dos_vga_bios + bios_ptr, vga_rom_14, sizeof vga_rom_14);
-  bios_ptr += sizeof vga_rom_14;
-
-  vgaemu_bios.font_16 = bios_ptr;
-  memcpy(dos_vga_bios + bios_ptr, vga_rom_16, sizeof vga_rom_16);
-  bios_ptr += sizeof vga_rom_16;
-
-  vgaemu_bios.font_14_alt = bios_ptr;
-  memcpy(dos_vga_bios + bios_ptr, vga_rom_14_alt, sizeof vga_rom_14_alt);
-  bios_ptr += sizeof vga_rom_14_alt;
-
-  vgaemu_bios.font_16_alt = bios_ptr;
-  memcpy(dos_vga_bios + bios_ptr, vga_rom_16_alt, sizeof vga_rom_16_alt);
-  bios_ptr += sizeof vga_rom_16_alt;
-  
   vgaemu_bios.functionality = bios_ptr;
   memcpy(dos_vga_bios + bios_ptr, &vgaemu_bios_functionality_table,
       sizeof vgaemu_bios_functionality_table);
@@ -286,15 +209,34 @@ void vbe_init(vgaemu_display_type *vedt)
   dos_vga_bios[2] = (bios_ptr + ((1 << 9) - 1)) >> 9;
   vgaemu_bios.pages = (bios_ptr + ((1 << 12) - 1)) >> 12;
 
+  if (config.vbios_file) {
+    /* EXPERIMENTAL: load and boot the Bochs BIOS */
+    int fd = open(config.vbios_file, O_RDONLY);
+    if (fd != -1) {
+      read(fd, (void*)0xc0000, 32768);
+      close(fd);
+      vgaemu_bios.pages = 8;
+      config.vbios_post = 1;
+    }
+  }
+
+  memcheck_addtype('V', "VGAEMU Video BIOS");
+  memcheck_reserve('V', (size_t)dos_vga_bios, vgaemu_bios.pages << 12);
+
   if(!config.X_pm_interface) {
     v_printf("VBE: vbe_init: protected mode interface disabled\n");
   }
 
   v_printf(
-    "VBE: vbe_init: %d pages for VGA BIOS, vga.mem.base = 0x%x\n",
-    vgaemu_bios.pages, (unsigned) vga.mem.base
+    "VBE: vbe_init: %d pages for VGA BIOS, vga.mem.base = %p\n",
+    vgaemu_bios.pages, vga.mem.base
   );
 
+  if(!vedt)
+    return;
+
+  vbe_screen = *vedt;
+
   v_printf(
     "VBE: vbe_init: real display: bits/pixel = %d, color bits (rgb) = %d/%d/%d\n",
     vbe_screen.bits, vbe_screen.r_bits, vbe_screen.g_bits, vbe_screen.b_bits
@@ -931,6 +873,7 @@ int vbe_scan_length(unsigned sub_func, unsigned scan_len)
         }
         vga.scan_len = u0;
         vga.reconfig.display = 1;
+	vga.reconfig.re_init = 1;
         break;
 
       case 1:	/* get current config */
@@ -947,6 +890,7 @@ int vbe_scan_length(unsigned sub_func, unsigned scan_len)
         }
         vga.scan_len = u0;
         vga.reconfig.display = 1;
+	vga.reconfig.re_init = 1;
         break;
 
       case 3:	/* get maximum config */
@@ -1153,7 +1097,7 @@ int vbe_palette_data(unsigned sub_func, unsigned len, unsigned first, unsigned c
     sub_func &= ~0x40;
     ofs = (_SI << 16) + _DI;
     base = dpmi_GetSegmentBaseAddress(_BP);
-    buffer = (unsigned char *) (base + ofs);
+    buffer = (unsigned char *)(uintptr_t) (base + ofs);
 #ifdef DEBUG_VBE
     v_printf(
       "VBE: [0x%02x.%u] vbe_palette_data: called via pm interface, es.sel = 0x%04x, es.base = 0x%08x, es.ofs = 0x%08x\n",
