@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
  
 /*
@@ -256,10 +256,7 @@
 #endif
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
-#include <features.h>
-#if GLIBC_VERSION_CODE == 2000
-#include <sigcontext.h>
-#endif
+#include <signal.h>
 #include <sys/types.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -270,6 +267,7 @@
 
 #include "cpu.h"		/* root@sjoerd: for context structure */
 #include "emu.h"
+#include "int.h"
 #include "dpmi.h"
 #include "port.h"
 #include "video.h"
@@ -279,16 +277,11 @@
 #include "vgaemu.h"
 #include "priv.h"
 #include "mapping.h"
+#include "utilities.h"
 
 /* table with video mode definitions */
 #include "vgaemu_modelist.h"
 
-/* from dosext/dpmi/dpmi.c */
-unsigned long dpmi_GetSegmentBaseAddress(unsigned short);
-
-/* env/video/video.c */
-int load_file(char *name, int foffset, char *mstart, int msize);
-
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 /*
  * functions local to this file
@@ -442,6 +435,10 @@ int VGA_emulate_outb(ioport_t port, Bit8u value)
       if(!vga.config.mono_port) CRTC_write_value(value);
       break;
 
+    case COLOR_SELECT:			/* 0x3d9 */
+      if(!vga.config.mono_port) Misc_set_color_select(value);
+      break;
+
     case FEATURE_CONTROL_W:		/* 0x3da */
       if(!vga.config.mono_port) Misc_set_feature_ctrl(value);
       break;
@@ -616,7 +613,7 @@ static Bit32u color2pixels[16] = {0,0xff,0xff00,0xffff,0xff0000,0xff00ff,0xffff0
 
 #define VGALatch	vga.latch
 
-unsigned char Logical_VGA_read(unsigned offset)
+static unsigned char Logical_VGA_read(unsigned offset)
 {
   
   switch (ReadMode) {
@@ -731,7 +728,7 @@ static Bit32u Logical_VGA_CalcNewVal (unsigned char value)
   return new_val;
 }
 
-void Logical_VGA_write(unsigned offset, unsigned char value)
+static void Logical_VGA_write(unsigned offset, unsigned char value)
 {
   unsigned vga_page;
   unsigned char *p;
@@ -765,6 +762,73 @@ void Logical_VGA_write(unsigned offset, unsigned char value)
   
 }
 
+unsigned char vga_read(const unsigned char *addr)
+{
+  if (!vga.inst_emu)
+    return READ_BYTE(addr);
+  return Logical_VGA_read((uintptr_t)addr - vga.mem.bank_base);
+}
+
+unsigned short vga_read_word(const unsigned short *addr)
+{
+  if (!vga.inst_emu)
+    return READ_WORD(addr);
+  return vga_read((const unsigned char *)addr) |
+	 vga_read((const unsigned char *)addr + 1) << 8;
+}
+
+void vga_write(unsigned char *addr, unsigned char val)
+{
+  if (!vga.inst_emu) {
+    WRITE_BYTE(addr, val);
+    return;
+  }
+  Logical_VGA_write((uintptr_t)addr - vga.mem.bank_base, val);
+}
+
+void vga_write_word(unsigned short *addr, unsigned short val)
+{
+  if (!vga.inst_emu) {
+    WRITE_WORD(addr, val);
+    return;
+  }
+  vga_write((unsigned char *)addr, val & 0xff);
+  vga_write((unsigned char *)addr + 1, val >> 8);
+}
+
+void memcpy_to_vga(void *dst, const void *src, size_t len)
+{
+  int i;
+  if (!vga.inst_emu) {
+    MEMCPY_2DOS(dst, src, len);
+    return;
+  }
+  for (i = 0; i < len; i++)
+    vga_write((unsigned char *)dst + i, READ_BYTE((unsigned char *)src + i));
+}
+
+void memcpy_from_vga(void *dst, const void *src, size_t len)
+{
+  int i;
+  if (!vga.inst_emu) {
+    MEMCPY_2UNIX(dst, src, len);
+    return;
+  }
+  for (i = 0; i < len; i++) {
+    WRITE_BYTE((unsigned char *)dst + i, vga_read((unsigned char *)src + i));
+  }
+}
+
+void vga_memcpy(void *dst, const void *src, size_t len)
+{
+  int i;
+  if (!vga.inst_emu) {
+    MEMCPY_DOS2DOS(dst, src, len);
+    return;
+  }
+  for (i = 0; i < len; i++)
+    vga_write((unsigned char *)dst + i, vga_read((unsigned char *)src + i));
+}
 
 /*
  * DANG_BEGIN_FUNCTION vga_emu_fault
@@ -904,7 +968,7 @@ int vga_emu_fault(struct sigcontext_struct *scp, int pmode)
   if(vga_page < vga.mem.pages) {
     vga.mem.dirty_map[vga_page] = 1;
 #ifdef X86_EMULATOR
-    if (config.cpuemu>1) {
+    if (config.cpuemu>1 && _cs == getsegment(cs)) {
 	error("VGAEmu: CPU emulation collision, should not be here\n");
 	leavedos(0x4945);
     }
@@ -914,11 +978,22 @@ int vga_emu_fault(struct sigcontext_struct *scp, int pmode)
       vga_emu_adjust_protection(vga_page, page_fault);
     }  
     else {
+#if 1
+      /* XXX Hack: dosemu touched the protected page of video mem, which is
+       * a bug. However for the video modes that do not require instremu, we
+       * can allow that access. For the planar modes we cant and will fail :(
+       */
+      if(pmode && _cs == getsegment(cs)) {
+	error("BUG: dosemu touched the protected video memory!!!\n");
+	return False;
+      }
+#endif
+
       /* A VGA mode PL4/PL2 video memory access has been trapped 
        * while we are using X.  Leave the display page read/write-protected
        * so that each instruction that accesses it can be trapped and
        * simulated. */
-        instr_emu(scp, pmode);
+        instr_emu(scp, pmode, 0);
     }
   }
   return True;
@@ -987,7 +1062,7 @@ int vga_emu_protect_page(unsigned page, int prot)
     page, prot == RW ? "RW" : prot == RO ? "RO" : "NONE"
   );
 
-  i = mprotect_mapping(MAPPING_VGAEMU, (void *) (page << 12), 1 << 12, sys_prot);
+  i = mprotect_mapping(MAPPING_VGAEMU, (void *) (uintptr_t)(page << 12), 1 << 12, sys_prot);
 
   if(i == -1) {
     sys_prot = 0xfe;
@@ -1192,7 +1267,7 @@ int vga_emu_adjust_protection(unsigned page, unsigned mapped_page)
 
 static int vga_emu_map(unsigned mapping, unsigned first_page)
 {
-  int i;
+  void *i;
   unsigned u;
   vga_mapping_type *vmt;
   int prot;
@@ -1221,11 +1296,11 @@ static int vga_emu_map(unsigned mapping, unsigned first_page)
       prot = VGA_EMU_RW_PROT;
   }
 
-  i = (int) mmap_mapping(MAPPING_VGAEMU | MAPPING_ALIAS,
-    (caddr_t) (vmt->base_page << 12), vmt->pages << 12,
+  i = alias_mapping(MAPPING_VGAEMU,
+    (void *)(uintptr_t) (vmt->base_page << 12), vmt->pages << 12,
     prot, vga.mem.base + (first_page << 12));
 
-  if(i == -1) {
+  if(i == MAP_FAILED) {
     prot = 0xfe;
     switch(errno) {
       case EINVAL: prot = 0xfd; break;
@@ -1238,7 +1313,7 @@ static int vga_emu_map(unsigned mapping, unsigned first_page)
     vgaemu_update_prot_cache(vmt->base_page + u, prot);
   }
 
-  if(i == -1) return 3;
+  if(i == MAP_FAILED) return 3;
 
   vmt->first_page = first_page;
 
@@ -1258,7 +1333,7 @@ static int vga_emu_map(unsigned mapping, unsigned first_page)
 
 static int vgaemu_unmap(unsigned page)
 {
-  int i;
+  void *i;
 
   if(
     page < 0xa0 ||
@@ -1266,14 +1341,14 @@ static int vgaemu_unmap(unsigned page)
     (!vga.config.mono_support && page >= 0xb0 && page < 0xb8)
   ) return 1;
 
-  *((volatile char *)(vga.mem.scratch_page << 12));
+  *((volatile char *)vga.mem.scratch_page);
 
-  i = (int) mmap_mapping(MAPPING_VGAEMU | MAPPING_ALIAS,
+  i = alias_mapping(MAPPING_VGAEMU,
     (caddr_t) (page << 12), 1 << 12,
-    VGA_EMU_RW_PROT, (void *) (vga.mem.scratch_page  << 12)
+    VGA_EMU_RW_PROT, vga.mem.scratch_page)
   );
 
-  if(i == -1) return 3;
+  if(i == MAP_FAILED) return 3;
 
   return vga_emu_protect_page(page, RO);
 }
@@ -1284,19 +1359,21 @@ static int vgaemu_unmap(unsigned page)
  */
 void vgaemu_reset_mapping()
 {
-  int i, prot, page;
+  void *i;
+  int prot, page;
 
-  memset((void *) (vga.mem.scratch_page << 12), 0xff, 1 << 12);
+  if (!Video->update_screen) return;
+  memset(vga.mem.scratch_page, 0xff, 1 << 12);
 
   prot = VGA_EMU_RW_PROT;
   for(page = 0xa0; page < 0xc0; page++) {
-    i = (int) mmap_mapping(MAPPING_VGAEMU | MAPPING_ALIAS,
-      (void *) (page << 12), 1 << 12,
-      prot, (void *) (vga.mem.scratch_page << 12)
+    i = alias_mapping(MAPPING_VGAEMU,
+      (void *)(uintptr_t) (page << 12), 1 << 12,
+      prot, vga.mem.scratch_page
     );
   }
 
-  if(i == -1) {
+  if(i == MAP_FAILED) {
     prot = 0xfe;
     switch(errno) {
       case EINVAL: prot = 0xfd; break;
@@ -1310,117 +1387,9 @@ void vgaemu_reset_mapping()
   }
 }
 
-
-/*
- * DANG_BEGIN_FUNCTION vga_emu_init
- *
- * description:
- * vga_emu_init() must be called before using the VGAEmu functions.
- * It is only called from env/video/X.c::X_init() at the moment.
- * This function basically initializes the global variable `vga' and
- * allocates the VGA memory.
- *
- * It does in particular *not* map any memory into the range
- * 0xa0000 - 0xc0000, this is done as part of a VGA mode switch.
- *
- * There should be an accompanying vga_emu_done().
- *
- * arguments:
- * vedt - Pointer to struct describing the type of display we are actually
- *        attached to.
- *
- * DANG_END_FUNCTION                        
- *
- */     
-
-int vga_emu_init(vgaemu_display_type *vedt)
+static void vgaemu_register_ports(void)
 {
-  int i;
-  vga_mapping_type vmt = {0, 0, 0};
   emu_iodev_t io_device;
-  static unsigned char *lfb_base = NULL;
-
-  open_mapping(MAPPING_VGAEMU);
-
-  /* clean it up - just in case */
-  memset(&vga, 0, sizeof vga);
-
-  vga.mode = vga.VGA_mode = vga.VESA_mode = 0;
-
-  vga.config.mono_support = config.dualmon ? 0 : 1;
-
-  if(config.vgaemu_memsize)
-    vga.mem.size = config.vgaemu_memsize << 10;
-  else
-    vga.mem.size = 1024 << 10;
-
-  /* force 256k granularity to prevent possible problems
-   * (with 4-plane-modes, to be precise)
-   */
-  vga.mem.size = (vga.mem.size + ~(-1 << 18)) & (-1 << 18);
-  vga.mem.pages = vga.mem.size >> 12;
-
-  vga.mem.base = alloc_mapping(MAPPING_VGAEMU, vga.mem.size+ (1 << 12), 0);
-  if(!vga.mem.base) {
-    vga_msg("vga_emu_init: not enough memory (%u k)\n", vga.mem.size >> 10);
-    return 1;
-  }
-  vga.mem.scratch_page = (unsigned) (vga.mem.base + vga.mem.size) >> 12;
-  memset((void *) (vga.mem.scratch_page << 12), 0xff, 1 << 12);
-  vga_msg("vga_emu_init: scratch_page at 0x%08x\n", vga.mem.scratch_page << 12);
-
-  if(config.X_lfb) {
-    lfb_base = alloc_mapping(MAPPING_VGAEMU, vga.mem.size, 0);
-    if(!lfb_base) {
-      vga_msg("vga_emu_init: not enough memory (%u k)\n", vga.mem.size >> 10);
-    }
-  }
-
-  if(lfb_base == NULL) {
-    vga_msg("vga_emu_init: linear frame buffer (lfb) disabled\n");
-  }
-
-  if((vga.mem.dirty_map = (unsigned char *) malloc(vga.mem.pages)) == NULL) {
-    vga_msg("vga_emu_init: not enough memory for dirty map\n");
-    return 1;
-  }
-  dirty_all_video_pages();		/* all need an update */
-
-  if(
-    (vga.mem.prot_map0 = (unsigned char *) malloc(0xc0 - 0xa0)) == NULL ||
-    (vga.mem.prot_map1 = (unsigned char *) malloc(vga.mem.pages)) == NULL
-  ) {
-    vga_msg("vga_emu_init: not enough memory for protection map\n");
-    return 1;
-  }
-  memset(vga.mem.prot_map0, 0xff, 0x20);
-  memset(vga.mem.prot_map1, 0xff, vga.mem.pages);
-
-  /*
-   * vga.mem.map _must_ contain only valid mappings - otherwise _bad_ things will happen!
-   * cf. vga_emu_protect()
-   * -- sw
-   */
-
-  for(i = 0; i < VGAEMU_MAX_MAPPINGS; i++) vga.mem.map[i] = vmt;
-
-  vga.mem.bank = vga.mem.bank_pages = 0;
-
-  if(lfb_base != NULL) {
-    vga.mem.lfb_base_page = (unsigned) lfb_base >> 12;
-  }
-
-  vga_emu_setup_mode_table();
-
-  vbe_init(vedt);
-
-  /*
-   * Make the VGA-BIOS ROM read-only; some dirty programs try to write to the ROM!
-   *
-   * Note: Unknown instructions will cause the write protection to be removed.
-   */
-  vga_msg("vga_emu_init: protecting ROM area 0xc0000 - 0x%05x\n", (0xc0 + vgaemu_bios.pages) << 12);
-  for(i = 0; i < vgaemu_bios.pages; i++) vga_emu_protect_page(0xc0 + i, RO);
 
   /* register VGA ports */
   io_device.read_portb = VGA_emulate_inb;
@@ -1455,11 +1424,165 @@ int vga_emu_init(vgaemu_display_type *vedt)
    * if something is missing. -- sw
    */
   if(vga.config.mono_support) {
-    io_device.handler_name = "VGAEmu Mono/Hercules Card";
+    io_device.handler_name = "VGAEmu Mono/Hercules Card Range 0";
     io_device.start_addr = 0x3b0;
+    io_device.end_addr = 0x3bb;
+    port_register_handler(io_device, 0);
+
+    io_device.handler_name = "VGAEmu Mono/Hercules Card Range 1";
+    io_device.start_addr = 0x3bf;
     io_device.end_addr = 0x3bf;
     port_register_handler(io_device, 0);
   }
+}
+
+/*
+ * DANG_BEGIN_FUNCTION vga_emu_init
+ *
+ * description:
+ * vga_emu_init() must be called before using the VGAEmu functions.
+ * It is only called from env/video/X.c::X_init() at the moment.
+ * This function basically initializes the global variable `vga' and
+ * allocates the VGA memory.
+ *
+ * It does in particular *not* map any memory into the range
+ * 0xa0000 - 0xc0000, this is done as part of a VGA mode switch.
+ *
+ * There should be an accompanying vga_emu_done().
+ *
+ * arguments:
+ * vedt - Pointer to struct describing the type of display we are actually
+ *        attached to.
+ *
+ * DANG_END_FUNCTION                        
+ *
+ */     
+
+int vga_emu_init(int src_modes, ColorSpaceDesc *csd)
+{
+  int i;
+  vga_mapping_type vmt = {0, 0, 0};
+  static unsigned char *lfb_base = NULL;
+
+  /* clean it up - just in case */
+  memset(&vga, 0, sizeof vga);
+
+  vga.mode = vga.VGA_mode = vga.VESA_mode = 0;
+  vga.config.video_off = 1;
+  vga.config.standard = 1;
+  vga.mem.plane_pages = 0x10;	/* 16 pages = 64k */
+  vga.dac.bits = 6;
+
+  vga.config.mono_support = config.dualmon ? 0 : 1;
+
+  if (!Video->update_screen) {
+    vga_emu_setup_mode_table();
+    if (Video->update_cursor) {
+      vgaemu_register_ports();
+      memcpy((void *) GFX_CHARS, vga_rom_08, 128 * 8);
+      vbe_init(NULL);
+      for(i = 0; i < vgaemu_bios.pages; i++) vga_emu_protect_page(0xc0 + i, RO);
+    }
+    return 0;
+  }
+
+  open_mapping(MAPPING_VGAEMU);
+
+  if(config.vgaemu_memsize)
+    vga.mem.size = config.vgaemu_memsize << 10;
+  else
+    vga.mem.size = 1024 << 10;
+
+  /* force 256k granularity to prevent possible problems
+   * (with 4-plane-modes, to be precise)
+   */
+  vga.mem.size = (vga.mem.size + ~(-1 << 18)) & (-1 << 18);
+  vga.mem.pages = vga.mem.size >> 12;
+
+  vga.mem.base = alloc_mapping(MAPPING_VGAEMU, vga.mem.size+ (1 << 12), 0);
+  if(!vga.mem.base) {
+    vga_msg("vga_emu_init: not enough memory (%u k)\n", vga.mem.size >> 10);
+    return 1;
+  }
+  vga.mem.scratch_page = vga.mem.base + vga.mem.size;
+  memset(vga.mem.scratch_page, 0xff, 1 << 12);
+  vga_msg("vga_emu_init: scratch_page at %p\n", vga.mem.scratch_page);
+
+  if(config.X_lfb) {
+    lfb_base = mmap_mapping(MAPPING_VGAEMU | MAPPING_SCRATCH, (void *)-1,
+			    vga.mem.size, PROT_READ|PROT_WRITE|PROT_EXEC, 0);
+    if(!lfb_base) {
+      vga_msg("vga_emu_init: not enough memory (%u k)\n", vga.mem.size >> 10);
+    }
+  }
+
+  if(lfb_base == NULL) {
+    vga_msg("vga_emu_init: linear frame buffer (lfb) disabled\n");
+  }
+
+  if((vga.mem.dirty_map = (unsigned char *) malloc(vga.mem.pages)) == NULL) {
+    vga_msg("vga_emu_init: not enough memory for dirty map\n");
+    return 1;
+  }
+  dirty_all_video_pages();		/* all need an update */
+
+  if(
+    (vga.mem.prot_map0 = malloc(vgaemu_bios.pages + 0xc0 - 0xa0)) == NULL ||
+    (vga.mem.prot_map1 = (unsigned char *) malloc(vga.mem.pages)) == NULL
+  ) {
+    vga_msg("vga_emu_init: not enough memory for protection map\n");
+    return 1;
+  }
+  memset(vga.mem.prot_map0, 0xff, vgaemu_bios.pages + 0x20);
+  memset(vga.mem.prot_map1, 0xff, vga.mem.pages);
+
+  /*
+   * vga.mem.map _must_ contain only valid mappings - otherwise _bad_ things will happen!
+   * cf. vga_emu_protect()
+   * -- sw
+   */
+
+  for(i = 0; i < VGAEMU_MAX_MAPPINGS; i++) vga.mem.map[i] = vmt;
+
+  vga.mem.bank = vga.mem.bank_pages = 0;
+
+  if(lfb_base != NULL) {
+    vga.mem.lfb_base_page = (uintptr_t) lfb_base >> 12;
+    memcheck_addtype('e', "VGAEMU LFB");
+    register_hardware_ram('e', (size_t)lfb_base, vga.mem.size);
+  }
+
+  vga_emu_setup_mode_table();
+
+  if(csd != NULL) {
+    vgaemu_display_type vedt;
+
+    vedt.src_modes = src_modes;
+    vedt.bits = csd->bits;
+    vedt.bytes = csd->bytes;
+    vedt.r_mask = csd->r_mask;
+    vedt.g_mask = csd->g_mask;
+    vedt.b_mask = csd->b_mask;
+    vedt.r_shift = csd->r_shift;
+    vedt.g_shift = csd->g_shift;
+    vedt.b_shift = csd->b_shift;
+    vedt.r_bits = csd->r_bits;
+    vedt.g_bits = csd->g_bits;
+    vedt.b_bits = csd->b_bits;
+    vbe_init(&vedt);
+  } else {
+    vbe_init(NULL);
+  }
+
+  /*
+   * Make the VGA-BIOS ROM read-only; some dirty programs try to write to the ROM!
+   *
+   * Note: Unknown instructions will cause the write protection to be removed.
+   */
+  vga_msg("vga_emu_init: protecting ROM area 0xc0000 - 0x%05x\n", (0xc0 + vgaemu_bios.pages) << 12);
+  for(i = 0; i < vgaemu_bios.pages; i++) vga_emu_protect_page(0xc0 + i, RO);
+
+  vgaemu_register_ports();
 
   /*
    * init the ROM-BIOS font (the VGA fonts are added in vbe_init())
@@ -1499,17 +1622,12 @@ int vga_emu_init(vgaemu_display_type *vedt)
 #endif
 
   vga_msg(
-    "vga_emu_init: memory: %u kbyte at 0x%x (lfb at 0x%x); %ssupport for mono modes\n",
-    vga.mem.size >> 10, (unsigned) vga.mem.base,
+    "vga_emu_init: memory: %u kbyte at %p (lfb at %#x); %ssupport for mono modes\n",
+    vga.mem.size >> 10, vga.mem.base,
     vga.mem.lfb_base_page << 12,
     vga.config.mono_support ? "" : "no "
   );
 
-  /*
-   * Set some mode; this initializes the DAC, CRTC, etc. as well.
-   */
-  set_video_mode(3);		/* initialize some mode */
-
   return 0;
 }
 
@@ -2051,6 +2169,8 @@ int vga_emu_setmode(int mode, int width, int height)
         vmi2 = vmi;
       }
     }
+    if(vmi == NULL && Video->setmode == NULL)
+      vmi2 = vga_emu_find_mode(mode, vmi);
     vmi = vmi2;
   }
 
@@ -2073,6 +2193,7 @@ int vga_emu_setmode(int mode, int width, int height)
   vga.mode_type = vmi->type;
   vga.width = vmi->width;
   vga.height = vmi->height;
+  vga.line_compare = vmi->height;
   vga.scan_len = (vmi->width + 3) & ~3;	/* dword aligned */
   vga.text_width = vmi->text_width;
   vga.text_height = vmi->text_height;
@@ -2122,12 +2243,9 @@ int vga_emu_setmode(int mode, int width, int height)
    */
 
   if(!(mode & 0x8000) && !(mode < 0x100 && (mode & 0x80))) {
-    memset((void *) vga.mem.base, 0, vga.mem.size);
-    if(vga.mode_class == TEXT) {
-      unsigned *p = (unsigned *) vga.mem.base;
-      int i;
-      
-      for(i = 0; i < 0x2000; i++) p[i] = 0x07200720;
+    if(vga.mode_class != TEXT && vga.mem.base) {
+      /* should be moved to BIOS */
+      memset((void *)vga.mem.base, 0, vga.mem.size);
     }
   }
 
@@ -2162,17 +2280,19 @@ int vga_emu_setmode(int mode, int width, int height)
   vga.mem.map[VGAEMU_MAP_BANK_MODE].base_page = vmi->buffer_start >> 8;
   vgaemu_map_bank();
 
-  if(vga.color_bits >= 8 && (vga.mode & 0xffff) >= 0x100 && vga.mem.lfb_base_page) {
-    vga.mem.map[VGAEMU_MAP_LFB_MODE].base_page = vga.mem.lfb_base_page;
-    vga.mem.map[VGAEMU_MAP_LFB_MODE].first_page = 0;
-    vga.mem.map[VGAEMU_MAP_LFB_MODE].pages = vga.mem.pages;
-    vga_emu_map(VGAEMU_MAP_LFB_MODE, 0);	/* map the VGA memory to LFB */
-  }
-  else {
-    vga.mem.map[VGAEMU_MAP_LFB_MODE].base_page = 0;
-    vga.mem.map[VGAEMU_MAP_LFB_MODE].first_page = 0;
-    vga.mem.map[VGAEMU_MAP_LFB_MODE].pages = 0;
-    // unmap ???
+  vga.mem.map[VGAEMU_MAP_LFB_MODE].base_page = 0;
+  vga.mem.map[VGAEMU_MAP_LFB_MODE].first_page = 0;
+  vga.mem.map[VGAEMU_MAP_LFB_MODE].pages = 0;
+  vga.mem.wrap = vmi->buffer_len * 1024;
+  // unmap ???
+
+  if(vga.color_bits >= 8 && (vga.mode & 0xffff) >= 0x100) {
+    vga.mem.wrap = vga.mem.size;
+    if(vga.mem.lfb_base_page) {
+      vga.mem.map[VGAEMU_MAP_LFB_MODE].base_page = vga.mem.lfb_base_page;
+      vga.mem.map[VGAEMU_MAP_LFB_MODE].pages = vga.mem.pages;
+      vga_emu_map(VGAEMU_MAP_LFB_MODE, 0);	/* map the VGA memory to LFB */
+    }
   }
 
   vga_msg("vga_emu_setmode: setting up components...\n");
@@ -2198,6 +2318,7 @@ int vgaemu_map_bank()
   int j, k0, k1;
 #endif
 
+  if(!Video->update_screen) return False;
   if((vga.mem.bank + 1) * vga.mem.bank_pages > vga.mem.pages) {
     vga_msg("vgaemu_map_bank: invalid bank %d\n", vga.mem.bank);
     return False;
@@ -2209,6 +2330,8 @@ int vgaemu_map_bank()
   }
 
   vga.mem.map[VGAEMU_MAP_BANK_MODE].pages = vga.mem.bank_pages;
+  vga.mem.bank_base = vga.mem.map[VGAEMU_MAP_BANK_MODE].base_page << 12;
+  vga.mem.bank_len = vga.mem.map[VGAEMU_MAP_BANK_MODE].pages << 12;
 
   if(vga.mem.write_plane) {
     first = vga.mem.plane_pages * vga.mem.write_plane;
@@ -2299,47 +2422,10 @@ int vga_emu_set_textsize(int width, int height)
 
 void dirty_all_video_pages()
 {
-  memset(vga.mem.dirty_map, 1, vga.mem.pages);
+  if (vga.mem.dirty_map)
+    memset(vga.mem.dirty_map, 1, vga.mem.pages);
 }
 
-
-/*
- * DANG_BEGIN_FUNCTION vga_emu_set_text_page
- *
- * description:
- * Set visible text page.
- *
- * `vga.display_start' is set to `page' * `page_size'.
- * This function works only in text modes.
- *
- * arguments:
- * page      - Number of the text page.
- * page_size - Size of one text page.
- *
- * DANG_END_FUNCTION                        
- *
- */     
-
-int vga_emu_set_text_page(unsigned page, unsigned page_size)
-{
-  if(vga.mode_class != TEXT) {
-    vga_msg("vga_emu_set_text_page: not in text mode\n");
-    return 1;
-  }
-
-  if((page + 1) * page_size > vga.mem.size) {
-    vga_msg("vga_emu_set_text_page: page number %d to high\n", page);
-    return 2;
-  }
-
-  vga_msg("vga_emu_set_text_page %d\n", page);
-
-  vga.display_start = page * page_size;
-
-  return 0;
-}
-
-
 /*
  * DANG_BEGIN_FUNCTION dirty_all_vga_colors
  *
@@ -2480,8 +2566,6 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
   static char *txt1[] = { "byte", "odd/even (word)", "chain4 (dword)" };
   static char *txt2[] = { "byte", "word", "dword" };
 
-  if(!vga.config.standard) return;
-
   switch(what) {
     case CFG_SEQ_ADDR_MODE:
       u0 = vga.seq.addr_mode;
@@ -2493,25 +2577,43 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
         vga.mem.planes = u1; vga.reconfig.mem = 1;   
         vga_msg("vgaemu_adj_cfg: mem reconfig (%u planes)\n", u1);
       }
-      if(msg || u != u0) vga_msg("vgaemu_adj_cfg: seq.addr_mode = %s\n", txt1[u]); 
+      if(msg || u != u0) vga_msg("vgaemu_adj_cfg: seq.addr_mode = %s\n", txt1[u]);
+      if (vga.mode_class == TEXT && vga.width < 2048) {
+        int horizontal_display_end = vga.crtc.data[0x1] + 1;
+	int horizontal_blanking_start = vga.crtc.data[0x2] + 1;
+	int multiplier = 9 - (vga.seq.data[1] & 1);
+	int width = min(horizontal_display_end, horizontal_blanking_start) *
+	  multiplier;
+	if ((vga.width != width) || (vga.char_width != multiplier)) {
+	  vga.width = width;
+	  vga.char_width = multiplier;
+	  vga.text_width = horizontal_display_end;
+	  vga.reconfig.display = 1;
+	  vga.reconfig.re_init = 1;
+	}
+      }
     break;
 
     case CFG_CRTC_ADDR_MODE:
+      if (vga.color_bits > 8)
+	return;
       u0 = vga.crtc.addr_mode;
       u1 = vga.scan_len;
       u = vga.crtc.data[0x17] & 0x40 ? 0 : 1;
       u = vga.crtc.data[0x14] & 0x40 ? 2 : u;
+      if (vga.scan_len > (255 << (u + 1)))
+	return;
+      if (vga.scan_len & ((1 << (u + 1)) - 1))
+	return;
       vga.crtc.addr_mode = u;
-      vga.display_start = (vga.crtc.data[0x0d] + (vga.crtc.data[0x0c] << 8)) << 
+      if (vga.config.standard)
+	vga.display_start = (vga.crtc.data[0x0d] + (vga.crtc.data[0x0c] << 8)) << 
 	      vga.crtc.addr_mode;
       /* this shift should really be a rotation, depending on mode control bit 5 */
-      screen_adr = SCREEN_ADR(0) + vga.display_start;
       vga.crtc.cursor_location =  (vga.crtc.data[0x0f] + (vga.crtc.data[0x0e] << 8)) <<
               vga.crtc.addr_mode;
       vga.scan_len = vga.crtc.data[0x13] << (vga.crtc.addr_mode + 1);
-      if (vga.scan_len == 0) vga.scan_len = 160; 
-      cursor_row = (vga.crtc.cursor_location - vga.display_start) / vga.scan_len;
-      cursor_col = ((vga.crtc.cursor_location - vga.display_start) % vga.scan_len) / 2;
+      if (vga.scan_len == 0) vga.scan_len = 160;
       if (u1 != vga.scan_len) vga.reconfig.mem = 1;
       if(msg || u != u0) vga_msg("vgaemu_adj_cfg: crtc.addr_mode = %s, scan_len = %d\n",
 	txt2[u], vga.scan_len);
@@ -2526,7 +2628,12 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       int vertical_multiplier;
       int vertical_blanking_start;
       int vertical_blanking_end;
+      int char_height;
       int height;
+
+      if (vga.height > 1024)
+	return;
+
       vertical_total = 
 	      vga.crtc.data[0x6] + 
 	      ((vga.crtc.data[0x7] & 0x1) << (8 - 0)) +
@@ -2534,7 +2641,7 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       vertical_retrace_start = 
 	      vga.crtc.data[0x10] +
 	      ((vga.crtc.data[0x7] & 0x4) << (8 - 2)) +
-	      ((vga.crtc.data[0x7] & 0xF0) << (9 - 7));
+	      ((vga.crtc.data[0x7] & 0x80) << (9 - 7));
       vertical_retrace_end = vga.crtc.data[0x11]  & 0x0F;
       vertical_display_end = 
 	      vga.crtc.data[0x12] +
@@ -2546,9 +2653,12 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
 	      ((vga.crtc.data[0x9] & 0x20) << (9 - 5));
       vertical_blanking_end =
 	      vga.crtc.data[0x16] & 0x7F;
-      vertical_multiplier = ((vga.crtc.data[0x9] & 0x1F) +1) <<
-	      ((vga.crtc.data[0x9] & 0x80) >> 7);
-      height = (vertical_display_end +1) / vertical_multiplier;
+      char_height = (vga.crtc.data[0x9] & 0x1f) + 1;
+      vertical_multiplier = char_height << ((vga.crtc.data[0x9] & 0x80) >> 7);
+      /* see VGADOC: CGA is special for reg 9 */
+      if(vga.mode_type == CGA) vertical_multiplier = char_height;
+      height = (min(vertical_blanking_start, vertical_display_end) + 1) /
+	vertical_multiplier;
       vga_msg("vgaemu_adj_cfg: vertical_total = %d\n", vertical_total);
       vga_msg("vgaemu_adj_cfg: vertical_retrace_start = %d\n", vertical_retrace_start);
       vga_msg("vgaemu_adj_cfg: vertical_retrace_end = %d\n", vertical_retrace_end);
@@ -2557,9 +2667,31 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       vga_msg("vgaemu_adj_cfg: vertical_display_end = %d\n", vertical_display_end);
       vga_msg("vgaemu_adj_cfg: vertical_multiplier = %d\n", vertical_multiplier);
       vga_msg("vgaemu_adj_cfg: height = %d\n", height);
-      if (vga.height != height) {
+      if (vga.line_compare != vga.crtc.line_compare / vertical_multiplier) {
+        vga.line_compare = vga.crtc.line_compare / vertical_multiplier;
+        dirty_all_video_pages();
+      }
+      if (vga.mode_class == TEXT) {
+        height *= char_height;
+      } else {
+        char_height = height / 25;
+        if (char_height >= 16)
+          char_height = 16;
+        else if (char_height >= 14)
+          char_height = 14;
+        else
+          char_height = 8;
+      }
+      /* By Eric (eric@coli.uni-sb.de):                        */
+      /* Required for 80x100 CGA "text graphics" with 8x2 font */
+      if (vga.height != height || vga.char_height != char_height) {
         vga.height = height;
+	vga.text_height = height / char_height;
+        vga.char_height = char_height;
+        vga_msg("vgaemu_adj_cfg: text_height=%d height=%d char_height=%d\n",
+                height, vertical_display_end+1, vga.char_height);
         vga.reconfig.display = 1;
+	vga.reconfig.re_init = 1;
       }
       break;
     }
@@ -2573,15 +2705,26 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       int horizontal_blanking_end;
       int multiplier;
       int width;
+
+      if (vga.width >= 2048)
+	return;
+
       /* everything below is in characters */
       horizontal_total = vga.crtc.data[0x0] +5;
       horizontal_display_end = vga.crtc.data[0x1] + 1;
-      horizontal_blanking_start = vga.crtc.data[0x2];
+      horizontal_blanking_start = vga.crtc.data[0x2] + 1;
       horizontal_blanking_end = vga.crtc.data[0x3] & 0xF;
       horizontal_retrace_start = vga.crtc.data[0x4];
       horizontal_retrace_end = vga.crtc.data[0x5] & 0xF;
-      multiplier = (vga.mode_class == TEXT) ? 9 : (vga.mode_type == P8) ? 4 : 8;
-      width = horizontal_display_end * multiplier;
+      if (vga.mode_class == TEXT) {
+        multiplier = 9 - (vga.seq.data[1] & 1);
+      } else {
+	multiplier = (vga.color_bits == 8 && vga.config.standard) ? 4 : 8;
+      }
+      if (vga.width % multiplier != 0) /* special user defined mode? */
+	return;
+      width = min(horizontal_display_end, horizontal_blanking_start) *
+	multiplier;
       vga_msg("vgaemu_adj_cfg: horizontal_total = %d\n", horizontal_total);
       vga_msg("vgaemu_adj_cfg: horizontal_retrace_start = %d\n", horizontal_retrace_start);
       vga_msg("vgaemu_adj_cfg: horizontal_retrace_end = %d\n", horizontal_retrace_end);
@@ -2590,12 +2733,91 @@ void vgaemu_adj_cfg(unsigned what, unsigned msg)
       vga_msg("vgaemu_adj_cfg: horizontal_display_end = %d\n", horizontal_display_end);
       vga_msg("vgaemu_adj_cfg: multiplier = %d\n", multiplier);
       vga_msg("vgaemu_adj_cfg: width = %d\n", width);
-      if (vga.width != width) {
-	 vga.width = width;
-	 vga.reconfig.display = 1;
+      if ((vga.width != width) || (vga.char_width != multiplier)) {
+        vga.width = width;
+        vga.char_width = (multiplier >= 8) ? multiplier : 8;
+	vga.text_width = min(horizontal_display_end, horizontal_blanking_start);
+        vga.reconfig.display = 1;
+	vga.reconfig.re_init = 1;
       }
       break;
     }
+    case CFG_CRTC_LINE_COMPARE:
+    {
+      int vertical_multiplier;
+
+      if (vga.height >= 1024)
+	return;
+
+      vga.crtc.line_compare =
+              vga.crtc.data[0x18] +
+              ((vga.crtc.data[0x7] & 0x10) << (8 - 4)) +
+              ((vga.crtc.data[0x9] & 0x40) << (9 - 6));
+      vertical_multiplier = ((vga.crtc.data[0x9] & 0x1F) +1) <<
+	      ((vga.crtc.data[0x9] & 0x80) >> 7);
+      vga_msg("vgaemu_adj_cfg: line_compare = %d\n", vga.crtc.line_compare);
+      if (vga.line_compare != vga.crtc.line_compare / vertical_multiplier) {
+        vga.line_compare = vga.crtc.line_compare / vertical_multiplier;
+        dirty_all_video_pages();
+      }
+      break;
+    }
+    case CFG_MODE_CONTROL:
+    {
+      int oldclass, old_color_bits;
+
+      if (vga.color_bits > 8)
+	return;
+
+      oldclass = vga.mode_class;
+      vga.mode_class = (vga.gfx.data[6] & 1) ? GRAPH : TEXT;
+      vga.pixel_size = 4;
+      if (vga.mode_class == TEXT) {
+	vga.mode_type = vga.config.mono_port ? TEXT_MONO : TEXT;
+      } else {
+	unsigned char gdata = vga.gfx.data[5];
+	unsigned char cdata = vga.crtc.data[0x17];
+	unsigned char adata = vga.attr.data[0x12];
+
+	vga.pixel_size = 1;
+	if (gdata & 0x40) {
+	  vga.mode_type = P8;
+	  vga.pixel_size = 8;
+	} else if (gdata & 0x20) {
+	  vga.mode_type = CGA;
+	  vga.pixel_size = 2;
+	} else if (!(cdata & 1)) {
+	  vga.mode_type = CGA;
+	} else if (!(cdata & 2)) {
+	  vga.mode_type = HERC;
+	} else if (adata == 5) {
+	  vga.mode_type = PL2;
+	  vga.pixel_size = 2;
+	} else if (adata == 1) {
+	  vga.mode_type = PL1;
+	} else {
+	  vga.mode_type = PL4;
+	  vga.pixel_size = 4;
+	}
+      }
+      old_color_bits = vga.color_bits;
+      vga.color_bits = vga.pixel_size;
+      vga.inst_emu = (vga.mode_type==PL4 || vga.mode_type==PL2 ||
+	(vga.color_bits == 8 && 
+	 ((vga.seq.map_mask | 0xf8) & (vga.seq.map_mask - 1))))
+	? EMU_ALL_INST : 0;
+      if (oldclass != vga.mode_class) {
+	vgaemu_adj_cfg(CFG_SEQ_ADDR_MODE, 0);
+	vgaemu_adj_cfg(CFG_CRTC_WIDTH, 0);
+	vgaemu_adj_cfg(CFG_CRTC_HEIGHT, 0);
+      } else if (old_color_bits != vga.color_bits) {
+	vgaemu_adj_cfg(CFG_CRTC_WIDTH, 0);
+      }
+      vga.mem.wrap = (vga.mode_type == CGA || vga.mode_class == TEXT ?
+		      32 : 64) * 1024;
+      vga.reconfig.re_init = 1;
+      break;
+    }
     default:
       vga_msg("vgaemu_adj_cfg: unknown item %u\n", what);
   }
@@ -2653,7 +2875,7 @@ unsigned vgaemu_xy2ofs(unsigned x, unsigned y)
 
 void vgaemu_move_vga_mem(unsigned dst, unsigned src, unsigned len)
 {
-  void *dp, *sp;
+  char *dp, *sp;
 
   if(len == 0) return;
 
@@ -2786,7 +3008,7 @@ void vgaemu_put_char(int x, int y, unsigned char c, unsigned char attr)
     src = (READ_WORD(0x43 * 4 + 2) << 4) + READ_WORD(0x43 * 4);
   }
   src += height * c;
-  font = (unsigned char *) src;
+  font = (unsigned char *)(uintptr_t) src;
 
   ofs = vgaemu_xy2ofs(start_x, start_y);
   vga_msg("vgaemu_put_char: src 0x%x, ofs 0x%x, height %u\n", src, ofs, height);
