@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 #ifndef DOSEMU_DEBUG_H
@@ -11,8 +11,11 @@
 #include <stdarg.h>
 #include "config.h"
 #include "extern.h"
+#include "machcompat.h"
 
-#if GCC_VERSION_CODE >= 2005
+/* Some older gccs accept the format attribute too,
+ * but not for the function pointers. :( */
+#if GCC_VERSION_CODE >= 3002
 # define FORMAT(T,A,B)  __attribute__((format(T,A,B)))
 #else
 # define FORMAT(T,A,B)
@@ -70,16 +73,15 @@ EXTERN struct debug_class debug_save[DEBUG_CLASSES];
 EXTERN struct debug_class debug[DEBUG_CLASSES];
 
 int log_printf(int, const char *,...) FORMAT(printf, 2, 3);
+int vlog_printf(int, const char *,va_list);
 
-void p_dos_str(char *,...) FORMAT(printf, 1, 2);
+int p_dos_str(const char *,...) FORMAT(printf, 1, 2);
 
 #if 0  /* set this to 1, if you want dosemu to honor the -D flags */
  #define NO_DEBUGPRINT_AT_ALL
 #endif
 #undef DEBUG_LITE
 
-extern FILE *dbg_fd;
-
 EXTERN int shut_debug INIT(0);
 
 #ifndef NO_DEBUGPRINT_AT_ALL
@@ -177,6 +179,7 @@ extern int register_debug_class(
 extern int unregister_debug_class(int letter);
 extern void print_debug_usage(FILE *stream);
 extern int set_debug_level(int letter, int level);
+extern inline int debug_level(int letter);
 extern inline int debug_level(int letter)
 {
 	if (letter >= DEBUG_CLASSES) {
@@ -189,17 +192,17 @@ extern inline int debug_level(int letter)
 #else
 
 
-extern int inline parse_debugflags(const char *s, unsigned char flag) { return 0; }
-extern int inline SetDebugFlagsHelper(char *debugStr) { return 0; }
-extern int inline GetDebugFlagsHelper(char *debugStr) { debugStr[0] = '\0'; return 0; }
-extern int inline register_debug_class(
+extern inline int parse_debugflags(const char *s, unsigned char flag) { return 0; }
+extern inline int SetDebugFlagsHelper(char *debugStr) { return 0; }
+extern inline int GetDebugFlagsHelper(char *debugStr) { debugStr[0] = '\0'; return 0; }
+extern inline int register_debug_class(
 	int letter, void (*change_level)(int level), char *help_text)
 {
 	return 0;
 }
-extern int inline unregister_debug_class(int letter) { return 0; }
-extern void inline print_debug_usage(FILE *stream) { return; }
-extern int inline set_debug_level(int letter, int level) { return 0; }
+extern inline int unregister_debug_class(int letter) { return 0; }
+extern inline void print_debug_usage(FILE *stream) { return; }
+extern inline int set_debug_level(int letter, int level) { return 0; }
 extern inline int debug_level(int letter) { return 0; }
 #endif
 
