@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* dos emulator, Matthias Lautner 
@@ -12,38 +12,29 @@
 #ifndef EMU_H
 #define EMU_H
 
-#include <features.h>
-#include <sys/types.h>
-#include <setjmp.h>
-#if GLIBC_VERSION_CODE == 2000
-#include <sigcontext.h>
-#endif
-#include <signal.h> 
-
 #include "config.h"
 
-#if defined(HAVE_KEYBOARD_V1) && (HAVE_KEYBOARD_V1 > 1)
-  #error "Sorry, wrong keyboard code version for this DOSEMU version"
-#endif
-
-#if defined(HAVE_UNICODE_KEYB) && (HAVE_UNICODE_KEYB != 2)
-  #error "Sorry, wrong unicode keyboard code version for this DOSEMU version"
-#endif
+#include <sys/types.h>
+#include <sys/ioctl.h>
+#include <setjmp.h>
+#include <signal.h> 
 
 #include "types.h"
 #include "extern.h"
 #include "machcompat.h"
 #include "cpu.h"
-#include "vm86plus.h"
 #include "priv.h"
+#include "mouse.h"
 
 #include "extern.h"
 
 #if 1 /* Set to 1 to use Silly Interrupt generator */
+#ifdef __i386__
 #define SIG 1
 typedef struct { int fd; int irq; } SillyG_t;
 extern SillyG_t *SillyG;
 #endif
+#endif
 
 #define inline __inline__
 
@@ -63,19 +54,16 @@ EXTERN struct vm86plus_struct vm86s INIT ( {
 } );
 
 EXTERN volatile sig_atomic_t signal_pending INIT(0);
+EXTERN volatile int fault_cnt INIT(0);
 EXTERN fd_set fds_sigio, fds_no_sigio;
-EXTERN unsigned int use_sigio INIT(0);
 EXTERN unsigned int not_use_sigio INIT(0);
 EXTERN int terminal_pipe;
 EXTERN int terminal_fd INIT(-1);
 EXTERN int running_kversion INIT(0);
 
-EXTERN char *cstack[16384];
+EXTERN char *(*cstack)[16384];
 
-/* this is DEBUGGING code! */
-EXTERN int sizes INIT(0);
-
-EXTERN int screen, screen_mode;
+EXTERN int screen_mode;
 
 /* number of highest vid page - 1 */
 EXTERN int max_page INIT(7);
@@ -110,27 +98,20 @@ EXTERN int console_fd INIT(-1);
 EXTERN int mem_fd INIT(-1);
 EXTERN int in_readkeyboard;
 
-/* X-pipes */
-EXTERN int keypipe;
-EXTERN int mousepipe;
+EXTERN volatile int in_vm86 INIT(0);
 
-EXTERN int in_vm86 INIT(0);
-
-EXTERN int li, co;	/* lines, columns */
 EXTERN int scanseq;
-EXTERN int cursor_row;
-EXTERN int cursor_col;
 
-#if 0
 void dos_ctrl_alt_del(void);	/* disabled */
-#endif
-extern jmp_buf NotJEnv;
+extern sigjmp_buf NotJEnv;
 
 EXTERN void run_vm86(void);
 EXTERN void loopstep_run_vm86(void);
-EXTERN void     vm86_GP_fault();
+EXTERN void vm86_GP_fault(void);
 
 EXTERN void do_call_back(Bit32u codefarptr);
+EXTERN void callback_return(void);
+EXTERN void do_intr_call_back(int intno);
 
 #define NOWAIT  0
 #define WAIT    1
@@ -143,25 +124,8 @@ void getKeys(void);
 
      void char_out(u_char, int);
 
-     struct ioctlq {
-       int fd, req, param3;
-       int queued;
-     };
-
-     void do_queued_ioctl(void);
-     int queue_ioctl(int, int, int), do_ioctl(int, int, int);
      void keybuf_clear(void);
 
-     int set_ioperm(int, int, int);
-
- 
-EXTERN u_char in_sighandler, in_ioctl;
-/* one-entry queue ;-( for ioctl's */
-EXTERN struct ioctlq iq INIT({0, 0, 0, 0}); 
-EXTERN u_char in_ioctl INIT(0);
-EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
-
-
 /* this macro can be safely wrapped around a system call with no side
  * effects; using a feature of GCC, it returns the same value as the
  * function call argument inside.
@@ -177,57 +141,17 @@ EXTERN struct ioctlq curi INIT({0, 0, 0, 0});
 
 #define SILENT_DOS_SYSCALL(sc) sc
 
-#if 0
+#if 1
 #define RPT_SYSCALL(sc) ({ int s_tmp, s_err; \
    do { \
 	  s_tmp = sc; \
 	  s_err = errno; \
-	  if (errno == EINTR) {\
-	    g_printf("Recursive run_irqs() RPT_SYSCALL()\n"); \
-	    handle_signals(); \
-	/*    run_irqs(); */ \
-	  } \
       } while ((s_tmp == -1) && (s_err == EINTR)); \
   s_tmp; })
 #else
 #define RPT_SYSCALL(sc) (sc)
 #endif
 
-#define RPT_SYSCALL2(sc) ({ int s_tmp; \
-   do { \
-	  s_tmp = sc; \
-	  s_err = errno; \
-	  if (errno == EINTR) {\
-	    g_printf("Recursive run_irqs() RPT_SYSCALL2()\n"); \
-	    handle_signals(); \
-	/*    run_irqs(); */ \
-	  } \
-      } while ((s_tmp == -1) ); \
-  s_tmp; })
-
-#ifndef FALSE
-#define FALSE	0
-#endif
-#ifndef TRUE
-#define TRUE	1
-#endif
-
-#ifndef True
-#define True  TRUE
-#endif
-#ifndef False
-#define False FALSE
-#endif
-
-#ifndef DOSEMU_BOOL_IS_CHAR
-typedef unsigned char boolean;
-#define DOSEMU_BOOL_IS_CHAR
-#endif
-
-/* would like to have this in memory.h (but emu.h is included before memory.h !) */
-#define HARDWARE_RAM_START 0xc8000
-#define HARDWARE_RAM_STOP  0xf0000
-
 typedef struct vesamode_type_struct {
   struct vesamode_type_struct *next;
   unsigned width, height, color_bits;
@@ -239,30 +163,31 @@ typedef struct vesamode_type_struct {
 
 #ifdef X86_EMULATOR
        boolean cpuemu;
+       boolean cpusim;
 #endif
        int CPUSpeedInMhz;
        /* for video */
-       boolean console;
        boolean console_video;
-       boolean graphics;
        boolean vga;
        boolean X;
+       boolean X_fullscreen;
        u_short cardtype;
        u_short chipset;
        boolean pci;
        boolean pci_video;
-       u_short gfxmemsize;		/* for SVGA card, in K */
+       long gfxmemsize;		/* for SVGA card, in K */
        /* u_short term_method; */	/* Terminal method: ANSI or NCURSES */
        u_short term_color;		/* Terminal color support on or off */
        /* u_short term_updatelines; */	/* Amount to update at a time */
        u_short term_updatefreq;		/* Terminal update frequency */
-       u_short term_charset;		/* Terminal Character set */
        u_short term_esc_char;	        /* ASCII value used to access slang help screen */
+       char    *xterm_title;	        /* xterm/putty window title */
        /* u_short term_corner; */       /* Update char at lower-right corner */
        u_short X_updatelines;           /* Amount to update at a time */
        u_short X_updatefreq;            /* X update frequency */
        char    *X_display;              /* X server to use (":0") */
        char    *X_title;                /* X window title */
+       int X_title_show_appname;        /* show name of running app in caption */
        char    *X_icon_name;
        char    *X_font;
        char    *X_mgrab_key;		/* KeySym name to activate mouse grab */
@@ -278,111 +203,119 @@ typedef struct vesamode_type_struct {
        int     X_mode13fact;            /* initial size factor for mode 0x13 */
        int     X_winsize_y;             /* initial window height */
        unsigned X_gamma;		/* gamma correction value */
-       u_short vgaemu_memsize;		/* for VGA emulation */
+       u_long vgaemu_memsize;		/* for VGA emulation */
        vesamode_type *vesamode_list;	/* chained list of VESA modes */
        int     X_lfb;			/* support VESA LFB modes */
        int     X_pm_interface;		/* support protected mode interface */
+       int     X_background_pause;	/* pause xdosemu if it loses focus */
        boolean fullrestore;
        boolean force_vt_switch;         /* in case of console_video force switch to emu VT at start */
        int     dualmon;
 
-       u_short usesX;  /* !=0 if dosemu owns an X window */
-
-       boolean console_keyb;
+       int     console_keyb;
        boolean kbd_tty;
        boolean X_keycode;	/* use keycode field of event structure */
        boolean exitearly;
+       boolean quiet;
        int     realcpu;
-       boolean mathco, smp, cpummx;
+       boolean mathco, smp, cpuprefetcht0, cpufxsr;
        boolean ipxsup;
-       boolean vnet;
+       long    ipx_net;
+       int     vnet;
        char   *netdev;
        boolean pktdrv;
        boolean dosbanner;
-       boolean allowvideoportaccess;
        boolean emuretrace;
        boolean rdtsc;
-       boolean timers;
-       boolean mouse_flag;
        boolean mapped_bios;	/* video BIOS */
        char *vbios_file;	/* loaded VBIOS file */
        boolean vbios_copy;
        int vbios_seg;           /* VGA-BIOS-segment for mapping */
        int vbios_size;          /* size of VGA-BIOS (64K for vbios_seg=0xe000
        						     32K for vbios_seg=0xc000) */
+       boolean vbios_post;
 
        boolean bootdisk;	/* Special bootdisk defined */
        int  fastfloppy;
        char *emusys;		/* map CONFIG.SYS to CONFIG.EMU */
        char *emuini;           /* map system.ini to  system.EMU */
+       char *install;          /* directory to point ~/.dosemu/drives/c to */
 
        u_short speaker;		/* 0 off, 1 native, 2 emulated */
        u_short fdisks, hdisks;
        u_short num_lpt;
        u_short num_ser;
-       u_short num_mice;
+       mouse_t mouse;
 
        int pktflags;		/* global flags for packet driver */
 
-       unsigned int update, freq;	/* temp timer magic */
-       unsigned int wantdelta, realdelta;
+       int update, freq;	/* temp timer magic */
        unsigned long cpu_spd;		/* (1/speed)<<32 */
        unsigned long cpu_tick_spd;	/* (1.19318/speed)<<32 */
 
-       unsigned int hogthreshold;
+       int hogthreshold;
 
-       int mem_size, xms_size, ems_size, dpmi, max_umb;
-       int secure;
+       int mem_size, ext_mem, xms_size, ems_size, max_umb;
        unsigned int ems_frame;
-       char must_spare_hardware_ram;
-       char hardware_pages[ ((HARDWARE_RAM_STOP-HARDWARE_RAM_START) >> 12)+1 ];
-
+       int dpmi, pm_dos_api, no_null_checks;
+       unsigned long dpmi_base;
 
        int sillyint;            /* IRQ numbers for Silly Interrupt Generator 
        				   (bitmask, bit3..15 ==> IRQ3 .. IRQ15) */
 
        struct keytable_entry *keytable;
        struct keytable_entry *altkeytable;
-       int toggle_mask;
 
        unsigned short detach;
-       unsigned char *debugout;
-       unsigned char *pre_stroke;  /* pointer to keyboard pre strokes */
-       unsigned char *pre_stroke_mem;
+       char *debugout;
+       char *pre_stroke;        /* pointer to keyboard pre strokes */
 
        /* Lock File business */
+       boolean full_file_locks;
        char *tty_lockdir;	/* The Lock directory  */
        char *tty_lockfile;	/* Lock file pretext ie LCK.. */
        boolean tty_lockbinary;	/* Binary lock files ? */
 
+       /* LFN support */
+       boolean lfn;
+
        /* type of mapping driver */
        char *mappingdriver;
 
-       /* list of arbitrary features
-        * (at minimum 1, will be increased when needed)
-        * The purpose of these parameters is to switch between code
-        * or code pieces that in principal do the same, but for some
-        * unknown reasons behave different between machines.
+       /* List of temporary hacks
+        * (at minimum 16, will be increased when needed)
+        *
         * If a 'features' becomes obsolete (problem solved) it will
         * remain dummy for a while before re-used.
         *
         * NOTE: 'features' are not subject to permanent documentation!
-        *       They should be considered 'temporary hacks'
         *
         * Currently assigned:
         *
-        *   features[0]		use new(0) or old(1) Slangcode
+        *   (none)
         */
-       int features[1];
+       int features[16];
+
+       /* Time mode is TM_BIOS / TM_PIT / TM_LINUX, see iodev.h */
+       int timemode;
 
        /* Sound emulation */
-       int sound; 
-       __u16 sb_base;
-       __u8 sb_dma;
-       __u8 sb_irq;
+       int sound;
+       uint16_t sb_base;
+       uint8_t sb_dma;
+       uint8_t sb_hdma;
+       uint8_t sb_irq;
        char *sb_dsp;
        char *sb_mixer;
-       __u16 mpu401_base;
+       uint16_t mpu401_base;
+       char *sound_driver;
+       /* OSS-specific options */
+       int oss_min_frags;
+       int oss_max_frags;
+       int oss_stalled_frags;
+       int oss_do_post;
+       int oss_min_extra_frags;
+       int oss_dac_freq;
 
        /* joystick */
        char *joy_device[2];
@@ -393,6 +326,9 @@ typedef struct vesamode_type_struct {
        
        int joy_granularity;	/* the higher, the less sensitive - for wobbly joysticks */
        int joy_latency;		/* delay between nonblocking linux joystick reads */
+
+       int cli_timeout;		/* cli timeout hack */
+       int pic_watchdog;        /* pic watchdog reschedule hack */      
      }
 
 config_t;
@@ -411,10 +347,6 @@ config_t;
 #define IO_WRITE 2
 #define IO_RDWR	 (IO_READ | IO_WRITE)
 
-#undef cli
-#undef sti
-EXTERN void cli(void);
-EXTERN void sti(void);
 EXTERN int port_readable(unsigned short);
 EXTERN int port_writeable(unsigned short);
 EXTERN unsigned char read_port(unsigned short);
@@ -422,121 +354,61 @@ EXTERN int write_port(unsigned int, unsigned short);
 EXTERN __inline__ void parent_nextscan(void);
 EXTERN __inline__ void disk_close(void);
 EXTERN void cpu_setup(void);
+EXTERN void cpu_reset(void);
 EXTERN void real_run_int(int);
 #define run_int do_int
 EXTERN int mfs_redirector(void);
-EXTERN void int10(void);
-EXTERN void int13(u_char);
-EXTERN void int14(u_char);
-EXTERN void int16(u_char);
-EXTERN void int17(u_char);
+EXTERN int mfs_lfn(void);
+EXTERN int int10(void);
+EXTERN int int13(void);
+EXTERN int int16(void);
+EXTERN int int17(void);
 EXTERN void io_select(fd_set);
+EXTERN void io_select_init(void);
 EXTERN int pd_receive_packet(void);
 EXTERN int printer_tick(u_long);
-EXTERN int printer_tick(u_long);
 EXTERN void floppy_tick(void);
 EXTERN void open_kmem(void);
 EXTERN void close_kmem(void);
-EXTERN void CloseNetworkLink(int);
 EXTERN int parse_config(char *, char *);
+EXTERN void prepare_dexe_load(char *name);
 EXTERN void disk_init(void);
+EXTERN void disk_reset(void);
 EXTERN void serial_init(void);
+EXTERN void serial_reset(void);
 EXTERN void close_all_printers(void);
-EXTERN void release_ports(void);
 EXTERN void serial_close(void);
 EXTERN void disk_close_all(void);
 EXTERN void init_all_printers(void);
 EXTERN int mfs_inte6(void);
+EXTERN int mfs_helper(state_t *regs);
 EXTERN void pkt_helper(void);
 EXTERN short pop_word(struct vm86_regs *);
-EXTERN void ems_init(void);
-EXTERN int GetDebugFlagsHelper(char *, int);
-EXTERN int SetDebugFlagsHelper(char *);
 EXTERN void leavedos(int) NORETURN;
-EXTERN void add_to_io_select(int, unsigned char);
-EXTERN void remove_from_io_select(int, unsigned char);
-EXTERN void sigquit(int);
+EXTERN void add_to_io_select(int, u_char, void(*)(void));
+EXTERN void remove_from_io_select(int, u_char);
 #ifdef __linux__
-EXTERN void sigalrm(int, struct sigcontext_struct);
-EXTERN void sigio(int, struct sigcontext_struct);
-EXTERN int dosemu_sigaction(int sig, struct sigaction *new, struct sigaction *old);
+EXTERN void SIG_init(void);
+EXTERN void SIG_close(void);
 #endif
 
+EXTERN unsigned long int stack_init_top INIT(0xffffffff);
+EXTERN unsigned long int stack_init_bot INIT(0xffffffff);
+
 /* signals for Linux's process control of consoles */
-#define SIG_RELEASE     SIGWINCH
+#define SIG_RELEASE     SIGUSR1
 #define SIG_ACQUIRE     SIGUSR2
 
- /* DANG_BEGIN_REMARK
-  * We assume system call restarting... under linux 0.99pl8 and earlier,
-  * this was the default.  SA_RESTART was defined in 0.99pl8 to explicitly
-  * request restarting (and thus does nothing).  However, if this ever
-  * changes, I want to be safe
-  * DANG_END_REMARK
-  */
-#ifndef SA_RESTART
-#define SA_RESTART 0
-#error SA_RESTART Not defined
-#endif
-
-/* DANG_BEGIN_FUNCTION NEWSETQSIG
- *
- * arguments:
- * sig - the signal to have a handler installed to.
- * fun - the signal handler function to install
- *
- * description:
- *  All signals that wish to be handled properly in context with the
- * execution of vm86() mode, and signals that wish to use non-reentrant
- * functions should add themselves to the ADDSET_SIGNALS_THAT_QUEUE define
- * and use SETQSIG(). To that end they will also need to be set up in an
- * order such as SIGIO.
- *
- * DANG_END_FUNCTION
- *
- */
-#define ADDSET_SIGNALS_THAT_QUEUE(x) \
-do { \
-       sigaddset(x, SIGIO); \
-       sigaddset(x, SIGALRM); \
-       sigaddset(x, SIG_RELEASE); \
-       sigaddset(x, SIG_ACQUIRE); \
-} while(0)
-
-#ifdef __linux__
-#if GLIBC_VERSION_CODE
-#define SignalHandler __sighandler_t
-#endif
-#define NEWSETQSIG(sig, fun)	sa.sa_handler = (__sighandler_t)fun; \
-			/* Point to the top of the stack, minus 4 \
-			   just in case, and make it aligned  */ \
-			sa.sa_restorer = \
-			(void (*)()) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
-					sa.sa_flags = SA_RESTART ; \
-					sigemptyset(&sa.sa_mask); \
-					ADDSET_SIGNALS_THAT_QUEUE(&sa.sa_mask); \
-					dosemu_sigaction(sig, &sa, NULL);
-
-#define SETSIG(sig, fun)	sa.sa_handler = (SignalHandler)fun; \
-					sa.sa_flags = SA_RESTART; \
-					sigemptyset(&sa.sa_mask); \
-					sigaddset(&sa.sa_mask, SIGALRM); \
-					sigaction(sig, &sa, NULL);
-
-#define NEWSETSIG(sig, fun) \
-			sa.sa_handler = (__sighandler_t) fun; \
-			/* Point to the top of the stack, minus 4 \
-			   just in case, and make it aligned  */ \
-			sa.sa_restorer = \
-			(void (*)()) (((unsigned int)(cstack) + sizeof(cstack) - 4) & ~3); \
-			sa.sa_flags = SA_RESTART; \
-			sigemptyset(&sa.sa_mask); \
-			sigaddset(&sa.sa_mask, SIGALRM); \
-			dosemu_sigaction(sig, &sa, NULL);
-#endif
-
 EXTERN inline void SIGNAL_save( void (*signal_call)(void) );
-EXTERN inline void handle_signals(void);
+extern void handle_signals(void);
+extern void handle_signals_force_reentry(void);
 
+extern void addset_signals_that_queue(sigset_t *x);
+extern void registersig(int sig, void (*handler)(struct sigcontext_struct *));
+extern void init_handler(struct sigcontext_struct *scp);
+#ifdef __x86_64__
+extern int check_fix_fs_gs_base(unsigned char prefix);
+#endif
 
 /* 
  * DANG_BEGIN_REMARK
@@ -554,6 +426,7 @@ EXTERN struct config_info config;
  * DANG_END_REMARK
  */
 EXTERN int fatalerr INIT(0);
+EXTERN int in_leavedos;
 
 /*
  * DANG_BEGIN_REMARK
@@ -570,24 +443,38 @@ EXTERN int dosc_interface(void);
 EXTERN void dump_config_status(void *);
 EXTERN void signal_init(void);
 EXTERN void device_init(void);
-EXTERN void hardware_setup(void);
 EXTERN void memory_init(void);
+EXTERN void map_video_bios(void);
+EXTERN void stdio_init(void);
+EXTERN void time_setting_init(void);
 EXTERN void timer_interrupt_init(void);
+EXTERN void low_mem_init(void);
+EXTERN void print_version(void);
 EXTERN void keyboard_flags_init(void);
 EXTERN void video_config_init(void);
+EXTERN void video_post_init(void);
+EXTERN void video_mem_setup(void);
 EXTERN void printer_init(void);
+EXTERN void printer_mem_setup(void);
 EXTERN void video_close(void);
 EXTERN void hma_exit(void);
 EXTERN void ems_helper(void);
 EXTERN boolean_t ems_fn(struct vm86_regs *);
-EXTERN void serial_helper(void);
-EXTERN void mouse_helper(void);
-EXTERN void cdrom_helper(void);
+EXTERN void mouse_helper(struct vm86_regs *);
+EXTERN void cdrom_helper(unsigned char *, unsigned char *);
+EXTERN int mscdex(void);
 EXTERN void boot(void);
+EXTERN void do_liability_disclaimer_prompt(int stage, int prompt);
+EXTERN void install_dos(int post_boot);
 EXTERN int pkt_int(void);
+EXTERN int ipx_int7a(void);
 EXTERN void read_next_scancode_from_queue (void);
 EXTERN unsigned short detach (void);
+EXTERN void disallocate_vt (void);
+EXTERN void restore_vt (unsigned short vt);
 EXTERN void HMA_init(void);
+EXTERN void HMA_MAP(int HMA);
+EXTERN void hardware_run(void);
 
 extern char *Path_cdrom[];
 
