@@ -1,8 +1,8 @@
 /* 
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /*
@@ -136,6 +136,8 @@
 #define CRTC_INDEX		0x3d4
 #define CRTC_DATA		0x3d5
 
+#define COLOR_SELECT		0x3d9
+
 #define INPUT_STATUS_1		0x3da
 #define FEATURE_CONTROL_W	0x3da
 
@@ -148,6 +150,8 @@
 #define CFG_CRTC_ADDR_MODE      2
 #define CFG_CRTC_HEIGHT         3
 #define CFG_CRTC_WIDTH          4
+#define CFG_CRTC_LINE_COMPARE   5
+#define CFG_MODE_CONTROL        6
 
 
 /*
@@ -262,10 +266,12 @@ typedef struct {
 typedef struct {
   unsigned char *base;			/* base address of VGA memory */
   unsigned size;			/* size of memory in bytes */
+  unsigned wrap;			/* wrapping point */
   unsigned pages;			/* dto in pages */
-  int fd;				/* file descriptor for "/proc/self/mem" */
+  unsigned bank_base;			/* banked base in bytes */
+  unsigned bank_len;			/* banked length in bytes */
   unsigned lfb_base_page;		/* lfb base page, 0 -> no lfb support */
-  unsigned scratch_page;		/* for unmapped areas */
+  void *scratch_page;			/* for unmapped areas */
   vga_mapping_type map[VGAEMU_MAX_MAPPINGS];	/* all the mappings */
   unsigned bank_pages;			/* size of a bank in pages */
   unsigned bank;			/* selected bank */
@@ -317,6 +323,9 @@ typedef struct {
 typedef struct {
   unsigned addr_mode;
   unsigned cursor_location;
+  unsigned short cursor_shape;
+  unsigned line_compare;
+  unsigned char readonly;
   unsigned char index;
   unsigned char data[CRTC_MAX_INDEX + 1];
   unsigned char dirty[CRTC_MAX_INDEX + 1];
@@ -336,6 +345,9 @@ typedef struct {
   unsigned char mode;
   unsigned char index;
   unsigned char data[SEQ_MAX_INDEX + 1];
+  /* Helpful offsets for font processing    */
+  /* depends on map_select which is data[3] */
+  unsigned fontofs[2];
 } vga_seq_type;
 
 
@@ -402,6 +414,7 @@ typedef struct {
   vga_reconfig_type reconfig;		/* indicate when essential things have changed */
   int width;				/* in pixels */
   int height;				/* dto */
+  int line_compare;			/* dto */
   int scan_len;				/* in bytes */
   int text_width;
   int text_height;
@@ -457,6 +470,7 @@ typedef struct {
 extern vga_type vga;
 extern vgaemu_bios_type vgaemu_bios;
 
+struct ColorSpaceDesc;
 
 /*
  * Functions defined in env/video/vgaemu.c.
@@ -470,7 +484,7 @@ Bit16u VGA_emulate_inw(ioport_t);
 int vga_emu_fault(struct sigcontext_struct *, int pmode);
 #define VGA_EMU_FAULT(scp,code,pmode) vga_emu_fault(scp,pmode)
 #endif
-int vga_emu_init(vgaemu_display_type *);
+int vga_emu_init(int src_modes, struct ColorSpaceDesc *);
 void vga_emu_done(void);
 int vga_emu_update(vga_emu_update_type *);
 int vgaemu_switch_plane(unsigned);
@@ -480,7 +494,6 @@ int vga_emu_setmode(int, int, int);
 int vgaemu_map_bank(void);
 int vga_emu_set_textsize(int, int);
 void dirty_all_video_pages(void);
-int vga_emu_set_text_page(unsigned, unsigned);
 void dirty_all_vga_colors(void);
 int changed_vga_colors(DAC_entry *);
 void vgaemu_adj_cfg(unsigned, unsigned);
@@ -488,8 +501,13 @@ void vgaemu_scroll(int x0, int y0, int x1, int y1, int n, unsigned char attr);
 void vgaemu_put_char(int x, int y, unsigned char c, unsigned char attr);
 void vgaemu_put_pixel(int x, int y, unsigned char page, unsigned char attr);
 unsigned char vgaemu_get_pixel(int x, int y, unsigned char page);
-unsigned char Logical_VGA_read(unsigned offset);
-void Logical_VGA_write(unsigned offset, unsigned char value);
+unsigned char vga_read(const unsigned char *addr);
+void vga_write(unsigned char *addr, unsigned char val);
+unsigned short vga_read_word(const unsigned short *addr);
+void vga_write_word(unsigned short *addr, unsigned short val);
+void memcpy_to_vga(void *dst, const void *src, size_t len);
+void memcpy_from_vga(void *dst, const void *src, size_t len);
+void vga_memcpy(void *dst, const void *src, size_t len);
 /* for cpuemu: */
 int vga_emu_protect_page(unsigned, int);
 int vga_emu_adjust_protection(unsigned, unsigned);
@@ -579,6 +597,7 @@ void Misc_init(void);
 void Misc_set_misc_output(unsigned char);
 unsigned char Misc_get_misc_output(void);
 void Misc_set_feature_ctrl(unsigned char);
+void Misc_set_color_select(unsigned char);
 unsigned char Misc_get_feature_ctrl(void);
 unsigned char Misc_get_input_status_0(void);
 unsigned char Misc_get_input_status_1(void);
@@ -599,7 +618,9 @@ unsigned char Herc_get_mode_ctrl(void);
  */
 
 unsigned instr_len(unsigned char *);
-void instr_emu(struct sigcontext_struct *scp, int pmode);
+void instr_emu(struct sigcontext_struct *scp, int pmode, int cnt);
+int decode_modify_segreg_insn(struct sigcontext_struct *scp, int pmode,
+    unsigned int *new_val);
 
 /*
  * whether we emulate only writes to VGA memory or read & write
@@ -619,4 +640,19 @@ extern unsigned char vga_rom_16[256 * 16];
 extern unsigned char vga_rom_14_alt[1];
 extern unsigned char vga_rom_16_alt[1];
 
+/*
+ * The actual VGA/VBE BIOS is in vesabios.S and vesabios_pm.S. These are
+ * some labels pointing to interesting code portions. They are used in
+ * vbe_init() to setup the struct vgaemu_bios which holds all relevant
+ * info about our BIOS.
+ */
+
+extern void vgaemu_bios_start(void);
+extern void vgaemu_bios_prod_name(void);
+extern void vgaemu_bios_win_func(void);
+extern void vgaemu_bios_end(void);
+extern void vgaemu_bios_pm_interface(void);
+extern void vgaemu_bios_pm_interface_end(void);
+
+
 #endif	/* !defined __VGAEMU_H */
