@@ -1,14 +1,14 @@
 /*
+ * Written: 10/29/93 by Tim Bird (original code)
  * All modifications in this file to the original code are
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /***********************************************
  * File: LREDIR.C
  *  Program for Linux DOSEMU disk redirector functions
- * Written: 10/29/93 by Tim Bird
  * Fixes: 1/7/95 by Tim Josling TEJ: check correct # args
  *      :                            display read/write status correctly
  * Changes: 3/19/95 by Kang-Jin Lee
@@ -57,8 +57,12 @@
 #include "config.h"
 #include "emu.h"
 #include "memory.h"
+#include "msetenv.h"
 #include "doshelpers.h"
-#include "../coopthreads/coopthreads.h"
+#include "utilities.h"
+#include "lredir.h"
+#include "builtins.h"
+#include "disks.h"
 
 #define printf	com_printf
 #define	intr	com_intr
@@ -90,15 +94,18 @@ typedef unsigned int uint16;
 #define KEYWORD_DEL   "DELETE"
 #define KEYWORD_DEL_COMPARE_LENGTH  3
 
+#define KEYWORD_HELP   "HELP"
+#define KEYWORD_HELP_COMPARE_LENGTH  4
+
 #define DEFAULT_REDIR_PARAM   0
 
 #include "doserror.h"
 
-FAR_PTR /* char far * */
+static FAR_PTR /* char far * */
 GetListOfLists(void)
 {
     FAR_PTR LOL;
-    struct REGPACK preg;
+    struct REGPACK preg = REGPACK_INIT;
 
     preg.r_ax = DOS_GET_LIST_OF_LISTS;
     intr(0x21, &preg);
@@ -106,11 +113,11 @@ GetListOfLists(void)
     return (LOL);
 }
 
-FAR_PTR /* char far * */
+static FAR_PTR /* char far * */
 GetSDAPointer(void)
 {
     FAR_PTR SDA;
-    struct REGPACK preg;
+    struct REGPACK preg = REGPACK_INIT;
 
     preg.r_ax = DOS_GET_SDA_POINTER;
     intr(0x21, &preg);
@@ -123,25 +130,32 @@ GetSDAPointer(void)
  * InitMFS - call Emulator to initialize MFS
  ********************************************/
 /* tej - changed return type to void as nothing returned */
-void InitMFS(void)
+static void InitMFS(void)
 {
     FAR_PTR LOL;
     FAR_PTR SDA;
-    struct REGPACK preg;
+    unsigned char _osmajor;
+    unsigned char _osminor;
+    state_t preg;
 
     LOL = GetListOfLists();
     SDA = GetSDAPointer();
 
-    /* get DOS version into CX */
-    preg.r_cx = _osmajor | (_osminor <<8);
+    /* now get the DOS version */
+    HI(ax) = 0x30;
+    call_msdos();
+    _osmajor = LO(ax);
+    _osminor = HI(ax);
 
-    preg.r_dx = FP_OFF16(LOL);
-    preg.r_es = FP_SEG16(LOL);
-    preg.r_si = FP_OFF16(SDA);
-    preg.r_ds = FP_SEG16(SDA);
-    preg.r_bx = 0x500;
-    preg.r_ax = DOS_HELPER_MFS_HELPER;
-    intr(DOS_HELPER_INT, &preg);
+    /* get DOS version into CX */
+    preg.ecx = _osmajor | (_osminor <<8);
+
+    preg.edx = FP_OFF16(LOL);
+    preg.es = FP_SEG16(LOL);
+    preg.esi = FP_OFF16(SDA);
+    preg.ds = FP_SEG16(SDA);
+    preg.ebx = 0x500;
+    mfs_helper(&preg);
 }
 
 /********************************************
@@ -163,11 +177,11 @@ void InitMFS(void)
  *  It is not actually saved and returned as specified by the redirector
  *  specification.  This type of usage is common among commercial redirectors.
  ********************************************/
-uint16 RedirectDevice(char *deviceStr, char *resourceStr, uint8 deviceType,
+static uint16 RedirectDevice(char *deviceStr, char *resourceStr, uint8 deviceType,
                       uint16 deviceParameter)
 {
     char slashedResourceStr[MAX_RESOURCE_PATH_LENGTH];
-    struct REGPACK preg;
+    struct REGPACK preg = REGPACK_INIT;
     char *dStr, *sStr;
 
     /* prepend the resource string with slashes */
@@ -214,28 +228,28 @@ uint16 RedirectDevice(char *deviceStr, char *resourceStr, uint8 deviceType,
  * NOTES:
  *
  ********************************************/
-uint16 GetRedirection(uint16 redirIndex, char *deviceStr, char *resourceStr,
+static uint16 GetRedirection(uint16 redirIndex, char *deviceStr, char *resourceStr,
                       uint8 * deviceType, uint16 * deviceParameter)
 {
     uint16 ccode;
     uint8 deviceTypeTemp;
     char slashedResourceStr[MAX_RESOURCE_PATH_LENGTH];
-    struct REGPACK preg;
+    struct REGPACK preg = REGPACK_INIT;
     char *dStr, *sStr;
 
-    dStr = com_strdup(deviceStr);
+    dStr = lowmem_alloc(16);
     preg.r_ds = FP_SEG(dStr);
     preg.r_si = FP_OFF(dStr);
-    sStr = com_strdup(slashedResourceStr);
+    sStr = lowmem_alloc(128);
     preg.r_es = FP_SEG(sStr);
     preg.r_di = FP_OFF(sStr);
     preg.r_bx = redirIndex;
     preg.r_ax = DOS_GET_REDIRECTION;
     intr(0x21, &preg);
     strcpy(deviceStr,dStr);
-    com_strfree(dStr);
+    lowmem_free(dStr, 16);
     strcpy(slashedResourceStr,sStr);
-    com_strfree(sStr);
+    lowmem_free(sStr, 128);
 
     ccode = preg.r_ax;
     deviceTypeTemp = preg.r_bx & 0xff;       /* save device type before C ruins it */
@@ -264,9 +278,9 @@ uint16 GetRedirection(uint16 redirIndex, char *deviceStr, char *resourceStr,
  * NOTES:
  *
  ********************************************/
-uint16 CancelRedirection(char *deviceStr)
+static uint16 CancelRedirection(char *deviceStr)
 {
-    struct REGPACK preg;
+    struct REGPACK preg = REGPACK_INIT;
     char *dStr;
 
     dStr = com_strdup(deviceStr);
@@ -291,7 +305,7 @@ uint16 CancelRedirection(char *deviceStr)
  *  I show the read-only attribute for each drive
  *    which is returned in deviceParam.
  *************************************/
-void
+static void
 ShowMyRedirections(void)
 {
     int driveCount;
@@ -304,10 +318,8 @@ ShowMyRedirections(void)
     redirIndex = 0;
     driveCount = 0;
 
-    ccode = GetRedirection(redirIndex, deviceStr, resourceStr,
-                           &deviceType, &deviceParam);
-
-    while (ccode == CC_SUCCESS) {
+    while ((ccode = GetRedirection(redirIndex, deviceStr, resourceStr,
+                           &deviceType, &deviceParam)) == CC_SUCCESS) {
       /* only print disk redirections here */
       if (deviceType == REDIR_DISK_TYPE) {
         if (driveCount == 0) {
@@ -318,7 +330,9 @@ ShowMyRedirections(void)
 
         /* read attribute is returned in the device parameter */
         if (deviceParam & 0x80) {
-          switch (deviceParam & 0x7f) { /* tej 1/7/95 */
+	  if ((deviceParam & 0x7f) > 1)
+	    printf("CDROM:%d ", (deviceParam & 0x7f) - 1);
+          switch ((deviceParam & 0x7f) != 0) {
             case READ_ONLY_DRIVE_ATTRIBUTE:
               printf("attrib = READ ONLY\n");
               break;
@@ -330,8 +344,6 @@ ShowMyRedirections(void)
       }
 
       redirIndex++;
-      ccode = GetRedirection(redirIndex, deviceStr, resourceStr,
-                             &deviceType, &deviceParam);
     }
 
     if (driveCount == 0) {
@@ -339,7 +351,7 @@ ShowMyRedirections(void)
     }
 }
 
-void
+static void
 DeleteDriveRedirection(char *deviceStr)
 {
     uint16 ccode;
@@ -356,6 +368,55 @@ DeleteDriveRedirection(char *deviceStr)
     }
 }
 
+static int FindRedirectionByDevice(char *deviceStr, char *resourceStr)
+{
+    uint16 redirIndex = 0, deviceParam, ccode;
+    uint8 deviceType;
+    char dStr[MAX_DEVICE_STRING_LENGTH];
+    char dStrSrc[MAX_DEVICE_STRING_LENGTH];
+
+    snprintf(dStrSrc, MAX_DEVICE_STRING_LENGTH, "%s", deviceStr);
+    strupr(dStrSrc);
+    while ((ccode = GetRedirection(redirIndex, dStr, resourceStr,
+                           &deviceType, &deviceParam)) == CC_SUCCESS) {
+      if (strcmp(dStrSrc, dStr) == 0)
+        break;
+      redirIndex++;
+    }
+
+    return ccode;
+}
+
+static int FindFATRedirectionByDevice(char *deviceStr, char *resourceStr)
+{
+    struct DINFO *di;
+    char *dir;
+    fatfs_t *f;
+    unsigned long serial;
+    if (!(di = (struct DINFO *)lowmem_alloc(sizeof(struct DINFO))))
+	return 0;
+    LWORD(eax) = 0x6900;
+    LWORD(ebx) = toupper(deviceStr[0]) - 'A' + 1;
+    REG(ds) = FP_SEG32(di);
+    LWORD(edx) = FP_OFF32(di);
+    call_msdos();
+    if (REG(eflags) & CF) {
+	lowmem_free((void *)di, sizeof(struct DINFO));
+	printf("error retrieving serial, %#x\n", LWORD(eax));
+	return -1;
+    }
+    serial = READ_DWORD(&di->serial);
+    lowmem_free((void *)di, sizeof(struct DINFO));
+    f = get_fat_fs_by_serial(serial);
+    if (!f || !(dir = f->dir)) {
+	printf("error identifying FAT volume\n");
+	return -1;
+    }
+    strcpy(resourceStr, "LINUX\\FS");
+    strcat(resourceStr, dir);
+    return CC_SUCCESS;
+}
+
 /********************************************
  * Check wether we are running DosC (FreeDos)
  * and check wether this version can cope with redirection
@@ -366,12 +427,14 @@ DeleteDriveRedirection(char *deviceStr)
  *  otherwise returns the DosC 'build' number
  *
  ********************************************/
-uint16 CheckForDosc(void)
+/* no longer used -- Bart */
+#if 0
+static uint16 CheckForDosc(void)
 {
-    struct REGPACK preg;
+    struct REGPACK preg = REGPACK_INIT;
 
     preg.r_ax = 0xdddc;
-    intr(DOS_HELPER_INT, &preg);
+    dos_helper_r(&preg);
 
     if (preg.r_ax == 0xdddc) {
       return 0;
@@ -380,11 +443,13 @@ uint16 CheckForDosc(void)
       return (preg.r_bx);
     }
 }
+#endif
 
 int lredir_main(int argc, char **argv)
 {
     uint16 ccode;
     uint16 deviceParam;
+    int carg;
 #if 0
     unsigned long dversion;
 #endif
@@ -404,13 +469,19 @@ int lredir_main(int argc, char **argv)
     }
 
     /* tej one parm is either error or HELP/-help etc */
-    if (argc == 2) {
-      printf("Usage: LREDIR [drive: LINUX\\FS\\path [R] | HELP]\n");
+    if (argc == 2 && strncmpi(argv[1], KEYWORD_HELP, KEYWORD_HELP_COMPARE_LENGTH) == 0) {
+      printf("Usage: LREDIR [[drive:] LINUX\\FS\\path [R] | [C [n]] | HELP]\n");
       printf("Redirect a drive to the Linux file system.\n\n");
       printf("LREDIR X: LINUX\\FS\\tmp\n");
       printf("  Redirect drive X: to /tmp of Linux file system for read/write\n");
       printf("  If R is specified, the drive will be read-only\n");
+      printf("  If C is specified, (read-only) CDROM n is used (n=1 by default)\n");
       printf("  ${home} represents user's home directory\n\n");
+      printf("  If drive is not specified, the next available drive will be used.");
+      printf("LREDIR X: Y:\n");
+      printf("  Redirect drive X: to where the drive Y: is redirected.\n");
+      printf("  If F is specified, the path for Y: is taken from its emulated "
+    	     "FAT volume.\n\n");
       printf("LREDIR DEL drive:\n");
       printf("  delete a drive redirection\n\n");
       printf("LREDIR\n");
@@ -420,21 +491,57 @@ int lredir_main(int argc, char **argv)
       return(0);
     }
 
-    if (strncmpi(argv[1], KEYWORD_DEL, KEYWORD_DEL_COMPARE_LENGTH) == 0) {
+    if (strncmpi(argv[1], KEYWORD_DEL, KEYWORD_DEL_COMPARE_LENGTH) == 0) {    
       DeleteDriveRedirection(argv[2]);
       return(0);
     }
 
     /* assume the command is to redirect a drive */
     /* read the drive letter and resource string */
-    strcpy(deviceStr, argv[1]);
-    strcpy(resourceStr, argv[2]);
+    carg = 3;
+    if (argc > 2 && argv[2][1] == ':') {
+      /* lredir c: d: */
+      strcpy(deviceStr, argv[1]);
+      if ((argc > 3 && toupper(argv[3][0]) == 'F') ||
+    	((ccode = FindRedirectionByDevice(argv[2], resourceStr)) != CC_SUCCESS)) {
+        if ((ccode = FindFATRedirectionByDevice(argv[2], resourceStr)) != CC_SUCCESS) {
+          printf("Error: unable to find redirection for drive %s\n", argv[2]);
+	  goto MainExit;
+	}
+      }
+    } else {
+      if (argc > 1 && argv[1][1] != ':') {
+	int nextDrive;
+        strcpy(resourceStr, argv[1]);
+	nextDrive = find_drive(resourceStr);
+	if (nextDrive == -26) {
+		printf("Cannot redirect (maybe no drives available).");
+		return(0);
+	} else if (nextDrive == -27) {
+		printf("Cannot canonicalize drive root path.\n");
+		return(0);
+	}
+        deviceStr[0] = -nextDrive + 'A';
+	deviceStr[1] = ':';
+	deviceStr[2] = '\0';
+	carg = 2;
+      } else if (argc > 2) {
+        strcpy(deviceStr, argv[1]);
+        strcpy(resourceStr, argv[2]);
+      }
+    }
     deviceParam = DEFAULT_REDIR_PARAM;
 
-    if (argc > 3) {
-      if (toupper(argv[3][0]) == 'R') {
+    if (argc > carg) {
+      if (toupper(argv[carg][0]) == 'R') {
         deviceParam = 1;
       }
+      if (toupper(argv[carg][0]) == 'C') {
+	int cdrom = 1;
+	if (argc > carg+1 && argv[carg+1][0] >= '1' && argv[carg+1][0] <= '4')
+	  cdrom = argv[carg+1][0] - '0';
+        deviceParam = 1 + cdrom;
+      }
     }
 
     /* upper-case both strings */
@@ -457,9 +564,14 @@ int lredir_main(int argc, char **argv)
              ccode, decode_DOS_error(ccode), deviceStr, resourceStr);
       goto MainExit;
     }
+    if ((argc <= 2 || argv[2][1] != ':') && argc >= 2 && argv[1][1] != ':')
+	msetenv("DOSEMU_LASTREDIR", deviceStr);
 
-    printf("%s = %s  attrib = ", deviceStr, resourceStr);
-    switch (deviceParam) {
+    printf("%s = %s", deviceStr, resourceStr);
+    if (deviceParam > 1)
+      printf(" CDROM:%d", deviceParam - 1);
+    printf(" attrib = ");
+    switch (deviceParam != 0) {
       case READ_ONLY_DRIVE_ATTRIBUTE:
         printf("READ ONLY\n");
         break;
