@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 
@@ -9,8 +9,10 @@
 #define _EMU_KEYB_CLNT_H
 
 #include "config.h"
+
 #include "emu.h"
 #include "keyboard.h"
+#include "translate.h"
 
 int keyb_client_init(void);
 void keyb_client_reset(void);
@@ -20,6 +22,9 @@ void keyb_client_set_leds(t_modifiers modifiers);
 
 int paste_text(const char *text, int len);
 
+/* this should really go somewhere else ... */
+void handle_slang_keys(Boolean make, t_keysym key);
+
 /* For the current sigio handler, this still has to be defined here. */
 EXTERN int kbd_fd INIT(-1);
 
@@ -31,12 +36,15 @@ struct keyboard_client {
   void   (*close)(void);
   void   (*run)(void);         /* check if keys are ready and process them */
   void   (*set_leds)(t_modifiers modifiers);
+  void   (*handle_keys)(Boolean make, t_keysym key);
+  struct keyboard_client *next;
 };
 
+void register_keyboard_client(struct keyboard_client *keyboard);
+
 EXTERN struct keyboard_client *Keyboard INIT(NULL);
 extern struct keyboard_client Keyboard_raw;
 extern struct keyboard_client Keyboard_slang;
-extern struct keyboard_client Keyboard_X;
 extern struct keyboard_client Keyboard_none;
 
 extern void  dos_slang_redraw(void);
