@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 /* 
  * DANG_BEGIN_MODULE
@@ -20,9 +20,14 @@
  * DANG_END_CHANGELOG
  */
 
-#include "slang.h" 
+#include <slang.h>
 #include "dosemu_debug.h"
 #include "emu.h"
+#include "init.h"
+#include "video.h"
+#include "vc.h"
+#include "keyb_clients.h"
+#include "utilities.h"
 
 #include "env_term.h"
 
@@ -77,3 +82,18 @@ void term_close(void)
 	if (--term_initialized != 0) 
 		return;
 }
+
+CONSTRUCTOR(static void init(void))
+{
+	if (!Video) {
+		Video = &Video_term;
+		v_printf("VID: Video set to Video_term\n");
+	}
+	register_keyboard_client(&Keyboard_raw);
+	register_keyboard_client(&Keyboard_slang);
+#ifdef USE_GPM
+	if (!config.vga && config.mouse.intdrv && on_console())
+		load_plugin("gpm");
+#endif
+	register_mouse_client(&Mouse_xterm);
+}
