@@ -1,7 +1,7 @@
 /* 
- * (C) Copyright 1992, ..., 2002 the "DOSEMU-Development-Team".
+ * (C) Copyright 1992, ..., 2007 the "DOSEMU-Development-Team".
  *
- * for details see file COPYING in the DOSEMU distribution
+ * for details see file COPYING.DOSEMU in the DOSEMU distribution
  */
 
 /* 
@@ -24,23 +24,56 @@
  * John E. Davis (Nov 17, 1994).
  */
 
+/* Some notes about how various versions of the SLang library are used,
+ * and UTF8. Right now (May 2005) there are three SLangs in mainstream use:
+ * a) slang 1.4.x
+ * b) slang 1.4.x + utf8 patch
+ * c) slang 2.0
+ *
+ * a) works with 8 bit character sets, but not with utf8
+ *    workarounds: upgrade to 2.0 or don't use utf8
+ * b) works with 8 bit and utf8, but has problems when the external
+ *    charset=cp437 (*)
+ * c) is ideal. It works, no problems.
+ *
+ * (*): b) ignores the setting of SLsmg_Display_Eight_Bit.
+ * any character between 0x80 and 0x9f is not displayed.
+ * in any case $_external_charset="cp437" is rarely necessary, perhaps
+ * only if you like to use an xterm with the vga font (but then, why not
+ * use xdosemu?)
+ *
+ * On the Linux console in non-utf8 mode a special trick is used to
+ * be able to display almost all cp437 characters: a reconstruction
+ * of the "alternate character string" that takes advantage of the
+ * fact that ACS=cp437 here. All cp437 characters with unicode>=256
+ * are placed in this table.
+ */
+
 #include <stdio.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
 #include <sys/ioctl.h>
+#include <limits.h>
 #include <termios.h>
 #include <errno.h>
+#include <slang.h>
 
 #include "bios.h"
 #include "emu.h"
 #include "memory.h"
 #include "video.h" 
+#include "serial.h"
 #include "keyboard.h"
-#include "slang.h"
+#include "keyb_clients.h"
 #include "env_term.h"
 #include "translate.h"
+#include "dosemu_charset.h"
+#include "vgaemu.h"
+#include "vgatext.h"
+
+struct text_system Text_term;
 
 /* The interpretation of the DOS attributes depend upon if the adapter is 
  * color or not.
@@ -64,7 +97,6 @@
  *      001          000           Underline
  *     anything else is invalid.
  */    
-extern int no_local_video;
 
 static int BW_Attribute_Map[256];
 static int Color_Attribute_Map[256];
@@ -72,18 +104,31 @@ static int Color_Attribute_Map[256];
 static int *Attribute_Map;
 /* if negative, char is invisible */
 
-int cursor_blink = 1;
-static unsigned char The_Charset[256];
+/* The layout of one charset element is:
+   mb0 mb1 mb2 len
+   as we never need more than 3 characters to represent a DOS character
+   in any multibyte charset that SLang supports (utf8);
+   this allows an efficient memcpy of 4 (gcc makes that "mov") followed
+   by an increase by len bytes of the buffer pointer in the draw_string
+   routine
+   mb1 != 0 in utf8 mode means that mb1 is in the alternate character set.
+ */
+static unsigned char The_Charset[256][4];
 static int slang_update (void);
+static void term_write_nchars_8bit(unsigned char *text, int len, Bit8u attr);
+static void term_write_nchars_utf8(unsigned char *text, int len, Bit8u attr);
+static void (*term_write_nchars)(unsigned char *, int, Bit8u) = term_write_nchars_utf8;
 
 static int Slsmg_is_not_initialized = 1;
-static int Use_IBM_Codes = 0;
 
 /* I think this is what is assumed. */
 static int Rows = 25;
 static int Columns = 80;
 
-void get_screen_size (void)
+/* sliding window for terminals < 25 lines */
+static int DOSemu_Terminal_Scroll_Min = 0;
+
+static void get_screen_size (void)
 {
   struct winsize ws;		/* buffer for TIOCSWINSZ */
 
@@ -91,6 +136,12 @@ void get_screen_size (void)
    SLtt_Screen_Cols = 0;
    if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) >= 0) 
      {
+        if (ws.ws_row > MAX_LINES || ws.ws_col > MAX_COLUMNS)
+	  {
+	    error("Screen size is too large: %dx%d, max is %dx%d\n",
+		  ws.ws_col, ws.ws_row, MAX_COLUMNS, MAX_LINES);
+	    leavedos(0x63);
+	  }
 	SLtt_Screen_Rows = ws.ws_row;
 	SLtt_Screen_Cols = ws.ws_col;
      }
@@ -103,12 +154,82 @@ void get_screen_size (void)
    Rows = SLtt_Screen_Rows;
    Columns = SLtt_Screen_Cols;
    if (Rows < 25) Rows = 25;
+   vga.text_width = Columns;
+   vga.scan_len = 2 * Columns;
+   vga.text_height = Rows;
+}
+
+/* bitmap of cp437 characters < 32 that are always control characters
+   according to the Linux kernel */
+#define CTRL_ALWAYS 0x0800f501
+static t_unicode acs_to_uni[128];
+
+/* construct an easy lookup array to figure out the relationship
+   between vt100 (acs) characters and unicode.
+   If the acs charset is the same as cp437 (as it is on the Linux
+   console) we replace the terminal acs lookup string with something
+   that can reach all the funny characters; there are around 100
+   of them
+*/
+static void construct_acs_table(void)
+{
+	struct char_set *charset;
+	struct char_set_state state;
+	char *smacs = SLtt_tgetstr ("as");
+	char *smpch = SLtt_tgetstr ("S2");
+	t_unicode uni;
+
+	if (smacs && smpch && strcmp(smacs, smpch) == 0) {
+		int i, j = 1;
+		char *cp437_acs = malloc(128*2);
+		charset = lookup_charset("cp437");
+		for (i = 1; i < 256; i++) {
+			unsigned char c = i;
+			if ((c >= ' ') || !((CTRL_ALWAYS >> c) & 1)) {
+				init_charset_state(&state, charset);
+				charset_to_unicode(&state, &uni, &c, 1);
+				if (uni >= 256) {
+					cp437_acs[(j-1)*2] = j;
+					cp437_acs[(j-1)*2+1] = c;
+					acs_to_uni[j] = uni;
+					j++;
+				}
+				cleanup_charset_state(&state);
+			}
+		}
+		cp437_acs[(j-1)*2] = '\0';
+		SLtt_Graphics_Char_Pairs = strdup(cp437_acs);
+		free(cp437_acs);
+	} else if (SLtt_Graphics_Char_Pairs) {
+		char *p;
+		charset = lookup_charset("vt100");
+		for (p = SLtt_Graphics_Char_Pairs; *p; p += 2) {
+			init_charset_state(&state, charset);
+			charset_to_unicode(&state, &uni, p, 1);
+			if (uni >= 256) {
+				acs_to_uni[(unsigned char)p[0]] = uni;
+			}
+			cleanup_charset_state(&state);
+		}
+	}
+}
+
+/* check if c is an approximation of uni */
+static int uni_approx(struct char_set *charset, t_unicode uni, unsigned char c)
+{
+        struct char_set_state state;
+        t_unicode u;
+        size_t result;
+
+        init_charset_state(&state, charset);
+        result = charset_to_unicode(&state, &u, &c, 1);
+        cleanup_charset_state(&state);
+        return (result == 1 && u != uni);
 }
 
 static void set_char_set (void)
 {
-	struct char_set *term_charset = trconfig.output_charset;
-	struct char_set *display_charset = trconfig.video_mem_charset;
+	struct char_set *term_charset, *display_charset;
 	/* For now neither encoding can be a stateful encoding... */
 	/* The video charset can never be a stateful encoding
 	 * it is a hardware limitation.
@@ -117,12 +238,15 @@ static void set_char_set (void)
 	struct char_set_state display_state;
 	int i;
 
+	term_charset = trconfig.output_charset;
+	display_charset = trconfig.video_mem_charset;
+
 	/* Initial don't allow the high control characters. */
 	SLsmg_Display_Eight_Bit = 0xA0;
 	/* Build the translate tables */
 	v_printf("mapping internal characters to terminal characters:\n");
 	for(i= 0; i <= 0xff; i++) {
-		unsigned char buff[1];
+		unsigned char buff[MB_LEN_MAX + 1];
 		t_unicode uni;
 		size_t result;
 
@@ -130,15 +254,30 @@ static void set_char_set (void)
 		init_charset_state(&display_state, display_charset);
 
 		buff[0] = i;
+		buff[1] = '\0';
 		result = charset_to_unicode(&display_state, &uni, buff, 1);
-		result = unicode_to_charset(&term_state, uni, The_Charset + i, 1);
-		v_printf("mapping: %c -> %04x -> %c\n", i, uni, The_Charset[i]);
+		result = unicode_to_charset(&term_state, uni, buff, MB_LEN_MAX);
+		if (result < 1 || result >= 4)
+			result = 1;
+		buff[3] = result;
+		if (result == 1 && SLtt_Graphics_Char_Pairs && uni >= 0x100 &&
+		    uni_approx(term_charset, uni, buff[0])) {
+			char *p;
+			for (p = SLtt_Graphics_Char_Pairs; *p; p += 2)
+				if (acs_to_uni[(unsigned char)p[0]] == uni) {
+					buff[1] = p[0];
+					break;
+				}
+		}
+		memcpy(The_Charset + i, buff, 4);
+		v_printf("mapping: %x -> %04x -> %.*s (len=%zu,acs=%x)\n", i, uni,
+			 (int)result, buff, result, result == 1 && buff[1] ? buff[1] : 0);
 
 		/* If we have any non control charcters in 0x80 - 0x9f
 		 * set up  the slang code up so we can send them. 
 		 */
-		if ((The_Charset[i] >= 0x80) && ((The_Charset[i] <= 0x9F)) &&
-			(((uni >= 0x20) && (uni < 0x80)) || (uni > 0x9f))) {
+		if (result > 1 || (buff[0] >= 0x80 && buff[0] <= 0x9f
+		    && (((uni >= 0x20) && (uni < 0x80)) || (uni > 0x9f)))) {
 			/* Allow us to use chars 0x80 to 0x9F */
 			SLsmg_Display_Eight_Bit = 0x80;
 		}
@@ -149,36 +288,65 @@ static void set_char_set (void)
 	/* Slang should filter out the control sequences for us... 
 	 * So don't worry about characters 0x00 - 0x1f && 0x80 - 0x9f
 	 */
-	if (trconfig.output_charset == lookup_charset("cp437")) {
-		Use_IBM_Codes = 1;
-		/* Should we be testing TERM here??? */
-		/* The following turns on the IBM character set mode of virtual console
-		 * The same code is echoed twice, then just in case the escape code
-		 * not recognized and was printed, erase it with spaces.
-		 */
-		SLtt_write_string ("\033(U\033(U\r        \r");
-	}
 }
 
+int using_xterm(void)
+{
+   char *term = getenv("TERM");
+
+   if (term == NULL)
+      return 0;
+   
+   return !strncmp("xterm", term, 5) ||
+           !strncmp("rxvt", term, 4) ||
+           !strcmp("dtterm", term);
+}
+
+static int term_change_config(unsigned item, void *buf)
+{
+   static char title_appname[TITLE_APPNAME_MAXLEN];
+
+   switch (item) {
+   case CHG_TITLE_APPNAME:
+      snprintf (title_appname, TITLE_APPNAME_MAXLEN, "%s", (char *) buf);
+      if (config.xterm_title && strlen(config.xterm_title)) {
+	printf("\x1b]2;");
+	printf(config.xterm_title, (char *)buf);
+	printf("\7");
+      }
+      return 0;
+   case GET_TITLE_APPNAME:
+      snprintf (buf, TITLE_APPNAME_MAXLEN, "%s", title_appname);
+      return 0;
+   } 
+   return 100;   
+}
+
+static void sigwinch(struct sigcontext_struct *scp)
+{
+  get_screen_size();
+}
+
+#if SLANG_VERSION < 20000
+/* replacement function to deal with old slangs */
+static int SLutf8_enable(int mode)
+{
+  if (mode != -1)
+    return mode;
+  return MB_CUR_MAX > 1;
+}
+#endif
+
 /* The following initializes the terminal.  This should be called at the
  * startup of DOSEMU if it's running in terminal mode.
  */ 
-int
-terminal_initialize()
+static int terminal_initialize(void)
 {
    SLtt_Char_Type sltt_attr, fg, bg, attr, color_sltt_attr, bw_sltt_attr;
    int is_color = config.term_color;
    int rotate[8];
 
    v_printf("VID: terminal_initialize() called \n");
-   /* I do not know why this routine is called if our update is not
-    * called.  Oh well.... 
-    */
-   if (config.console_video) 
-     {
-	Slsmg_is_not_initialized = 1;
-	return 0;
-     }
    Slsmg_is_not_initialized = 0;
    
    /* This maps (r,g,b) --> (b,g,r) */
@@ -193,28 +361,49 @@ terminal_initialize()
    else
      Video_term.update_screen = NULL;
 
+   if (using_xterm())
+     Video_term.change_config = term_change_config;
+
    term_init();
 
    get_screen_size ();
 
-   /* Now set the variables li and co back to their dos defaults. */
+   /* respond to resize events unless we're running on the Linux console
+      with raw keyboard: then SIGWINCH = SIG_RELEASE ! */
+   if (!config.console_keyb) {
+     registersig(SIGWINCH, sigwinch);
+   }
 
-#if 0
-   if (li < 25)
-     {
-        li = Rows;
-        co = Columns;
-     }
-#endif
-   
+   /* initialize VGA emulator */
+   use_bitmap_font = FALSE;
+   config.X_updatelines = Rows;
+   if (vga_emu_init(0, NULL)) {
+     error("X: X_init: VGAEmu init failed!\n");
+     leavedos(99);
+   }
+   vga.text_width = Columns;
+   vga.scan_len = 2 * Columns;
+   vga.text_height = Rows;
+   register_text_system(&Text_term);
+   vga_emu_setmode(video_mode, Columns, Rows);
+
+#if SLANG_VERSION < 20000
    SLtt_Use_Blink_For_ACS = 1;
+#endif
    SLtt_Blink_Mode = 1;
-   
+
    SLtt_Use_Ansi_Colors = is_color;
    
    if (is_color) Attribute_Map = Color_Attribute_Map;
    else Attribute_Map = BW_Attribute_Map;
 
+   if (!SLutf8_enable(
+	strstr("utf8", trconfig.output_charset->names[0]) ? 1 :
+	strstr("default", trconfig.output_charset->names[0]) ? -1 : 0)) {
+      construct_acs_table();
+      term_write_nchars = term_write_nchars_8bit;
+   }
+
 #if SLANG_VERSION < 10000
    if (!SLsmg_init_smg ())
 #else
@@ -267,9 +456,12 @@ terminal_initialize()
    
    /* object 0 is special.  It is normal video.  Lets fix that now. */   
    BW_Attribute_Map[0x7] = Color_Attribute_Map[0x7] = 0;
+   BW_Attribute_Map[0] = Color_Attribute_Map[0] = 7;
    
    SLtt_set_color_object (0, 0x000700);
    SLtt_set_mono (0, NULL, 0x000700);
+   SLtt_set_color_object (7, 0);
+   SLtt_set_mono (7, NULL, 0);
    
    SLsmg_refresh ();
    /* This goes after the refresh because it might try 
@@ -279,7 +471,7 @@ terminal_initialize()
    return 0;
 }
 
-void terminal_close (void)
+static void terminal_close (void)
 {
    v_printf("VID: terminal_close() called\n");
    if (Slsmg_is_not_initialized == 0)
@@ -287,25 +479,21 @@ void terminal_close (void)
 	SLsmg_gotorc (SLtt_Screen_Rows - 1, 0);
 	SLsmg_refresh ();
 	SLsmg_reset_smg ();
-	if (Use_IBM_Codes) 
-	  {
-	     SLtt_write_string ("\n\033(B\033(B\r         \r");
-	     SLtt_flush_output ();
-	  }
-	else putc ('\n', stdout);
+	putc ('\n', stdout);
 	Slsmg_is_not_initialized = 1;
 	term_close();
      }
 }
 
-void
-v_write(int fd, unsigned char *ch, int len)
+#if 0 /* unused -- Bart */
+static void v_write(int fd, unsigned char *ch, int len)
 {
-  if (!config.console_video && !config.usesX)
+  if (!config.console_video)
     DOS_SYSCALL(write(fd, ch, len));
   else
     error("(video) v_write deferred for console_video\n");
 }
+#endif
 
 static char *Help[] = 
 {
@@ -378,18 +566,13 @@ static void show_help (void)
  */
 static int slang_update (void)
 {
-   register unsigned short *line, *prev_line, *line_max, char_attr;
-   int i, pn, sn, row_len = Columns * 2;
-   unsigned char buf[256], *bufp;
-   int last_obj = 1000, this_obj;
-   int changed = 0;
-   int imin, imax;
-   
+   int changed, imin, cursor_row, cursor_col;
+
    static int last_row, last_col, help_showing;
    static const char *last_prompt = NULL;
-   
-   SLtt_Blink_Mode = char_blink;
-   
+
+   SLtt_Blink_Mode = (vga.attr.data[0x10] & 0x8) != 0;
+
    if (DOSemu_Slang_Show_Help) 
      {
 	if (help_showing == 0) show_help ();
@@ -398,60 +581,27 @@ static int slang_update (void)
      }
    help_showing = 0;
    
+   cursor_row = (vga.crtc.cursor_location - vga.display_start) / vga.scan_len;
+   cursor_col = ((vga.crtc.cursor_location - vga.display_start) % vga.scan_len) / 2;
+   imin = Rows - SLtt_Screen_Rows;
    if (((DOSemu_Terminal_Scroll == 0) && 
 	(cursor_row < SLtt_Screen_Rows))
        || (DOSemu_Terminal_Scroll == -1))
      {
 	imin = 0;
-	imax = SLtt_Screen_Rows;
      }
-   else 
-     {
-	imax = Rows;
-	imin = imax - SLtt_Screen_Rows;
-	if (imin < 0) imin = 0;
-     }
-   
-   pn = 0;
-   sn = imin * Columns;
-   for (i = imin; i < imax; i++)
-     {
-#if 0
-	if (memcmp(screen_adr + sn, prev_screen + pn, row_len))
-#else
-	if (MEMCMP_DOS_VS_UNIX(screen_adr + sn, prev_screen + pn, row_len))
-#endif
-	  {
-	     line = screen_adr + sn;
-	     prev_line = prev_screen + pn;
-	     line_max = line + Columns;
-	     bufp = buf;
-	     SLsmg_gotorc (i - imin, 0);
-	     while (line < line_max)
-	       {
-		  /* *prev_line = char_attr = *line++; */
-		  *prev_line = char_attr = READ_WORD(line++);
-		  prev_line++;
-		  this_obj = Attribute_Map[(unsigned int) (char_attr >> 8)];
-		  if (this_obj != last_obj)
-		    {
-		       if (bufp != buf) 
-		       	 SLsmg_write_nchars ((char *) buf, (int) (bufp - buf));
-		       bufp = buf;
-		       SLsmg_set_color (abs(this_obj));
-		       last_obj = this_obj;
-		    }
-		  /* take care of invisible character */
-		  if (this_obj < 0) char_attr = (unsigned short) ' ';
-		  *bufp++ = The_Charset [char_attr & 0xFF];
-		  changed = 1;
-	       }
-	     SLsmg_write_nchars ((char *) buf, (int) (bufp - buf));
-	  }
-	pn += Columns;
-	sn += Columns;
-     }
-   
+
+   changed = 1;
+   vga.text_width = Columns;
+   vga.scan_len = 2 * Columns;
+   vga.text_height = Rows;
+   if (imin != DOSemu_Terminal_Scroll_Min) {
+      DOSemu_Terminal_Scroll_Min = imin;
+      redraw_text_screen();
+   } else {
+      changed = update_text_screen();
+   }
+
    if (changed || (last_col != cursor_col) || (last_row != cursor_row)
        || (DOSemu_Keyboard_Keymap_Prompt != last_prompt))
      {
@@ -473,7 +623,7 @@ static int slang_update (void)
 	       }
 	     else last_col -= 1;
 	  }
-	else if (cursor_blink == 0) last_row = last_col = 0;
+	else if (vga.crtc.cursor_shape & 0x6000) last_row = last_col = 0;
 	else
 	  {
 	     last_row = cursor_row - imin;
@@ -487,13 +637,83 @@ static int slang_update (void)
    return 1;
 }
 
+static void term_write_nchars_8bit(unsigned char *text, int len, Bit8u attr)
+{
+   unsigned char buf[len + 1];
+   unsigned char *bufp, *text_end;
+
+   text_end = text + len;
+
+#if SLANG_VERSION < 20000
+   /* switch off blinking for bright backgrounds */
+   if ((attr & 0x80) && !(vga.attr.data[0x10] & 0x8)) {
+      attr &= ~0x80;
+      SLsmg_set_color (Attribute_Map[attr]);
+   }
+   SLtt_Use_Blink_For_ACS = (attr & 0x80) >> 7;
+   /* we can't use the ACS when blinking */
+   if (SLtt_Use_Blink_For_ACS) {
+      for (bufp = buf; text < text_end; bufp++, text++)
+         *bufp = The_Charset[*text][0];
+      SLsmg_write_nchars(buf, bufp - buf);
+      SLsmg_refresh ();
+      return;
+   }
+#endif
+
+   while (text < text_end) {
+      for (bufp = buf; text < text_end; bufp++, text++) {
+	 if (The_Charset[*text][1] != '\0') break;
+         *bufp = The_Charset[*text][0];
+      }
+      SLsmg_write_nchars(buf, bufp - buf);
+      if (text >= text_end) break;
+      /* print ACS characters */
+      for (bufp = buf; text < text_end; bufp++, text++) {
+	 unsigned char ch = The_Charset[*text][1];
+	 if (ch == '\0') break;
+         *bufp = ch;
+      }
+      SLsmg_set_char_set(1);
+      SLsmg_write_nchars(buf, bufp - buf);
+      SLsmg_set_char_set(0);
+   }
+}
+
+static void term_write_nchars_utf8(unsigned char *text, int len, Bit8u attr)
+{
+   unsigned char buf[(len + 1) * 3];
+   unsigned char *bufp, *text_end = text + len;
+
+   for (bufp = buf; text < text_end; bufp += bufp[3], text++)
+      memcpy(bufp, The_Charset + *text, 4);
+   SLsmg_write_nchars(buf, bufp - buf);
+}
+
+static void term_draw_string(int x, int y, unsigned char *text, int len, Bit8u attr)
+{
+   int this_obj = Attribute_Map[attr];
+
+   y -= DOSemu_Terminal_Scroll_Min;
+   if (y < 0 || y >= SLtt_Screen_Rows) return;
+   SLsmg_gotorc (y, x);
+   SLsmg_set_color (abs(this_obj));
+
+   /* take care of invisible character */
+   if (this_obj < 0) {
+      unsigned char buf[len];
+      memset(buf, ' ', len);
+      SLsmg_write_nchars(buf, len);
+   } else
+      term_write_nchars(text, len, attr);
+}
+
 void dos_slang_redraw (void)
 {
    if (Slsmg_is_not_initialized) return;
    
-   SLsmg_cls ();
-   memset ((char *) prev_screen, 0xFF, 2 * Rows * Columns);
-   slang_update ();
+   redraw_text_screen();
+   SLsmg_refresh ();
 }
 
 void dos_slang_suspend (void)
@@ -516,7 +736,7 @@ void dos_slang_smart_set_mono (void)
    register unsigned short *s, *smax;
    
    Attribute_Map = BW_Attribute_Map;
-   s = screen_adr;
+   s = (unsigned short *)(vga.mem.base + vga.display_start);
    smax = s + Rows * Columns;
    
    for (i = 0; i < 256; i++) attr_count[i] = 0;
@@ -556,19 +776,32 @@ void dos_slang_smart_set_mono (void)
    SLsmg_cls ();
 }
 
+static void term_draw_text_cursor(int x, int y, Bit8u attr, int first, int last, Boolean focus)
+{
+}
 
-
+static void term_set_text_palette(DAC_entry color)
+{
+}
 
 #define term_setmode NULL
 #define term_update_cursor NULL
 
 struct video_system Video_term = {
-   0,                /* is_mapped */
+   NULL,
    terminal_initialize, 
    terminal_close,      
    term_setmode,      
    slang_update,
-   term_update_cursor
+   term_update_cursor,
+   NULL,
+   NULL
 };
 
-
+struct text_system Text_term =
+{
+   term_draw_string, 
+   NULL,
+   term_draw_text_cursor,
+   term_set_text_palette,
+};
